/**
 * @file config.h
 *
 * @brief Global configuration and parameters.
 *
 * Copyright (c) 2007, 2008, 2009, 2010 Georg Wittenburg, Freie Universitaet
 * Berlin (wittenbu@inf.fu-berlin.de). All rights reserved.
 * Please refer to the COPYING file for detailed licensing information.
 *
 * @author Georg Wittenburg
 *
 * $Id: config.h 3721 2010-08-23 13:21:05Z gwitten $
 */

#ifndef CONFIG_H
#define CONFIG_H



/**
 * @brief Name of the executable.
 */
#define SRVPLACE_NAME		"srvplace"

/**
 * @brief Version of the executable.
 */
#define SRVPLACE_VERSION	"1.0"



/*
 * Globally configure the framework for certain types of evaluation.
 */

/**
 * @brief Global switch to configure the framework for the evaluation of
 * routing protocols.
 */
#define C_ROUTING_EVALUATION

/**
 * @brief Global switch to configure the framework for the evaluation of
 * internal routing protocols.
 *
 * Requires C_ROUTING_EVALUATION to be defined.
 */
#define C_INTERNAL_ROUTING_EVALUATION

/**
 * @brief Global switch to configure the framework for the evaluation of
 * service placement algorithms.
 *
 * Requires C_INTERNAL_ROUTING_EVALUATION to be defined.
 */
#define C_SERVICE_PLACEMENT_EVALUATION

/**
 * @brief Global switch to configure the framework for the evaluation of the
 * SPi service placement system.
 *
 * Requires C_SERVICE_PLACEMENT_EVALUATION to be defined.
 */
// #define C_SERVICE_PLACEMENT_OPTIMIZED_FOR_SPI_EVALUATION

/**
 * @brief Global switch to configure the framework for the evaluation of the
 * ASG service placement system.
 *
 * Requires C_SERVICE_PLACEMENT_EVALUATION to be defined.
 */
// #define C_SERVICE_PLACEMENT_OPTIMIZED_FOR_ASG_EVALUATION

/**
 * @brief Global switch to configure the framework for the evaluation on the
 * testbed.
 */
// #define C_TESTBED_EVALUATION

/**
 * @brief (Artificial) single-hop link reliability.
 *
 * Allows to artificially reduce the reliability of links. Defines the
 * probability of receiving a packet; the other packets are dropped.
 */
// #define C_ARTIFICIAL_LINK_RELIABILITY 1.0
// #define C_ARTIFICIAL_LINK_RELIABILITY 0.9
// #define C_ARTIFICIAL_LINK_RELIABILITY 0.8
// #define C_ARTIFICIAL_LINK_RELIABILITY 0.7



#ifdef C_INTERNAL_ROUTING_EVALUATION
	#ifndef C_ROUTING_EVALUATION
		#error Support for internal routing evaluation requires support for routing evaluation.
	#endif	// C_ROUTING_EVALUATION
#endif	// C_INTERNAL_ROUTING_EVALUATION

#ifdef C_SERVICE_PLACEMENT_EVALUATION
	#ifndef C_INTERNAL_ROUTING_EVALUATION
		#error Support for service placement evaluation requires support for internal routing evaluation.
	#endif	// C_INTERNAL_ROUTING_EVALUATION
#endif	// C_SERVICE_PLACEMENT_EVALUATION

#ifdef C_SERVICE_PLACEMENT_OPTIMIZED_FOR_SPI_EVALUATION
	#ifndef C_SERVICE_PLACEMENT_EVALUATION
		#error Support for service placement optimized for SPi requires support for service placement evaluation.
	#endif	// C_SERVICE_PLACEMENT_EVALUATION
	#ifdef C_SERVICE_PLACEMENT_OPTIMIZED_FOR_ASG_EVALUATION
		#error Cannot optimize for both SPi and ASG service placement architectures.
	#endif	// C_SERVICE_PLACEMENT_OPTIMIZED_FOR_ASG_EVALUATION
#endif	// C_SERVICE_PLACEMENT_OPTIMIZED_FOR_SPI_EVALUATION

#ifdef C_SERVICE_PLACEMENT_OPTIMIZED_FOR_ASG_EVALUATION
	#ifndef C_SERVICE_PLACEMENT_EVALUATION
		#error Support for service placement optimized for ASG requires support for service placement evaluation.
	#endif	// C_SERVICE_PLACEMENT_EVALUATION
	#ifdef C_SERVICE_PLACEMENT_OPTIMIZED_FOR_SPI_EVALUATION
		#error Cannot optimize for both SPi and ASG service placement architectures.
	#endif	// C_SERVICE_PLACEMENT_OPTIMIZED_FOR_SPI_EVALUATION
#endif	// C_SERVICE_PLACEMENT_OPTIMIZED_FOR_ASG_EVALUATION

#ifdef C_TESTBED_EVALUATION
	#ifdef ARCH_NS2
		#error Support for the testbed conflicts with support for ns-2.
	#endif	// ARCH_NS2
	#ifdef C_ARTIFICIAL_LINK_RELIABILITY
		#error Support for the testbed conflicts with support for artificial link reliability.
	#endif	// C_ARTIFICIAL_LINK_RELIABILITY
#endif	// C_TESTBED_EVALUATION

#ifdef ARCH_NS2
	#ifdef C_ARTIFICIAL_LINK_RELIABILITY
		#warning Compiling for ns-2 with support for artificial link reliability...
	#endif	// C_ARTIFICIAL_LINK_RELIABILITY
#endif	// ARCH_NS2



/*
 * Conditionally use some features of the platform.
 */

/**
 * @brief Global constant to enable debugging.
 */
#define C_ENABLE_DEBUG

/**
 * @brief Debug level for the main component.
 */
#define C_MAIN_DEBUG_LEVEL	C_MSG_INFO

/**
 * @brief Global constant to enable consistency checks.
 *
 * NOTE: List-specific checks are disabled in list.c.
 */
#define C_ENABLE_CONSISTENCY_CHECKS

/**
 * @brief Interval of periodic measurements in seconds.
 *
 * Set to 0.0 to disable periodic measurements.
 */
#define C_PERIODIC_MEASUREMENT_INTERVAL		1.0

/**
 * @brief Enable output of tracing information for NAM.
 *
 * Undefined except when specifically generating NAM output.
 */
#ifdef ARCH_NS2
	#define C_ENABLE_NAM_TRACE_OUTPUT
#endif	// ARCH_NS2

/**
 * @brief Add simulated processing time on each node.
 *
 * Undefined for routing evaluation.
 * Defined for routing and SPi evaluation.
 */
#ifdef C_SERVICE_PLACEMENT_EVALUATION
	#ifdef ARCH_NS2
		#define C_ADD_SIMULATED_PROCESSING_TIME
	#endif	// ARCH_NS2
#endif	// C_SERVICE_PLACEMENT_EVALUATION

/**
 * @brief Use internal (built-in) link monitoring or not.
 */
// #define C_USE_INTERNAL_LINK_MONITORING

/**
 * @brief Use internal (built-in) routing or not.
 */
#ifdef C_INTERNAL_ROUTING_EVALUATION
	#define C_USE_INTERNAL_ROUTING
#endif	// C_INTERNAL_ROUTING_EVALUATION

/**
 * @brief Enable sending of periodic beacons.
 */
#ifdef C_USE_INTERNAL_ROUTING
	#define C_ENABLE_BEACONING
#endif	// C_USE_INTERNAL_ROUTING

/**
 * @brief Enable optimization for unreliable links.
 */
#if defined (C_TESTBED_EVALUATION) || defined (C_ARTIFICIAL_LINK_RELIABILITY)
	#define C_SPI_OPTIMIZED_FOR_UNRELIABLE_LINKS
#endif	// defined (C_TESTBED_EVALUATION) || defined (C_ARTIFICIAL_LINK_RELIABILITY)

/**
 * @brief Use ETX link metric or not.
 */
#ifndef ARCH_NS2
	#define C_USE_ETX_LINK_METRIC
#endif	// ARCH_NS2

#ifdef C_USE_ETX_LINK_METRIC
	#ifdef C_USE_INTERNAL_LINK_MONITORING
		#error Support for ETX conflicts with internal link monitoring since they use the same types of packets.
	#endif	// C_USE_INTERNAL_LINK_MONITORING
	#ifndef C_ENABLE_BEACONING
		#error Support for ETX requires support for beaconing.
	#endif	// C_ENABLE_BEACONING
#endif	// C_USE_ETX_LINK_METRIC

#ifdef C_SPI_OPTIMIZED_FOR_UNRELIABLE_LINKS
	#ifndef C_USE_ETX_LINK_METRIC
		#error Optimization for unreliable links requires support for ETX.
	#endif	// C_USE_ETX_LINK_METRIC
#endif	// C_SPI_OPTIMIZED_FOR_UNRELIABLE_LINKS

/**
 * @brief Use network mapping or not.
 */
#ifdef C_SERVICE_PLACEMENT_EVALUATION
	#define C_USE_NETWORK_MAPPING
#endif	// C_SERVICE_PLACEMENT_EVALUATION

#ifdef C_USE_NETWORK_MAPPING
	#ifndef C_USE_INTERNAL_ROUTING
		#error Support for network mapping requires built-in routing.
	#endif	// C_USE_INTERNAL_ROUTING
#endif	// C_USE_NETWORK_MAPPING

/**
 * @brief Enable SPi-specific optimizations in various components.
 */
#ifdef C_SERVICE_PLACEMENT_OPTIMIZED_FOR_SPI_EVALUATION
	#define C_ENABLE_SPI_OPTIMIZATIONS
#endif	// C_SERVICE_PLACEMENT_OPTIMIZED_FOR_SPI_EVALUATION

/**
 * @brief Enable ASG-specific optimizations in various components.
 */
#ifdef C_SERVICE_PLACEMENT_OPTIMIZED_FOR_ASG_EVALUATION
	#define C_ENABLE_ASG_OPTIMIZATIONS
#endif	// C_SERVICE_PLACEMENT_OPTIMIZED_FOR_ASG_EVALUATION

/**
 * @brief Flag whether or not to assume that links are bi-directional.
 *
 * NOTE: Some of the clustering code relies on this assumption and will break
 * if uncommented. It is probably just a question of getting all the corner
 * cases right and should be considered a TODO.
 */
#define C_ASSUME_BIDIRECTIONAL_LINKS

/**
 * @brief Flag whether or not to transmit all packets via broadcast.
 *
 * The PDR for unicast packets is very bad on some wireless links, hence
 * packets can be transmitted as broadcasts and later filtered using the
 * internal addressing scheme.
 */
#ifndef ARCH_NS2
// 	#define C_TRANSMIT_PACKETS_VIA_BROADCAST
#endif	// ARCH_NS2



/*
 * Set global constants for the entire framework.
 */

/**
 * @brief Maximal number of nodes in the network.
 */
#define C_MAXIMAL_NUMBER_OF_NODES	100

/**
 * @brief Maximum transmission unit (MTU) in bytes.
 *
 * Value for service placement evaluation:	4096
 * Value for routing evaluation:		1500
 *
 * NOTE: A more suitable value for SPi would be 1,500 bytes, but with that
 * size, we would have to implement fragmentation of packets for larger
 * scenarious.
 */
#if defined (ARCH_NS2) && defined (C_SERVICE_PLACEMENT_EVALUATION)
	#define C_MTU			4096
#else
	#define C_MTU			1500
#endif	// ARCH_NS2 && C_SERVICE_PLACEMENT_EVALUATION

/**
 * @brief Maximal network diameter (in hops).
 *
 * NOTE: Changing this value changes the output format and requires adaptations
 * in the scripts that process the output. Current value is 20.
 *
 * Recommended values: AODV RFC: 35; DYMO RFC draft: 10
 * Used values: AODV (ns-2): 30; AODV-UU: RFC; DYMO-UM: RFC
 */
#define C_MAXIMAL_NETWORK_DIAMETER	20

/**
 * @brief Maximal number of expected neighbors.
 */
#ifdef ARCH_NS2
	#define C_MAXIMAL_NODE_DEGREE	20
#else
	#define C_MAXIMAL_NODE_DEGREE	30
#endif	// ARCH_NS2

/**
 * @brief Default time-to-live (TTL) for transmitted packets.
 */
#define C_DEFAULT_TTL			C_MAXIMAL_NETWORK_DIAMETER

/**
 * @brief Expected traversal time for a single node.
 *
 * [Adapted from RFC 3561 (AODV)] A conservative estimate of the average one
 * hop traversal time for packets and should include queuing delays, interrupt
 * processing times and transfer times.
 *
 * Recommended values: AODV RFC: 0.04; DYMO RFC draft: n/a
 * Used values: AODV (ns-2): 0.03; AODV-UU: RFC; DYMO-UM: n/a
 */
#define C_NODE_TRAVERSAL_TIME		0.04

/**
 * @brief Expected single-hop round trip time.
 *
 * Expected round trip time (RTT) for single-hop packet transmission in
 * seconds.
 */
#define C_SINGLE_HOP_RTT		(2 * C_NODE_TRAVERSAL_TIME)

/**
 * @brief Expected network traversal time time.
 *
 * Expected time for a packet to traverse the entire network.
 */
#define C_NETWORK_TRAVERSAL_TIME	(2 * C_NODE_TRAVERSAL_TIME * C_MAXIMAL_NETWORK_DIAMETER)

/**
 * @brief Expected time to establish a new route in seconds.
 */
#define C_PATH_DISCOVERY_TIME		(2 * C_NETWORK_TRAVERSAL_TIME)

/**
 * @brief Expected time to allow for all remaining packets in the network to be
 * delivered.
 */
#define C_NETWORK_FLUSH_TIME		(10 * C_NETWORK_TRAVERSAL_TIME)

/**
 * @brief Timeout for higherlevel protocols.
 *
 * It should allow for enough time for RUDP to transmit and acknowledge a
 * packet under normal network conditions.
 */
#define C_TIMEOUT			((C_RUDP_PACKET_RETRIES + 1) * C_RUDP_MAX_BACKOFF_TIME)

/**
 * @brief Unicast message sent timeout in seconds.
 *
 * Adapted from DYMO draft (Sec. 6 and 7.2), used in link monitoring.
 *
 * Recommended values: AODV RFC: (NODE_TRAVERSAL_TIME + 10); DYMO RFC draft: 1
 * Used values: AODV (ns-2): n/a; AODV-UU: RFC; DYMO-UM: n/a
 */
#define C_UNICAST_MESSAGE_SENT_TIMEOUT	(C_NODE_TRAVERSAL_TIME + 10)

#ifdef C_USE_INTERNAL_LINK_MONITORING
/**
 * @brief Maximal number of unanswered packets before a link is reported as
 * broken.
 */
#define INTERNAL_LINK_MONITORING_MAX_UNANSWERED_PACKETS		10
#endif	// C_USE_INTERNAL_LINK_MONITORING

#ifdef C_ENABLE_BEACONING
/**
 * @brief Beacon interval in seconds.
 */
#ifdef C_USE_ETX_LINK_METRIC
	#define C_BEACON_INTERVAL	1	// Suggested by ETX paper.
#else
	#define C_BEACON_INTERVAL	10	// Sufficient to keep routing information stable.
#endif	// C_USE_ETX_LINK_METRIC
#endif	// C_ENABLE_BEACONING

#ifdef C_USE_ETX_LINK_METRIC
/**
 * @brief Size of the window for counting beacons in seconds.
 */
#define C_ETX_WINDOW_SIZE		10	// Suggested by ETX paper.

/**
 * @brief A threshold value below which links can be considered "good".
 */
#define C_ETX_GOOD_LINK_THRESHOLD	2.0

/**
 * @brief A threshold value above which links can be considered broken.
 */
#define C_ETX_BROKEN_LINK_THRESHOLD	100.0
#endif	// C_USE_ETX_LINK_METRIC

/**
 * @brief Maximal supported number of different services.
 *
 * Each service may have multiple instances.
 */
#define C_MAX_SERVICES	4

#endif	// CONFIG_H
