/**
 * @file dymo.c
 *
 * @brief Dynamic MANET On-demand (DYMO) routing protocol.
 *
 * This is an implementation of the Dynamic MANET On-demand (DYMO) routing
 * protocol as specified in
 * I. Chakeres and C. Perkins. Dynamic MANET On-demand (DYMO) Routing. IETF
 * Internet Draft, revision 16, December 2008.
 * (http://www.ietf.org/internet-drafts/draft-ietf-manet-dymo-16.txt)
 *
 * The implementation is feature complete and compliant with the revision of
 * the draft cited above EXCEPT for the following MANDATORY (!) features:
 * - We use the regular broadcast address (e.g. 255.255.255.255) rather than
 *   LL-MANET-ROUTERS (224.0.0.109) (Sec. 4.2.1)
 *   -> This is purely convenience and should be OK for all experiments.
 * - No Action is taken after OwnSeqNum loss (Sec. 5.1.4)
 *   -> For simulations, this is a non-issue. However, it must be implemented
 *      for any non-trivial deployment.
 *
 * Excerpts from the draft are used throughout the code for documentation.
 *
 * Copyright (c) 2007, 2008, 2009, 2010 Georg Wittenburg, Freie Universitaet
 * Berlin (wittenbu@inf.fu-berlin.de). All rights reserved.
 * Please refer to the COPYING file for detailed licensing information.
 *
 * @author Georg Wittenburg
 *
 * $Id: dymo.c 3640 2010-07-06 14:40:37Z gwitten $
 */

#include <float.h>
#include <math.h>

#include "../config.h"
#include "../macros.h"
#include "../types.h"
#include "../vars.h"
#include "../main.h"
#include "../util/debug.h"
#include "../util/list.h"
#include "../util/callback.h"
#include "../util/net.h"
#include "dymo.h"
#include "dymo_types.h"
#include "dymo_vars.h"



#define DYMO_PACKET_TYPE_RREQ	0x0	///< DYMO RREQ packet type.
#define DYMO_PACKET_TYPE_RREP	0x1	///< DYMO RREP packet type.
#define DYMO_PACKET_TYPE_RERR	0x2	///< DYMO RERR packet type.

/**
 * @brief IP destination address for (most) DYMO messages.
 *
 * 4.2.1. Generalized Packet and Message Structure
 *
 * [...]
 *
 * Most DYMO messages are sent with the IP destination address set to the
 * link-local multicast address LL-MANET-ROUTERS [I-D.ietf-manet-iana] unless
 * otherwise stated. Therefore, all DYMO routers SHOULD subscribe to
 * LL-MANET-ROUTERS [I-D.ietf-manet-iana] for receiving control packets. Note
 * that multicast packets may be sent via unicast. For example, this may occur
 * for certain link-types (non broadcast mediums), improved robustness, or
 * manually configured router adjacency.
 *
 * [...]
 *
 * NOTE: According to RFC-ietf-manet-iana-07.txt, the correct address is
 * 224.0.0.109, however, we're using the broadcast address to simplify the
 * integration.
 */
#define DYMO_LL_MANET_ROUTERS			C_BROADCAST_ADDRESS



/*
 * 6. Configuration Parameters and Other Administrative Options
 */

/**
 * @brief Message hop limit in hops.
 *
 * Recommended values: AODV RFC: 35; DYMO RFC draft: 10
 * Used values: AODV (ns-2): 30; AODV-UU: RFC; DYMO-UM: RFC
 */
#define DYMO_MSG_HOPLIMIT			C_MAXIMAL_NETWORK_DIAMETER

/**
 * @brief Route timeout in seconds.
 *
 * Recommended values: AODV RFC: n/a; DYMO RFC draft: 5
 * Used values: AODV (ns-2): n/a; AODV-UU: n/a; DYMO-UM: 3
 */
#define DYMO_ROUTE_TIMEOUT			5

/**
 * @brief Minimal route age timeout in seconds.
 *
 * Recommended values: AODV RFC: n/a; DYMO RFC draft: 1
 * Used values: AODV (ns-2): n/a; AODV-UU: n/a; DYMO-UM: n/a
 */
#define DYMO_ROUTE_AGE_MIN_TIMEOUT		1

/**
 * @brief Maximal route sequence number age timeout in seconds.
 *
 * Recommended values: AODV RFC: n/a; DYMO RFC draft: 60
 * Used values: AODV (ns-2): n/a; AODV-UU: n/a; DYMO-UM: n/a
 *
 * NOTE: The default value of this required feature leads to bad performance
 * under light load. DYMO-UM does not implement this timeout at all, so we
 * increase the value in order to have comparable results.
 */
#define DYMO_ROUTE_SEQNUM_AGE_MAX_TIMEOUT	300

#ifdef DYMO_DISCARD_UNUSED_ROUTES_AFTER_TIMEOUT
/**
 * @brief Route use timeout in seconds.
 *
 * Recommended values: AODV RFC: (5 * 3); DYMO RFC draft: DYMO_ROUTE_TIMEOUT
 * Used values: AODV (ns-2): n/a; AODV-UU: 10; DYMO-UM: 3
 */
#define DYMO_ROUTE_USED_TIMEOUT			DYMO_ROUTE_TIMEOUT
#endif	// DYMO_DISCARD_UNUSED_ROUTES_AFTER_TIMEOUT

/**
 * @brief Route delete timeout in seconds.
 *
 * Recommended values: AODV RFC: 3, (2 * 3) for RREPs; DYMO RFC draft: (2 * DYMO_ROUTE_TIMEOUT)
 * Used values: AODV (ns-2): 10; AODV-UU: 10, (2 * 10) for RREPs; DYMO-UM: (5 * DYMO_ROUTE_TIMEOUT)
 */
#define DYMO_ROUTE_DELETE_TIMEOUT		(5 * DYMO_ROUTE_TIMEOUT)

/**
 * @brief RREQ wait time in seconds.
 *
 * Recommended values: AODV RFC: NET_TRAVERSAL_TIME; DYMO RFC draft: 2
 * Used values: AODV (ns-2): 10; AODV-UU: RING_TRAVERSAL_TIME; DYMO-UM: 1
 *
 * NOTE: Our value differs vastly from the draft.
 */
#define DYMO_ROUTE_RREQ_WAIT_TIME		C_PATH_DISCOVERY_TIME

/**
 * @brief RREQ tries.
 *
 * Recommended values: AODV RFC: 2; DYMO RFC draft: 3
 * Used values: AODV (ns-2): 3; AODV-UU: RFC; DYMO-UM: RFC
 */
#define DYMO_RREQ_TRIES				3

// NOTE: Active link monitoring is not part of the routing protocol and hence
// UNICAST_MESSAGE_SENT_TIMEOUT is defined elsewhere.



/*
 * RFC3561 (AODV) - 10. Configuration Parameters
 */

/**
 * @brief Network diameter.
 *
 * Recommended values: AODV RFC: 35; DYMO RFC draft: n/a
 * Used values: AODV (ns-2): 30; AODV-UU: RFC; DYMO-UM: 10
 */
#define AODV_NET_DIAMETER		C_MAXIMAL_NETWORK_DIAMETER

/**
 * @brief Time required for forwarding a packet in seconds.
 *
 * Recommended values: AODV RFC: 0.04; DYMO RFC draft: n/a
 * Used values: AODV (ns-2): 0.03; AODV-UU: RFC; DYMO-UM: n/a
 */
#define AODV_NODE_TRAVERSAL_TIME	C_NODE_TRAVERSAL_TIME

/**
 * @brief Time required for disseminating a packet through the entire network in seconds.
 *
 * Recommended values: AODV RFC: (2 * AODV_NODE_TRAVERSAL_TIME * AODV_NET_DIAMETER); DYMO RFC draft: n/a
 * Used values: AODV (ns-2): 10; AODV-UU: RFC; DYMO-UM: n/a
 */
#define AODV_NET_TRAVERSAL_TIME		(2 * AODV_NODE_TRAVERSAL_TIME * AODV_NET_DIAMETER)

/**
 * @brief Timeout buffer to compensate for congestion.
 *
 * Recommended values: AODV RFC: 2; DYMO RFC draft: n/a
 * Used values: AODV (ns-2): n/a; AODV-UU: RFC; DYMO-UM: n/a
 */
#define AODV_TIMEOUT_BUFFER		2

/**
 * @brief Initial TTL value for RREQs
 *
 * Recommended values: AODV RFC: 1; DYMO RFC draft: n/a
 * Used values: AODV (ns-2): 5; AODV-UU: RFC; DYMO-UM: n/a
 */
#define AODV_TTL_START			1

/**
 * @brief TTL increment for RREQs
 *
 * Recommended values: AODV RFC: 2; DYMO RFC draft: n/a
 * Used values: AODV (ns-2): RFC; AODV-UU: RFC; DYMO-UM: n/a
 */
#define AODV_TTL_INCREMENT		2

/**
 * @brief Threshold for network-wide dissemination
 *
 * Recommended values: AODV RFC: 7; DYMO RFC draft: n/a
 * Used values: AODV (ns-2): RFC; AODV-UU: RFC; DYMO-UM: n/a
 */
#define AODV_TTL_THRESHOLD		7



/*
 * Additional configuration options.
 */

/**
 * @brief Size of packet queue in packets.
 *
 * Recommended values: AODV RFC: n/a; DYMO RFC draft: n/a
 * Used values: AODV (ns-2): 64; AODV-UU: 512; DYMO-UM: 512
 */
#define DYMO_BUFFER_SIZE_PACKETS	512

/**
 * @brief Use intermediate RREPs.
 */
#define DYMO_USE_INTERMEDIATE_RREPS

/**
 * @brief Use additional routes in RREPs.
 */
#define DYMO_USE_ADDITIONAL_ROUTES_IN_RREPS

/**
 * @brief Use additional routes in RERRs.
 */
#define DYMO_USE_ADDITIONAL_ROUTES_IN_RERRS



#if defined (C_ENABLE_SPI_OPTIMIZATIONS) || defined (C_ENABLE_ASG_OPTIMIZATIONS)

	/**
	* @brief Use overheard broadcast packets to add routes.
	*
	* NOTE: This functionality is not part of the RFC, but improves performance
	* tremendously.
	*/
	#define DYMO_USE_OVERHEARD_BROADCAST_TO_ADD_ROUTES

#ifdef C_USE_ETX_LINK_METRIC
	/**
	* @brief ETX threshold for using overheard routes.
	*/
	#define DYMO_USE_OVERHEARD_BROADCAST_ETX_THRESHOLD	C_ETX_GOOD_LINK_THRESHOLD
#endif	// C_USE_ETX_LINK_METRIC

#endif	// (C_ENABLE_SPI_OPTIMIZATIONS || C_ENABLE_ASG_OPTIMIZATIONS)



/**
 * @brief Constant for unknown DYMO sequence numer.
 *
 * NOTE: We are using the value zero (0) because it is not a valid value
 * according to Sec. 5.1.3.
 */
#define DYMO_UNKNOWN_SEQUENCE_NUMBER	0

/**
 * @brief Constant for unknown DYMO distance metric.
 *
 * NOTE: We are using the value zero (0) because it is not a valid value
 * according various sections of the draft.
 */
#define DYMO_UNKNOWN_DISTANCE		0

/**
 * @brief Constant for maximal DYMO distance metric.
 *
 * This depends on the type of DYMO_route_t.dist.
 */
#define DYMO_MAX_DISTANCE		0xFFFF

#define DYMO_ROUTE_INTERNAL_ERROR	0x0	///< Internal error.
#define DYMO_ROUTE_STALE		0x1	///< DYMO stale route.
#define DYMO_ROUTE_LOOP_POSSIBLE	0x2	///< DYMO route with possible loop.
#define DYMO_ROUTE_INFERIOR		0x3	///< DYMO inferior route.
#define DYMO_ROUTE_SUPERIOR		0x4	///< DYMO superior route.



/*
 * Interface to main module.
 */

/**
 * @brief Initialize the DYMO module.
 */
void DYMO_init() {
	C_DYMO_own_seq_num = 0;
	C_DYMO_routing_table = NULL;
	list_alloc(&C_DYMO_routing_table);
	C_DYMO_packet_queue = NULL;
	list_alloc(&C_DYMO_packet_queue);
	C_DYMO_transmitted_RREQs = 0;
	C_DYMO_transmitted_RREPs = 0;
	C_DYMO_transmitted_RERRs = 0;
	C_DYMO_transmitted_unicast_packets = 0;
	C_DYMO_received_unicast_packets = 0;
	C_DYMO_transmitted_broadcast_packets = 0;
	C_DYMO_received_broadcast_packets = 0;
}

/**
 * @brief Start the DYMO module.
 */
void DYMO_start() {
}

static void remove_and_free_packet_from_queue(DYMO_queued_packet_t* packet);

/**
 * @brief Stop the DYMO module.
 */
void DYMO_stop() {
	DYMO_queued_packet_t* queued_packet = NULL;
	while(!list_empty(C_DYMO_packet_queue)) {
		queued_packet = (DYMO_queued_packet_t*) list_head(C_DYMO_packet_queue);

		C_DEBUG(DYMO_DEBUG_LEVEL, C_MSG_INFO, "DYMO: Aborting transmission of packet to %s with sequence number %d...\n", addr_to_str(queued_packet->packet.header->destination_address), queued_packet->packet.header->seqno);
		remove_and_free_packet_from_queue(queued_packet);
	}
}

/**
 * @brief Finalize the DYMO module.
 */
void DYMO_finalize() {
	DYMO_route_t* route = NULL;
	list_iterator_t routes_iterator;

	list_iterator_init(C_DYMO_routing_table, &routes_iterator);
	while((route = (DYMO_route_t*) list_iterator_next(&routes_iterator)) != NULL) {
		if(route->minimum_age_delete_timeout_callback_id != 0)
			callback_cancel(route->minimum_age_delete_timeout_callback_id);
		if(route->maximum_sequence_number_delete_timeout_callback_id != 0)
			callback_cancel(route->maximum_sequence_number_delete_timeout_callback_id);
#ifdef DYMO_DISCARD_UNUSED_ROUTES_AFTER_TIMEOUT
		if(route->recently_used_timeout_callback_id != 0)
			callback_cancel(route->recently_used_timeout_callback_id);
#endif	// DYMO_DISCARD_UNUSED_ROUTES_AFTER_TIMEOUT
		if(route->delete_information_timeout_callback_id != 0)
			callback_cancel(route->delete_information_timeout_callback_id);
	}
	list_free_including_data(&C_DYMO_routing_table, free);

	if(C_check_consistency_on_shutdown) {
		C_ABORT_ON(!list_empty(C_DYMO_packet_queue), "DYMO: There are still %d packet(s) in the packet queue.\n", list_size(C_DYMO_packet_queue));
	}
	list_free(&C_DYMO_packet_queue);
}



/*
 * Utility funcions.
 */

/**
 * @brief Get the appropriatly incremented local sequence number.
 *
 * 5.1.2. Numerical Operations on OwnSeqNum
 *
 * When ThisNode increments its OwnSeqNum (as described in Section 5.3) it MUST
 * do so by treating the sequence number value as an unsigned number.
 *
 * 5.1.3. OwnSeqNum Rollover
 *
 * If the sequence number has been assigned to be the largest possible number
 * representable as a 16-bit unsigned integer (i.e., 65,535), then the sequence
 * number SHOULD be set to one (1) incremented.
 */
static uint16_t get_own_seq_num() {
	C_DYMO_own_seq_num++;

	if(C_DYMO_own_seq_num == 0)
		C_DYMO_own_seq_num = 1;

	return C_DYMO_own_seq_num;
}



/*
 * Routing table related functions.
 */

/**
 * @brief Check whether address of two routing table entries match.
 *
 * @param item Routing table entry to check.
 * @param cmp_item Reference routing table entry.
 * @return <code>true</code> if address two routing table entries are equal,
 * <code>false</code> otherwise.
 */
static bool matches_route_address(const void* item, const void* cmp_item) {
	DYMO_route_t* route = (DYMO_route_t*) item;
	DYMO_route_t* cmp_route = (DYMO_route_t*) cmp_item;
	return (route->address == cmp_route->address);
}

/**
 * @brief Minimum delete timeout.
 *
 * 5.2.3.1. Minimum Delete Timeout (ROUTE_AGE_MIN)
 *
 * When a DYMO router transmits a RM, other DYMO routers expect the
 * transmitting DYMO router to have a forwarding route to the RM originator.
 * After updating a route table entry, it should be maintained for at least
 * ROUTE_AGE_MIN. Failure to maintain the information might result in lost
 * messages/packets, or in the worst case scenario several duplicate messages.
 *
 * After the ROUTE_AGE_MIN timeout a route can safely be deleted.
 *
 * @param parameter Pointer to routing table entry.
 */
static void route_minimum_age_delete_timeout(callback_parameter_t parameter) {
	DYMO_route_t* route = (DYMO_route_t*) parameter;

	C_ABORT_ON(!list_contains(C_DYMO_routing_table, route), "DYMO: Trying to flag route as safe to delete, but route at %p is not contained in the routing table.\n", route);

	C_DEBUG(DYMO_DEBUG_LEVEL, C_MSG_VERBOSE, "DYMO: Flagging route to %s as safe to delete...\n", addr_to_str(route->address));
	route->minimum_age_delete_timeout_callback_id = 0;
	route->can_be_deleted = true;
}

/**
 * @brief Maximum sequence number delete timeout.
 *
 * 5.2.3.2. Maximum Sequence Number Delete Timeout (ROUTE_SEQNUM_AGE_MAX)
 *
 * Sequence number information is time sensitive, and MUST be deleted after a
 * time in order to ensure loop-free routing.
 *
 * After the ROUTE_AGE_MAX timeout a route's sequence number information MUST
 * be discarded.
 *
 * @param parameter Pointer to routing table entry.
 */
static void route_maximum_sequence_number_delete_timeout(callback_parameter_t parameter) {
	DYMO_route_t* route = (DYMO_route_t*) parameter;

	C_ABORT_ON(!list_contains(C_DYMO_routing_table, route), "DYMO: Trying to discard sequence number of route, but route at %p is not contained in the routing table.\n", route);

	C_DEBUG(DYMO_DEBUG_LEVEL, C_MSG_VERBOSE, "DYMO: Discarding sequence number of route to %s...\n", addr_to_str(route->address));
	route->maximum_sequence_number_delete_timeout_callback_id = 0;
	route->seq_num = DYMO_UNKNOWN_SEQUENCE_NUMBER;
}

static void route_delete_information_timeout(callback_parameter_t parameter);

#ifdef DYMO_DISCARD_UNUSED_ROUTES_AFTER_TIMEOUT
/**
 * @brief Recently used timeout.
 *
 * 5.2.3.3. Recently Used Timeout (ROUTE_USED)
 *
 * When a route is used to forward data packets, this timer is set to expire
 * after ROUTE_USED_TIMEOUT. This operation is also discussed in Section 5.5.2.
 *
 * If a route has not been used recently, then a timer for ROUTE_DELETE is set
 * to ROUTE_DELETE_TIMEOUT.
 *
 * @param parameter Pointer to routing table entry.
 */
static void route_recently_used_timeout(callback_parameter_t parameter) {
	DYMO_route_t* route = (DYMO_route_t*) parameter;

	C_ABORT_ON(!list_contains(C_DYMO_routing_table, route), "DYMO: Trying to start deletion timer on unused route, but route at %p is not contained in the routing table.\n", route);

	C_DEBUG(DYMO_DEBUG_LEVEL, C_MSG_DEBUG, "DYMO: Route to %s has not recently been used, starting timer for deletion...\n", addr_to_str(route->address));
	route->recently_used_timeout_callback_id = 0;
	if(route->delete_information_timeout_callback_id == 0)
		route->delete_information_timeout_callback_id = callback_add(route_delete_information_timeout, route, DYMO_ROUTE_DELETE_TIMEOUT);
}
#endif	// DYMO_DISCARD_UNUSED_ROUTES_AFTER_TIMEOUT

/**
 * @brief Delete Information timeout.
 *
 * 5.2.3.4. Delete Information Timeout (ROUTE_DELETE)
 *
 * As time progresses the likelihood that old routing information is useful
 * decreases, especially if the network nodes are mobile. Therefore, old
 * information should be deleted.
 *
 * After the ROUTE_DELETE timeout, the routing table entry should be deleted.
 * If a forwarding route exists, it should be removed and the Route.Forwarding
 * flag unset.
 *
 * @param parameter Pointer to routing table entry.
 */
static void route_delete_information_timeout(callback_parameter_t parameter) {
	DYMO_route_t* route = (DYMO_route_t*) parameter;

	C_ABORT_ON(!list_contains(C_DYMO_routing_table, route), "DYMO: Trying to delete route, but route at %p is not contained in the routing table.\n", route);
	C_ABORT_ON(route->minimum_age_delete_timeout_callback_id != 0, "DYMO: Trying to delete route to %s, but the minimum age timer is still running.\n", addr_to_str(route->address));
	C_ABORT_ON(!route->can_be_deleted, "DYMO: Trying to delete route to %s, but it is flagged as not to be deleted.\n", addr_to_str(route->address));

	C_DEBUG(DYMO_DEBUG_LEVEL, C_MSG_DEBUG, "DYMO: Deletion timeout for route to %s has expired, deleting route...\n", addr_to_str(route->address));
	route->delete_information_timeout_callback_id = 0;

	// Cancel any other active timers on this routing table entry.
	if(route->maximum_sequence_number_delete_timeout_callback_id != 0) {
		callback_cancel(route->maximum_sequence_number_delete_timeout_callback_id);
		route->maximum_sequence_number_delete_timeout_callback_id = 0;
	}
#ifdef DYMO_DISCARD_UNUSED_ROUTES_AFTER_TIMEOUT
	if(route->recently_used_timeout_callback_id != 0) {
		callback_cancel(route->recently_used_timeout_callback_id);
		route->recently_used_timeout_callback_id = 0;
	}
#endif	// DYMO_DISCARD_UNUSED_ROUTES_AFTER_TIMEOUT

	list_remove(C_DYMO_routing_table, route);
	FREE(route);
}

/**
 * @brief Initialize a route.
 *
 * @param route Pointer to the route to initialize.
 * @param seq_num New sequence number for this route.
 * @param next_hop_address New next hop address for this route.
 * @param dist New distance for this route.
 */
static void init_route(DYMO_route_t* route, uint16_t seq_num, C_addr_t next_hop_address, uint16_t dist) {
	C_ABORT_ON(route == NULL, "Trying to initialize a route at NULL.\n");
	C_ABORT_ON(route->address == C_BROADCAST_ADDRESS, "Trying to initialize a route to the BROADCAST address.\n");
	C_ABORT_ON(seq_num == DYMO_UNKNOWN_SEQUENCE_NUMBER, "Trying to set sequence number of route to %s to unknown.\n", addr_to_str(route->address));
	C_ABORT_ON(next_hop_address == C_BROADCAST_ADDRESS, "Trying to set next hop of route to %s to BROADCAST address.\n", addr_to_str(route->address));
	C_ABORT_ON(dist == DYMO_UNKNOWN_DISTANCE, "Trying to set distance of route to %s to unknown.\n", addr_to_str(route->address));

	route->seq_num = seq_num;
	route->next_hop_address = next_hop_address;
	route->dist = dist;
	route->forwarding = true;
	route->broken = false;

	route->minimum_age_delete_timeout_callback_id = callback_add(route_minimum_age_delete_timeout, route, DYMO_ROUTE_AGE_MIN_TIMEOUT);
	route->maximum_sequence_number_delete_timeout_callback_id = callback_add(route_maximum_sequence_number_delete_timeout, route, DYMO_ROUTE_SEQNUM_AGE_MAX_TIMEOUT);
#ifdef DYMO_DISCARD_UNUSED_ROUTES_AFTER_TIMEOUT
	route->recently_used_timeout_callback_id = 0;
#endif	// DYMO_DISCARD_UNUSED_ROUTES_AFTER_TIMEOUT
	route->delete_information_timeout_callback_id = 0;

	route->can_be_deleted = false;
}

/**
 * @brief Reset routing timers.
 *
 * This function resets the ROUTE_USED and ROUTE_DELETE timeouts of a route
 * after a data packet has been sent or received.
 *
 * NOTE: Contrary to the draft, we also re-enable forwarding on this route.
 *
 * @param address Address of the target node of the route.
 * @return <code>true</code> if timers of a route were updated,
 * <code>false</code> otherwise.
 */
static bool reset_routing_timers(C_addr_t address) {
	DYMO_route_t cmp_route;
	DYMO_route_t* route = NULL;

	memset(&cmp_route, 0, sizeof(DYMO_route_t));
	cmp_route.address = address;
	route = (DYMO_route_t*) list_find(C_DYMO_routing_table, &cmp_route, matches_route_address);
	if(route == NULL) {
		C_DEBUG(DYMO_DEBUG_LEVEL, C_MSG_VERBOSE, "DYMO: Trying to reset timers for route to %s, but no route to that node is available...\n", addr_to_str(address));
		return false;
	} else {
		C_DEBUG(DYMO_DEBUG_LEVEL, C_MSG_VERBOSE, "DYMO: Resetting timers for route to %s...\n", addr_to_str(address));
#ifdef DYMO_DISCARD_UNUSED_ROUTES_AFTER_TIMEOUT
		if(route->recently_used_timeout_callback_id != 0)
			callback_cancel(route->recently_used_timeout_callback_id);
		route->recently_used_timeout_callback_id = callback_add(route_recently_used_timeout, route, DYMO_ROUTE_USED_TIMEOUT);
#endif	// DYMO_DISCARD_UNUSED_ROUTES_AFTER_TIMEOUT
		if(route->delete_information_timeout_callback_id != 0) {
			callback_cancel(route->delete_information_timeout_callback_id);
			route->delete_information_timeout_callback_id = 0;
		}

		// NOTE: Not covered in the draft (see above).
		if(route->broken) {
			C_DEBUG(DYMO_DEBUG_LEVEL, C_MSG_DEBUG, "DYMO: Re-enabling forwarding on route to %s...\n", addr_to_str(route->address));
			route->forwarding = true;
			route->broken = false;
		}
		return true;
	}
}

/**
 * @brief Classify new routing information.
 *
 * 5.2.1. Judging Routing Information's Usefulness
 *
 * Given a route table entry (Route.SeqNum, Route.Dist, and Route.Broken) and
 * new incoming routing information for a particular node in a RM (Node.SeqNum,
 * Node.Dist, and RM message type - RREQ/RREP), the quality of the new routing
 * information is evaluated to determine its usefulness. Incoming routing
 * information is classified as follows:
 * [...]
 *
 * @param route Pointer to existing routing table entry.
 * @param node_seq_num New sequence number of node.
 * @param node_dist New distance to node.
 * @param routing_message_type Type of routing message, either RREQ or RREP.
 * @return Numeric identifier for the quality of the new routing information.
 */
static uint8_t classify_routing_information(DYMO_route_t* route, uint16_t node_seq_num, uint16_t node_dist, uint8_t routing_message_type) {
	C_ABORT_ON(route == NULL, "DYMO: Trying to classify route at NULL...\n");
	C_ABORT_ON((routing_message_type != DYMO_PACKET_TYPE_RREQ) && (routing_message_type != DYMO_PACKET_TYPE_RREP), "DYMO: Trying to classify route, but packet type is neither RREQ nor RREP...\n");

	if(node_seq_num < route->seq_num) {	// NOTE: This is equivalent to the draft which specifies "node_seq_num - route->seq_num < 0" using signed 16-bit arithmetic.
		C_DEBUG(DYMO_DEBUG_LEVEL, C_MSG_VERBOSE, "DYMO: Classified route to %s as stale (old seqnum = %d, old dist = %d, new seqnum = %d, new dist = %d).\n", addr_to_str(route->address), route->seq_num, route->dist, node_seq_num, node_dist);
		return DYMO_ROUTE_STALE;
	} else if(
		(node_seq_num == route->seq_num)
		&& (
			(node_dist == DYMO_UNKNOWN_DISTANCE)
			|| (route->dist == DYMO_UNKNOWN_DISTANCE)
			|| (node_dist > (route->dist + 1))
		)
	) {
		C_DEBUG(DYMO_DEBUG_LEVEL, C_MSG_VERBOSE, "DYMO: Classified route to %s as possibly having a loop (old seqnum = %d, old dist = %d, new seqnum = %d, new dist = %d).\n", addr_to_str(route->address), route->seq_num, route->dist, node_seq_num, node_dist);
		return DYMO_ROUTE_LOOP_POSSIBLE;
	} else if(
		(node_seq_num == route->seq_num)
		&& (
			(
				(node_dist == (route->dist + 1))
				&& (route->broken == false)
			) || (
				(node_dist == route->dist)
				&& (routing_message_type == DYMO_PACKET_TYPE_RREQ)
				&& (route->broken == false)
			)
		)
	) {
		C_DEBUG(DYMO_DEBUG_LEVEL, C_MSG_VERBOSE, "DYMO: Classified route to %s as inferior (old seqnum = %d, old dist = %d, new seqnum = %d, new dist = %d).\n", addr_to_str(route->address), route->seq_num, route->dist, node_seq_num, node_dist);
		return DYMO_ROUTE_INFERIOR;
	} else if(
		(node_seq_num > route->seq_num)	// NOTE: This is equivalent to the draft which specifies "node_seq_num - route->seq_num > 0" using signed 16-bit arithmetic.
		|| (
			(node_seq_num == route->seq_num)
			&& (
				(node_dist < route->dist)
				|| (
					(node_dist == (route->dist + 1))
					&& (route->broken == true)
				) || (
					(node_dist == route->dist)
					&& (
						(routing_message_type == DYMO_PACKET_TYPE_RREP)
						|| (route->broken == true)
					)
				)
			)
		)
	) {
		C_DEBUG(DYMO_DEBUG_LEVEL, C_MSG_VERBOSE, "DYMO: Classified route to %s as superior (old seqnum = %d, old dist = %d, new seqnum = %d, new dist = %d).\n", addr_to_str(route->address), route->seq_num, route->dist, node_seq_num, node_dist);
		return DYMO_ROUTE_SUPERIOR;
	}

	C_ABORT_ON(true, "DYMO: Route classification failed.\n");
	return DYMO_ROUTE_INTERNAL_ERROR;
}

/**
 * @brief Update routing table and indicate whether the new information is
 * superior.
 *
 * (see Sec. 5.2.2)
 *
 * If a new route has been established, process_queued_packets() should be
 * called after all control packets have been sent to send any data packets.
 *
 * @param node_block Pointer to node block to process.
 * @param next_hop_address Address of the next hop for this route.
 * @param routing_message_type Identifier of routing message type.
 * @return <code>true</code> if the new routing information is superior
 * compared to the previous one, <code>false</code> otherwise.
 */
static bool update_routing_information(const DYMO_node_block_t* node_block, C_addr_t next_hop_address, uint8_t routing_message_type) {
	DYMO_route_t* route = NULL;
	DYMO_route_t cmp_route;

	if(node_block->address == C_LOCAL_ADDRESS)
		return false;

	memset(&cmp_route, 0, sizeof(DYMO_route_t));
	cmp_route.address = node_block->address;
	if((route = (DYMO_route_t*) list_find(C_DYMO_routing_table, &cmp_route, matches_route_address)) == NULL) {

		C_DEBUG(DYMO_DEBUG_LEVEL, C_MSG_INFO, "DYMO: Adding new route to %s with next hop %s...\n", addr_to_str(node_block->address), addr_to_str(next_hop_address));
		MALLOC(route, DYMO_route_t);
		memset(route, 0, sizeof(DYMO_route_t));	// Make sure the checks for the timer fields below fail.
		route->address = node_block->address;
		list_add(C_DYMO_routing_table, route);

	} else if(route->seq_num == DYMO_UNKNOWN_SEQUENCE_NUMBER) {

		C_DEBUG(DYMO_DEBUG_LEVEL, C_MSG_INFO, "DYMO: Updating existing route to %s with expired sequence number, next hop is %s (seq_num = %d)...\n", addr_to_str(node_block->address), addr_to_str(next_hop_address), node_block->seq_num);
		// Nothing to do.

	} else if(classify_routing_information(route, node_block->seq_num, node_block->dist, routing_message_type) == DYMO_ROUTE_SUPERIOR) {

		if(next_hop_address == route->next_hop_address)
			C_DEBUG(DYMO_DEBUG_LEVEL, C_MSG_DEBUG, "DYMO: Updating route to %s, next hop is still %s (distance = %d)...\n", addr_to_str(node_block->address), addr_to_str(route->next_hop_address), node_block->dist);
		/* This might look a bit scary, but in fact it's just normal.
		else if(node_block->dist > (route->dist + DYMO_MSG_HOPLIMIT))
			C_DEBUG(DYMO_DEBUG_LEVEL, C_MSG_WARN, "DYMO: Updating route to %s, old next hop was %s, new next hop is %s, but distance increases from %d to %d...\n", addr_to_str(node_block->address), addr_to_str(route->next_hop_address), addr_to_str(next_hop_address), route->dist, node_block->dist);
		*/
		else
			C_DEBUG(DYMO_DEBUG_LEVEL, C_MSG_INFO, "DYMO: Updating route to %s, old next hop was %s, new next hop is %s (distance = %d)...\n", addr_to_str(node_block->address), addr_to_str(route->next_hop_address), addr_to_str(next_hop_address), node_block->dist);
		// Nothing to do.

	} else {

		C_DEBUG(DYMO_DEBUG_LEVEL, C_MSG_VERBOSE, "DYMO: Not updating route to %s, next hop remains %s...\n", addr_to_str(node_block->address), addr_to_str(route->next_hop_address));
		return false;

	}

	// NOTE: The timeout fields are re-initialized in
	// init_route() below.
	if(route->minimum_age_delete_timeout_callback_id != 0)
		callback_cancel(route->minimum_age_delete_timeout_callback_id);
	if(route->maximum_sequence_number_delete_timeout_callback_id != 0)
		callback_cancel(route->maximum_sequence_number_delete_timeout_callback_id);
#ifdef DYMO_DISCARD_UNUSED_ROUTES_AFTER_TIMEOUT
	if(route->recently_used_timeout_callback_id != 0)
		callback_cancel(route->recently_used_timeout_callback_id);
#endif	// DYMO_DISCARD_UNUSED_ROUTES_AFTER_TIMEOUT
	if(route->delete_information_timeout_callback_id != 0)
		callback_cancel(route->delete_information_timeout_callback_id);

	init_route(route, node_block->seq_num, next_hop_address, node_block->dist);
	return true;
}



/*
 * Packet queue related functions.
 */

/**
 * @brief Check whether the destination address of two queued packets match.
 *
 * @param item Pointer to queued packet to check.
 * @param cmp_item Pointer to reference queued packet.
 * @return <code>true</code> if the destination address two queued packets are
 * equal, <code>false</code> otherwise.
 */
static bool matches_packet_destination_address(const void* item, const void* cmp_item) {
	DYMO_queued_packet_t* packet = (DYMO_queued_packet_t*) item;
	DYMO_queued_packet_t* cmp_packet = (DYMO_queued_packet_t*) cmp_item;
	return (packet->packet.header->destination_address == cmp_packet->packet.header->destination_address);
}

/**
 * @brief Remove a queued packet from the packet queue and free its resources.
 *
 * @param packet Pointer to a queued packet.
 */
static void remove_and_free_packet_from_queue(DYMO_queued_packet_t* packet) {
	C_ABORT_ON(packet == NULL, "DYMO: Trying to free queued packet at NULL...\n");

	C_DEBUG(DYMO_DEBUG_LEVEL, C_MSG_VERBOSE, "DYMO: Removing packet #%d from %s to %s from queue...\n", packet->packet.header->seqno, addr_to_str(packet->packet.header->source_address), addr_to_str(packet->packet.header->destination_address));
	if(packet->callback_id != 0)
		callback_cancel(packet->callback_id);
	list_remove(C_DYMO_packet_queue, packet);
	free_header(&packet->packet.header);
	FREE(packet->packet.payload);
	FREE(packet);
}

/**
 * @brief Remove all packets to a specific destination from packet queue.
 *
 * @param address Pointer to a queued packet.
 */
static void remove_and_free_addressed_packets_from_queue(C_addr_t address) {
	DYMO_queued_packet_t* queued_packet = NULL;
	DYMO_queued_packet_t cmp_queued_packet;
	net_header_t cmp_header;

	C_DEBUG(DYMO_DEBUG_LEVEL, C_MSG_DEBUG, "DYMO: Removing all packets to %s from queue...\n", addr_to_str(address));
	memset(&cmp_queued_packet, 0, sizeof(DYMO_queued_packet_t));
	memset(&cmp_header, 0, sizeof(net_header_t));
	cmp_queued_packet.packet.header = &cmp_header;
	cmp_header.destination_address = address;
	while((queued_packet = (DYMO_queued_packet_t*) list_find(C_DYMO_packet_queue, &cmp_queued_packet, matches_packet_destination_address)) != NULL)
		remove_and_free_packet_from_queue(queued_packet);
}

/**
 * @brief Check whether packets need to be dropped from the packet queue.
 *
 * Drop the oldest packet of the lowest priority from the packet queue until
 * there is enough space available to accomodate a given number of pending
 * packets or bytes.
 *
 * NOTE: The current implementation only takes the number of packets into
 * account. A fixed size buffer with pending_bytes and DYMO_BUFFER_SIZE_BYTES
 * needs to be implemented for this to be useful on memory contrained devices.
 *
 * @param pending_packets Number of pending packets.
 * @param pending_bytes Number of pending bytes.
 */
static void purge_packet_queue(uint16_t pending_packets, uint16_t pending_bytes) {
	DYMO_queued_packet_t* packet_to_purge = NULL;
	uint8_t priority = PACKET_PRIORITY_MIN;
	DYMO_queued_packet_t* packet = NULL;
	list_iterator_t packet_iterator;

	C_DEBUG(DYMO_DEBUG_LEVEL, C_MSG_VERBOSE, "DYMO: Purging packet from queue to make room for %d new packet(s) with %d bytes.\n", pending_packets, pending_bytes);

	while((list_size(C_DYMO_packet_queue) + pending_packets) > DYMO_BUFFER_SIZE_PACKETS) {

		// Find lowest priority, oldest packet in queue.
		packet_to_purge = NULL;
		for(priority = PACKET_PRIORITY_MIN; priority <= PACKET_PRIORITY_MAX; priority++) {
			list_iterator_init(C_DYMO_packet_queue, &packet_iterator);
			while((packet = (DYMO_queued_packet_t*) list_iterator_next(&packet_iterator)) != NULL) {
				if(
					(packet->packet.header->priority == priority)
					&& (
						(packet_to_purge == NULL)
						|| (packet->time < packet_to_purge->time)
					)
				) {
					packet_to_purge = packet;
				}
			}
		}
		C_ABORT_ON(packet_to_purge == NULL, "DYMO: No packet found in non-empty packet queue (%d).\n", list_size(C_DYMO_packet_queue));

		// Remove from packet queue, cancel callback and free memory.
		if(packet_to_purge->packet.header->priority == PACKET_PRIORITY_MIN)
			C_DEBUG(DYMO_DEBUG_LEVEL, C_MSG_INFO, "DYMO: Purging packet #%d to %s with priority 0x%x from queue.\n", packet_to_purge->packet.header->seqno, addr_to_str(packet_to_purge->packet.header->destination_address), packet_to_purge->packet.header->priority);
		else
			C_DEBUG(DYMO_DEBUG_LEVEL, C_MSG_WARN, "DYMO: Purging high-priority packet #%d to %s with priority 0x%x from queue.\n", packet_to_purge->packet.header->seqno, addr_to_str(packet_to_purge->packet.header->destination_address), packet_to_purge->packet.header->priority);
		remove_and_free_packet_from_queue(packet_to_purge);
	}
}

/**
 * @brief Process queued packets.
 *
 * @param target_node_address Address of the target node.
 * @param next_hop_address Address of the next hop on the route.
 */
static void process_queued_packets(C_addr_t target_node_address, C_addr_t next_hop_address) {
	DYMO_queued_packet_t* packet = NULL;
	DYMO_queued_packet_t cmp_queued_packet;
	net_header_t cmp_header;

	memset(&cmp_queued_packet, 0, sizeof(DYMO_queued_packet_t));
	memset(&cmp_header, 0, sizeof(net_header_t));
	cmp_queued_packet.packet.header = &cmp_header;
	cmp_queued_packet.packet.header->destination_address = target_node_address;
	while((packet = (DYMO_queued_packet_t*) list_find(C_DYMO_packet_queue, &cmp_queued_packet, matches_packet_destination_address)) != NULL) {

		C_DEBUG(DYMO_DEBUG_LEVEL, C_MSG_INFO, "DYMO: Routing queued packet #%d from %s to %s via next hop %s...\n", packet->packet.header->seqno, addr_to_str(packet->packet.header->source_address), addr_to_str(packet->packet.header->destination_address), addr_to_str(next_hop_address));

		packet->packet.header->next_hop_address = next_hop_address;
		C_send(packet->packet.header, packet->packet.payload);
		remove_and_free_packet_from_queue(packet);
	}
}



/*
 * Packet handling functions.
 */

/**
 * @brief Send a DYMO RREQ packet.
 *
 * See Sec. 5.3.1.
 *
 * @param target_node_address Address of the target node.
 * @param originator_node_address Address of the originator node.
 * @param originator_node_seq_num Sequence number of the originator node.
 * @param originator_node_dist Distance to the originator node.
 * @param hop_limit Hop limit for the RREQ packet.
 */
static void send_DYMO_RREQ(C_addr_t target_node_address, C_addr_t originator_node_address, uint16_t originator_node_seq_num, uint16_t originator_node_dist, uint8_t hop_limit) {
	DYMO_RREQ_packet_t* rreq = NULL;
	net_header_t* header = NULL;
	DYMO_route_t* route = NULL;
	DYMO_route_t cmp_route;

	C_ABORT_ON(target_node_address == C_BROADCAST_ADDRESS, "DYMO: Trying to send a RREQ to BROADCAST.\n");
	C_ABORT_ON(originator_node_address == C_BROADCAST_ADDRESS, "DYMO: Trying to send a RREQ from BROADCAST.\n");
	C_ABORT_ON(originator_node_seq_num == DYMO_UNKNOWN_SEQUENCE_NUMBER, "DYMO: Trying to send a RREQ with invalid sequence number.\n");
	C_ABORT_ON(originator_node_dist == DYMO_UNKNOWN_DISTANCE, "DYMO: Trying to send a RREQ with invalid distance.\n");

	C_DEBUG(DYMO_DEBUG_LEVEL, C_MSG_DEBUG, "DYMO: Broadcasting RREQ #%d from %s looking for %s...\n", originator_node_seq_num, addr_to_str(originator_node_address), addr_to_str(target_node_address));

	memset(&cmp_route, 0, sizeof(DYMO_route_t));
	cmp_route.address = target_node_address;
	MALLOC(rreq, DYMO_RREQ_packet_t);
	rreq->target_node.address = target_node_address;
	if((route = (DYMO_route_t*) list_find(C_DYMO_routing_table, &cmp_route, matches_route_address)) == NULL) {
		rreq->target_node.seq_num = DYMO_UNKNOWN_SEQUENCE_NUMBER;
		rreq->target_node.dist = DYMO_UNKNOWN_DISTANCE;
	} else {	// After the check above, we know that route->forwarding == false
		rreq->target_node.seq_num = route->seq_num;
		rreq->target_node.dist = route->dist;			// NOTE: The draft doesn't say anything about this, but it seems as if it was intended this way.
	}
	rreq->originator_node.address = originator_node_address;	// NOT IMPLEMENTED: Prefix handling.
	rreq->originator_node.seq_num = originator_node_seq_num;
	rreq->originator_node.dist = originator_node_dist;
	rreq->hop_limit = hop_limit;

	// NOTE: According to Sec. 5.3.2 of the draft, the destination address
	// is set to NextHopAddress for unicast RREQs. But why would one send a
	// RREQ if the next hop to the target node is known?
	alloc_header(&header, DYMO_LL_MANET_ROUTERS, PACKET_PRIORITY_SIGNALLING_ROUTING, APP_TYPE_DYMO, DYMO_PACKET_TYPE_RREQ, sizeof(DYMO_RREQ_packet_t));
	header->ttl = 255;	// Ensures packets have not passed through any intermediate routers.
	C_send(header, rreq);
	free_header(&header);

	C_DYMO_transmitted_RREQs++;
	FREE(rreq);
}

/**
 * @brief Send a DYMO RREP packet.
 *
 * See Sec. 5.3.2.
 *
 * @param target_node_address Address of the target node.
 * @param originator_node_address Address of the originator node.
 * @param originator_node_seq_num Sequence number of the originator node.
 * @param originator_node_dist Distance to the originator node.
 * @param hop_limit Hop limit for the RREP packet.
 * @param next_hop_address Address of the next hop node to which this RREP is to be sent.
 */
static void send_DYMO_RREP(C_addr_t target_node_address, C_addr_t originator_node_address, uint16_t originator_node_seq_num, uint16_t originator_node_dist, uint8_t hop_limit, C_addr_t next_hop_address) {
	DYMO_RREP_packet_t* rrep = NULL;
#ifdef DYMO_USE_ADDITIONAL_ROUTES_IN_RREPS
	DYMO_node_block_t* additional_node_array = NULL;
	size_t additional_node_array_size = (list_size(C_DYMO_routing_table) - 1) * sizeof(DYMO_node_block_t);	// At least the route to the target node wont get transmitted, hence - 1.
	DYMO_route_t* route = NULL;
	list_iterator_t routes_iterator;
	unsigned int current_node_index = 0;
#endif	// DYMO_USE_ADDITIONAL_ROUTES_IN_RREPS
	net_header_t* header = NULL;

	C_ABORT_ON(target_node_address == C_BROADCAST_ADDRESS, "DYMO: Trying to send a RREP to BROADCAST.\n");
	C_ABORT_ON(originator_node_seq_num == DYMO_UNKNOWN_SEQUENCE_NUMBER, "DYMO: Trying to send a RREP with invalid sequence number.\n");
	C_ABORT_ON(originator_node_dist == DYMO_UNKNOWN_DISTANCE, "DYMO: Trying to send a RREP with invalid distance.\n");
	C_ABORT_ON(next_hop_address == C_BROADCAST_ADDRESS, "DYMO: Trying to send a RREP to next hop BROADCAST.\n");

	C_DEBUG(DYMO_DEBUG_LEVEL, C_MSG_DEBUG, "DYMO: Sending RREP #%d from %s to %s via %s...\n", originator_node_seq_num, addr_to_str(originator_node_address), addr_to_str(target_node_address), addr_to_str(next_hop_address));

	MALLOC(rrep, DYMO_RREP_packet_t);
	rrep->target_node.address = target_node_address;
	rrep->originator_node.address = originator_node_address;	// NOT IMPLEMENTED: Prefix handling.
	rrep->originator_node.seq_num = originator_node_seq_num;
	rrep->originator_node.dist = originator_node_dist;
	rrep->hop_limit = hop_limit;

	// Set fields whose value is not specified to sane defaults.
	rrep->target_node.seq_num = DYMO_UNKNOWN_SEQUENCE_NUMBER;
	rrep->target_node.dist = DYMO_UNKNOWN_DISTANCE;

#ifdef DYMO_USE_ADDITIONAL_ROUTES_IN_RREPS
	/*
	 * Add additional information to routing message (see Sec. 5.3.5)
	 */
	get_own_seq_num();	// The draft specifies that we SHOULD increment our own sequence number.
	MALLOC_ARRAY(additional_node_array, DYMO_node_block_t, (list_size(C_DYMO_routing_table) - 1));
	list_iterator_init(C_DYMO_routing_table, &routes_iterator);
	while((route = (DYMO_route_t*) list_iterator_next(&routes_iterator)) != NULL) {
		C_ABORT_ON(current_node_index > (list_size(C_DYMO_routing_table) - 1), "DYMO: Trying to add more additional routes than available (%d > %d).\n", current_node_index, (list_size(C_DYMO_routing_table) - 1));
		C_ABORT_ON(route->address == C_LOCAL_ADDRESS, "DYMO: Found own address in routing table while adding additional routes to RREP.\n");
		if(
			(route->seq_num != DYMO_UNKNOWN_SEQUENCE_NUMBER)
			&& (route->dist != DYMO_UNKNOWN_DISTANCE)
#ifdef DYMO_USE_OVERHEARD_BROADCAST_TO_ADD_ROUTES
			// Make sure we don't propagate routing information
			// based on overheard packets.
			&& (route->dist < DYMO_MAX_DISTANCE)
#endif	// DYMO_USE_OVERHEARD_BROADCAST_TO_ADD_ROUTES
			&& (route->address != target_node_address)
		) {
			additional_node_array[current_node_index].address = route->address;
			additional_node_array[current_node_index].seq_num = route->seq_num;
			additional_node_array[current_node_index].dist = route->dist;
			// NOTE: According to the draft, we should MaxAge
			// information, but it doesn't say where to get it.
			current_node_index++;
		}
	}
	// Adjust size of node array to only transmit the valid part of the array.
	additional_node_array_size = current_node_index * sizeof(DYMO_node_block_t);
	C_DEBUG(DYMO_DEBUG_LEVEL, C_MSG_DEBUG, "DYMO: Added %d routes to RREP...\n", current_node_index);
#endif	// DYMO_USE_ADDITIONAL_ROUTES_IN_RREPS

	alloc_header(&header, next_hop_address, PACKET_PRIORITY_SIGNALLING_ROUTING, APP_TYPE_DYMO, DYMO_PACKET_TYPE_RREP, sizeof(DYMO_RREP_packet_t));
	header->next_hop_address = next_hop_address;
	header->ttl = 255;	// Ensures packets have not passed through any intermediate routers.
#ifndef DYMO_USE_ADDITIONAL_ROUTES_IN_RREPS
	C_send(header, rrep);
#else
	if(current_node_index == 0)
		C_send(header, rrep);
	else
		C_send2(header, rrep, sizeof(DYMO_RREP_packet_t), additional_node_array, additional_node_array_size);
#endif	// DYMO_USE_ADDITIONAL_ROUTES_IN_RREPS
	free_header(&header);

#ifdef DYMO_USE_ADDITIONAL_ROUTES_IN_RREPS
	FREE(additional_node_array);
#endif	// DYMO_USE_ADDITIONAL_ROUTES_IN_RREPS

	C_DYMO_transmitted_RREPs++;
	FREE(rrep);
}

/**
 * @brief Low-level send a DYMO RERR packet with a given array of node blocks.
 *
 * @param destination_address Address of the destination node.
 * @param node_array Pointer to the array of unreachable nodes.
 * @param node_array_size Size of the array of unreachable nodes in bytes.
 * @param hop_limit Hop limit of this packet.
 */
static void send_DYMO_RERR_array(C_addr_t destination_address, const DYMO_node_block_t* node_array, size_t node_array_size, uint8_t hop_limit) {
	DYMO_RERR_packet_t* rerr = NULL;
	net_header_t* header = NULL;
	DYMO_route_t* route = NULL;
	DYMO_route_t cmp_route;
	C_addr_t next_hop_address;

	memset(&cmp_route, 0, sizeof(DYMO_route_t));
	cmp_route.address = destination_address;
	if(destination_address == DYMO_LL_MANET_ROUTERS) {
		next_hop_address = DYMO_LL_MANET_ROUTERS;
	} else if((route = (DYMO_route_t*) list_find(C_DYMO_routing_table, &cmp_route, matches_route_address)) != NULL) {
		next_hop_address = route->next_hop_address;
	} else {
		// NOTE: This situation might warrant a WARNing.
		C_DEBUG(DYMO_DEBUG_LEVEL, C_MSG_INFO, "DYMO: No route to unicast destination %s of RERR, sending multicast RERR instead.\n", addr_to_str(destination_address));
		destination_address = next_hop_address = DYMO_LL_MANET_ROUTERS;
	}

	C_DEBUG(DYMO_DEBUG_LEVEL, C_MSG_DEBUG, "DYMO: Sending RERR packet to %s via %s...\n", addr_to_str(destination_address), addr_to_str(next_hop_address));

	MALLOC(rerr, DYMO_RERR_packet_t);
	rerr->hop_limit = hop_limit;

	// NOTE: Contrary to the draft (ll. 1489-1493) we don't set
	// IP.DestinationAddress to NextHopAddress but rather to
	// DYMO_LL_MANET_ROUTERS. In DYMO, a RERR doesn't carry any destination
	// information but is propagated via broadcast as long as it removes an
	// entry from the local routing table.
	alloc_header(&header, DYMO_LL_MANET_ROUTERS, PACKET_PRIORITY_SIGNALLING_ROUTING, APP_TYPE_DYMO, DYMO_PACKET_TYPE_RERR, sizeof(DYMO_RERR_packet_t));
	header->ttl = 255;	// Ensures packets have not passed through any intermediate routers.
	C_send2(header, rerr, sizeof(DYMO_RERR_packet_t), node_array, node_array_size);
	free_header(&header);

	C_DYMO_transmitted_RERRs++;
	FREE(rerr);
}

/**
 * @brief Send a DYMO RERR packet.
 *
 * See Sec. 5.5.3.
 *
 * @param destination_address Address of the destination node.
 * @param unreachable_node_address Address of the unreachable node.
 * @param unreachable_node_seq_num Last known sequence number of the unreachable node.
 * @param unreachable_node_next_hop_address  Address of the next hop to the unreachable node.
 * @param hop_limit Hop limit of this packet.
 */
static void send_DYMO_RERR(C_addr_t destination_address, C_addr_t unreachable_node_address, uint16_t unreachable_node_seq_num, C_addr_t unreachable_node_next_hop_address, uint8_t hop_limit) {
	DYMO_node_block_t* node_array = NULL;
	size_t node_array_size = 0;
#ifdef DYMO_USE_ADDITIONAL_ROUTES_IN_RERRS
	unsigned int current_node_index = 0;
	DYMO_route_t* route = NULL;
	list_iterator_t routes_iterator;

	MALLOC_ARRAY(node_array, DYMO_node_block_t, (list_size(C_DYMO_routing_table) + 1));
	node_array_size = (list_size(C_DYMO_routing_table) + 1) * sizeof(DYMO_node_block_t);

	node_array[0].address = unreachable_node_address;
	node_array[0].seq_num = unreachable_node_seq_num;
	node_array[0].dist = DYMO_UNKNOWN_DISTANCE;	// Not used for RERRs.
	current_node_index++;

	// Add additional information about other affected routes.
	if(unreachable_node_next_hop_address != C_BROADCAST_ADDRESS) {
		list_iterator_init(C_DYMO_routing_table, &routes_iterator);
		while((route = (DYMO_route_t*) list_iterator_next(&routes_iterator)) != NULL) {
			if(
				(route->address != unreachable_node_address)
				&& (route->next_hop_address == unreachable_node_next_hop_address)
			) {
				C_DEBUG(DYMO_DEBUG_LEVEL, C_MSG_VERBOSE, "DYMO: Next hop on route to %s is also %s.\n", addr_to_str(route->address), addr_to_str(route->next_hop_address));
				node_array[current_node_index].address = route->address;
				node_array[current_node_index].seq_num = route->seq_num;
				node_array[current_node_index].dist = DYMO_UNKNOWN_DISTANCE;	// Not used for RERRs.
				current_node_index++;
			}
		}
		C_DEBUG(DYMO_DEBUG_LEVEL, C_MSG_DEBUG, "DYMO: Added %d routes to RERR...\n", current_node_index);
	}
	C_ABORT_ON(current_node_index == 0, "DYMO: Trying to send RERR, but current_node_index is zero.\n", current_node_index, list_size(C_DYMO_routing_table));
	C_ABORT_ON(current_node_index > (list_size(C_DYMO_routing_table) + 1), "DYMO: Trying to send RERR, but current_node_index is larger than the size of the routing table (%d > %d + 1).\n", current_node_index, list_size(C_DYMO_routing_table));

	// Adjust size of node array to only transmit the valid part of the array.
	node_array_size = current_node_index * sizeof(DYMO_node_block_t);
#else
	MALLOC_ARRAY(node_array, DYMO_node_block_t, 1);
	node_array_size = sizeof(DYMO_node_block_t);
	node_array[0].address = unreachable_node_address;
	node_array[0].seq_num = unreachable_node_seq_num;
	node_array[0].dist = DYMO_UNKNOWN_DISTANCE;	// Not used for RERRs.
#endif	// DYMO_USE_ADDITIONAL_ROUTES_IN_RERRS

	send_DYMO_RERR_array(destination_address, node_array, node_array_size, hop_limit);

	FREE(node_array);
}

/**
 * @brief Calculate new distance metric to be used in routing messages.
 *
 * @param current_dist The current distance.
 * @param address The address of the other node.
 */
static uint16_t calculate_distance(uint16_t current_dist, C_addr_t address) {
#ifdef C_USE_ETX_LINK_METRIC
	uint16_t increment = 0;
	uint16_t new_dist = 0;
	float etx = calculate_ETX(address);

	if(etx == FLT_MAX)
		return (current_dist + DYMO_MSG_HOPLIMIT);

	increment = ((int) (10.0 * etx)) + 1;
	if(increment > DYMO_MSG_HOPLIMIT)
		increment = DYMO_MSG_HOPLIMIT;
	new_dist = current_dist + increment;

	C_ABORT_ON(new_dist <= current_dist, "DYMO: Incremented distance metric is smaller than original metric, possible wrap-around (%d <= %d).\n", new_dist, current_dist);
	C_ABORT_ON(new_dist == DYMO_UNKNOWN_DISTANCE, "DYMO: Incremented distance metric equals unknown distance.\n");
	C_ABORT_ON(new_dist == DYMO_MAX_DISTANCE, "DYMO: Incremented distance metric equals maximum distance.\n");

	return new_dist;
#else
	return (current_dist + 1);
#endif	// C_USE_ETX_LINK_METRIC
}

/**
 * @brief Handle incoming DYMO RREQ packet.
 *
 * @param header Pointer to header of incoming DYMO RREQ packet.
 * @param rreq Pointer to route request.
 */
static void handle_DYMO_RREQ(const net_header_t* header, const DYMO_RREQ_packet_t* rreq) {
#ifdef DYMO_USE_INTERMEDIATE_RREPS
	DYMO_route_t* route_to_target_of_RREQ = NULL;
#endif	// DYMO_USE_INTERMEDIATE_RREPS
	DYMO_route_t* route_to_originator_of_RREQ = NULL;
	DYMO_route_t cmp_route;
	uint16_t new_originator_node_dist = calculate_distance(rreq->originator_node.dist, header->source_address);
	uint8_t new_hop_limit = rreq->hop_limit - 1;

	C_ABORT_ON(header->ttl != 255, "DYMO: Received RREQ #%d from %s looking for %s via %s with TTL %d instead of 255.\n", rreq->originator_node.seq_num, addr_to_str(rreq->originator_node.address), addr_to_str(rreq->target_node.address), addr_to_str(header->source_address), header->ttl);
	C_ABORT_ON(rreq->originator_node.address == C_BROADCAST_ADDRESS, "DYMO: Received RREQ #%d from %s looking for %s via %s from broadcast address.\n", rreq->originator_node.seq_num, addr_to_str(rreq->originator_node.address), addr_to_str(rreq->target_node.address), addr_to_str(header->source_address));
	C_ABORT_ON(rreq->target_node.address == C_BROADCAST_ADDRESS, "DYMO: Received RREQ #%d from %s looking for %s via %s looking for broadcast address.\n", rreq->originator_node.seq_num, addr_to_str(rreq->originator_node.address), addr_to_str(rreq->target_node.address), addr_to_str(header->source_address));

	if(rreq->originator_node.address == C_LOCAL_ADDRESS) {
		C_DEBUG(DYMO_DEBUG_LEVEL, C_MSG_DEBUG, "DYMO: Dropping RREQ #%d from %s looking for %s via %s from local node.\n", rreq->originator_node.seq_num, addr_to_str(rreq->originator_node.address), addr_to_str(rreq->target_node.address), addr_to_str(header->source_address));
		return;
	} else if(!update_routing_information(&rreq->originator_node, header->source_address, DYMO_PACKET_TYPE_RREQ)) {
		C_DEBUG(DYMO_DEBUG_LEVEL, C_MSG_DEBUG, "DYMO: Received RREQ #%d from %s looking for %s via %s, but the routing information is not superior, dropping RREQ...\n", rreq->originator_node.seq_num, addr_to_str(rreq->originator_node.address), addr_to_str(rreq->target_node.address), addr_to_str(header->source_address));
		return;
	}

	// Reset the routing timers for the originator of the RREQ to avoid
	// deleting the route before it is ever used by data packets.
	// NOTE: This is not explicitly stated in the draft.
	reset_routing_timers(rreq->originator_node.address);

	memset(&cmp_route, 0, sizeof(DYMO_route_t));
	cmp_route.address = rreq->target_node.address;
	if(rreq->target_node.address == C_LOCAL_ADDRESS) {

		// Local node is target of RREQ, send RREP.

		cmp_route.address = rreq->originator_node.address;
		route_to_originator_of_RREQ = (DYMO_route_t*) list_find(C_DYMO_routing_table, &cmp_route, matches_route_address);
		C_ABORT_ON(route_to_originator_of_RREQ == NULL, "DYMO: Trying to send a RREP responding to RREQ #%d from %s looking for local node received via %s, but no route to originator is known.\n", rreq->originator_node.seq_num, addr_to_str(rreq->originator_node.address), addr_to_str(header->source_address));
		C_ABORT_ON(header->source_address != route_to_originator_of_RREQ->next_hop_address, "DYMO: Trying to send a RREP responding to RREQ #%d from %s looking for local node received via %s, but last hop the RREQ does not match the first hop on the route to the originator (%s != %s).\n", rreq->originator_node.seq_num, addr_to_str(rreq->originator_node.address), addr_to_str(header->source_address), addr_to_str(header->source_address), addr_to_str(route_to_originator_of_RREQ->next_hop_address));

		C_DEBUG(DYMO_DEBUG_LEVEL, C_MSG_NOTE, "DYMO: Received RREQ #%d from %s looking for local node via %s, sending RREP...\n", rreq->originator_node.seq_num, addr_to_str(rreq->originator_node.address), addr_to_str(header->source_address));

		// The draft talks quite a bit about sequence number handling at this point (ll. 1071-1091), but there actually seems to be no reason to simply ALWAYS increment the sequence number.
		// Set OrigNode.Dist to a number greater than zero (0) and less than 65,535.
		send_DYMO_RREP(rreq->originator_node.address, C_LOCAL_ADDRESS, get_own_seq_num(), 1, DYMO_MSG_HOPLIMIT, route_to_originator_of_RREQ->next_hop_address);

#ifdef DYMO_USE_INTERMEDIATE_RREPS
	} else if(
		((route_to_target_of_RREQ = (DYMO_route_t*) list_find(C_DYMO_routing_table, &cmp_route, matches_route_address)) != NULL)
		&& route_to_target_of_RREQ->forwarding
		&& (rreq->target_node.seq_num != DYMO_UNKNOWN_SEQUENCE_NUMBER)
		&& (route_to_target_of_RREQ->seq_num >= rreq->target_node.seq_num)	// NOTE: This is equivalent to the draft which specifies "route->seq_num - rreq->target_node.seq_num >= 0" using signed 16-bit arithmetic.
#ifdef DYMO_USE_OVERHEARD_BROADCAST_TO_ADD_ROUTES
		// Make sure we don't propagate routing information based on
		// overheard packets.
		&& (route_to_target_of_RREQ->dist < DYMO_MAX_DISTANCE)
#endif	// DYMO_USE_OVERHEARD_BROADCAST_TO_ADD_ROUTES
	) {

		// Local node has route to target of RREQ, send itermediate RREP (Sec. 5.3.3).

		cmp_route.address = rreq->originator_node.address;
		route_to_originator_of_RREQ = (DYMO_route_t*) list_find(C_DYMO_routing_table, &cmp_route, matches_route_address);
		C_ABORT_ON(route_to_originator_of_RREQ == NULL, "DYMO: Trying to send a RREP, but route to source of RREQ %s is unknown.\n", addr_to_str(rreq->originator_node.address));

		C_DEBUG(DYMO_DEBUG_LEVEL, C_MSG_NOTE, "DYMO: Received RREQ #%d from %s looking for known target %s via %s, sending intermediate RREPs to both nodes...\n", rreq->originator_node.seq_num, addr_to_str(rreq->originator_node.address), addr_to_str(rreq->target_node.address), addr_to_str(header->source_address));

		// Reset the routing timers for the target node of the
		// intermediate RREP to avoid deleting the route before it is
		// ever used by data packets. The originator is handled above.
		// NOTE: This is not explicitly stated in the draft.
		reset_routing_timers(rreq->target_node.address);

		get_own_seq_num();	// The draft specifies that we SHOULD increment our own sequence number.

		// Send RREP to originator of RREQ.
		// NOTE: Although not specified in the draft, we increment the
		// distance found in routing before sending. Otherwise, the
		// route would be shorter than a route based on a RREP sent
		// directly by the destination node.
		send_DYMO_RREP(rreq->originator_node.address, rreq->target_node.address, route_to_target_of_RREQ->seq_num, calculate_distance(route_to_target_of_RREQ->dist, route_to_target_of_RREQ->next_hop_address), DYMO_MSG_HOPLIMIT, route_to_originator_of_RREQ->next_hop_address);

		// Send RREP to target of RREQ.
		// NOTE: Although not specified in the draft, we increment the
		// distance found in routing before sending. Otherwise, the
		// route would be shorter than a route based on a RREP sent
		// directly by the destination node.
		send_DYMO_RREP(rreq->target_node.address, rreq->originator_node.address, route_to_originator_of_RREQ->seq_num, calculate_distance(route_to_originator_of_RREQ->dist, header->source_address), DYMO_MSG_HOPLIMIT, route_to_target_of_RREQ->next_hop_address);
#endif	// DYMO_USE_INTERMEDIATE_RREPS

	} else if(new_originator_node_dist == 0) {

		// NOTE: The draft specifies this condition as "greater than 65,535". However, since we are working with 16-bit values, the variable would wrap-around to zero.
		C_DEBUG(DYMO_DEBUG_LEVEL, C_MSG_ERROR, "DYMO: Received RREQ #%d from %s looking for unknown target %s via %s, dropping due to exceeded distance field...\n", rreq->originator_node.seq_num, addr_to_str(rreq->originator_node.address), addr_to_str(rreq->target_node.address), addr_to_str(header->source_address));

	} else if(new_hop_limit == 0) {

		C_DEBUG(DYMO_DEBUG_LEVEL, C_MSG_DEBUG, "DYMO: Received RREQ #%d from %s looking for unknown target %s via %s, dropping due to hop limit reaching zero...\n", rreq->originator_node.seq_num, addr_to_str(rreq->originator_node.address), addr_to_str(rreq->target_node.address), addr_to_str(header->source_address));

	} else {

		// Local node has no route to target of RREQ, forward RREQ.

		C_DEBUG(DYMO_DEBUG_LEVEL, C_MSG_INFO, "DYMO: Received RREQ #%d from %s looking for unknown target %s via %s, propagating RREQ...\n", rreq->originator_node.seq_num, addr_to_str(rreq->originator_node.address), addr_to_str(rreq->target_node.address), addr_to_str(header->source_address));
		send_DYMO_RREQ(rreq->target_node.address, rreq->originator_node.address, rreq->originator_node.seq_num, new_originator_node_dist, new_hop_limit);

	}

	process_queued_packets(rreq->originator_node.address, header->source_address);
}

/**
 * @brief Handle incoming DYMO RREP packet.
 *
 * @param header Pointer to header of incoming DYMO RREP packet.
 * @param rrep Pointer to route reply.
 * @param additional_node_array Pointer to array of additional nodes.
 * @param additional_node_array_size Size of array of additional nodes in bytes.
 */
static void handle_DYMO_RREP(const net_header_t* header, const DYMO_RREP_packet_t* rrep, const DYMO_node_block_t* additional_node_array, size_t additional_node_array_size) {
#ifdef DYMO_USE_ADDITIONAL_ROUTES_IN_RREPS
	unsigned int current_node_index = 0;
	DYMO_node_block_t node_block;
#endif	// DYMO_USE_ADDITIONAL_ROUTES_IN_RREPS
	DYMO_route_t* route = NULL;
	DYMO_route_t cmp_route;
	uint16_t new_originator_node_dist = calculate_distance(rrep->originator_node.dist, header->source_address);
	uint8_t new_hop_limit = rrep->hop_limit - 1;

	C_ABORT_ON(header->source_address != header->last_hop_address, "DYMO: Received RREP #%d from %s for %s with a route to %s, but source address does not match the last hop address (%s).\n", rrep->originator_node.seq_num, addr_to_str(header->source_address), addr_to_str(rrep->target_node.address), addr_to_str(rrep->originator_node.address), addr_to_str(header->last_hop_address));
	C_ABORT_ON(header->ttl != 255, "DYMO: Received RREP #%d from %s for %s with a route to %s with TTL %d instead of 255.\n", rrep->originator_node.seq_num, addr_to_str(header->source_address), addr_to_str(rrep->target_node.address), addr_to_str(rrep->originator_node.address), header->ttl);
	C_ABORT_ON(rrep->originator_node.address == C_BROADCAST_ADDRESS, "DYMO: Received RREP #%d from %s for %s with a route to %s from broadcast address.\n", rrep->originator_node.seq_num, addr_to_str(header->source_address), addr_to_str(rrep->target_node.address), addr_to_str(rrep->originator_node.address));
	C_ABORT_ON(rrep->target_node.address == C_BROADCAST_ADDRESS, "DYMO: Received RREP #%d from %s for %s with a route to %s looking for broadcast address.\n", rrep->originator_node.seq_num, addr_to_str(header->source_address), addr_to_str(rrep->target_node.address), addr_to_str(rrep->originator_node.address));

	if(rrep->originator_node.address == C_LOCAL_ADDRESS) {
		C_DEBUG(DYMO_DEBUG_LEVEL, C_MSG_DEBUG, "DYMO: Dropping RREP #%d from %s for %s with a route to %s from local node.\n", rrep->originator_node.seq_num, addr_to_str(header->source_address), addr_to_str(rrep->target_node.address), addr_to_str(rrep->originator_node.address));
		return;
	} else if(!update_routing_information(&rrep->originator_node, header->source_address, DYMO_PACKET_TYPE_RREP)) {
		C_DEBUG(DYMO_DEBUG_LEVEL, C_MSG_DEBUG, "DYMO: Received RREP #%d from %s for %s with a route to %s, but the routing information is not superior, dropping RREP...\n", rrep->originator_node.seq_num, addr_to_str(header->source_address), addr_to_str(rrep->target_node.address), addr_to_str(rrep->originator_node.address));
		return;
	}

#ifdef DYMO_USE_ADDITIONAL_ROUTES_IN_RREPS
	// Update routing information for additional addresses contained in
	// this routing message and check packet queue.
	// NOTE: Rather than checking the validity of all additional node
	// blocks, we generate a new array when propagating the RREP.
	for(current_node_index = 0; current_node_index < (additional_node_array_size / sizeof(DYMO_node_block_t)); current_node_index++) {
		memcpy(&node_block, &additional_node_array[current_node_index], sizeof(DYMO_node_block_t));
		node_block.dist = calculate_distance(node_block.dist, header->source_address);
		if(node_block.dist == DYMO_UNKNOWN_DISTANCE)
			continue;
		if(update_routing_information(&node_block, header->source_address, DYMO_PACKET_TYPE_RREP))
			process_queued_packets(node_block.address, header->source_address);
	}
#endif	// DYMO_USE_ADDITIONAL_ROUTES_IN_RREPS

	// Reset the routing timers for the originator of the RREP to avoid
	// deleting the route before it is ever used by data packets.
	// NOTE: This is not explicitly stated in the draft.
	reset_routing_timers(rrep->originator_node.address);

	memset(&cmp_route, 0, sizeof(DYMO_route_t));
	cmp_route.address = rrep->target_node.address;
	if(rrep->target_node.address == C_LOCAL_ADDRESS) {

		// Route established.

		C_DEBUG(DYMO_DEBUG_LEVEL, C_MSG_NOTE, "DYMO: Received RREP #%d from %s for %s with a route to %s, route established.\n", rrep->originator_node.seq_num, addr_to_str(header->source_address), addr_to_str(rrep->target_node.address), addr_to_str(rrep->originator_node.address));

	} else if(new_originator_node_dist == 0) {

		// NOTE: The draft specifies this condition as "greater than 65,535". However, since we are working with 16-bit values, the variable would wrap-around to zero.
		C_DEBUG(DYMO_DEBUG_LEVEL, C_MSG_ERROR, "DYMO: Received RREP #%d from %s for %s with a route to %s, not forwarding due to exceeded distance field...\n", rrep->originator_node.seq_num, addr_to_str(header->source_address), addr_to_str(rrep->target_node.address), addr_to_str(rrep->originator_node.address));

	} else if(new_hop_limit == 0) {

		C_DEBUG(DYMO_DEBUG_LEVEL, C_MSG_ERROR, "DYMO: Received RREP #%d from %s for %s with a route to %s, not forwarding due to hop limit reaching zero...\n", rrep->originator_node.seq_num, addr_to_str(header->source_address), addr_to_str(rrep->target_node.address), addr_to_str(rrep->originator_node.address));

	} else if((route = (DYMO_route_t*) list_find(C_DYMO_routing_table, &cmp_route, matches_route_address)) != NULL) {

		if(route->forwarding) {

#ifdef C_USE_NETWORK_MAPPING
			unsigned int i = 0;
			for(i = 0; i < header->hop_count; i++) {
				if(header->hops[i] == route->next_hop_address) {
					C_DEBUG(DYMO_DEBUG_LEVEL, C_MSG_WARN, "DYMO: Received RREP packet #%d from %s to %s which has previously been routed by the next hop %s, dropping packet...\n", header->seqno, addr_to_str(header->source_address), addr_to_str(header->destination_address), addr_to_str(route->next_hop_address));
					return;
				}
			}
#endif	// C_USE_NETWORK_MAPPING

			// Reset the routing timers for the target node of the
			// RREP to avoid deleting the route before it is ever
			// used by data packets. The originator is handled above.
			// NOTE: This is not explicitly stated in the draft.
			reset_routing_timers(rrep->target_node.address);

			// Forward RREP to next hop on established route.
			C_DEBUG(DYMO_DEBUG_LEVEL, C_MSG_INFO, "DYMO: Received RREP #%d from %s for %s with a route to %s, forwarding RREP via %s....\n", rrep->originator_node.seq_num, addr_to_str(header->source_address), addr_to_str(rrep->target_node.address), addr_to_str(rrep->originator_node.address), addr_to_str(route->next_hop_address));
			send_DYMO_RREP(rrep->target_node.address, rrep->originator_node.address, rrep->originator_node.seq_num, new_originator_node_dist, new_hop_limit, route->next_hop_address);

		} else {

			// Next hop for RREP unknown, send RERR.
	
			// This situation may occur if a node has sent an
			// intermediate RREP without any prior RREQ passing
			// through this node.
	
			// NOTE: This situation might warrant a WARNing.
			C_DEBUG(DYMO_DEBUG_LEVEL, C_MSG_INFO, "DYMO: Received RREP #%d from %s for %s with a route to %s, but route to target node is broken. Sending RERR and not forwarding packet...\n", rrep->originator_node.seq_num, addr_to_str(header->source_address), addr_to_str(rrep->target_node.address), addr_to_str(rrep->originator_node.address));
			send_DYMO_RERR(rrep->originator_node.address, rrep->target_node.address, route->seq_num, route->next_hop_address, DYMO_MSG_HOPLIMIT);

		}

	} else {

		// Next hop for RREP unknown, send RERR.

		// This situation may occur if a node has sent an intermediate
		// RREP without any prior RREQ passing through this node.

		// NOTE: This situation might warrant a WARNing.
		C_DEBUG(DYMO_DEBUG_LEVEL, C_MSG_INFO, "DYMO: Received RREP #%d from %s for %s with a route to %s, but route to target node is unknown. Sending RERR and not forwarding packet...\n", rrep->originator_node.seq_num, addr_to_str(header->source_address), addr_to_str(rrep->target_node.address), addr_to_str(rrep->originator_node.address));
		send_DYMO_RERR(rrep->originator_node.address, rrep->target_node.address, DYMO_UNKNOWN_SEQUENCE_NUMBER, C_BROADCAST_ADDRESS, DYMO_MSG_HOPLIMIT);

	}

	process_queued_packets(rrep->originator_node.address, header->source_address);
}

/**
 * @brief Handle node blocks in an incoming DYMO RERR packet.
 *
 * @param header Pointer to header of incoming DYMO RERR packet.
 * @param rerr Pointer to route error.
 * @param unreachable_node Pointer to node block of an unreachable node.
 * @return <code>true</code> if this node block should be forwarded,
 * <code>false</code> otherwise.
 */
static bool process_RERR_node_block(const net_header_t* header, const DYMO_RERR_packet_t* rerr, DYMO_node_block_t* unreachable_node) {
	DYMO_route_t* route = NULL;
	DYMO_route_t cmp_route;
	uint16_t new_unreachable_node_seq_num = 0;

	memset(&cmp_route, 0, sizeof(DYMO_route_t));
	cmp_route.address = unreachable_node->address;
	if((route = (DYMO_route_t*) list_find(C_DYMO_routing_table, &cmp_route, matches_route_address)) == NULL) {
		C_DEBUG(DYMO_DEBUG_LEVEL, C_MSG_DEBUG, "DYMO: Processing RERR block from %s for %s informing about a broken route to %s but no route to %s is known, invalidating block...\n", addr_to_str(header->source_address), addr_to_str(header->destination_address), addr_to_str(unreachable_node->address), addr_to_str(unreachable_node->address));
		unreachable_node->address = C_BROADCAST_ADDRESS;
		return false;
	}

	if(
		(unreachable_node->address != C_BROADCAST_ADDRESS)
		&& (route->next_hop_address == header->source_address)
		// NOT IMPLEMENTED: Consider next hop interface.
		&& (
			(route->seq_num == DYMO_UNKNOWN_SEQUENCE_NUMBER)
			|| (unreachable_node->seq_num == DYMO_UNKNOWN_SEQUENCE_NUMBER)
			|| (route->seq_num <= unreachable_node->seq_num)	// NOTE: This is equivalent to the draft which specifies "route->seq_num - unreachable_node->seq_num <= 0" using signed 16-bit arithmetic.
		)
	) {

		if(route->broken) {
			C_ABORT_ON(route->forwarding, "DYMO: Processing RERR block from %s for %s informing about a broken route to %s, but route is already tagged as broken while forwarding is still enabled.\n", addr_to_str(header->source_address), addr_to_str(header->destination_address), addr_to_str(unreachable_node->address));
			C_ABORT_ON(route->delete_information_timeout_callback_id == 0, "DYMO: Processing RERR block from %s for %s informing about a broken route to %s, but route is already tagged as broken while no deletion timer is running.\n", addr_to_str(header->source_address), addr_to_str(header->destination_address), addr_to_str(unreachable_node->address));

			C_DEBUG(DYMO_DEBUG_LEVEL, C_MSG_DEBUG, "DYMO: Processing RERR block from %s for %s informing about a broken route to %s, but route is already tagged as broken. Invalidating block...\n", addr_to_str(header->source_address), addr_to_str(header->destination_address), addr_to_str(unreachable_node->address));
			unreachable_node->address = C_BROADCAST_ADDRESS;
			return false;
		}

		C_DEBUG(DYMO_DEBUG_LEVEL, C_MSG_DEBUG, "DYMO: Processing RERR block from %s for %s informing about a broken route to %s...\n", addr_to_str(header->source_address), addr_to_str(header->destination_address), addr_to_str(unreachable_node->address));

		route->forwarding = false;
		route->broken = true;

		if(route->delete_information_timeout_callback_id == 0) {
			C_DEBUG(DYMO_DEBUG_LEVEL, C_MSG_DEBUG, "DYMO: Starting timer for removal of route to %s...\n", addr_to_str(route->address));
			route->delete_information_timeout_callback_id = callback_add(route_delete_information_timeout, route, DYMO_ROUTE_DELETE_TIMEOUT);
		}

		if(
			(
				(route->seq_num == 0)
				|| (route->seq_num == DYMO_UNKNOWN_SEQUENCE_NUMBER)
			)
			&& (unreachable_node->seq_num != DYMO_UNKNOWN_SEQUENCE_NUMBER)
		) {

			C_DEBUG(DYMO_DEBUG_LEVEL, C_MSG_VERBOSE, "DYMO: Setting sequence number of route to %s to %d...\n", addr_to_str(route->address), unreachable_node->seq_num);

			route->seq_num = unreachable_node->seq_num;
		}

		new_unreachable_node_seq_num = unreachable_node->seq_num;
		if(
			(route->seq_num != DYMO_UNKNOWN_SEQUENCE_NUMBER)
			&& (unreachable_node->seq_num == DYMO_UNKNOWN_SEQUENCE_NUMBER)
		) {
			new_unreachable_node_seq_num = route->seq_num;
		}

		return true;

	} else {

		C_DEBUG(DYMO_DEBUG_LEVEL, C_MSG_DEBUG, "DYMO: Processing RERR block from %s for %s informing about a broken route to %s, invalidating block because processing criteria don't match...\n", addr_to_str(header->source_address), addr_to_str(header->destination_address), addr_to_str(unreachable_node->address));
		unreachable_node->address = C_BROADCAST_ADDRESS;
		return false;

	}
}

/**
 * @brief Handle incoming DYMO RERR packet.
 *
 * See Sec. 5.5.4.
 *
 * @param header Pointer to header of incoming DYMO RERR packet.
 * @param rerr Pointer to route error.
 * @param node_array Pointer to array of additional nodes.
 * @param node_array_size Size of array of additional nodes in bytes.
 */
static void handle_DYMO_RERR(const net_header_t* header, const DYMO_RERR_packet_t* rerr, const DYMO_node_block_t* node_array, size_t node_array_size) {
	DYMO_node_block_t* new_node_array = NULL;
	unsigned int i = 0;
	bool forward_rerr = false;
	uint8_t new_hop_limit = 0;

	C_ABORT_ON(header->ttl != 255, "DYMO: Received RERR from %s for %s with TTL %d instead of 255.\n", addr_to_str(header->source_address), addr_to_str(header->destination_address), header->ttl);

	C_DEBUG(DYMO_DEBUG_LEVEL, C_MSG_DEBUG, "DYMO: Received RERR from %s for %s...\n", addr_to_str(header->source_address), addr_to_str(header->destination_address));

	MALLOC_ARRAY(new_node_array, DYMO_node_block_t, (node_array_size / sizeof(DYMO_node_block_t)));
	memcpy(new_node_array, node_array, node_array_size);

	for(i = 0; i < (node_array_size / sizeof(DYMO_node_block_t)); i++) {
		if(new_node_array[i].address != C_BROADCAST_ADDRESS)
			forward_rerr |= process_RERR_node_block(header, rerr, &new_node_array[i]);
	}

	if(forward_rerr) {
		new_hop_limit = rerr->hop_limit - 1;
		if(new_hop_limit > 1) {
			C_DEBUG(DYMO_DEBUG_LEVEL, C_MSG_INFO, "DYMO: Received RERR from %s for %s, propagating RERR...\n", addr_to_str(header->source_address), addr_to_str(header->destination_address));
			send_DYMO_RERR_array(header->destination_address, new_node_array, node_array_size, new_hop_limit);
		} else {
			C_DEBUG(DYMO_DEBUG_LEVEL, C_MSG_DEBUG, "DYMO: Received RERR from %s for %s, dropping due to hop limit reaching zero...\n", addr_to_str(header->source_address), addr_to_str(header->destination_address));
		}

	}

	FREE(new_node_array);
}

/**
 * @brief DYMO packet classifier.
 *
 * @param header Pointer to header of an incoming packet.
 * @param payload Pointer to payload of an incoming packet.
 */
void DYMO_recv(const net_header_t* header, const void* payload) {
	DYMO_node_block_t* additional_node_array = NULL;
	size_t additional_node_array_size;

	C_ABORT_ON(header->app != APP_TYPE_DYMO, "DYMO: Received non-DYMO packet.\n");

	switch(header->type) {
	case DYMO_PACKET_TYPE_RREQ:
		C_ABORT_ON(header->payload_length != sizeof(DYMO_RREQ_packet_t), "DYMO: Received RREQ packet with wrong payload length (%d != %d).\n", header->payload_length, sizeof(DYMO_RREQ_packet_t));
		handle_DYMO_RREQ(header, (DYMO_RREQ_packet_t*) payload);
		break;
	case DYMO_PACKET_TYPE_RREP:
		C_ABORT_ON(header->payload_length < sizeof(DYMO_RREP_packet_t), "DYMO: Received RREP packet with wrong payload length (%d < %d).\n", header->payload_length, sizeof(DYMO_RREP_packet_t));
		additional_node_array = ((DYMO_node_block_t*) (((char*) payload) + sizeof(DYMO_RREP_packet_t)));
		additional_node_array_size = header->payload_length - sizeof(DYMO_RREP_packet_t);
		handle_DYMO_RREP(header, (DYMO_RREP_packet_t*) payload, additional_node_array, additional_node_array_size);
		break;
	case DYMO_PACKET_TYPE_RERR:
		C_ABORT_ON(header->payload_length < sizeof(DYMO_RERR_packet_t), "DYMO: Received RERR packet with wrong payload length (%d < %d).\n", header->payload_length, sizeof(DYMO_RERR_packet_t));
		additional_node_array = ((DYMO_node_block_t*) (((char*) payload) + sizeof(DYMO_RERR_packet_t)));
		additional_node_array_size = header->payload_length - sizeof(DYMO_RERR_packet_t);
		handle_DYMO_RERR(header, (DYMO_RERR_packet_t*) payload, additional_node_array, additional_node_array_size);
		break;
	default:
		C_ABORT_ON(true, "DYMO: Received packet of unknown type 0x%X from %s.\n", header->type, addr_to_str(header->source_address));
	}
}



/*
 * Interface provided by this module.
 */

/**
 * @brief Timeout callback for sending a packet via DYMO.
 *
 * @param parameter Queued packet for which the timeout ocurred.
 */
void DYMO_send_timeout(callback_parameter_t parameter) {
	DYMO_queued_packet_t* queued_packet = (DYMO_queued_packet_t*) parameter;
	C_time_t current_rreq_wait_time = 0;

	queued_packet->callback_id = 0;

	if(queued_packet->tries < DYMO_RREQ_TRIES) {

#ifdef DYMO_USE_EXPANDING_RING_SEARCH
		queued_packet->ttl += AODV_TTL_INCREMENT;
		if(queued_packet->ttl > AODV_TTL_THRESHOLD)
			queued_packet->ttl = AODV_NET_DIAMETER;

		current_rreq_wait_time = 2 * AODV_NODE_TRAVERSAL_TIME * (queued_packet->ttl + AODV_TIMEOUT_BUFFER);
		if(queued_packet->ttl == AODV_NET_DIAMETER) {
			current_rreq_wait_time = AODV_NET_TRAVERSAL_TIME * pow(2, queued_packet->tries);
			queued_packet->tries++;
		}

		C_DEBUG(DYMO_DEBUG_LEVEL, C_MSG_INFO, "DYMO: Route discovery timeout expired for packet #%d from %s to %s, retrying with TTL = %d, timeout in %.2fs...\n", queued_packet->packet.header->seqno, addr_to_str(queued_packet->packet.header->source_address), addr_to_str(queued_packet->packet.header->destination_address), queued_packet->ttl, current_rreq_wait_time);

		queued_packet->callback_id = callback_add(DYMO_send_timeout, queued_packet, current_rreq_wait_time);
		// OrigNode.Dist set to a number greater than zero (0).
		send_DYMO_RREQ(queued_packet->packet.header->destination_address, C_LOCAL_ADDRESS, get_own_seq_num(), 1, queued_packet->ttl);
#else
		current_rreq_wait_time = DYMO_ROUTE_RREQ_WAIT_TIME * pow(2, queued_packet->tries);
		C_DEBUG(DYMO_DEBUG_LEVEL, C_MSG_INFO, "DYMO: Route discovery timeout expired for packet #%d from %s to %s, retrying with TTL = %d, timeout in %.2fs...\n", queued_packet->packet.header->seqno, addr_to_str(queued_packet->packet.header->source_address), addr_to_str(queued_packet->packet.header->destination_address), DYMO_MSG_HOPLIMIT, current_rreq_wait_time);

		queued_packet->tries++;
		queued_packet->callback_id = callback_add(DYMO_send_timeout, queued_packet, current_rreq_wait_time);
		// OrigNode.Dist set to a number greater than zero (0).
		send_DYMO_RREQ(queued_packet->packet.header->destination_address, C_LOCAL_ADDRESS, get_own_seq_num(), 1, DYMO_MSG_HOPLIMIT);
#endif	// DYMO_USE_EXPANDING_RING_SEARCH

	} else {

		// NOTE: This situation might warrant a WARNing.
		C_DEBUG(DYMO_DEBUG_LEVEL, C_MSG_INFO, "DYMO: Route discovery timeout expired for packet #%d from %s to %s and no retries left, dropping all packets to destination...\n", queued_packet->packet.header->seqno, addr_to_str(queued_packet->packet.header->source_address), addr_to_str(queued_packet->packet.header->destination_address));


		remove_and_free_addressed_packets_from_queue(queued_packet->packet.header->destination_address);

		// NOT IMPLEMENTED: Send ICMP Destination Unreachable to source.

	}
}

/**
 * @brief Check whether a queued packet has a lower priority than another one.
 *
 * @param item Pointer to queued packet to check.
 * @param cmp_item Pointer to reference queued packet.
 * @return <code>true</code> if the queued packet to check has a lower
 * priority than the reference queued packet, <code>false</code> otherwise.
 */
static bool has_lower_priority(const void* item, const void* cmp_item) {
	DYMO_queued_packet_t* queued_packet = (DYMO_queued_packet_t*) item;
	DYMO_queued_packet_t* cmp_queued_packet = (DYMO_queued_packet_t*) cmp_item;
	return (queued_packet->packet.header->priority < cmp_queued_packet->packet.header->priority);
}

/**
 * @brief Route a packet via DYMO.
 *
 * @param header Pointer to the header of the packet to send.
 * @param payload Pointer to the payload of the packet to send.
 */
void DYMO_route(net_header_t* header, const void* payload) {
	DYMO_queued_packet_t* queued_packet = NULL;
	DYMO_queued_packet_t cmp_queued_packet;
	net_header_t cmp_header;
	DYMO_route_t* route = NULL;
	DYMO_route_t cmp_route;
#ifdef DYMO_USE_EXPANDING_RING_SEARCH
	ttl_t current_rreq_ttl = 0;
#endif	// DYMO_USE_EXPANDING_RING_SEARCH

	if(header == NULL) {
		C_DEBUG(DYMO_DEBUG_LEVEL, C_MSG_ERROR, "DYMO: Trying to send packet without header.\n");
		return;
	} else if((header->payload_length > 0) && (payload == NULL)) {
		C_DEBUG(DYMO_DEBUG_LEVEL, C_MSG_ERROR, "DYMO: Trying to send non-empty packet without payload.\n");
		return;
	}

	memset(&cmp_queued_packet, 0, sizeof(DYMO_queued_packet_t));
	memset(&cmp_header, 0, sizeof(net_header_t));
	cmp_queued_packet.packet.header = &cmp_header;
	cmp_header.destination_address = header->destination_address;
	memset(&cmp_route, 0, sizeof(DYMO_route_t));
	cmp_route.address = header->destination_address;
	if(header->destination_address == C_BROADCAST_ADDRESS) {

		C_DEBUG(DYMO_DEBUG_LEVEL, C_MSG_INFO, "DYMO: Sending packet #%d from %s to %s...\n", header->seqno, addr_to_str(header->source_address), addr_to_str(header->destination_address));
		C_send(header, payload);

	} else if(
		((route = (DYMO_route_t*) list_find(C_DYMO_routing_table, &cmp_route, matches_route_address)) != NULL)
		&& route->forwarding
	) {

#ifdef C_USE_ETX_LINK_METRIC
		C_DEBUG(DYMO_DEBUG_LEVEL, C_MSG_INFO, "DYMO: Routing packet #%d (0x%X/0x%X) from %s to %s via next hop %s (ETX %.4f)...\n", header->seqno, header->app, header->type, addr_to_str(header->source_address), addr_to_str(header->destination_address), addr_to_str(route->next_hop_address), calculate_ETX(route->next_hop_address));
#else
		C_DEBUG(DYMO_DEBUG_LEVEL, C_MSG_INFO, "DYMO: Routing packet #%d (0x%X/0x%X) from %s to %s via next hop %s...\n", header->seqno, header->app, header->type, addr_to_str(header->source_address), addr_to_str(header->destination_address), addr_to_str(route->next_hop_address));
#endif	// C_USE_ETX_LINK_METRIC
		reset_routing_timers(header->destination_address);

		header->next_hop_address = route->next_hop_address;
		C_send(header, payload);

	} else if((queued_packet = (DYMO_queued_packet_t*) list_find(C_DYMO_packet_queue, &cmp_queued_packet, matches_packet_destination_address)) != NULL) {
		queued_packet = NULL;	// We're just interested in the content of the queue, not in the packet.

		C_DEBUG(DYMO_DEBUG_LEVEL, C_MSG_NOTE, "DYMO: Queueing packet #%d from %s to %s, route discovery is already in progress...\n", header->seqno, addr_to_str(header->source_address), addr_to_str(header->destination_address));

		purge_packet_queue(1, (sizeof(net_header_t) + header->payload_length));
		MALLOC(queued_packet, DYMO_queued_packet_t);
		alloc_header_from_header(&queued_packet->packet.header, header);
		if(header->payload_length > 0) {
			void* p = NULL;
			MALLOC_ARRAY(p, char, header->payload_length);
			memcpy(p, payload, header->payload_length);
			queued_packet->packet.payload = p;
		} else {
			queued_packet->packet.payload = NULL;
		}
#ifdef DYMO_USE_EXPANDING_RING_SEARCH
		queued_packet->ttl = AODV_TTL_START;
#endif	// DYMO_USE_EXPANDING_RING_SEARCH
		queued_packet->tries = 1;

		list_insert_where(C_DYMO_packet_queue, queued_packet, has_lower_priority);
		queued_packet->callback_id = 0;	// There is already a timer running for the other queued packet.

	} else {

		C_DEBUG(DYMO_DEBUG_LEVEL, C_MSG_NOTE, "DYMO: Queueing packet #%d from %s to %s and starting route discovery...\n", header->seqno, addr_to_str(header->source_address), addr_to_str(header->destination_address));

		purge_packet_queue(1, (sizeof(net_header_t) + header->payload_length));
		MALLOC(queued_packet, DYMO_queued_packet_t);
		alloc_header_from_header(&queued_packet->packet.header, header);
		if(header->payload_length > 0) {
			void* p = NULL;
			MALLOC_ARRAY(p, char, header->payload_length);
			memcpy(p, payload, header->payload_length);
			queued_packet->packet.payload = p;
		} else {
			queued_packet->packet.payload = NULL;
		}
#ifdef DYMO_USE_EXPANDING_RING_SEARCH
		queued_packet->ttl = AODV_TTL_START;
		queued_packet->tries = 0;
#else
		queued_packet->tries = 1;
#endif	// DYMO_USE_EXPANDING_RING_SEARCH
		list_insert_where(C_DYMO_packet_queue, queued_packet, has_lower_priority);

#ifdef DYMO_USE_EXPANDING_RING_SEARCH
		current_rreq_ttl = AODV_TTL_START;
		if(route != NULL) {
			C_ABORT_ON(route->forwarding, "DYMO: Preparing discovery, but route to %s is enabled for forwarding.\n", route->address);
			current_rreq_ttl = route->dist + AODV_TTL_INCREMENT;
		}

		// According to Sec. 10 of AODV RFC: RING_TRAVERSAL_TIME
		// = (2 * NODE_TRAVERSAL_TIME * (TTL_VALUE + TIMEOUT_BUFFER))
		queued_packet->callback_id = callback_add(DYMO_send_timeout, queued_packet, (2 * AODV_NODE_TRAVERSAL_TIME * (current_rreq_ttl + AODV_TIMEOUT_BUFFER)));
		// OrigNode.Dist set to a number greater than zero (0).
		send_DYMO_RREQ(header->destination_address, C_LOCAL_ADDRESS, get_own_seq_num(), 1, current_rreq_ttl);
#else
		queued_packet->callback_id = callback_add(DYMO_send_timeout, queued_packet, DYMO_ROUTE_RREQ_WAIT_TIME);
		// OrigNode.Dist set to a number greater than zero (0).
		send_DYMO_RREQ(header->destination_address, C_LOCAL_ADDRESS, get_own_seq_num(), 1, DYMO_MSG_HOPLIMIT);
#endif	// DYMO_USE_EXPANDING_RING_SEARCH

	}

	if(header->destination_address != C_BROADCAST_ADDRESS)
		C_DYMO_transmitted_unicast_packets++;
	else
		C_DYMO_transmitted_broadcast_packets++;
}

/**
 * @brief Route, deliver or drop a received packet.
 *
 * 5.5.2. Updating Route Lifetimes During Packet Forwarding
 *
 * To avoid removing the forwarding route to reach the IP.SourceAddress,
 * ThisNode SHOULD set a timeout (ROUTE_USED) to ROUTE_USED_TIMEOUT for the
 * route to the IP.SourceAddress upon receiving a data packet. If a timer for
 * ROUTE_DELETE is set, it is removed.
 *
 * To avoid removing the forwarding route to the IP.DestinationAddress that is
 * being used, ThisNode SHOULD set a timeout (ROUTE_USED) to ROUTE_USED_TIMEOUT
 * for the route to the IP.DestinationAddress upon sending a data packet. If a
 * timer for ROUTE_DELETE is set, it is removed.
 *
 * @param header Pointer to the header of the received packet.
 * @param payload Pointer to the payload of the received packet.
 */
void DYMO_route_recv(const net_header_t* header, const void* payload) {
	DYMO_route_t* route = NULL;
	DYMO_route_t cmp_route;
	net_header_t* new_header = NULL;

	if(!reset_routing_timers(header->source_address)) {
#ifdef DYMO_USE_OVERHEARD_BROADCAST_TO_ADD_ROUTES
#ifdef C_USE_ETX_LINK_METRIC
		if(calculate_ETX(header->last_hop_address) < DYMO_USE_OVERHEARD_BROADCAST_ETX_THRESHOLD) {
#endif	// C_USE_ETX_LINK_METRIC
			// Add a routes to the source of this packet on
			// reliable links, but use worst possible metrics to
			// make sure that the route can be replaced by real
			// routing messages.
			C_DEBUG(DYMO_DEBUG_LEVEL, C_MSG_INFO, "DYMO: Adding new route to %s with next hop %s based on overheard broadcast...\n", addr_to_str(header->source_address), addr_to_str(header->last_hop_address));
			MALLOC(route, DYMO_route_t);
			route->address = header->source_address;
			list_add(C_DYMO_routing_table, route);
			init_route(route, 1, header->last_hop_address, DYMO_MAX_DISTANCE);
#ifdef C_USE_ETX_LINK_METRIC
		}
#endif	// C_USE_ETX_LINK_METRIC
#endif	// DYMO_USE_OVERHEARD_BROADCAST_TO_ADD_ROUTES
	}

	memset(&cmp_route, 0, sizeof(DYMO_route_t));
	cmp_route.address = header->destination_address;
	if((header->destination_address == C_LOCAL_ADDRESS) || (header->destination_address == C_BROADCAST_ADDRESS)) {

		// Received a packet for the local node, deliver it.

		C_DEBUG(DYMO_DEBUG_LEVEL, C_MSG_DEBUG, "DYMO: Received packet #%d from %s to %s via %s...\n", header->seqno, addr_to_str(header->source_address), addr_to_str(header->destination_address), addr_to_str(header->last_hop_address));
		C_handle_packet(header, payload);

		// Do stats, but don't count DYMO control packets.
		if(header->app != APP_TYPE_DYMO) {
			if(header->destination_address != C_BROADCAST_ADDRESS)
				C_DYMO_received_unicast_packets++;
			else
				C_DYMO_received_broadcast_packets++;
		}

	} else if(header->ttl == 0) {

		// Received a packet with zero TTL, drop it.

		// NOTE: This situation might warrant a WARNing.
		C_DEBUG(DYMO_DEBUG_LEVEL, C_MSG_INFO, "DYMO: Received packet #%d from %s to %s with zero TTL, dropping packet...\n", header->seqno, addr_to_str(header->source_address), addr_to_str(header->destination_address));

	} else if((route = (DYMO_route_t*) list_find(C_DYMO_routing_table, &cmp_route, matches_route_address)) != NULL) {

		if(route->forwarding) {

#ifdef C_USE_NETWORK_MAPPING
			unsigned int i = 0;
			for(i = 0; i < header->hop_count; i++) {
				if(header->hops[i] == route->next_hop_address) {
					// NOTE: This situation might warrant a WARNing.
					C_DEBUG(DYMO_DEBUG_LEVEL, C_MSG_INFO, "DYMO: Received packet #%d from %s to %s which has previously been routed by the next hop %s, dropping packet...\n", header->seqno, addr_to_str(header->source_address), addr_to_str(header->destination_address), addr_to_str(route->next_hop_address));
					return;
				}
			}
#endif	// C_USE_NETWORK_MAPPING

			// Received a packet for which a route is known, forward it.
#ifdef C_USE_ETX_LINK_METRIC
			C_DEBUG(DYMO_DEBUG_LEVEL, C_MSG_INFO, "DYMO: Routing packet #%d (0x%X/0x%X) from %s to %s (TTL = %d) via next hop %s (ETX %.4f)...\n", header->seqno, header->app, header->type, addr_to_str(header->source_address), addr_to_str(header->destination_address), header->ttl, addr_to_str(route->next_hop_address), calculate_ETX(route->next_hop_address));
#else
			C_DEBUG(DYMO_DEBUG_LEVEL, C_MSG_INFO, "DYMO: Routing packet #%d (0x%X/0x%X) from %s to %s (TTL = %d) via next hop %s...\n", header->seqno, header->app, header->type, addr_to_str(header->source_address), addr_to_str(header->destination_address), header->ttl, addr_to_str(route->next_hop_address));
#endif	// C_USE_ETX_LINK_METRIC
	
			reset_routing_timers(header->destination_address);
	
			alloc_header_from_header(&new_header, header);
			new_header->last_hop_address = C_LOCAL_ADDRESS;
			new_header->next_hop_address = route->next_hop_address;
			new_header->ttl--;
			C_send(new_header, payload);
			free_header(&new_header);

		} else {

			// Received a packet for which no valid route is known, send RERR.

			// NOTE: This situation might warrant a WARNing.
			C_DEBUG(DYMO_DEBUG_LEVEL, C_MSG_INFO, "DYMO: Route to %s for packet #%d from %s is broken, sending RERR and dropping packet...\n", addr_to_str(header->destination_address), header->seqno, addr_to_str(header->source_address));
			send_DYMO_RERR(header->source_address, header->destination_address, route->seq_num, route->next_hop_address, DYMO_MSG_HOPLIMIT);

		}

	} else {

		// Received a packet for which no route is known, send RERR.

		// NOTE: This situation might warrant a WARNing.
		C_DEBUG(DYMO_DEBUG_LEVEL, C_MSG_INFO, "DYMO: No route to %s for packet #%d from %s, sending RERR and dropping packet...\n", addr_to_str(header->destination_address), header->seqno, addr_to_str(header->source_address));
		send_DYMO_RERR(header->source_address, header->destination_address, DYMO_UNKNOWN_SEQUENCE_NUMBER, C_BROADCAST_ADDRESS, DYMO_MSG_HOPLIMIT);

	}
}

/**
 * @brief Remove routes via a particular next hop.
 *
 * From Sec. 5.5:
 * In addition to inability to forward a data packet, a RERR SHOULD be
 * issued immediately after detecting a broken link of an forwarding
 * route to quickly notify DYMO routers that a link break occurred and
 * that certain routes are no longer available.  If the route with the
 * broken link has not been used recently (indicated by ROUTE_USED), the
 * RERR SHOULD NOT be generated.
 *
 * From Sec. 5.5.1:
 * Upon determining that a link is broken or the next-hop is unreachable,
 * ThisNode MUST remove the affected forwarding routes (those with an
 * unreachable next-hop) and unset the Route.Forwarding flag. ThisNode also
 * flags the associated routes in DYMO's routing table as Broken. For each
 * broken route a timer for ROUTE_DELETE is set to ROUTE_DELETE_TIMEOUT.
 *
 * @param next_hop_address Address of node to which DLL link has become broken.
 */
void DYMO_handle_unreachable_next_hop(C_addr_t next_hop_address) {
	DYMO_route_t* route = NULL;
	list_iterator_t route_iterator;

	C_DEBUG(DYMO_DEBUG_LEVEL, C_MSG_INFO, "DYMO: Detected broken link to %s, purging routing table...\n", addr_to_str(next_hop_address));

	list_iterator_init(C_DYMO_routing_table, &route_iterator);
	while((route = (DYMO_route_t*) list_iterator_next(&route_iterator)) != NULL) {
		if(route->next_hop_address == next_hop_address) {

			if(route->broken) {
				C_DEBUG(DYMO_DEBUG_LEVEL, C_MSG_VERBOSE, "DYMO: Broken route to %s is affected by broken link to %s, skipping...\n", addr_to_str(route->address), addr_to_str(next_hop_address));
				continue;
			}
			
			C_DEBUG(DYMO_DEBUG_LEVEL, C_MSG_DEBUG, "DYMO: Route to %s is affected by broken link to %s, disabling forwarding...\n", addr_to_str(route->address), addr_to_str(next_hop_address));
			route->forwarding = false;
			route->broken = true;

			if(route->delete_information_timeout_callback_id == 0) {
				C_DEBUG(DYMO_DEBUG_LEVEL, C_MSG_DEBUG, "DYMO: Route to %s is affected by broken link to %s, starting timer for deletion...\n", addr_to_str(route->address), addr_to_str(next_hop_address));
				route->delete_information_timeout_callback_id = callback_add(route_delete_information_timeout, route, DYMO_ROUTE_DELETE_TIMEOUT);
			}
		}
	}
	send_DYMO_RERR(C_BROADCAST_ADDRESS, next_hop_address, DYMO_UNKNOWN_SEQUENCE_NUMBER, next_hop_address, DYMO_MSG_HOPLIMIT);
}

/**
 * @brief Return the MTU size supported by DYMO.
 *
 * @return Maximal Transmission Unit (MTU) size in bytes.
 */
size_t DYMO_max_payload_size() {
	// DYMO is reusing the fields in the DLL header.
	return C_max_dll_payload_size();
}



/*
 * Interface provided by this module for direct commands.
 */

/**
 * @brief DYMO command function.
 *
 * @param argc Number of command parameters.
 * @param argv Array of strings used as command parameters.
 * @return <code>true</code> if the command was executed successfully,
 * <code>false</code> otherwise.
 */
bool DYMO_command(int argc, const char*const* argv) {
	return false;
}
