/**
 * @file dymo_types.h
 *
 * @brief Dynamic MANET On-demand (DYMO) routing protocol (types).
 *
 * Copyright (c) 2007, 2008, 2009, 2010 Georg Wittenburg, Freie Universitaet
 * Berlin (wittenbu@inf.fu-berlin.de). All rights reserved.
 * Please refer to the COPYING file for detailed licensing information.
 *
 * @author Georg Wittenburg
 *
 * $Id: dymo_types.h 3591 2010-06-15 15:09:05Z gwitten $
 */

#ifndef DYMO_TYPES_H
#define DYMO_TYPES_H

#include "../types.h"



/**
 * @brief Constant to enable expanding ring search when sending RREQs.
 *
 * Sec 5.3.1 of the draft states that expanding ring search according to
 * RFC3561 MAY be enabled.
 *
 * DYMO-UM does not use expanding ring search.
 */
// #define DYMO_USE_EXPANDING_RING_SEARCH

/**
 * @brief Constant to enable discarding unused routes.
 *
 * Discarding supposedly unused routes results in a low PDR for low data rate
 * traffic on the testbed. Hence, we make it optional to use this feature.
 */
#ifndef C_TESTBED_EVALUATION	// Don't discard unused routes if running on the testbed.
	#define DYMO_DISCARD_UNUSED_ROUTES_AFTER_TIMEOUT
#endif	// C_TESTBED_EVALUATION



/**
 * @brief Routing table entry.
 *
 * 4.1. Route Table Entry
 *
 * The route table entry is a conceptual data structure. Implementations may
 * use any internal representation that conforms to the semantics of a route as
 * specified in this document.
 *
 * Conceptually, a route table entry has the following fields:
 * [SEE COMMENTS BELOW]
 *
 * Not including optional information may cause performance degradation, but it
 * will not cause the protocol to operate incorrectly.
 *
 * In addition to a route table data structure, each route table entry may have
 * several timers associated with the information. These timers/timeouts are
 * discussed in Section 5.2.3.
 */
typedef struct {
	C_addr_t	address;		///< (Route.Address) The IP (host or network) destination address of the node(s) associated with the routing table entry.
	// NOT_USED				///< (Route.Prefix) Indicates that the associated address is a network address, rather than a host address. The value is the length of the netmask/prefix.
	uint16_t	seq_num;		///< (Route.SeqNum) The DYMO SeqNum associated with this routing information.
	C_addr_t	next_hop_address;	///< (Route.NextHopAddress) The IP address of the adjacent DYMO router on the path toward the Route.Address.
	// NOT_USED				///< (Route.NextHopInterface) The interface used to send packets toward the Route.Address.
	bool		forwarding;		///< (Route.Forwarding) A flag indicating whether this Route can be used for forwarding data packets. This flag MAY be provided for management and monitoring.
	bool		broken;			///< (Route.Broken) A flag indicating whether this Route is broken. This flag is set if the next-hop becomes unreachable or in response to processing a RERR.
	uint16_t	dist;			///< (Route.Dist, optional) A metric indicating the distance traversed before reaching the Route.Address node.
	callback_id_t	minimum_age_delete_timeout_callback_id;			///< Callback ID of minimum age delete timeout
	callback_id_t	maximum_sequence_number_delete_timeout_callback_id;	///< Callback ID of maximum sequence number delete timeout
#ifdef DYMO_DISCARD_UNUSED_ROUTES_AFTER_TIMEOUT
	callback_id_t	recently_used_timeout_callback_id;			///< Callback ID of recently used timeout
#endif	// DYMO_DISCARD_UNUSED_ROUTES_AFTER_TIMEOUT
	callback_id_t	delete_information_timeout_callback_id;			///< Callback ID of delete information timeout
	bool		can_be_deleted;		///< Extra (not mentioned in draft) flag to note whether a route may be deleted
} DYMO_route_t;

/**
 * @brief Packet queue entry.
 */
typedef struct {
	C_packet_t	packet;		///< Packet to be transmitted.
	C_time_t	time;		///< Time at which the packet was enqueued.
#ifdef DYMO_USE_EXPANDING_RING_SEARCH
	ttl_t		ttl;		///< TTL of the last RREQ that was sent for this packet.
#endif	// DYMO_USE_EXPANDING_RING_SEARCH
	uint8_t		tries;		///< Current number of tries at finding a route for this packet.
	callback_id_t	callback_id;	///< ID of the timeout callback.
} DYMO_queued_packet_t;



/*
 * Packet types.
 */

/**
 * @brief Node blocks to be used in routing messages.
 */
typedef struct {
	C_addr_t	address;	///< (AddBlk.Node.Address) The IP address of the Node.
	uint16_t	seq_num;	///< (Node.AddTLV.SeqNum, optional) The last known DYMO sequence number of the Node.
	uint16_t	dist;		///< (Node.AddTLV.Dist, optional) The last known Distance to the Node
} DYMO_node_block_t;

/**
 * @brief Routing message.
 *
 * 4.2.2. Routing Messages (RM) - RREQ & RREP
 *
 * Routing Messages (RMs) are used to disseminate routing information. There
 * are two DYMO message types that are considered to be routing messages (RMs):
 * RREQ and RREP. They contain very similar information and function, but have
 * slightly different processing rules. The main difference between the two
 * messages is that RREQ messages generally solicit a RREP, whereas a RREP is
 * the response to RREQ.
 *
 * RM creation and processing are described in Section 5.3.
 *
 * A RM requires the following information:
 * [SEE COMMENTS BELOW]
 */
typedef struct {
	// SET IN PACKET HEADER				///< (IP.SourceAddress) The IP address of the node currently sending this packet. This field is generally filled automatically by the operating system and should not require special handling.
	// SET IN PACKET HEADER				///< (IP.DestinationAddress) The IP address of the packet destination. For multicast RREQ the IP.DestinationAddress is set to LL-MANET ROUTERS. For unicast RREQ and RREP the IP.DestinationAddress is set to the NextHopAddress toward the RREP TargetNode.
	// NOT_USED					///< (UDP.DestinationPort) By default, the UDP destination port is set to MANET.
	uint8_t			hop_limit;		///< (MsgHdr.HopLimit) The remaining number of hops this message is allowed to traverse.
	DYMO_node_block_t	target_node;		///< (AddBlk/AddTLV.TargetNode) Node block for the TargetNode. In a RREQ the TargetNode is the destination for which route discovery is being performed. In a RREP the TargetNode is the RREQ OrigNode. The TargetNode address is the first address in a routing message.
	DYMO_node_block_t	originator_node;	///< (AddBlk/AddTLV.OrigNode) Node block for the OrigNode. In a RREQ the OrigNode is the source. In a RREP the OrigNode is the RREQ TargetNode for which a RREP is being generated.  This address is the second address in the message for RREQ.
	// (AddBlk/AddTLV.AdditionalNode) is implemented as secondary payload.
} DYMO_routing_message_t;

/**
 * @brief DYMO RREQ packet.
 */
typedef DYMO_routing_message_t DYMO_RREQ_packet_t;

/**
 * @brief DYMO RREP packet.
 */
typedef DYMO_routing_message_t DYMO_RREP_packet_t;

/**
 * @brief DYMO RERR packet.
 *
 * 4.2.3. Route Error (RERR)
 *
 * A RERR message is used to disseminate the information that a route is not
 * available for one or more particular IP addresses.
 *
 * RERR creation and processing are described in Section 5.5.
 *
 * A RERR requires the following information:
 * [SEE COMMENTS BELOW]
 */
typedef struct {
	// SET IN PACKET HEADER				///< (IP.SourceAddress) The IP address of the node currently sending this packet. This field is generally filled automatically by the operating system and should not require special handling.
	// SET IN PACKET HEADER				///< (IP.DestinationAddress) For multicast RERR messages, The IP address is set to LL-MANET-ROUTERS. For unicast RERR messages, The IP address is set to the NextHopAddress.
	// NOT_USED					///< (UDP.DestinationPort) By default, the UDP destination port is set to MANET.
	uint8_t			hop_limit;		///< (MsgHdr.HopLimit) The remaining number of hops this message is allowed to traverse.
	// (AddBlk/AddTLV.UnreachableNode) is implemented as secondary payload.
} DYMO_RERR_packet_t;

#endif	// DYMO_TYPES_H
