/**
 * @file interface.c
 *
 * @brief Service Placement interface implementation.
 *
 * Copyright (c) 2007, 2008, 2009, 2010 Georg Wittenburg, Freie Universitaet
 * Berlin (wittenbu@inf.fu-berlin.de). All rights reserved.
 * Please refer to the COPYING file for detailed licensing information.
 *
 * @author Georg Wittenburg
 *
 * $Id: interface.c 3571 2010-06-09 14:23:55Z gwitten $
 */

#include <float.h>
#include <string.h>

#include "config.h"
#include "macros.h"
#include "types.h"
#include "vars.h"
#include "interface.h"
#include "main.h"
#include "util/debug.h"
#include "util/list.h"
#include "util/callback.h"
#include "util/mapping.h"
#include "util/net.h"
#include "util/stats.h"
#include "dymo/dymo.h"
#include "icmp-light/icmp-light.h"
#include "rudp-light/rudp-light.h"
#include "slp-light/slp-light.h"
#include "service/service.h"
#include "placement/placement.h"
#include "interference/interference.h"



/*
 * Interface functions to be called by underlying system.
 */

/**
 * @brief Global initialization function called by the underlying system.
 *
 * @param local_address The local address of this node.
 * @param broadcast_address The broadcast address of the network.
 */
void C_init(C_addr_t local_address, C_addr_t broadcast_address) {
	C_state = C_STATE_INITIALIZING;

	C_LOCAL_ADDRESS = local_address;
	C_BROADCAST_ADDRESS = broadcast_address;
	C_addr_to_str_buffer_index = 0;
	C_DEBUG(C_MAIN_DEBUG_LEVEL, C_MSG_INFO, "Initializing %s with broadcast %s... \n", addr_to_str(C_LOCAL_ADDRESS), addr_to_str(C_BROADCAST_ADDRESS));

	C_ABORT_ON(C_LOCAL_ADDRESS == C_BROADCAST_ADDRESS, "Local node is using the broadcast address (%s == %s).\n", addr_to_str(C_LOCAL_ADDRESS), addr_to_str(C_BROADCAST_ADDRESS));
	if(!is_valid_address(C_LOCAL_ADDRESS)) {
		C_DEBUG(C_MAIN_DEBUG_LEVEL, C_MSG_ERROR, "Local node is using an invalid address (%s). This may be due to an overly restrictive configuration of valid address ranges (see net.c:is_valid_address()).\n", addr_to_str(C_LOCAL_ADDRESS));
	}

	// Initialize global variables.
	C_init_time = __C_time();
	C_stop_time = 0.0;
	C_check_consistency_on_shutdown = true;
	C_callbacks = NULL;
	list_alloc(&C_callbacks);
	C_next_callback_id = 0;
	C_registered_callbacks = 0;
	C_executed_callbacks = 0;
	C_sequence_number = 0;
	C_send_callback_id = 0;
	memset(C_received_packet_cache, 0, C_RECEIVED_PACKET_CACHE_SIZE * sizeof(packet_fingerprint_t));

	C_known_nodes = NULL;
	list_alloc(&C_known_nodes);
	register_known_node(C_LOCAL_ADDRESS);
	C_network_map = NULL;
	C_network_map_may_need_purging = true;
	MALLOC_ARRAY(C_network_map, C_time_t, (C_MAXIMAL_NUMBER_OF_NODES * C_MAXIMAL_NUMBER_OF_NODES));
	memset(C_network_map, 0, (C_MAXIMAL_NUMBER_OF_NODES * C_MAXIMAL_NUMBER_OF_NODES) * sizeof(C_time_t));
	C_beacon_callback_id = 0;

	C_edge_cost_matrix = NULL;
	C_distance_matrix = NULL;

	C_periodic_statistics_measurement_callback_id = 0;
	C_periodic_statistics_measurement_samples = 0;
	C_node_degree_running_sum = 0;
	C_average_node_degree = 0.0;
	C_node_distance_running_sum = 0.0;
	C_average_node_distance = 0.0;
	memset(C_unicast_packets_per_hop_count, 0, C_MAXIMAL_NETWORK_DIAMETER * sizeof(uint32_t));
	C_transmitted_unicast_packets = 0;
	C_unicast_transmission_failures = 0;
	C_received_unicast_packets = 0;
	C_transmitted_broadcast_packets = 0;
	C_received_broadcast_packets = 0;
	C_transmitted_bytes = 0;
	C_received_bytes = 0;
	C_transmitted_datarate_stat = NULL;
	alloc_datarate_stat(&C_transmitted_datarate_stat);
	C_transmitted_datarate	= 0;
	C_received_datarate_stat = NULL;
	alloc_datarate_stat(&C_received_datarate_stat);
	C_received_datarate = 0;
	C_network_mapping_overhead = 0;

	C_DEBUG_ERRORs = 0;
	C_DEBUG_WARNs = 0;

	// Initialize components.
	DYMO_init();
	ICMP_init();
	RUDP_init();
	SLP_init();
	SERVICE_init();
	PLACEMENT_init();
	INTERFERENCE_init();

	C_state = C_STATE_INITIALIZED;

#ifndef ARCH_NS2
	print_statistics(false, true, true);
#endif	// ARCH_NS2

	if(C_PERIODIC_MEASUREMENT_INTERVAL > 0.0)
		C_periodic_statistics_measurement_callback_id = callback_add(periodic_statistics_measurement_callback, NULL, C_PERIODIC_MEASUREMENT_INTERVAL);

#ifdef C_USE_NETWORK_MAPPING
#ifdef C_ENABLE_BEACONING
	C_beacon_callback_id = callback_add(beacon_callback, NULL, __C_random(C_BEACON_INTERVAL));
#endif	// C_ENABLE_BEACONING
#endif	// C_USE_NETWORK_MAPPING
}

/**
 * @brief Global start function called by the underlying system.
 */
void C_start() {
	C_DEBUG(C_MAIN_DEBUG_LEVEL, C_MSG_NOTE, "Starting %s...\n", addr_to_str(C_LOCAL_ADDRESS));

#ifdef C_USE_NETWORK_MAPPING
	C_network_map_may_need_purging = true;
#endif	// C_USE_NETWORK_MAPPING

	// Start components.
	DYMO_start();
	ICMP_start();
	RUDP_start();
	SLP_start();
	SERVICE_start();
	PLACEMENT_start();
	INTERFERENCE_start();

	C_state = C_STATE_RUNNING;
}

/**
 * @brief Global stop function called by the underlying system.
 */
void C_stop() {
	C_DEBUG(C_MAIN_DEBUG_LEVEL, C_MSG_NOTE, "Stopping %s...\n", addr_to_str(C_LOCAL_ADDRESS));

	C_stop_time = __C_time();

#ifdef C_USE_NETWORK_MAPPING
	C_network_map_may_need_purging = true;
#endif	// C_USE_NETWORK_MAPPING

	// Stop components.
	INTERFERENCE_stop();
	PLACEMENT_stop();
	SERVICE_stop();
	SLP_stop();
	RUDP_stop();
	ICMP_stop();
	DYMO_stop();

	C_state = C_STATE_STOPPED;
}

/**
 * @brief Global finalization function called by the underlying system.
 */
void C_finalize() {
	C_DEBUG(C_MAIN_DEBUG_LEVEL, C_MSG_DEBUG, "Finalizing %s...\n", addr_to_str(C_LOCAL_ADDRESS));

	if(C_send_callback_id != 0)
		callback_cancel(C_send_callback_id);
#ifdef C_USE_NETWORK_MAPPING
	C_network_map_may_need_purging = true;
#ifdef C_ENABLE_BEACONING
	if(C_beacon_callback_id != 0)
		callback_cancel(C_beacon_callback_id);
#endif	// C_ENABLE_BEACONING
#endif	// C_USE_NETWORK_MAPPING

	// NOTE: The periodic measurement callback is stopped at this point to
	// allow for processing of data even after the node has been stopped.
	if(C_periodic_statistics_measurement_callback_id != 0) {
		callback_cancel(C_periodic_statistics_measurement_callback_id);
		C_periodic_statistics_measurement_callback_id = 0;
		// Make sure the measurements are updated one last time.
		periodic_statistics_measurement();
	}

	// Destroy components.
	INTERFERENCE_finalize();
	PLACEMENT_finalize();
	SERVICE_finalize();
	SLP_finalize();
	RUDP_finalize();
	ICMP_finalize();
	DYMO_finalize();

	C_state = C_STATE_FINALIZED;

	if(!list_empty(C_callbacks)) {
		C_DEBUG(C_MAIN_DEBUG_LEVEL, C_MSG_WARN, "Cancelling %d callback(s) in queue, last one due in %.2f seconds.\n", list_size(C_callbacks), (((callback_t*) list_tail(C_callbacks))->time - __C_time()));
		while(!list_empty(C_callbacks))
			callback_cancel(((callback_t*) list_head(C_callbacks))->id);
	}

	if(C_check_consistency_on_shutdown) {
		C_ABORT_ON(C_executed_callbacks != C_registered_callbacks, "Callback counters don't add up (%d != %d).\n", C_executed_callbacks, C_registered_callbacks);
	}

#ifndef ARCH_NS2
	C_DEBUG(C_MAIN_DEBUG_LEVEL, (C_MSG_INFO | C_MSG_CONT), "Run complete (%.2f).\n\n", (C_time() - C_init_time));
	print_statistics(true, true, true);
#endif	// ARCH_NS2

	free_datarate_stat(&C_transmitted_datarate_stat);
	free_datarate_stat(&C_received_datarate_stat);
	list_free_including_data(&C_known_nodes, free_node_list);
	if(C_network_map != NULL)
		FREE(C_network_map);
	if(C_edge_cost_matrix != NULL)
		FREE(C_edge_cost_matrix);
	if(C_distance_matrix != NULL)
		FREE(C_distance_matrix);

	list_free_including_data(&C_callbacks, free);
}

/**
 * @brief Global packet reception function called by the underlying system.
 *
 * @param header Pointer to the header of the received packet.
 * @param payload Pointer to the payload of the received packet.
 */
void C_recv(net_header_t* header, const void* payload) {
#ifdef C_USE_NETWORK_MAPPING
#ifdef C_ENABLE_CONSISTENCY_CHECKS
	unsigned int i = 0;
#endif	// C_ENABLE_CONSISTENCY_CHECKS
	C_addr_t* hops_including_current_node = NULL;
#endif	// C_USE_NETWORK_MAPPING

#ifdef C_ARTIFICIAL_LINK_RELIABILITY
	if(__C_random(1.0) > C_ARTIFICIAL_LINK_RELIABILITY)
		return;
#endif	// C_ARTIFICIAL_LINK_RELIABILITY

	C_ABORT_ON(!is_valid_address(header->source_address), "Received packet with invalid source address (%s).\n", addr_to_str(header->source_address));
	C_ABORT_ON(!is_valid_address(header->destination_address), "Received packet with invalid destination address (%s).\n", addr_to_str(header->destination_address));
	C_ABORT_ON(!is_valid_address(header->last_hop_address), "Received packet with invalid last hop address (%s).\n", addr_to_str(header->last_hop_address));
	C_ABORT_ON(!is_valid_address(header->next_hop_address), "Received packet with invalid next hop address (%s).\n", addr_to_str(header->next_hop_address));

	// Update statistics counters.
	if(header->destination_address == C_BROADCAST_ADDRESS)
		C_received_broadcast_packets++;
	else
		C_received_unicast_packets++;
	C_received_bytes += (sizeof(net_header_t) + header->payload_length);
	add_bytes_to_datarate_stat(C_received_datarate_stat, (sizeof(net_header_t) + header->payload_length));

	if((header->next_hop_address != C_LOCAL_ADDRESS) && (header->next_hop_address != C_BROADCAST_ADDRESS)) {
		C_DEBUG(C_MAIN_DEBUG_LEVEL, C_MSG_VERBOSE, "Dropping packet #%d from %s to %s (%d bytes, received from %s after %d hops) that is to be handled by next hop %s...\n", header->seqno, addr_to_str(header->source_address), addr_to_str(header->destination_address), (sizeof(net_header_t) + header->payload_length), addr_to_str(header->last_hop_address), header->hop_count, addr_to_str(header->next_hop_address));
		return;
	}

#ifdef C_USE_INTERNAL_ROUTING
	C_ABORT_ON(header->hop_count >= C_MAXIMAL_NETWORK_DIAMETER, "Received packet #%d from %s to %s via last hop %s (%d bytes) after %d hops, which exceeds the maximal network diameter (%d).\n", header->seqno, addr_to_str(header->source_address), addr_to_str(header->destination_address), addr_to_str(header->last_hop_address), (sizeof(net_header_t) + header->payload_length), header->hop_count, C_MAXIMAL_NETWORK_DIAMETER);
#else
	C_ABORT_ON((header->destination_address != C_LOCAL_ADDRESS) && (header->destination_address != C_BROADCAST_ADDRESS), "Received packet #%d from %s to %s via last hop %s (%d bytes, %d hops), but internal routing is disabled.\n", header->seqno, addr_to_str(header->source_address), addr_to_str(header->destination_address), addr_to_str(header->last_hop_address), (sizeof(net_header_t) + header->payload_length), header->hop_count);
	if(header->hop_count >= C_MAXIMAL_NETWORK_DIAMETER)
		C_DEBUG(C_MAIN_DEBUG_LEVEL, C_MSG_WARN, "Received packet #%d from %s to %s via last hop %s (%d bytes) after %d hops, which exceeds the maximal network diameter (%d).\n", header->seqno, addr_to_str(header->source_address), addr_to_str(header->destination_address), addr_to_str(header->last_hop_address), (sizeof(net_header_t) + header->payload_length), header->hop_count, C_MAXIMAL_NETWORK_DIAMETER);
#endif	// C_USE_INTERNAL_ROUTING

	if(header->source_address == C_LOCAL_ADDRESS) {
		C_DEBUG(C_MAIN_DEBUG_LEVEL, C_MSG_VERBOSE, "Dropping packet #%d from local node to %s (%d bytes, received from %s after %d hops)...\n", header->seqno, addr_to_str(header->destination_address), (sizeof(net_header_t) + header->payload_length), addr_to_str(header->last_hop_address), header->hop_count);
		return;
	} else if(
		(header->destination_address == C_BROADCAST_ADDRESS)
		&& check_for_duplicate(C_received_packet_cache, C_RECEIVED_PACKET_CACHE_SIZE, header))
	{
		C_DEBUG(C_MAIN_DEBUG_LEVEL, C_MSG_VERBOSE, "Dropping duplicate packet #%d from %s to %s (%d bytes, received from %s after %d hops)...\n", header->seqno, addr_to_str(header->source_address), addr_to_str(header->destination_address), (sizeof(net_header_t) + header->payload_length), addr_to_str(header->last_hop_address), header->hop_count);
		return;
	}

#ifdef C_USE_NETWORK_MAPPING
	C_network_map_may_need_purging = true;

	// Extract hops and neighbors.
	header->payload_length -= (header->hop_count + header->neighbor_count) * sizeof(C_addr_t);
	if(header->hop_count == 0)
		header->hops = NULL;
	else
		header->hops = (C_addr_t*) (((char*) payload) + header->payload_length);
	if(header->neighbor_count == 0)
		header->neighbors = NULL;
	else
		header->neighbors = (C_addr_t*) (((char*) payload) + (header->payload_length + (header->hop_count * sizeof(C_addr_t))));
#ifdef C_ENABLE_CONSISTENCY_CHECKS
	for(i = 0; i < header->hop_count; i++)
		C_ABORT_ON(!is_valid_address(header->hops[i]), "Received packet #%d from %s to %s via last hop %s (%d bytes) with invalid address for hop #%d (%s).\n", header->seqno, addr_to_str(header->source_address), addr_to_str(header->destination_address), addr_to_str(header->last_hop_address), (sizeof(net_header_t) + header->payload_length), i, addr_to_str(header->hops[i]));
	for(i = 0; i < header->neighbor_count; i++)
		C_ABORT_ON(!is_valid_address(header->neighbors[i]), "Received packet #%d from %s to %s via last hop %s (%d bytes) with invalid address for neighbor #%d (%s).\n", header->seqno, addr_to_str(header->source_address), addr_to_str(header->destination_address), addr_to_str(header->last_hop_address), (sizeof(net_header_t) + header->payload_length), i, addr_to_str(header->neighbors[i]));
#endif	// C_ENABLE_CONSISTENCY_CHECKS

	// Add local address to hop list.
	if(header->hop_count < C_MAXIMAL_NETWORK_DIAMETER) {
		MALLOC_ARRAY(hops_including_current_node, C_addr_t, (header->hop_count + 1));
		if(header->hop_count > 0)
			memcpy(hops_including_current_node, header->hops, (header->hop_count * sizeof(C_addr_t)));
		hops_including_current_node[header->hop_count] = C_LOCAL_ADDRESS;
		header->hops = hops_including_current_node;
	} else {
		C_DEBUG(C_MAIN_DEBUG_LEVEL, C_MSG_WARN, "Number of hops exceeds maximal expected number of hops (%d >= %d).\n", header->hop_count, C_MAXIMAL_NETWORK_DIAMETER);
	}
#endif	// C_USE_NETWORK_MAPPING

	// Update hop count.
	header->hop_count++;

	C_DEBUG(C_MAIN_DEBUG_LEVEL, ((header->app == APP_TYPE_INTERNAL) ? C_MSG_VERBOSE : C_MSG_DEBUG), "Received packet #%d (0x%X/0x%X) from %s to %s via last hop %s (%d bytes, %d hops)...\n", header->seqno, header->app, header->type, addr_to_str(header->source_address), addr_to_str(header->destination_address), addr_to_str(header->last_hop_address), (sizeof(net_header_t) + header->payload_length), header->hop_count);

	// Take care of network statistics, mapping and active link monitoring.
	register_received_packet(header);

	if(header->app != APP_TYPE_NONE) {
#ifdef C_USE_INTERNAL_ROUTING
		DYMO_route_recv(header, payload);
#else
		C_handle_packet(header, payload);
#endif	// C_USE_INTERNAL_ROUTING
	}

#ifdef C_USE_NETWORK_MAPPING
	FREE(hops_including_current_node);
#endif	// C_USE_NETWORK_MAPPING
}

/**
 * @brief Global function for reporting broken links called by the underlying
 * system.
 *
 * @param header Pointer to the header of the failed packet.
 * @param payload Pointer to the payload of the failed packet.
 */
void C_xmit_failure(net_header_t* header, const void* payload) {
	C_ABORT_ON(!is_valid_address(header->source_address), "Failed to transmit packet with invalid source address (%s).\n", addr_to_str(header->source_address));
	C_ABORT_ON(!is_valid_address(header->destination_address), "Failed to transmit packet with invalid destination address (%s).\n", addr_to_str(header->destination_address));
	C_ABORT_ON(!is_valid_address(header->last_hop_address), "Failed to transmit packet with invalid last hop address (%s).\n", addr_to_str(header->last_hop_address));
	C_ABORT_ON(!is_valid_address(header->next_hop_address), "Failed to transmit packet with invalid next hop address (%s).\n", addr_to_str(header->next_hop_address));
	C_ABORT_ON(header->last_hop_address != C_LOCAL_ADDRESS, "Failed to transmit packet whose last hop address is not the local node (%s != %s)\n", addr_to_str(header->source_address), addr_to_str(C_LOCAL_ADDRESS));

	C_DEBUG(C_MAIN_DEBUG_LEVEL, C_MSG_DEBUG, "Failed to transmit packet #%d from %s to %s via %s (%d bytes)...\n", header->seqno, addr_to_str(header->source_address), addr_to_str(header->destination_address), addr_to_str(header->next_hop_address), (sizeof(net_header_t) + header->payload_length));

#ifdef C_USE_NETWORK_MAPPING
	C_network_map_may_need_purging = true;
#endif	// C_USE_NETWORK_MAPPING

	if(header->next_hop_address != C_BROADCAST_ADDRESS) {
		C_report_broken_link(header->next_hop_address);

		// Update statistics counter.
		C_unicast_transmission_failures++;
	}
}

/**
 * @brief Global callback trigger function called by the underlying system.
 */
void C_execute_callback() {
	C_DEBUG(C_MAIN_DEBUG_LEVEL, C_MSG_VERBOSE, "Executing callback...\n");

#ifdef C_USE_NETWORK_MAPPING
	C_network_map_may_need_purging = true;
#endif	// C_USE_NETWORK_MAPPING

	callback_expire();
}

/**
 * @brief Global command function called by the underlying system.
 *
 * @param argc Length of the argument array.
 * @param argv Array of argument strings.
 * @return <code>true</code> if the command was processed successfully,
 * <code>false</code> otherwise.
 */
bool C_command(int argc, const char*const* argv) {
	if(argc >= 2) {
		C_DEBUG(C_MAIN_DEBUG_LEVEL, C_MSG_VERBOSE, "Received command \"%s\"...\n", argv[1]);
	} else {
		C_DEBUG(C_MAIN_DEBUG_LEVEL, C_MSG_WARN, "Received command without command string...\n");
	}

#ifdef C_USE_NETWORK_MAPPING
	C_network_map_may_need_purging = true;
#endif	// C_USE_NETWORK_MAPPING

	// Call into components.
	if(DYMO_command(argc, argv))
		return true;
	else if(ICMP_command(argc, argv))
		return true;
	else if(RUDP_command(argc, argv))
		return true;
	else if(SLP_command(argc, argv))
		return true;
	else if(SERVICE_command(argc, argv))
		return true;
	else if(PLACEMENT_command(argc, argv))
		return true;
	else if(INTERFERENCE_command(argc, argv))
		return true;
	else
		return false;
}



/*
 * Wrappers around functions provided by the underlying system.
 */

/**
 * @brief Wrapper function around low-level output function.
 *
 * @param format Format string.
 * @param ... Variable number of arguments matching the format string.
 */
void C_printf(const char *format, ...) {
	va_list argp;
	va_start(argp, format);
	__C_vprintf(format, argp);
	va_end(argp);
}

/**
 * @brief Wrapper function around low-level output function.
 *
 * @param format Format string.
 * @param ap Variable length argument structure.
 */
void C_vprintf(const char *format, va_list ap) {
	__C_vprintf(format, ap);
}

/**
 * @brief Wrapper function around low-level random function.
 *
 * @param max Maximal value.
 * @return Random value between 0 and max.
 */
double C_random(double max) {
	return __C_random(max);
}

/**
 * @brief Wrapper function around low-level time function.
 *
 * @return Current time.
 */
C_time_t C_time() {
	return __C_time();
}

/**
 * @brief Callback function for delayed sending of a packet.
 *
 * @param parameter Packet to be sent.
 */
static void C_send_callback(callback_parameter_t parameter) {
	C_packet_t* packet = (C_packet_t*) parameter;

	register_sent_packet(packet->header);
	__C_send(packet->header, packet->payload);

	free_header(&packet->header);
	FREE(packet->payload);
	FREE(packet);
	C_send_callback_id = 0;
}

/**
 * @brief Wrapper function around low-level send function.
 *
 * @param header Header of packet to send.
 * @param payload Payload of packet to send.
 */
void C_send(net_header_t* header, const void* payload) {
	C_packet_t* packet = NULL;
	C_time_t delay = 0;
#ifdef C_USE_NETWORK_MAPPING
	unsigned int i = 0;
	void* payload_with_mapping_information = NULL;
#endif	// C_USE_NETWORK_MAPPING

	C_ABORT_ON(header == NULL, "Trying to send packet without header.\n");
	C_ABORT_ON((header->payload_length > 0) && (payload == NULL), "Trying to send non-empty packet without payload.\n");
	C_ABORT_ON((sizeof(net_header_t) + header->payload_length) > C_MTU, "Trying to send packet exceeding the MTU (%d > %d).\n", (sizeof(net_header_t) + header->payload_length), C_MTU);
	C_ABORT_ON(!is_valid_address(header->source_address), "Trying to send packet with invalid source address (%s).\n", addr_to_str(header->source_address));
	C_ABORT_ON(!is_valid_address(header->destination_address), "Trying to send packet with invalid destination address (%s).\n", addr_to_str(header->destination_address));
	C_ABORT_ON(!is_valid_address(header->last_hop_address), "Trying to send packet with invalid last hop address (%s).\n", addr_to_str(header->last_hop_address));
	C_ABORT_ON(!is_valid_address(header->next_hop_address), "Trying to send packet with invalid next hop address (%s).\n", addr_to_str(header->next_hop_address));
	C_ABORT_ON(header->last_hop_address != C_LOCAL_ADDRESS, "Trying to send packet with last hop other than local node.\n");
	C_ABORT_ON(header->next_hop_address == C_LOCAL_ADDRESS, "Trying to send packet with the local node as next hop.\n");
	C_ABORT_ON((header->ttl > C_DEFAULT_TTL) && (header->ttl != 255), "Trying to send packet with TTL %d which is greater than the default TTL %d.\n", header->ttl, C_DEFAULT_TTL);	// TTL = 255 is used by DYMO.
	C_ABORT_ON(header->hop_count >= C_MAXIMAL_NETWORK_DIAMETER, "Trying to send packet which already has been sent over the number of hops of the longest expected route (%d >= %d).\n", header->hop_count, C_MAXIMAL_NETWORK_DIAMETER);
#ifdef C_USE_INTERNAL_ROUTING
	// The performance for unicast and broadcast packets on the DLL may
	// differ. Hence, we make sure that we only broadcast (DLL) packets
	// that are addressed to the broadcast (NET) address.
	C_ABORT_ON((header->next_hop_address == C_BROADCAST_ADDRESS) && (header->destination_address != C_BROADCAST_ADDRESS), "Trying to broadcast (DLL) a unicast (NET) packet.\n");
#endif	// C_USE_INTERNAL_ROUTING

#ifdef C_USE_NETWORK_MAPPING
	C_IF_DEBUG(C_MAIN_DEBUG_LEVEL, C_MSG_WARN) {
#ifdef C_ENABLE_BEACONING
#ifdef C_ASSUME_BIDIRECTIONAL_LINKS_FOT_NETWORK_MAP
		// Only warn on unknown links if beaconing is enabled, as
		// otherwise the mapping information gets stale too quickly.
		if(
			(header->next_hop_address != C_BROADCAST_ADDRESS)
			&& !exists_known_link(C_LOCAL_ADDRESS, header->next_hop_address)
		) {
			C_DEBUG(C_MAIN_DEBUG_LEVEL, C_MSG_WARN, "Sending packet #%d from %s to %s via next hop %s, but link to next hop is unknown...\n", header->seqno, addr_to_str(header->source_address), addr_to_str(header->destination_address), addr_to_str(header->next_hop_address));
		}
#endif	// C_ASSUME_BIDIRECTIONAL_LINKS_FOT_NETWORK_MAP
#endif	// C_ENABLE_BEACONING

		for(i = 0; i < header->hop_count; i++) {
			C_ABORT_ON(!is_valid_address(header->hops[i]), "Trying to send packet with invalid address for hop #%d (%s).\n", i, addr_to_str(header->hops[i]));
			if(header->hops[i] == header->next_hop_address) {
				C_DEBUG(C_MAIN_DEBUG_LEVEL, C_MSG_WARN, "Sending packet #%d from %s to %s via next hop %s, which has previously been forwarded by the next hop...\n", header->seqno, addr_to_str(header->source_address), addr_to_str(header->destination_address), addr_to_str(header->next_hop_address));
				break;
			}
		}
	}
#endif	// C_USE_NETWORK_MAPPING

	C_DEBUG(C_MAIN_DEBUG_LEVEL, ((header->app == APP_TYPE_INTERNAL) ? C_MSG_VERBOSE : C_MSG_DEBUG), "Sending packet #%d (0x%X/0x%X) from %s to %s via next hop %s (%d bytes, %d hops)...\n", header->seqno, header->app, header->type, addr_to_str(header->source_address), addr_to_str(header->destination_address), addr_to_str(header->next_hop_address), (sizeof(net_header_t) + header->payload_length), header->hop_count);

	// Add broadcast packets with TTL > 0 to the received packet cache to
	// ignore retransmissions of neighboring nodes.
	if((header->destination_address == C_BROADCAST_ADDRESS) && (header->ttl > 0))
		check_for_duplicate(C_received_packet_cache, C_RECEIVED_PACKET_CACHE_SIZE, header);

#ifdef C_USE_NETWORK_MAPPING
#ifdef C_ENABLE_CONSISTENCY_CHECKS
	for(i = 0; i < header->hop_count; i++)
		C_ABORT_ON(!is_valid_address(header->hops[i]), "Trying to send packet with invalid address for hop #%d (%s).\n", i, addr_to_str(header->hops[i]));
	for(i = 0; i < header->neighbor_count; i++)
		C_ABORT_ON(!is_valid_address(header->neighbors[i]), "Trying to send packet with invalid address for neighbor #%d (%s).\n", i, addr_to_str(header->neighbors[i]));
#endif	// C_ENABLE_CONSISTENCY_CHECKS

	// Store hops and neighbors.
	MALLOC_ARRAY(payload_with_mapping_information, char, (header->payload_length + ((header->hop_count + header->neighbor_count) * sizeof(C_addr_t))));
	memcpy(payload_with_mapping_information, payload, header->payload_length);
	memcpy(
		(((char*) payload_with_mapping_information) + header->payload_length),
		header->hops,
		(header->hop_count * sizeof(C_addr_t))
	);
	memcpy(
		(((char*) payload_with_mapping_information) + (header->payload_length + (header->hop_count * sizeof(C_addr_t)))),
		header->neighbors,
		(header->neighbor_count * sizeof(C_addr_t))
	);
#ifdef C_ENABLE_CONSISTENCY_CHECKS
	for(i = 0; i < (header->hop_count + header->neighbor_count); i++)
		C_ABORT_ON(!is_valid_address(*(((C_addr_t*) (((char*) payload_with_mapping_information) + header->payload_length)) + i)), "Trying to send packet with invalid address at position %d in payload (%s).\n", i, addr_to_str(*(((C_addr_t*) (((char*) payload_with_mapping_information) + header->payload_length)) + i)));
#endif	// C_ENABLE_CONSISTENCY_CHECKS
	header->payload_length += (header->hop_count + header->neighbor_count) * sizeof(C_addr_t);

	// Count how many fields in the header that are used for network
	// mapping.
	C_network_mapping_overhead += (header->hop_count + header->neighbor_count) * sizeof(C_addr_t);
#endif	// C_USE_NETWORK_MAPPING

#ifdef C_ADD_SIMULATED_PROCESSING_TIME
	// Add some processing time on simulated nodes.
	delay = ((C_time_t) C_NODE_TRAVERSAL_TIME / 2.0);
#else
	delay = 0;
#endif	// C_ADD_SIMULATED_PROCESSING_TIME

#ifdef C_USE_INTERNAL_ROUTING
	// Add some jitter to broadcast packets.
	if(header->destination_address == C_BROADCAST_ADDRESS)
		delay += ((C_time_t) C_NODE_TRAVERSAL_TIME) * C_random(10.0);
#endif	// C_USE_INTERNAL_ROUTING

	if(delay > 0) {
		MALLOC(packet, C_packet_t);
		alloc_header_from_header(&packet->header, header);
#ifndef C_USE_NETWORK_MAPPING
		MALLOC_ARRAY(packet->payload, char, header->payload_length);
		memcpy(packet->payload, payload, header->payload_length);
#else
		packet->payload = payload_with_mapping_information;
#endif	// C_USE_NETWORK_MAPPING
		C_send_callback_id = callback_add(C_send_callback, packet, delay);
	} else {
		register_sent_packet(header);
#ifndef C_USE_NETWORK_MAPPING
		__C_send(header, payload);
#else
		__C_send(header, payload_with_mapping_information);
		FREE(payload_with_mapping_information);
#endif	// C_USE_NETWORK_MAPPING
	}

#ifdef C_USE_ETX_LINK_METRIC
	// Register link as broken if indicated by high ETX value.
	if(
		(header->next_hop_address != C_BROADCAST_ADDRESS)
		&& exists_known_node(header->next_hop_address)
		&& (calculate_ETX(header->next_hop_address) > C_ETX_BROKEN_LINK_THRESHOLD)
	) {
		C_report_broken_link(header->next_hop_address);
	}
#endif	// C_USE_ETX_LINK_METRIC

	// Update statistics counters.
	if(header->destination_address == C_BROADCAST_ADDRESS)
		C_transmitted_broadcast_packets++;
	else
		C_transmitted_unicast_packets++;
	C_transmitted_bytes += (sizeof(net_header_t) + header->payload_length);
	add_bytes_to_datarate_stat(C_transmitted_datarate_stat, (sizeof(net_header_t) + header->payload_length));
}

/**
 * @brief Wrapper function around low-level send function with two payloads.
 *
 * @param header Header of packet to send.
 * @param payload1 Payload of packet to send.
 * @param payload_length1 Length of second payload in bytes.
 * @param payload2 Pointer to second payload.
 * @param payload_length2 Length of second payload in bytes.
 */
void C_send2(net_header_t* header, const void* payload1, size_t payload_length1, const void* payload2, size_t payload_length2) {
	void* payload = NULL;

	C_ABORT_ON(((payload1 == NULL) && (payload_length1 != 0)) || ((payload1 != NULL) && (payload_length1 == 0)) || ((payload2 == NULL) && (payload_length2 != 0)) || ((payload2 != NULL) && (payload_length2 == 0)), "Trying to assemble packet for sending, but payload parameters are inconsistent.\n");

	header->payload_length = payload_length1 + payload_length2;
	MALLOC_JOIN(payload, char, payload1, payload_length1, payload2, payload_length2);
	C_send(header, payload);
	FREE(payload);
}

/**
 * @brief Wrapper function around low-level callback activation function.
 *
 * @param time Time at which the next callback needs to be activated, or 0.0 if
 * no callback needs to be called.
 */
void C_activate_callback(C_time_t time) {
	C_time_t current_time = C_time();

	if(
		(time != 0.0)
		&& ((current_time - time) > 0.01)
	) {
#ifdef ARCH_NS2
		C_DEBUG(C_MAIN_DEBUG_LEVEL, C_MSG_ERROR, "Trying to activate callback %.2fs in the past (%f < %f).\n", (current_time - time), time, current_time);
		return;
#else
		// The currently active callback may occur in the past if
		// running on real hardware due to processing overhead.
		C_DEBUG(C_MAIN_DEBUG_LEVEL, C_MSG_DEBUG, "Activating callback %.2fs in the past (%.2f < %.2f).\n", (current_time - time), time, current_time);
#endif	// ARCH_NS2
	}
	__C_activate_callback(time);
}

#ifdef C_ENABLE_NAM_TRACE_OUTPUT
/**
 * @brief Wrapper function around low-level trace function.
 *
 * @param format Format string.
 * @param ... Variable number of arguments matching the format string.
 */
void C_trace(const char *format, ...) {
	va_list argp;
	va_start(argp, format);
	__C_trace(format, argp);
	va_end(argp);
}
#endif	// C_ENABLE_NAM_TRACE_OUTPUT



/*
 * Internal interface support functions.
 */

/**
 * @brief Internal packet reception function.
 *
 * This function may be called either when packets arrive from the underlying
 * system or if a (possibly encapsulated) packet needs to be processed again.
 *
 * @param header Pointer to the header of the received packet.
 * @param payload Pointer to the payload of the received packet.
 */
void C_handle_packet(const net_header_t* header, const void* payload) {
	C_ABORT_ON(!is_valid_address(header->source_address), "Trying to handle packet with invalid source address (%s).\n", addr_to_str(header->source_address));
	C_ABORT_ON(!is_valid_address(header->destination_address), "Trying to handle packet with invalid destination address (%s).\n", addr_to_str(header->destination_address));
	C_ABORT_ON(!is_valid_address(header->last_hop_address), "Trying to handle packet with invalid last hop address (%s).\n", addr_to_str(header->last_hop_address));
	C_ABORT_ON(!is_valid_address(header->next_hop_address), "Trying to handle packet with invalid next hop address (%s).\n", addr_to_str(header->next_hop_address));

	C_DEBUG(C_MAIN_DEBUG_LEVEL, C_MSG_VERBOSE, "Handling packet #%d from %s to %s (%d bytes, received from %s after %d hops)...\n", header->seqno, addr_to_str(header->source_address), addr_to_str(header->destination_address), (sizeof(net_header_t) + header->payload_length), addr_to_str(header->last_hop_address), header->hop_count);

	// Count unicast packets sorted by hop counts as received on this node.
	// NOTE: RUDP DATA packets are not counted since this function is
	// called again with the payload of these packets.
	if(
		(header->destination_address != C_BROADCAST_ADDRESS)
		&& !((header->app == APP_TYPE_RUDP) && (header->type == RUDP_PACKET_TYPE_DATA))
		&& (header->hop_count <= C_MAXIMAL_NETWORK_DIAMETER)
	) {
		C_unicast_packets_per_hop_count[header->hop_count]++;
	}

	switch(header->app) {
	case APP_TYPE_INTERNAL:
		switch(header->type) {
			case PACKET_TYPE_BEACON:
#ifdef C_ENABLE_BEACONING
				handle_BEACON(header, payload);
#endif	// C_ENABLE_BEACONING
				break;
			default:
				C_ABORT_ON(true, "Received internal packet of unknown type 0x%X from %s.\n", header->type, addr_to_str(header->source_address));
		}
		break;
	case APP_TYPE_DYMO:
		DYMO_recv(header, payload);
		break;
	case APP_TYPE_ICMP:
		ICMP_recv(header, payload);
		break;
	case APP_TYPE_RUDP:
		RUDP_recv(header, payload);
		break;
	case APP_TYPE_SLP:
		SLP_recv(header, payload);
		break;
	case APP_TYPE_SERVICE:
		SERVICE_recv(header, payload);
		break;
	case APP_TYPE_PLACEMENT:
		PLACEMENT_recv(header, payload);
		break;
	case APP_TYPE_INTERFERENCE:
		INTERFERENCE_recv(header, payload);
		break;
	default:
		C_DEBUG(C_MAIN_DEBUG_LEVEL, C_MSG_ERROR, "Received packet for unknown application 0x%X from %s.\n", header->app, addr_to_str(header->source_address));
	}
}

/**
 * @brief Internal packet routing function.
 *
 * Establish a route to the destination node and send the packet to the first
 * hop. A packet addressed to the local node is delivered locally.
 *
 * @param header Pointer to the header of the packet.
 * @param payload Pointer to the payload of the packet.
 */
void C_route(net_header_t* header, const void* payload) {
#ifdef C_USE_NETWORK_MAPPING
#ifdef C_ENABLE_CONSISTENCY_CHECKS
	unsigned int i = 0;
	for(i = 0; i < header->hop_count; i++)
		C_ABORT_ON(!is_valid_address(header->hops[i]), "Trying to route packet with invalid address for hop #%d (%s).\n", i, addr_to_str(header->hops[i]));
	for(i = 0; i < header->neighbor_count; i++)
		C_ABORT_ON(!is_valid_address(header->neighbors[i]), "Trying to route packet with invalid address for neighbor #%d (%s).\n", i, addr_to_str(header->neighbors[i]));
#endif	// C_ENABLE_CONSISTENCY_CHECKS
#endif	// C_USE_NETWORK_MAPPING

	// Make sure that unicast packets originated on this node. Broadcast
	// packets may have originated else where since they may be forwarded
	// by the application and the routing protocol doesn't establish a
	// route for them.
	C_ABORT_ON((header->destination_address != C_BROADCAST_ADDRESS) && (header->source_address != C_LOCAL_ADDRESS), "Trying to route unicast packet with source other than the local node (%s).\n", addr_to_str(header->source_address));

	C_DEBUG(C_MAIN_DEBUG_LEVEL, C_MSG_VERBOSE, "Routing packet #%d from %s to %s (%d bytes, received from %s after %d hops)...\n", header->seqno, addr_to_str(header->source_address), addr_to_str(header->destination_address), (sizeof(net_header_t) + header->payload_length), addr_to_str(header->last_hop_address), header->hop_count);

	if(header->destination_address == C_LOCAL_ADDRESS) {
		C_handle_packet(header, payload);
	} else {
		header->ttl--;
#ifdef C_USE_INTERNAL_ROUTING
		DYMO_route(header, payload);
#else
		C_send(header, payload);
#endif	// C_USE_INTERNAL_ROUTING
	}
}

/**
 * @brief Report a broken DLL link.
 *
 * @param address Address of the node to which the link is broken.
 */
void C_report_broken_link(C_addr_t address) {
	C_DEBUG(C_MAIN_DEBUG_LEVEL, C_MSG_DEBUG, "Link to %s is broken.\n", addr_to_str(address));
	register_broken_link(C_LOCAL_ADDRESS, address);
#ifdef C_USE_INTERNAL_ROUTING
	DYMO_handle_unreachable_next_hop(address);
#endif	// C_USE_INTERNAL_ROUTING
}

/**
 * @brief Calculate the maximal DLL (layer 2) payload size.
 *
 * @return Maximal payload size in bytes.
 */
size_t C_max_dll_payload_size() {
#ifdef C_USE_NETWORK_MAPPING
	return (C_MTU - sizeof(net_header_t) - ((C_MAXIMAL_NETWORK_DIAMETER + 1 + C_MAXIMAL_NODE_DEGREE) * sizeof(C_addr_t)));
#else
	return (C_MTU - sizeof(net_header_t));
#endif	// C_USE_NETWORK_MAPPING
}

/**
 * @brief Calculate the maximal NET (layer 3) payload size.
 *
 * @return Maximal payload size in bytes.
 */
size_t C_max_net_payload_size() {
#ifdef C_USE_INTERNAL_ROUTING
	return DYMO_max_payload_size();
#else
	return C_max_dll_payload_size();
#endif	// C_USE_INTERNAL_ROUTING
}
