/**
 * @file interference_local.c
 *
 * @brief Interference graphs - functions local to each node.
 *
 * Copyright (c) 2007, 2008, 2009, 2010 Georg Wittenburg, Freie Universitaet
 * Berlin (wittenbu@inf.fu-berlin.de). All rights reserved.
 * Please refer to the COPYING file for detailed licensing information.
 *
 * @author Georg Wittenburg
 *
 * $Id: interference_local.c 3292 2010-02-14 02:02:41Z gwitten $
 */

#include <stdlib.h>
#include <string.h>

#include "../config.h"
#include "../macros.h"
#include "../types.h"
#include "../vars.h"
#include "../main.h"
#include "../util/debug.h"
#include "../util/list.h"
#include "../util/callback.h"
#include "../util/net.h"
#include "../icmp-light/icmp-light.h"
#include "../rudp-light/rudp-light.h"
#include "interference.h"
#include "interference_types.h"
#include "interference_vars.h"
#include "interference_local.h"



/*
 * Utility functions.
 */

/**
 * @brief Allocate and initialize a local data structure.
 *
 * @param local_data Pointer to pointer to local data structure to allocate.
 * @param coordinator_address Address of the coordinator node.
 */
static void alloc_local_data(INTERFERENCE_local_data_t** local_data, C_addr_t coordinator_address) {
	MALLOC(*local_data, INTERFERENCE_local_data_t);
	(*local_data)->coordinator_address = coordinator_address;
	list_alloc(&(*local_data)->data_nodes);
	(*local_data)->connectivity_samples = 0;
	list_alloc(&(*local_data)->newly_discovered_nodes);
	(*local_data)->active_data_node = NULL;
	(*local_data)->max_connectivity_count = 0;
	(*local_data)->max_connectivity_node_address = C_BROADCAST_ADDRESS;
	(*local_data)->max2_connectivity_count = 0;
	(*local_data)->max2_connectivity_node_address = C_BROADCAST_ADDRESS;
	(*local_data)->active_data_node = NULL;
	(*local_data)->active_test_id = 0;
	(*local_data)->reference_node_address = C_BROADCAST_ADDRESS;
	(*local_data)->reference_connectivity_count = 0;
	(*local_data)->last_interference_packet_source = C_BROADCAST_ADDRESS;
	(*local_data)->current_test_invalid = false;
}

/*
 * @brief Free a local data structure.
 *
 * @param local_data Pointer to pointer to local data structure to free.
 */
/* Unused, but retained for interface completeness.
static void free_local_data(INTERFERENCE_local_data_t** local_data) {
	free_local_data_list(*local_data);
	*local_data = NULL;
}
*/

/**
 * @brief Free a local data structure.
 *
 * @param p Pointer to local data structure to free.
 */
void free_local_data_list(void* p) {
	INTERFERENCE_local_data_t* local_data = (INTERFERENCE_local_data_t*) p;
	list_free_including_data(&local_data->data_nodes, free);
	list_free_including_data(&local_data->newly_discovered_nodes, free);
	FREE(local_data);
}



/*
 * Packet handling functions.
 */

/**
 * @brief Send a INTERFERENCE REPLY packet.
 *
 * @param to Address of the destination node.
 * @param data_nodes List of nodes with data on network conditions.
 */
static void send_INTERFERENCE_REPLY(C_addr_t to, list_t* data_nodes) {
	INTERFERENCE_local_data_node_t* node_array = NULL;

	C_DEBUG(INTERFERENCE_DEBUG_LEVEL, C_MSG_DEBUG, "INTERFERENCE: Sending INTERFERENCE REPLY to %s...\n", addr_to_str(to));
	MALLOC_ARRAY(node_array, INTERFERENCE_local_data_node_t, list_size(data_nodes));
	list_to_array(data_nodes, node_array, sizeof(INTERFERENCE_local_data_node_t));
	RUDP_sendto(to, PACKET_PRIORITY_SIGNALLING_SERVICE, APP_TYPE_INTERFERENCE, INTERFERENCE_PACKET_TYPE_REPLY, node_array, (list_size(data_nodes) * sizeof(INTERFERENCE_local_data_node_t)), NULL, NULL);
	FREE(node_array);
}

/**
 * @brief Send a INTERFERENCE RESET packet.
 *
 * @param to Address of the destination node.
 * @param newly_discovered_nodes List of newly discovered nodes.
 */
static void send_INTERFERENCE_RESET(C_addr_t to, list_t* newly_discovered_nodes) {
	node_t* node_array = NULL;

	C_DEBUG(INTERFERENCE_DEBUG_LEVEL, C_MSG_DEBUG, "INTERFERENCE: Sending INTERFERENCE RESET to %s...\n", addr_to_str(to));
	MALLOC_ARRAY(node_array, node_t, list_size(newly_discovered_nodes));
	list_to_array(newly_discovered_nodes, node_array, sizeof(node_t));
	RUDP_sendto(to, PACKET_PRIORITY_SIGNALLING_SERVICE, APP_TYPE_INTERFERENCE, INTERFERENCE_PACKET_TYPE_RESET, node_array, (list_size(newly_discovered_nodes) * sizeof(node_t)), NULL, NULL);
	FREE(node_array);
}

/**
 * @brief RUDP callback for a INTERFERENCE REST REQUEST packet.
 *
 * @param parameter Pointer to RUDP transmission result.
 */
static void send_INTERFERENCE_TEST_REQUEST_callback(callback_parameter_t parameter) {
	RUDP_transmission_result_t* result = (RUDP_transmission_result_t*) parameter;
	INTERFERENCE_local_data_t* local_data = (INTERFERENCE_local_data_t*) result->parameter;

	if(result->error_code == C_ERROR_OK) {
		C_DEBUG(INTERFERENCE_DEBUG_LEVEL, C_MSG_DEBUG, "INTERFERENCE: INTERFERENCE TEST REQUEST was successful.\n");
	} else {
		C_DEBUG(INTERFERENCE_DEBUG_LEVEL, C_MSG_WARN, "INTERFERENCE: INTERFERENCE TEST REQUEST failed (%d), invalidating current test...\n", result->error_code);
		local_data->current_test_invalid = true;
	}
	FREE(result);
}

/**
 * @brief Send a INTERFERENCE TEST REQUEST packet.
 *
 * @param to Address of destination node.
 * @param test_id ID of test.
 * @param delay Delay before starting the test.
 * @param packets Number of test packets.
 * @param interval Time interval between test packets.
 * @param duration Duration of the test (used if <code>packets</code> is set to 0).
 * @param local_data Pointer to local test data.
 */
static void send_INTERFERENCE_TEST_REQUEST(C_addr_t to, uint16_t test_id, C_time_t delay, uint16_t packets, C_time_t interval, C_time_t duration, INTERFERENCE_local_data_t* local_data) {
	INTERFERENCE_TEST_REQUEST_packet_t test_request;
	RUDP_transmission_result_t* result = NULL;

	C_DEBUG(INTERFERENCE_DEBUG_LEVEL, C_MSG_DEBUG, "INTERFERENCE: Sending INTERFERENCE TEST REQUEST to %s...\n", addr_to_str(to));
	memset(&test_request, 0, sizeof(INTERFERENCE_TEST_REQUEST_packet_t));
	test_request.test_id = test_id;
	test_request.delay = delay;
	test_request.packets = packets;
	test_request.interval = interval;
	test_request.duration = duration;
	MALLOC(result, RUDP_transmission_result_t);
	result->parameter = local_data;
	RUDP_sendto(to, PACKET_PRIORITY_SIGNALLING_SERVICE, APP_TYPE_INTERFERENCE, INTERFERENCE_PACKET_TYPE_TEST_REQUEST, &test_request, sizeof(INTERFERENCE_TEST_REQUEST_packet_t), send_INTERFERENCE_TEST_REQUEST_callback, result);
}

/**
 * @brief Send a INTERFERENCE TEST DATA packet.
 *
 * @param test_id ID of test.
 * @param node_address Address of the node that requested the test.
 */
static void send_INTERFERENCE_TEST_DATA(uint16_t test_id, C_addr_t node_address) {
	INTERFERENCE_TEST_DATA_packet_t test_data;
	char* payload = NULL;

	C_DEBUG(INTERFERENCE_DEBUG_LEVEL, C_MSG_VERBOSE, "INTERFERENCE: Broadcasting INTERFERENCE TEST DATA as requested by %s...\n", addr_to_str(node_address));
	memset(&test_data, 0, sizeof(INTERFERENCE_TEST_DATA_packet_t));
	test_data.test_id = test_id;
	test_data.node_address = node_address;
	MALLOC_ARRAY(payload, char, (INTERFERENCE_INTERFERENCE_PAYLOAD_SIZE - sizeof(INTERFERENCE_TEST_DATA_packet_t)));
	C_sendto2(C_BROADCAST_ADDRESS, PACKET_PRIORITY_SIGNALLING_SERVICE, APP_TYPE_INTERFERENCE, INTERFERENCE_PACKET_TYPE_TEST_DATA, &test_data, sizeof(INTERFERENCE_TEST_DATA_packet_t), payload, (INTERFERENCE_INTERFERENCE_PAYLOAD_SIZE - sizeof(INTERFERENCE_TEST_DATA_packet_t)));
	FREE(payload);
}

static void collect_interference_data(INTERFERENCE_local_data_t* local_data);

/**
 * @brief Callback for gathering interference information.
 *
 * @param parameter Callback parameter.
 */
static void collect_interference_data_timeout(callback_parameter_t parameter) {
	INTERFERENCE_local_data_t* local_data = (INTERFERENCE_local_data_t*) parameter;

	if(local_data->last_interference_packet_source == local_data->reference_node_address) {
		C_DEBUG(INTERFERENCE_DEBUG_LEVEL, C_MSG_WARN, "INTERFERENCE: Last interference packet was received from reference %s instead of active %s.\n", addr_to_str(C_INTERFERENCE_local_data->reference_node_address), addr_to_str(C_INTERFERENCE_local_data->active_data_node->node.address));
	}

	if(!local_data->current_test_invalid) {
		local_data->active_data_node->interference = (((float) local_data->active_data_node->interference_count) / ((float) local_data->reference_connectivity_count));
		C_DEBUG(INTERFERENCE_DEBUG_LEVEL, C_MSG_INFO, "INTERFERENCE: Interference test with %s complete (%.0f%%).\n", addr_to_str(local_data->active_data_node->node.address), (local_data->active_data_node->interference * 100));
	} else {
		C_DEBUG(INTERFERENCE_DEBUG_LEVEL, C_MSG_WARN, "INTERFERENCE: Interference test with %s complete, but invalid. Skipping node...\n", addr_to_str(local_data->active_data_node->node.address));
		local_data->active_data_node->interference = -0.01f;
		local_data->current_test_invalid = false;
	}
	local_data->active_test_id++;
	local_data->last_interference_packet_source = C_BROADCAST_ADDRESS;
	collect_interference_data(local_data);
}

/**
 * @brief Check whether a local data node has no interference information.
 *
 * @param item Local data node to check.
 * @param UNUSED Unused.
 * @return <code>true</code> if the data node has no interference information,
 * <code>false</code> otherwise.
 */
static bool local_data_node_without_interference(const void* item, const void* UNUSED) {
	INTERFERENCE_local_data_node_t* data_node = (INTERFERENCE_local_data_node_t*) item;
	return (data_node->interference == -1);
}

/**
 * @brief Gather interference information.
 *
 * @param local_data Data structure for keeping the locally collected data.
 */
static void collect_interference_data(INTERFERENCE_local_data_t* local_data) {
	INTERFERENCE_local_data_node_t* data_node = NULL;
	data_node = (INTERFERENCE_local_data_node_t*) list_find(local_data->data_nodes, NULL, local_data_node_without_interference);
	if(data_node != NULL) {
		local_data->active_data_node = data_node;
		if(local_data->active_data_node->node.address != local_data->max_connectivity_node_address) {
			// Use best connected node as reference.
			local_data->reference_node_address = local_data->max_connectivity_node_address;
			local_data->reference_connectivity_count = local_data->max_connectivity_count;
		} else if(local_data->max2_connectivity_node_address != C_BROADCAST_ADDRESS) {
			// Use second best connected node as reference.
			local_data->reference_node_address = local_data->max2_connectivity_node_address;
			local_data->reference_connectivity_count = local_data->max2_connectivity_count;
		} else {
			// We don't have a second best connected node.
			C_DEBUG(INTERFERENCE_DEBUG_LEVEL, C_MSG_WARN, "INTERFERENCE: Skipping interference with %s, because no reference node is avaible.\n", addr_to_str(data_node->node.address));
			local_data->active_data_node->interference = -0.01f;
			collect_interference_data(local_data);
			return;
		}
		C_DEBUG(INTERFERENCE_DEBUG_LEVEL, C_MSG_INFO, "INTERFERENCE: Testing interference with %s using %s as reference...\n", addr_to_str(data_node->node.address), addr_to_str(local_data->reference_node_address));
		local_data->active_data_node->interference_count = local_data->reference_connectivity_count;
		send_INTERFERENCE_TEST_REQUEST(local_data->active_data_node->node.address, local_data->active_test_id, INTERFERENCE_INTERFERENCE_TEST_DELAY, 0, INTERFERENCE_INTERFERENCE_TEST_INTERVAL, INTERFERENCE_INTERFERENCE_TEST_DURATION, local_data);
		send_INTERFERENCE_TEST_REQUEST(local_data->reference_node_address, local_data->active_test_id, INTERFERENCE_INTERFERENCE_REFERENCE_DELAY, INTERFERENCE_INTERFERENCE_SAMPLES, INTERFERENCE_INTERFERENCE_REFERENCE_INTERVAL, 0, local_data);
		callback_add(collect_interference_data_timeout, local_data, (C_TIMEOUT + INTERFERENCE_INTERFERENCE_TEST_DURATION));
	} else {
		C_DEBUG(INTERFERENCE_DEBUG_LEVEL, C_MSG_INFO, "INTERFERENCE: Interference tests on %s complete, sending data to coordinator node...\n", addr_to_str(C_LOCAL_ADDRESS));
		send_INTERFERENCE_REPLY(local_data->coordinator_address, local_data->data_nodes);
		list_free_including_data(&local_data->data_nodes, free);
		list_free_including_data(&local_data->newly_discovered_nodes, free_node_list);
		FREE(local_data);
		C_INTERFERENCE_local_data = NULL;
	}
}

/**
 * @brief Check whether two local data nodes have the same address.
 *
 * @param item Data node to check.
 * @param cmp_item Reference data node.
 * @return <code>true</code> if the data nodes have the same address,
 * <code>false</code> otherwise.
 */
static bool cmp_local_data_node_addr(const void* item, const void* cmp_item) {
	INTERFERENCE_local_data_node_t* data_node = (INTERFERENCE_local_data_node_t*) item;
	INTERFERENCE_local_data_node_t* cmp_data_node = (INTERFERENCE_local_data_node_t*) cmp_item;
	return (data_node->node.address == cmp_data_node->node.address);
}

static void collect_connectivity_data(INTERFERENCE_local_data_t* local_data);

/**
 * @brief Callback for the echo request that gathers connectivity information.
 *
 * @param parameter The results of the echo request.
 */
static void collect_connectivity_data_callback(callback_parameter_t parameter) {
	ICMP_echo_request_result_t* echo_request_result = (ICMP_echo_request_result_t*) parameter;
	INTERFERENCE_local_data_t* local_data = (INTERFERENCE_local_data_t*) echo_request_result->parameter;

	switch(echo_request_result->error_code) {
	case C_ERROR_OK:
		while(!list_empty(echo_request_result->nodes)) {
			node_t* node = (node_t*) list_head(echo_request_result->nodes);
			INTERFERENCE_local_data_node_t* data_node = NULL;
			INTERFERENCE_local_data_node_t cmp_data_node;
			cmp_data_node.node.address = node->address;
			data_node = (INTERFERENCE_local_data_node_t*) list_find(local_data->data_nodes, &cmp_data_node, cmp_local_data_node_addr);
			if(data_node == NULL) {
				C_DEBUG(INTERFERENCE_DEBUG_LEVEL, C_MSG_DEBUG, "INTERFERENCE: Received connectivity test packet from %s, but it is not in the local data list...\n", addr_to_str(node->address));
				if(list_find(local_data->newly_discovered_nodes, node, matches_node_address) == NULL) {
					list_add(local_data->newly_discovered_nodes, node);
					list_remove(echo_request_result->nodes, node);
				} else {
					list_remove(echo_request_result->nodes, node);
					FREE(node);
				}
			} else {
				C_DEBUG(INTERFERENCE_DEBUG_LEVEL, C_MSG_DEBUG, "INTERFERENCE: Received connectivity test packet from %s.\n", addr_to_str(data_node->node.address));
				data_node->connectivity_count++;
				list_remove(echo_request_result->nodes, node);
				FREE(node);
			}
		}
		break;
	case C_ERROR_TIMEOUT:
		C_DEBUG(INTERFERENCE_DEBUG_LEVEL, C_MSG_DEBUG, "INTERFERENCE: Connectivity test packet from timed out.\n");
		break;
	default:
		C_ABORT_ON(true, "INTERFERENCE: Callback with unknown error code 0x%X.\n", echo_request_result->error_code);
	}
	list_free_including_data(&echo_request_result->nodes, free_node_list);
	FREE(echo_request_result);

	collect_connectivity_data(local_data);
}

/**
 * @brief Gather connectivity information.
 *
 * Send local ECHO broadcasts and count which nodes reply.
 *
 * @param local_data Data structure for keeping the locally collected data.
 */
static void collect_connectivity_data(INTERFERENCE_local_data_t* local_data) {
	if(local_data->connectivity_samples < INTERFERENCE_CONNECTIVITY_SAMPLES) {
		ICMP_echo_request_result_t* echo_request_result = NULL;
		local_data->connectivity_samples++;
		C_DEBUG(INTERFERENCE_DEBUG_LEVEL, C_MSG_DEBUG, "INTERFERENCE: Sending connectivity test packet #%d...\n", local_data->connectivity_samples);
		MALLOC(echo_request_result, ICMP_echo_request_result_t);
		list_alloc(&echo_request_result->nodes);
		echo_request_result->parameter = local_data;
		ICMP_send_echo_request(C_BROADCAST_ADDRESS, 1, collect_connectivity_data_callback, echo_request_result);
	} else {
		INTERFERENCE_local_data_node_t* data_node = NULL;
		list_iterator_t data_node_iterator;

		if(!list_empty(local_data->newly_discovered_nodes)) {
			C_DEBUG(INTERFERENCE_DEBUG_LEVEL, C_MSG_WARN, "INTERFERENCE: Connectivity tests at %s complete, %d new node(s) have been discovered: ", addr_to_str(C_LOCAL_ADDRESS), list_size(local_data->newly_discovered_nodes));
			list_print(INTERFERENCE_DEBUG_LEVEL, (C_MSG_WARN | C_MSG_CONT), local_data->newly_discovered_nodes, print_node_address);
			C_DEBUG(INTERFERENCE_DEBUG_LEVEL, (C_MSG_WARN | C_MSG_CONT), "\n");
			send_INTERFERENCE_RESET(local_data->coordinator_address, local_data->newly_discovered_nodes);
			list_free_including_data(&local_data->data_nodes, free);
			list_free_including_data(&local_data->newly_discovered_nodes, free_node_list);
			FREE(local_data);
			C_INTERFERENCE_local_data = NULL;
			return;
		}
		C_DEBUG(INTERFERENCE_DEBUG_LEVEL, C_MSG_INFO, "INTERFERENCE: Connectivity tests complete.\n");

		// Establish the reference node and its backup.
		list_iterator_init(local_data->data_nodes, &data_node_iterator);
		while((data_node = (INTERFERENCE_local_data_node_t*) list_iterator_next(&data_node_iterator)) != NULL) {
			if(
				(data_node->node.address != C_LOCAL_ADDRESS)
				&& (data_node->connectivity_count > local_data->max_connectivity_count)
			) {
				local_data->max_connectivity_count = data_node->connectivity_count;
				local_data->max_connectivity_node_address = data_node->node.address;
			}
		}
		list_iterator_init(local_data->data_nodes, &data_node_iterator);
		while((data_node = (INTERFERENCE_local_data_node_t*) list_iterator_next(&data_node_iterator)) != NULL) {
			if(
				(data_node->node.address != C_LOCAL_ADDRESS)
				&& (data_node->node.address != local_data->max_connectivity_node_address)
				&& (data_node->connectivity_count > local_data->max2_connectivity_count)
			) {
				local_data->max2_connectivity_count = data_node->connectivity_count;
				local_data->max2_connectivity_node_address = data_node->node.address;
			}
		}

		C_DEBUG(INTERFERENCE_DEBUG_LEVEL, C_MSG_INFO, "INTERFERENCE: Starting interference tests using %s as reference and %s as backup...\n", addr_to_str(local_data->max_connectivity_node_address), addr_to_str(local_data->max2_connectivity_node_address));
		collect_interference_data(local_data);
	}
}

/**
 * @brief Handle an incoming INTERFERENCE REQUEST packet.
 *
 * Start gathering information on network conditions.
 *
 * @param header Pointer to header of the incoming packet.
 * @param payload Pointer to the payload that contains a list of all nodes in
 * the network.
 */
void handle_INTERFERENCE_REQUEST(const net_header_t* header, const void* payload) {
	list_t* nodes = NULL;

	if(C_INTERFERENCE_local_data != NULL) {
		C_DEBUG(INTERFERENCE_DEBUG_LEVEL, C_MSG_ERROR, "INTERFERENCE: Received INTERFERENCE REQUEST from %s, but local graph measurement is already in progress...\n", addr_to_str(header->source_address));
		return;
	}

	C_DEBUG(INTERFERENCE_DEBUG_LEVEL, C_MSG_DEBUG, "INTERFERENCE: Received INTERFERENCE REQUEST from %s...\n", addr_to_str(header->source_address));
	list_alloc(&nodes);
	list_from_array(nodes, payload, sizeof(node_t), header->payload_length);
	alloc_local_data(&C_INTERFERENCE_local_data, header->source_address);
	while(!list_empty(nodes)) {
		node_t* node = (node_t*) list_head(nodes);
		INTERFERENCE_local_data_node_t* data_node;
		MALLOC(data_node, INTERFERENCE_local_data_node_t);
		memcpy(&data_node->node, node, sizeof(node_t));
		list_remove(nodes, node);
		FREE(node);
		if(data_node->node.address == C_LOCAL_ADDRESS) {
			data_node->connectivity_count = INTERFERENCE_CONNECTIVITY_SAMPLES;
			data_node->interference = 0;
			data_node->interference_count = 0;
		} else {
			data_node->connectivity_count = 0;
			data_node->interference = -1;
			data_node->interference_count = 0;
		}
		list_add(C_INTERFERENCE_local_data->data_nodes, data_node);
	}
	list_free_including_data(&nodes, free_node_list);

	C_DEBUG(INTERFERENCE_DEBUG_LEVEL, C_MSG_INFO, "INTERFERENCE: Starting connectivity tests...\n");
	collect_connectivity_data(C_INTERFERENCE_local_data);
}

/**
 * @brief Callback for iteratively sending test packets.
 *
 * @param parameter Callback parameter.
 */
void handle_INTERFERENCE_TEST_REQUEST_callback(callback_parameter_t parameter) {
	INTERFERENCE_test_t* test = (INTERFERENCE_test_t*) parameter;

	C_DEBUG(INTERFERENCE_DEBUG_LEVEL, C_MSG_VERBOSE, "INTERFERENCE: Preparing next TEST DATA packet for test #%d from %s...\n", test->id, addr_to_str(test->node_address));
	if((test->packets > 0) && (test->packet_count < test->packets)) {
		test->packet_count++;
		send_INTERFERENCE_TEST_DATA(test->id, test->node_address);
		callback_add(handle_INTERFERENCE_TEST_REQUEST_callback, test, test->interval);
	} else if(C_time() < test->end_time) {
		send_INTERFERENCE_TEST_DATA(test->id, test->node_address);
		callback_add(handle_INTERFERENCE_TEST_REQUEST_callback, test, test->interval);
	} else {
		C_DEBUG(INTERFERENCE_DEBUG_LEVEL, C_MSG_DEBUG, "INTERFERENCE: All packets for test #%d from %s have been sent.\n", test->id, addr_to_str(test->node_address));
		FREE(test);
	}
}

/**
 * @brief Handle INTERFERENCE TEST REQUEST packet.
 *
 * @param header Pointer to packet header.
 * @param test_request Pointer to test request.
 */
void handle_INTERFERENCE_TEST_REQUEST(const net_header_t* header, const INTERFERENCE_TEST_REQUEST_packet_t* test_request) {
	INTERFERENCE_test_t* test = NULL;

	C_DEBUG(INTERFERENCE_DEBUG_LEVEL, C_MSG_DEBUG, "INTERFERENCE: Received INTERFERENCE TEST REQUEST from %s, starting at %.2f...\n", addr_to_str(header->source_address), (C_time() + test_request->delay));
	MALLOC(test, INTERFERENCE_test_t);
	test->id = test_request->test_id;
	test->node_address = header->source_address;
	test->packets = test_request->packets;
	test->interval = test_request->interval;
	test->packet_count = 0;
	test->end_time = C_time() + test_request->duration;
	callback_add(handle_INTERFERENCE_TEST_REQUEST_callback, test, test_request->delay);
}

/**
 * @brief Handle INTERFERENCE TEST DATA packet.
 *
 * @param header Pointer to packet header.
 * @param test_data Pointer to test data.
 */
void handle_INTERFERENCE_TEST_DATA(const net_header_t* header, const INTERFERENCE_TEST_DATA_packet_t* test_data) {
	if(test_data->node_address != C_LOCAL_ADDRESS) {
		C_DEBUG(INTERFERENCE_DEBUG_LEVEL, C_MSG_VERBOSE, "INTERFERENCE: Ignoring INTERFERENCE TEST DATA from %s to %s...\n", addr_to_str(header->source_address), addr_to_str(test_data->node_address));
		return;
	} else if(C_INTERFERENCE_local_data == NULL) {
		C_DEBUG(INTERFERENCE_DEBUG_LEVEL, C_MSG_WARN, "INTERFERENCE: Received INTERFERENCE TEST DATA from %s, but no local graph measurement is active on this node.\n", addr_to_str(header->source_address));
		return;
	} else if(test_data->test_id != C_INTERFERENCE_local_data->active_test_id) {
		C_DEBUG(INTERFERENCE_DEBUG_LEVEL, C_MSG_WARN, "INTERFERENCE: Received INTERFERENCE TEST DATA from %s with wrong test ID (%d != %d)...\n", addr_to_str(header->source_address), test_data->test_id, C_INTERFERENCE_local_data->active_test_id);
		return;
	}

	if(header->source_address == C_INTERFERENCE_local_data->reference_node_address) {
		if(C_INTERFERENCE_local_data->last_interference_packet_source == C_BROADCAST_ADDRESS) {
			C_DEBUG(INTERFERENCE_DEBUG_LEVEL, C_MSG_WARN, "INTERFERENCE: Received INTERFERENCE TEST DATA from reference %s before seeing any packets from the active %s.\n", addr_to_str(C_INTERFERENCE_local_data->reference_node_address), addr_to_str(C_INTERFERENCE_local_data->active_data_node->node.address));
		} else if(C_INTERFERENCE_local_data->active_data_node->interference_count > 0) {
			C_DEBUG(INTERFERENCE_DEBUG_LEVEL, C_MSG_DEBUG, "INTERFERENCE: Received INTERFERENCE TEST DATA from reference %s...\n", addr_to_str(header->source_address));
			C_INTERFERENCE_local_data->active_data_node->interference_count--;
		} else {
			C_DEBUG(INTERFERENCE_DEBUG_LEVEL, C_MSG_WARN, "INTERFERENCE: Ignoring excessive INTERFERENCE TEST DATA from reference %s...\n", addr_to_str(header->source_address));
		}
	} else if(header->source_address == C_INTERFERENCE_local_data->active_data_node->node.address) {
		C_DEBUG(INTERFERENCE_DEBUG_LEVEL, C_MSG_VERBOSE, "INTERFERENCE: Received INTERFERENCE TEST DATA from active %s...\n", addr_to_str(header->source_address));
	} else {
		C_DEBUG(INTERFERENCE_DEBUG_LEVEL, C_MSG_ERROR, "INTERFERENCE: Received INTERFERENCE TEST DATA from invalid %s...\n", addr_to_str(header->source_address));
		return;
	}
	C_INTERFERENCE_local_data->last_interference_packet_source = header->source_address;
}
