/**
 * @file macros.h
 *
 * @brief Macro definitions.
 *
 * Copyright (c) 2007, 2008, 2009, 2010 Georg Wittenburg, Freie Universitaet
 * Berlin (wittenbu@inf.fu-berlin.de). All rights reserved.
 * Please refer to the COPYING file for detailed licensing information.
 *
 * @author Georg Wittenburg
 *
 * $Id: macros.h 3400 2010-03-10 17:27:20Z gwitten $
 */

#ifndef MACROS_H
#define MACROS_H



#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "config.h"



#ifdef C_ENABLE_DEBUG

	/**
	 * @brief Debug output function macro (enabled)
	 */
	#define C_DEBUG(debug_level, msg_type, ...)	debug_printf(debug_level, msg_type, __VA_ARGS__)

	/**
	 * @brief Debug conditional execution macro (enabled)
	 */
	#define C_IF_DEBUG(debug_level, msg_type)	if((msg_type & C_MSG_PRIORITY) >= debug_level)

	/**
	 * @brief Debug check, output and abort function macro (enabled)
	 */
	#define C_ABORT_ON(cond, ...)			if(cond) { \
								C_DEBUG(C_MAIN_DEBUG_LEVEL, C_MSG_FATAL, __VA_ARGS__); \
								fflush(NULL); \
								raise(SIGTRAP); \
								abort(); \
							}

#else

	/**
	 * @brief Debug output function macro (disabled)
	 */
	#define C_DEBUG(debug_level, msg_type, ...)	do { } while(0)

	/**
	 * @brief Debug conditional execution macro (disabled)
	 */
	#define C_IF_DEBUG(debug_level, msg_type)	if(0)

	/**
	 * @brief Debug check, output and abort function macro (disabled)
	 */
	#define C_ABORT_ON(cond, ...)			do { } while(0)

#endif	// C_ENABLE_DEBUG



/**
 * @brief Wrapper macro around malloc().
 *
 * Allocates memory and initializes it to 0x0. Aborts if no memory was
 * available or if pointer is not initialized to NULL.
 *
 * @param p Pointer at which memory is to be allocated.
 * @param type Type of data structure.
 */
#define MALLOC(p, type) \
	{ \
		C_ABORT_ON(p != NULL, "Trying to allocate memory for non-NULL pointer (%p).\n", p); \
		p = (type*) malloc(sizeof(type)); \
		if(p == NULL) { \
			printf("Unable to allocate memory.\n"); \
			raise(SIGTRAP); \
		} \
		memset(p, 0, sizeof(type)); \
	}

/**
 * @brief Wrapper macro around malloc() for allocating arrays.
 *
 * @param p Pointer at which memory is to be allocated.
 * @param type Type of data structure.
 * @param count Number of items of the data structure that are to fit into the
 * allocated memory.
 */
#define MALLOC_ARRAY(p, type, count) \
	{ \
		C_ABORT_ON(p != NULL, "Trying to allocate array memory for non-NULL pointer (%p).\n", p); \
		p = (type*) malloc(sizeof(type) * count); \
		if(p == NULL) { \
			printf("Unable to allocate array memory.\n"); \
			raise(SIGTRAP); \
		} \
		memset(p, 0, sizeof(type) * count); \
	}

/**
 * @brief Wrapper macro around free().
 *
 * Frees memory and sets the pointer to NULL.
 *
 * @param p Pointer to memory.
 */
#define FREE(p) \
	{ \
		free(p); \
		p = NULL; \
	}

/**
 * @brief Join two memory regions.
 *
 * Join two memory regions into a newly allocated memory region of the combined
 * size.
 *
 * @param p Pointer at which memory is to be allocated.
 * @param type Type of data structure.
 * @param p1 Pointer to first memory region.
 * @param size1 Size of first memory region.
 * @param p2 Pointer to second memory region.
 * @param size2 Size of second memory region.
 */
#define MALLOC_JOIN(p, type, p1, size1, p2, size2) \
	{ \
		char* tmp = NULL; \
		C_ABORT_ON(sizeof(type) > (size1 + size2), "Size of type of joined memory region is greater than allocated memory.\n"); \
		MALLOC_ARRAY(tmp, char, (size1 + size2)); \
		p = (type*) tmp; \
		if(size1 > 0) \
			memcpy(p, p1, size1); \
		if(size2 > 0) \
			memcpy(((char*) p) + size1, p2, size2); \
	}

#endif	// MACROS_H
