/**
 * @file asg_event_flow_tree.c
 *
 * @brief ASG event flow tree algorithm algorithm.
 *
 * This is an implementation of the "ASG event flow tree" service placement
 * algorithm proposed in:
 * Klaus Herrmann. Self-Organizing Infrastructures for Ambient Services. PhD
 * Thesis. Berlin University of Technology, July 2006.
 *
 * It is applicable to a single service with multiple service instances. These
 * service instances are migrated and replicated (and dissolved) iteratively to
 * optimal nodes (based on a partial network map) according to a set of rules.
 *
 * The ASG architecture also describes a mechanism that optimizes the
 * distribution of information about active service instances. Instead of
 * re-implementing the entire mechanism, we decided to simply enable our own
 * optimization in the SLP component when evaluating this algorithm.
 *
 * Similarly, we enable our own optimizations for service migration and routing
 * as closely related concepts are mentioned by Herrmann.
 *
 * Copyright (c) 2007, 2008, 2009, 2010 Georg Wittenburg, Freie Universitaet
 * Berlin (wittenbu@inf.fu-berlin.de). All rights reserved.
 * Please refer to the COPYING file for detailed licensing information.
 *
 * @author Georg Wittenburg
 *
 * $Id: asg_event_flow_tree.c 3616 2010-06-23 08:35:44Z gwitten $
 */

#include <math.h>

#include "../../config.h"
#include "../../macros.h"
#include "../../types.h"
#include "../../vars.h"
#include "../../main.h"
#include "../../util/callback.h"
#include "../../util/debug.h"
#include "../../util/list.h"
#include "../../util/net.h"
#include "../../slp-light/slp-light.h"
#include "../../service/service.h"
#include "../placement.h"
#include "../placement_types.h"
#include "../placement_vars.h"
#include "../placement_types.h"
#include "../asg_simple_algorithm/asg_simple_algorithm.h"
#include "asg_event_flow_tree.h"



#ifdef C_SERVICE_PLACEMENT_EVALUATION

#ifndef C_ENABLE_ASG_OPTIMIZATIONS
	#warning Compiling the ASG event flow tree algorithm without ASG optimizations...
#endif	// C_ENABLE_ASG_OPTIMIZATIONS



/**
 * @brief Debug level for this component.
 */
#define ASG_EVENT_FLOW_TREE_DEBUG_LEVEL		C_MSG_INFO



/**
 * @brief Interface to ASG event flow tree algorithm.
 */
PLACEMENT_algorithm_t asg_event_flow_tree_algorithm = {
	C_PLACEMENT_ALGORITHM_ASG_EVENT_FLOW_TREE,
	asg_eft_start,
	asg_eft_stop,
	NULL,
	NULL,
	NULL,
	NULL,
	asg_eft_report_known_service_instance,
	asg_eft_report_invalid_service_instance,
	asg_eft_get_synchronization_hosts
};

/**
 * @brief ASG event flow tree algorithm data.
 */
typedef struct {
	SERVICE_id_t*	service_id;			///< Service ID.
	list_t*		known_service_instances;	///< List of addresses of known service instances.
	callback_id_t	callback_id;			///< ID of the periodic callback.
} asg_eft_data_t;

/**
 * @brief Time interval between runs of the adaptation algorithm (in seconds).
 *
 * Herrmann gives the value of 5000 "ticks" for this parameter (p. 124).
 * However, earlier on he states that "there is no satisfactory mapping of
 * simulation time [ticks] to real time or vice versa." (p. 92).
 *
 * For our evaluation, we assume a value of 60s, which should provide a good
 * granularity for adaptation decisions in the range of request rates that we
 * are evaluating.
 */
#define ASG_EVENT_FLOW_TREE_ADAPTATION_INTERVAL	60



/**
 * @brief Calculcate the flow measure of a node (8.45, p. 117).
 *
 * @param node_stat Pointer to the node stat for which to calculate the flow
 * measure.
 * @return Flow measure of the node.
 */
static float asg_eft_calculate_flow_measure(SERVICE_node_stat_t* node_stat) {
	uint8_t depth = 0;
	uint32_t weight = 0;
	float flow_measure = 0.0;

	C_ABORT_ON((node_stat->node->minimal_hop_count > C_MAXIMAL_NETWORK_DIAMETER) && (node_stat->requests_forwarded == 0), "ASG/EFT: Distance in hops to %s is unknown, but it has forwarded %d request(s).\n", node_stat->node->address, node_stat->requests_forwarded);

	depth = node_stat->node->minimal_hop_count - 1;	// We know that the root node of the tree has a distance of 1.
	weight = node_stat->requests_forwarded;
	flow_measure = (((float) depth) * ((float) weight));

	C_DEBUG(ASG_EVENT_FLOW_TREE_DEBUG_LEVEL, C_MSG_VERBOSE, "ASG/EFT: Flow measure of %s is %.2f (%d * %d).\n", addr_to_str(node_stat->node->address), flow_measure, depth, weight);

	return flow_measure;
}

/**
 * @brief Find the node with the maximal flow measure in an event flow tree.
 *
 * @param service Pointer to the service for which to search an event flow tree.
 * @param root_node_address Address of the root node of the event flow tree.
 * @return Address of the node with the maximal flow measure in the given event
 * flow tree.
 */
static C_addr_t asg_eft_find_node_with_max_flow(SERVICE_service_t* service, C_addr_t root_node_address) {
	SERVICE_node_stat_t* node_stat = NULL;
	list_iterator_t node_stats_iterator;
	float maximal_flow_measure = 0.0;
	C_addr_t node_with_max_flow_address = C_BROADCAST_ADDRESS;

	C_DEBUG(ASG_EVENT_FLOW_TREE_DEBUG_LEVEL, C_MSG_VERBOSE, "ASG/EFT: Looking for node with maximal flow in event flow tree of %s...\n", addr_to_str(root_node_address));
	list_iterator_init(service->node_stats, &node_stats_iterator);
	while((node_stat = (SERVICE_node_stat_t*) list_iterator_next(&node_stats_iterator)) != NULL) {
		float flow_measure = 0.0;
		if(node_stat->last_hop_of_last_request == root_node_address) {
			flow_measure = asg_eft_calculate_flow_measure(node_stat);
			C_DEBUG(ASG_EVENT_FLOW_TREE_DEBUG_LEVEL, C_MSG_VERBOSE, "ASG/EFT: Flow measure of %s in event flow tree of %s is %.2f.\n", addr_to_str(node_stat->node->address), addr_to_str(root_node_address), flow_measure);
			if(flow_measure > maximal_flow_measure) {
				node_with_max_flow_address = node_stat->node->address;
				maximal_flow_measure = flow_measure;
			}
		}
	}
	if(node_with_max_flow_address == C_BROADCAST_ADDRESS) {
		// NOTE: According Herrmann's definition (8.45, p. 117), the
		// flow meassure for the root node of an event flow tree is
		// always 0. Hence, root nodes would never get selected from
		// their own tree. We work around this by returning the root
		// node if no alternatives were found.
		C_DEBUG(ASG_EVENT_FLOW_TREE_DEBUG_LEVEL, C_MSG_DEBUG, "ASG/EFT: Failed to establish node with maximal flow for event flow tree of service \"%s\", using root %s...\n", service->id, addr_to_str(root_node_address));
		return root_node_address;
	}

	C_DEBUG(ASG_EVENT_FLOW_TREE_DEBUG_LEVEL, C_MSG_DEBUG, "ASG/EFT: Node with maximal flow in event flow tree of %s is %s (%.2f flow).\n", addr_to_str(root_node_address), addr_to_str(node_with_max_flow_address), maximal_flow_measure);
	return node_with_max_flow_address;
}

/**
 * @brief Implementation of the idle rule.
 *
 * @param service Pointer to the service for which to check whether a migration
 * is necessary.
 * @return <code>true</code> if this service instance is idle,
 * <code>false</code> otherwise.
 */
static bool asg_eft_idle(SERVICE_service_t* service) {
	return asg_simple_idle(service);
}

/**
 * @brief Implementation of the replication rule.
 *
 * @param service Pointer to the service for which to check whether a migration
 * is necessary.
 * @param replication_target_address_1 Pointer to the address of the first
 * replication target node (return value).
 * @param replication_target_address_2 Pointer to the address of the second
 * replication target node (return value).
 */
static void asg_eft_get_replication_targets(SERVICE_service_t* service, C_addr_t* replication_target_address_1, C_addr_t* replication_target_address_2) {
	C_addr_t simple_replication_target_address_1 = C_BROADCAST_ADDRESS;
	C_addr_t simple_replication_target_address_2 = C_BROADCAST_ADDRESS;
	C_addr_t simple_migration_target_address = C_BROADCAST_ADDRESS;

	*replication_target_address_1 = C_BROADCAST_ADDRESS;
	*replication_target_address_2 = C_BROADCAST_ADDRESS;

	asg_simple_get_replication_targets(service, &simple_replication_target_address_1, &simple_replication_target_address_2);

	if(simple_replication_target_address_1 == C_BROADCAST_ADDRESS)
		return;

	C_DEBUG(ASG_EVENT_FLOW_TREE_DEBUG_LEVEL, C_MSG_DEBUG, "ASG/EFT: Using %s as root for event flow tree for first replication of service \"%s\"...\n", addr_to_str(simple_replication_target_address_1), service->id);
	*replication_target_address_1 = asg_eft_find_node_with_max_flow(service, simple_replication_target_address_1);

	simple_migration_target_address = asg_simple_get_migration_target(service, simple_replication_target_address_1);
	if(simple_migration_target_address == C_BROADCAST_ADDRESS) {
		C_DEBUG(ASG_EVENT_FLOW_TREE_DEBUG_LEVEL, C_MSG_INFO, "ASG/EFT: Replicating one instance of service \"%s\" to %s...\n", service->id, addr_to_str(*replication_target_address_1));
		*replication_target_address_2 = C_LOCAL_ADDRESS;
	} else {
		C_DEBUG(ASG_EVENT_FLOW_TREE_DEBUG_LEVEL, C_MSG_DEBUG, "ASG/EFT: Using %s as root for event flow tree for second replication of service \"%s\"...\n", addr_to_str(simple_replication_target_address_2), service->id);
		*replication_target_address_2 = asg_eft_find_node_with_max_flow(service, simple_migration_target_address);
		C_DEBUG(ASG_EVENT_FLOW_TREE_DEBUG_LEVEL, C_MSG_INFO, "ASG/EFT: Replicating two instances of service \"%s\" to %s and %s...\n", service->id, addr_to_str(*replication_target_address_1), addr_to_str(*replication_target_address_2));
	}
}

/**
 * @brief Implementation of the migration rule.
 *
 * NOTE: This function only implements the "simple migration condition" (8.27)
 * explicitly. However, since its decision is based on the data rate
 * calculation, it is stable enough and implementing the "advanced migration
 * condition" (8.31) is not necessary.
 *
 * @param service Pointer to the service for which to check whether a migration
 * is necessary.
 * @return Address of the migration target node, or C_BROADCAST_ADDRESS if no
 * migration is necessary.
 */
static C_addr_t asg_eft_get_migration_target(SERVICE_service_t* service) {
	C_addr_t simple_migration_target_address = asg_simple_get_migration_target(service, C_BROADCAST_ADDRESS);
	C_addr_t eft_migration_target_address = C_BROADCAST_ADDRESS;

	if(simple_migration_target_address == C_BROADCAST_ADDRESS) {
		C_DEBUG(ASG_EVENT_FLOW_TREE_DEBUG_LEVEL, C_MSG_INFO, "ASG/EFT: Instance of service \"%s\" is placed optimally on the current node.\n", service->id);
		return C_BROADCAST_ADDRESS;
	}

	C_DEBUG(ASG_EVENT_FLOW_TREE_DEBUG_LEVEL, C_MSG_DEBUG, "ASG/EFT: Using %s as root for event flow tree for migration of service \"%s\"...\n", addr_to_str(simple_migration_target_address), service->id);
	eft_migration_target_address = asg_eft_find_node_with_max_flow(service, simple_migration_target_address);

	C_DEBUG(ASG_EVENT_FLOW_TREE_DEBUG_LEVEL, C_MSG_INFO, "ASG/EFT: Migrating instance of service \"%s\" to %s...\n", service->id, addr_to_str(eft_migration_target_address));
	return eft_migration_target_address;
}

static void asg_eft_run(callback_parameter_t parameter);

/**
 * @brief Replication callback data.
 */
typedef struct {
	SERVICE_id_t	service_id[SERVICE_ID_LENGTH];	///< Service ID.
	C_addr_t	replication_target_address_2;	///< Address of the second replication target.
} asg_eft_replication_data_t;

/**
 * @brief Second service replication callback.
 *
 * @param parameter Replication result.
 */
static void asg_eft_replication_2_callback(callback_parameter_t parameter) {
	SERVICE_replication_result_t* replication_result = (SERVICE_replication_result_t*) parameter;
	asg_eft_replication_data_t* replication_data = (asg_eft_replication_data_t*) replication_result->parameter;
	SERVICE_service_t* service = SERVICE_get_service(replication_data->service_id);
	asg_eft_data_t* placement_data = NULL;

	C_ABORT_ON(service == NULL, "ASG/EFT: Trying to call second replication callback of ASG event flow tree algorithm for service \"%s\", but service is NULL.\n", replication_data->service_id);
	if(service->placement_data == NULL) {
		C_DEBUG(ASG_EVENT_FLOW_TREE_DEBUG_LEVEL, C_MSG_WARN, "ASG/EFT: Second replication of service \"%s\" failed (%d), but ASG event flow tree algorithm is not active anymore.\n", replication_data->service_id, replication_result->error_code);
		goto free_replication_2_parameters;
	}
	placement_data = (asg_eft_data_t*) service->placement_data;

	switch(replication_result->error_code) {
	case C_ERROR_OK:
		C_DEBUG(ASG_EVENT_FLOW_TREE_DEBUG_LEVEL, C_MSG_NOTE, "ASG/EFT: Second replication of service \"%s\" was successful. Shutting down service instance on this node...\n", replication_data->service_id);
		SERVICE_unlock_service(replication_data->service_id);
		SERVICE_shutdown_service(replication_data->service_id);
		break;
	case C_ERROR_PARAMETER:
		C_DEBUG(ASG_EVENT_FLOW_TREE_DEBUG_LEVEL, C_MSG_WARN, "ASG/EFT: Second replication of service \"%s\" failed because the service was found running on the destination node. Shutting down service instance on this node...\n", replication_data->service_id);
		SERVICE_unlock_service(replication_data->service_id);
		SERVICE_shutdown_service(replication_data->service_id);
		break;
	case C_ERROR_TIMEOUT:
		C_DEBUG(ASG_EVENT_FLOW_TREE_DEBUG_LEVEL, C_MSG_WARN, "ASG/EFT: Second replication of service \"%s\" timed out, service instance remains on the current node...\n", replication_data->service_id);
		SERVICE_unlock_service(replication_data->service_id);
		placement_data->callback_id = callback_add(asg_eft_run, service, ASG_EVENT_FLOW_TREE_ADAPTATION_INTERVAL);
		break;
	case C_ERROR_ABORTED:
		C_DEBUG(ASG_EVENT_FLOW_TREE_DEBUG_LEVEL, C_MSG_WARN, "ASG/EFT: Second replication of service \"%s\" has been aborted, service instance remains on the current node...\n", replication_data->service_id);
		SERVICE_unlock_service(replication_data->service_id);
		// Don't re-add the callback if aborted by the user.
		break;
	case C_ERROR_BUSY:
		C_DEBUG(ASG_EVENT_FLOW_TREE_DEBUG_LEVEL, C_MSG_ERROR, "ASG/EFT: Second replication of service \"%s\" failed because the local system is busy, service instance remains on the current node...\n", replication_data->service_id);
		SERVICE_unlock_service(replication_data->service_id);
		placement_data->callback_id = callback_add(asg_eft_run, service, ASG_EVENT_FLOW_TREE_ADAPTATION_INTERVAL);
		break;
	case C_ERROR_IN_PROGRESS:
		C_DEBUG(ASG_EVENT_FLOW_TREE_DEBUG_LEVEL, C_MSG_ERROR, "ASG/EFT: Second replication of service \"%s\" is already in progress, service instance remains on the current node...\n", replication_data->service_id);
		SERVICE_unlock_service(replication_data->service_id);
		placement_data->callback_id = callback_add(asg_eft_run, service, ASG_EVENT_FLOW_TREE_ADAPTATION_INTERVAL);
		break;
	case C_ERROR_COMM_ERROR:
		C_DEBUG(ASG_EVENT_FLOW_TREE_DEBUG_LEVEL, C_MSG_WARN, "ASG/EFT: Second replication of service \"%s\" failed due to a communication error, service instance remains on the current node...\n", replication_data->service_id);
		SERVICE_unlock_service(replication_data->service_id);
		placement_data->callback_id = callback_add(asg_eft_run, service, ASG_EVENT_FLOW_TREE_ADAPTATION_INTERVAL);
		break;
	case C_ERROR_REMOTE_BUSY:
		C_DEBUG(ASG_EVENT_FLOW_TREE_DEBUG_LEVEL, C_MSG_ERROR, "ASG/EFT: Second replication of service \"%s\" failed because the remote system refused to host the service, service instance remains on the current node...\n", replication_data->service_id);
		SERVICE_unlock_service(replication_data->service_id);
		placement_data->callback_id = callback_add(asg_eft_run, service, ASG_EVENT_FLOW_TREE_ADAPTATION_INTERVAL);
		break;
	default:
		C_ABORT_ON(true, "ASG/EFT: Second replication callback with unknown error code 0x%X.\n", replication_result->error_code);
	}

free_replication_2_parameters:
	FREE(replication_data);
	FREE(replication_result);
}

/**
 * @brief First service replication callback.
 *
 * @param parameter Replication result.
 */
static void asg_eft_replication_1_callback(callback_parameter_t parameter) {
	SERVICE_replication_result_t* replication_result = (SERVICE_replication_result_t*) parameter;
	asg_eft_replication_data_t* replication_data = (asg_eft_replication_data_t*) replication_result->parameter;
	SERVICE_service_t* service = SERVICE_get_service(replication_data->service_id);
	asg_eft_data_t* placement_data = NULL;

	C_ABORT_ON(service == NULL, "ASG/EFT: Trying to call first replication callback of ASG event flow tree algorithm for service \"%s\", but service is NULL.\n", replication_data->service_id);
	if(service->placement_data == NULL) {
		C_DEBUG(ASG_EVENT_FLOW_TREE_DEBUG_LEVEL, C_MSG_WARN, "ASG/EFT: First replication of service \"%s\" failed (%d), but ASG event flow tree algorithm is not active anymore.\n", replication_data->service_id, replication_result->error_code);
		goto free_replication_1_parameters;
	}
	placement_data = (asg_eft_data_t*) service->placement_data;

	switch(replication_result->error_code) {
	case C_ERROR_OK:
		C_DEBUG(ASG_EVENT_FLOW_TREE_DEBUG_LEVEL, C_MSG_NOTE, "ASG/EFT: First replication of service \"%s\" was successful.\n", replication_data->service_id);
		break;
	case C_ERROR_PARAMETER:
		C_DEBUG(ASG_EVENT_FLOW_TREE_DEBUG_LEVEL, C_MSG_WARN, "ASG/EFT: First replication of service \"%s\" failed because the service was found running on the destination node.\n", replication_data->service_id);
		SERVICE_reset_service_stats(replication_data->service_id);	// NOTE: This indicates that the stats that lead to this replication decision were wrong. Reset stats to take this into account.
		break;
	case C_ERROR_TIMEOUT:
		C_DEBUG(ASG_EVENT_FLOW_TREE_DEBUG_LEVEL, C_MSG_WARN, "ASG/EFT: First replication of service \"%s\" timed out.\n", replication_data->service_id);
		break;
	case C_ERROR_ABORTED:
		C_DEBUG(ASG_EVENT_FLOW_TREE_DEBUG_LEVEL, C_MSG_WARN, "ASG/EFT: First replication of service \"%s\" has been aborted.\n", replication_data->service_id);
		// Don't consider second replication if aborted by the user.
		goto free_replication_1_parameters;
	case C_ERROR_BUSY:
		C_DEBUG(ASG_EVENT_FLOW_TREE_DEBUG_LEVEL, C_MSG_ERROR, "ASG/EFT: First replication of service \"%s\" failed because the local system is busy.\n", replication_data->service_id);
		break;
	case C_ERROR_IN_PROGRESS:
		C_DEBUG(ASG_EVENT_FLOW_TREE_DEBUG_LEVEL, C_MSG_ERROR, "ASG/EFT: First replication of service \"%s\" is already in progress.\n", replication_data->service_id);
		break;
	case C_ERROR_COMM_ERROR:
		C_DEBUG(ASG_EVENT_FLOW_TREE_DEBUG_LEVEL, C_MSG_WARN, "ASG/EFT: First replication of service \"%s\" failed due to a communication error.\n", replication_data->service_id);
		break;
	case C_ERROR_REMOTE_BUSY:
		C_DEBUG(ASG_EVENT_FLOW_TREE_DEBUG_LEVEL, C_MSG_ERROR, "ASG/EFT: First replication of service \"%s\" failed because the remote system refused to host the service.\n", replication_data->service_id);
		break;
	default:
		C_ABORT_ON(true, "ASG/EFT: First replication callback with unknown error code 0x%X.\n", replication_result->error_code);
	}

	if(replication_data->replication_target_address_2 != C_LOCAL_ADDRESS) {
		C_DEBUG(ASG_EVENT_FLOW_TREE_DEBUG_LEVEL, C_MSG_INFO, "ASG/EFT: Starting second replication of service \"%s\" to %s...\n", replication_data->service_id, addr_to_str(replication_data->replication_target_address_2));
		SERVICE_replicate_service(replication_data->service_id, replication_data->replication_target_address_2, asg_eft_replication_2_callback, replication_result);
		return;		// Don't free parameters.
	} else {
		C_DEBUG(ASG_EVENT_FLOW_TREE_DEBUG_LEVEL, C_MSG_INFO, "ASG/EFT: Second replication of service \"%s\" is not necessary, service instance remains on the current node.\n", replication_data->service_id);
		SERVICE_unlock_service(replication_data->service_id);
		// The service demand has changed since this node now has a
		// neighboring service instance. Reset stats to take this into
		// account.
		SERVICE_reset_service_stats(replication_data->service_id);
		// Make sure the new service instance knows about this one for
		// synchronization.
		SLP_announce_service(replication_data->service_id, C_BROADCAST_ADDRESS);
		placement_data->callback_id = callback_add(asg_eft_run, service, ASG_EVENT_FLOW_TREE_ADAPTATION_INTERVAL);
	}

free_replication_1_parameters:
	FREE(replication_data);
	FREE(replication_result);
}

/**
 * @brief Service migration callback.
 *
 * @param parameter Migration result.
 */
static void asg_eft_migration_callback(callback_parameter_t parameter) {
	SERVICE_migration_result_t* migration_result = (SERVICE_migration_result_t*) parameter;
	SERVICE_id_t* service_id = (SERVICE_id_t*) migration_result->parameter;
	SERVICE_service_t* service = SERVICE_get_service(service_id);
	asg_eft_data_t* placement_data = NULL;

	if(migration_result->error_code != C_ERROR_OK) {
		C_ABORT_ON(service == NULL, "ASG/EFT: Trying to call migration callback of ASG event flow tree algorithm for service \"%s\", but service is NULL.\n", service_id);
		if(service->placement_data == NULL) {
			C_DEBUG(ASG_EVENT_FLOW_TREE_DEBUG_LEVEL, C_MSG_WARN, "ASG/EFT: Migration of service \"%s\" failed (%d), but ASG event flow tree algorithm is not active anymore.\n", service_id, migration_result->error_code);
			goto free_migration_parameters;
		}
		placement_data = (asg_eft_data_t*) service->placement_data;
	}

	switch(migration_result->error_code) {
	case C_ERROR_OK:
		C_DEBUG(ASG_EVENT_FLOW_TREE_DEBUG_LEVEL, C_MSG_NOTE, "ASG/EFT: Migration of service \"%s\" was successful.\n", service_id);
		// There is no need to unlock the service since no instance
		// remains on this node.
		break;
	case C_ERROR_PARAMETER:
		C_DEBUG(ASG_EVENT_FLOW_TREE_DEBUG_LEVEL, C_MSG_WARN, "ASG/EFT: Migration of service \"%s\" failed because the service was found running on the destination node. Shutting down service on this node...\n", service_id);
		SERVICE_unlock_service(service_id);
		SERVICE_shutdown_service(service_id);
		break;
	case C_ERROR_TIMEOUT:
		C_DEBUG(ASG_EVENT_FLOW_TREE_DEBUG_LEVEL, C_MSG_WARN, "ASG/EFT: Migration of service \"%s\" timed out.\n", service_id);
		SERVICE_unlock_service(service_id);
		placement_data->callback_id = callback_add(asg_eft_run, service, ASG_EVENT_FLOW_TREE_ADAPTATION_INTERVAL);
		break;
	case C_ERROR_ABORTED:
		C_DEBUG(ASG_EVENT_FLOW_TREE_DEBUG_LEVEL, C_MSG_WARN, "ASG/EFT: Migration of service \"%s\" has been aborted.\n", service_id);
		SERVICE_unlock_service(service_id);
		// Don't re-add the callback if aborted by the user.
		break;
	case C_ERROR_BUSY:
		C_DEBUG(ASG_EVENT_FLOW_TREE_DEBUG_LEVEL, C_MSG_ERROR, "ASG/EFT: Migration of service \"%s\" failed because the local system is busy.\n", service_id);
		SERVICE_unlock_service(service_id);
		placement_data->callback_id = callback_add(asg_eft_run, service, ASG_EVENT_FLOW_TREE_ADAPTATION_INTERVAL);
		break;
	case C_ERROR_IN_PROGRESS:
		C_DEBUG(ASG_EVENT_FLOW_TREE_DEBUG_LEVEL, C_MSG_ERROR, "ASG/EFT: Migration of service \"%s\" is already in progress.\n", service_id);
		SERVICE_unlock_service(service_id);
		placement_data->callback_id = callback_add(asg_eft_run, service, ASG_EVENT_FLOW_TREE_ADAPTATION_INTERVAL);
		break;
	case C_ERROR_COMM_ERROR:
		C_DEBUG(ASG_EVENT_FLOW_TREE_DEBUG_LEVEL, C_MSG_WARN, "ASG/EFT: Migration of service \"%s\" failed due to a communication error.\n", service_id);
		SERVICE_unlock_service(service_id);
		placement_data->callback_id = callback_add(asg_eft_run, service, ASG_EVENT_FLOW_TREE_ADAPTATION_INTERVAL);
		break;
	case C_ERROR_REMOTE_BUSY:
		C_DEBUG(ASG_EVENT_FLOW_TREE_DEBUG_LEVEL, C_MSG_ERROR, "ASG/EFT: Migration of service \"%s\" failed because the remote system refused to host the service.\n", service_id);
		SERVICE_unlock_service(service_id);
		placement_data->callback_id = callback_add(asg_eft_run, service, ASG_EVENT_FLOW_TREE_ADAPTATION_INTERVAL);
		break;
	default:
		C_ABORT_ON(true, "ASG/EFT: Migration callback with unknown error code 0x%X.\n", migration_result->error_code);
	}

free_migration_parameters:
	FREE(service_id);
	FREE(migration_result);
}

/**
 * @brief ASG event flow tree algorithm.
 *
 * @param parameter Pointer to service for which to run this algorithm.
 */
static void asg_eft_run(callback_parameter_t parameter) {
	SERVICE_id_t* service_id = (SERVICE_id_t*) parameter;
	SERVICE_service_t* service = SERVICE_get_running_service(service_id);
	asg_eft_data_t* placement_data = NULL;

	if(service == NULL) {
		C_DEBUG(ASG_EVENT_FLOW_TREE_DEBUG_LEVEL, C_MSG_ERROR, "ASG/EFT: Trying to run ASG event flow tree algorithm for service \"%s\", but service is not running.\n", service->id);
		return;
	}

	C_ABORT_ON(service->placement_data == NULL, "ASG/EFT: Trying to run ASG event flow tree algorithm for service \"%s\", but placement data is NULL.\n", service->id);
	placement_data = (asg_eft_data_t*) service->placement_data;
	placement_data->callback_id = 0;

	// Make sure all service instances know about this instance for
	// synchronization.
	SLP_announce_service(service->id, C_BROADCAST_ADDRESS);

	if(service->host_stat->request_count == 0) {
		C_DEBUG(ASG_EVENT_FLOW_TREE_DEBUG_LEVEL, C_MSG_WARN, "ASG/EFT: Trying to run ASG event flow tree algorithm for service \"%s\", but no service requests have been received.\n", service->id);
		placement_data->callback_id = callback_add(asg_eft_run, service, ASG_EVENT_FLOW_TREE_ADAPTATION_INTERVAL);
		return;
	}

	C_DEBUG(ASG_EVENT_FLOW_TREE_DEBUG_LEVEL, C_MSG_INFO, "ASG/EFT: Checking whether an adaptation is required for service \"%s\"...\n", service->id);

	/// This following corresponds to the code snippet show in Figure 8.5
	/// (page 101) of Herrmann.

	if(asg_eft_idle(service)) {
		C_DEBUG(ASG_EVENT_FLOW_TREE_DEBUG_LEVEL, C_MSG_NOTE, "ASG/EFT: Service \"%s\" is idle, shutting down service...\n", service->id);
		SERVICE_shutdown_service(service->id);
	} else {
		C_addr_t replication_target_address_1 = C_BROADCAST_ADDRESS;
		C_addr_t replication_target_address_2 = C_BROADCAST_ADDRESS;
		C_addr_t migration_target_address = C_BROADCAST_ADDRESS;

		asg_eft_get_replication_targets(service, &replication_target_address_1, &replication_target_address_2);
		if(
			(replication_target_address_1 != C_BROADCAST_ADDRESS)
			&& (replication_target_address_2 != C_BROADCAST_ADDRESS)
		) {
			SERVICE_replication_result_t* replication_result = NULL;
			asg_eft_replication_data_t* replication_data = NULL;

			if(replication_target_address_2 == C_LOCAL_ADDRESS)
				C_DEBUG(ASG_EVENT_FLOW_TREE_DEBUG_LEVEL, C_MSG_NOTE, "ASG/EFT: Replicating service \"%s\" to %s...\n", service->id, addr_to_str(replication_target_address_1));
			else
				C_DEBUG(ASG_EVENT_FLOW_TREE_DEBUG_LEVEL, C_MSG_NOTE, "ASG/EFT: Replicating service \"%s\" to %s and %s...\n", service->id, addr_to_str(replication_target_address_1), addr_to_str(replication_target_address_2));
			SERVICE_lock_service(service->id);
			MALLOC(replication_result, SERVICE_replication_result_t);
			MALLOC(replication_data, asg_eft_replication_data_t);
			SERVICE_copy_service_id(replication_data->service_id, service->id);
			replication_data->replication_target_address_2 = replication_target_address_2;
			replication_result->parameter = replication_data;
			SERVICE_replicate_service(service->id, replication_target_address_1, asg_eft_replication_1_callback, replication_result);
		} else if ((migration_target_address = asg_eft_get_migration_target(service)) != C_BROADCAST_ADDRESS) {
			SERVICE_migration_result_t* migration_result = NULL;

			C_DEBUG(ASG_EVENT_FLOW_TREE_DEBUG_LEVEL, C_MSG_NOTE, "ASG/EFT: Migrating service \"%s\" to %s...\n", service->id, addr_to_str(migration_target_address));
			SERVICE_lock_service(service->id);
			MALLOC(migration_result, SERVICE_migration_result_t);
			SERVICE_alloc_service_id((SERVICE_id_t**) &migration_result->parameter, service->id);
			SERVICE_migrate_service(service->id, migration_target_address, asg_eft_migration_callback, migration_result);
		} else {
			C_DEBUG(ASG_EVENT_FLOW_TREE_DEBUG_LEVEL, C_MSG_NOTE, "ASG/EFT: Nothing to do for service \"%s\"...\n", service->id);
			placement_data->callback_id = callback_add(asg_eft_run, service, ASG_EVENT_FLOW_TREE_ADAPTATION_INTERVAL);
		}
	}
}

/**
 * @brief Start ASG event flow tree algorithm.
 *
 * @param service Pointer to service for which to start this algorithm.
 */
void asg_eft_start(SERVICE_service_t* service) {
	asg_eft_data_t* placement_data = NULL;
	C_addr_t* address = NULL;

	C_ABORT_ON(service->placement_data != NULL, "ASG/EFT: Trying to start ASG event flow tree algorithm for service \"%s\", but placement data is not NULL.\n", service->id);

	C_DEBUG(ASG_EVENT_FLOW_TREE_DEBUG_LEVEL, C_MSG_INFO, "ASG/EFT: Starting ASG event flow tree algorithm for service \"%s\"...\n", service->id);
	MALLOC(service->placement_data, asg_eft_data_t);
	placement_data = (asg_eft_data_t*) service->placement_data;
	SERVICE_alloc_service_id(&placement_data->service_id, service->id);
	list_alloc(&placement_data->known_service_instances);
	MALLOC(address, C_addr_t);
	*address = C_LOCAL_ADDRESS;
	list_add(placement_data->known_service_instances, address);
	placement_data->callback_id = callback_add(asg_eft_run, placement_data->service_id, ASG_EVENT_FLOW_TREE_ADAPTATION_INTERVAL);
}

/**
 * @brief Stop ASG event flow tree algorithm.
 *
 * @param service Pointer to service for which to stop this algorithm.
 */
void asg_eft_stop(SERVICE_service_t* service) {
	asg_eft_data_t* placement_data = (asg_eft_data_t*) service->placement_data;

	C_ABORT_ON(placement_data == NULL, "ASG/EFT: Trying to stop ASG event flow tree algorithm for service \"%s\", but placement data is NULL.\n", service->id);

	C_DEBUG(ASG_EVENT_FLOW_TREE_DEBUG_LEVEL, C_MSG_INFO, "ASG/EFT: Stopping ASG event flow tree algorithm for service \"%s\"...\n", service->id);
	FREE(placement_data->service_id);
	list_free_including_data(&placement_data->known_service_instances, free);
	if(placement_data->callback_id != 0)
		callback_cancel(placement_data->callback_id);
	FREE(service->placement_data);
}

/**
 * @brief Report known service instance to ASG event flow tree algorithm.
 *
 * @param service Pointer to service for which report a known service instance.
 * @param host_address Address of the host of the known service instance.
 */
void asg_eft_report_known_service_instance(SERVICE_service_t* service, C_addr_t host_address) {
	asg_eft_data_t* placement_data = (asg_eft_data_t*) service->placement_data;
	C_addr_t* address = NULL;
	C_addr_t cmp_address;

	C_ABORT_ON(placement_data == NULL, "ASG/EFT: Trying to report known service instance of service \"%s\", but placement data is NULL.\n", service->id);

	C_DEBUG(ASG_EVENT_FLOW_TREE_DEBUG_LEVEL, C_MSG_DEBUG, "ASG/EFT: Handling report of known service instance of service \"%s\" at %s...\n", service->id, addr_to_str(host_address));

	cmp_address = host_address;
	if((address = (C_addr_t*) list_find(placement_data->known_service_instances, &cmp_address, matches_address)) == NULL) {
		C_DEBUG(ASG_EVENT_FLOW_TREE_DEBUG_LEVEL, C_MSG_INFO, "ASG/EFT: Adding %s to list of known service instances of service \"%s\"...\n", addr_to_str(host_address), service->id);
		MALLOC(address, C_addr_t);
		*address = host_address;
		list_add(placement_data->known_service_instances, address);
	}
}

/**
 * @brief Report invalid service instance to ASG event flow tree algorithm.
 *
 * @param service Pointer to service for which report a known service instance.
 * @param host_address Address of the host of the invalid service instance.
 */
void asg_eft_report_invalid_service_instance(SERVICE_service_t* service, C_addr_t host_address) {
	asg_eft_data_t* placement_data = (asg_eft_data_t*) service->placement_data;
	C_addr_t* address = NULL;
	C_addr_t cmp_address;

	C_ABORT_ON(placement_data == NULL, "ASG/EFT: Trying to report invalid service instance of service \"%s\", but placement data is NULL.\n", service->id);

	C_DEBUG(ASG_EVENT_FLOW_TREE_DEBUG_LEVEL, C_MSG_DEBUG, "ASG/EFT: Handling report of invalid service instance of service \"%s\" at %s...\n", service->id, addr_to_str(host_address));

	cmp_address = host_address;
	if((address = (C_addr_t*) list_find(placement_data->known_service_instances, &cmp_address, matches_address)) != NULL) {
		C_DEBUG(ASG_EVENT_FLOW_TREE_DEBUG_LEVEL, C_MSG_INFO, "ASG/EFT: Removing %s from list of known service instances of service \"%s\"...\n", addr_to_str(host_address), service->id);
		list_remove(placement_data->known_service_instances, address);
		FREE(address);
	}
}

/**
 * @brief Get synchronization hosts from ASG event flow tree algorithm.
 *
 * The synchronization host in the ASG context is called the "primary".
 * Herrmann states that "[t]he initial replica of a specific service type that
 * is installed in an ASG automatically becomes the primary" (p. 205). It is
 * however unclear how this primary service instance is identified if it
 * migrates, or replicates and shuts down afterwards. Hence, we just use the
 * instance with the smallest address as primary.
 *
 * @param service Pointer to service for which to get the synchronization hosts.
 * @param inter_host_update Flag whether this update is an inter-host update.
 * @return Pointer to a list of nodes to which synchronize changes to (must be
 * freed by the caller), or NULL if no synchronization is necessary.
 */
list_t* asg_eft_get_synchronization_hosts(const SERVICE_service_t* service, bool inter_host_update) {
	asg_eft_data_t* placement_data = (asg_eft_data_t*) service->placement_data;
	list_t* synchronization_hosts = NULL;
	C_addr_t primary_address = C_BROADCAST_ADDRESS;
	C_addr_t* instance_address = NULL;
	list_iterator_t instances_iterator;

	C_ABORT_ON(placement_data == NULL, "ASG/EFT: Trying to get synchronization hosts for service \"%s\", but placement data is NULL.\n", service->id);

	C_DEBUG(ASG_EVENT_FLOW_TREE_DEBUG_LEVEL, C_MSG_DEBUG, "ASG/EFT: Getting synchronization hosts for service \"%s\"...\n", service->id);
	list_iterator_init(placement_data->known_service_instances, &instances_iterator);
	while((instance_address = (C_addr_t*) list_iterator_next(&instances_iterator)) != NULL) {
		if(
			(primary_address == C_BROADCAST_ADDRESS)
			|| (*instance_address < primary_address)
		) {
			primary_address = *instance_address;
		}
	}
	C_ABORT_ON(primary_address == C_BROADCAST_ADDRESS, "ASG/EFT: Failed to establish \"primary\" for service \"%s\"...", service->id);

	C_DEBUG(ASG_EVENT_FLOW_TREE_DEBUG_LEVEL, C_MSG_DEBUG, "ASG/EFT: Using %s as \"primary\" for service \"%s\"...\n", addr_to_str(primary_address), service->id);

	list_alloc(&synchronization_hosts);
	if(primary_address == C_LOCAL_ADDRESS) {
		list_iterator_init(placement_data->known_service_instances, &instances_iterator);
		while((instance_address = (C_addr_t*) list_iterator_next(&instances_iterator)) != NULL) {
			if(*instance_address != primary_address)
				list_add(synchronization_hosts, instance_address);
		}
	} else if(!inter_host_update) {
		// Only synchronize regular updates on non-primary hosts.
		list_iterator_init(placement_data->known_service_instances, &instances_iterator);
		while((instance_address = (C_addr_t*) list_iterator_next(&instances_iterator)) != NULL) {
			if(*instance_address == primary_address) {
				// Using the item from the list avoids having
				// to allocate memory (which would be hard to
				// free later on).
				list_add(synchronization_hosts, instance_address);
				break;
			}
		}
	}
	list_sort(synchronization_hosts, node_address_less_than);

	if(inter_host_update)
		C_DEBUG(ASG_EVENT_FLOW_TREE_DEBUG_LEVEL, C_MSG_DEBUG, "ASG/EFT: Synchronization hosts for inter-host update of service \"%s\" are ", service->id);
	else
		C_DEBUG(ASG_EVENT_FLOW_TREE_DEBUG_LEVEL, C_MSG_DEBUG, "ASG/EFT: Synchronization hosts for update of service \"%s\" are ", service->id);
	list_print(ASG_EVENT_FLOW_TREE_DEBUG_LEVEL, (C_MSG_DEBUG | C_MSG_CONT), synchronization_hosts, print_node_address);
	C_DEBUG(ASG_EVENT_FLOW_TREE_DEBUG_LEVEL, (C_MSG_DEBUG | C_MSG_CONT), ".\n");

	return synchronization_hosts;
}

#else
	#warning Compiling without support for ASG event flow tree algorithm...
#endif	// C_SERVICE_PLACEMENT_EVALUATION
