/**
 * @file gcmi_clustering.c
 *
 * @brief Graph cost multiple instances (clustering).
 *
 * Copyright (c) 2007, 2008, 2009, 2010 Georg Wittenburg, Freie Universitaet
 * Berlin (wittenbu@inf.fu-berlin.de). All rights reserved.
 * Please refer to the COPYING file for detailed licensing information.
 *
 * @author Georg Wittenburg
 *
 * $Id: gcmi_clustering.c 3426 2010-03-18 15:00:10Z gwitten $
 */

#include <float.h>
#include <math.h>

#include "../../config.h"
#include "../../macros.h"
#include "../../types.h"
#include "../../vars.h"
#include "../../main.h"
#include "../../util/debug.h"
#include "../../util/list.h"
#include "../../util/mapping.h"
#include "../../util/net.h"
#include "../../service/service.h"
#include "../placement.h"
#include "../placement_types.h"
#include "../placement_vars.h"
#include "../graph_cost_single_instance/graph_cost_single_instance.h"
#include "graph_cost_multiple_instances.h"
#include "gcmi_cost_calculation.h"
#include "gcmi_clustering.h"



#ifdef C_SERVICE_PLACEMENT_EVALUATION



/**
 * @brief Debug level for this component.
 */
#define GCMI_CLUSTERING_DEBUG_LEVEL	C_MSG_WARN
// #define GCMI_CLUSTERING_DEBUG_LEVEL	C_MSG_VERBOSE



#ifdef C_ENABLE_CONSISTENCY_CHECKS
/**
 * @brief Consistency check for cluster data structures.
 *
 * Time complexity:	O(clusters * (
 *			  avg_nodes_in_cluster
 *			  + avg_adjacent_clusters * clusters
 *			) + nodes * clusters)
 *			= O(nodes * clusters)
 * Space complexity:	O(1)
 *
 * @param clusters Pointer to list of clusters.
 * @param node_id_to_cluster_map Array of pointers to clusters, indexed by node IDs.
 */
static void check_clusters(const list_t* clusters, cluster_t** node_id_to_cluster_map) {
	SERVICE_host_stat_t* host_stat = NULL;
	cluster_t* cluster = NULL;
	list_iterator_t clusters_iterator;
	bool node_list_contains_cluster_head = false;
	SERVICE_node_stat_t* node_stat = NULL;
	list_iterator_t node_stats_iterator;
	uint16_t node_id = C_UNKNOWN_NODE_ID;
	cluster_t* adjacent_cluster = NULL;
	list_iterator_t adjacent_clusters_iterator;
	float distance_to_cluster_head = FLT_MAX;

	SERVICE_alloc_host_stat(&host_stat, C_LOCAL_ADDRESS);

	list_iterator_init(clusters, &clusters_iterator);
	while((cluster = (cluster_t*) list_iterator_next(&clusters_iterator)) != NULL) {

		C_ABORT_ON(list_size(cluster->node_stats) < 1, "GCMI/CLUST: Cluster with head %s contains less than one node.\n", addr_to_str(cluster->host_stat->node->address));

		node_list_contains_cluster_head = false;
		SERVICE_init_host_stat(host_stat, cluster->host_stat->node->address);

		list_iterator_init(cluster->node_stats, &node_stats_iterator);
		while((node_stat = (SERVICE_node_stat_t*) list_iterator_next(&node_stats_iterator)) != NULL) {
			if(node_stat->node->address == cluster->host_stat->node->address) {
				C_ABORT_ON(node_list_contains_cluster_head, "GCMI/CLUST: Cluster head %s is contained multiple times in node list.\n", addr_to_str(cluster->host_stat->node->address));
				node_list_contains_cluster_head = true;
			}

			node_id = node_address_to_id(node_stat->node->address);
			C_ABORT_ON(node_id >= get_number_of_known_nodes(), "GCMI/CLUST: Node with invalid ID %d found in cluster.\n", node_id);
			C_ABORT_ON(node_id_to_cluster_map[node_id] != cluster, "GCMI/CLUST: Node ID to cluster mapping is inconsistent with clusters (%p != %p).\n", node_id_to_cluster_map[node_id], cluster);

			SERVICE_add_node_stat_to_host_stat(host_stat, node_stat);
		}
		C_ABORT_ON(!node_list_contains_cluster_head, "GCMI/CLUST: Cluster head %s is not contained in node list.\n", addr_to_str(cluster->host_stat->node->address));

#ifdef ARCH_NS2
		C_ABORT_ON(SERVICE_host_stat_to_cost(cluster->host_stat) != SERVICE_host_stat_to_cost(host_stat), "GCMI/CLUST: Cluster with head %s reports inconsitent host statistics (%.1f != %.1f).\n", addr_to_str(cluster->host_stat->node->address), SERVICE_host_stat_to_cost(cluster->host_stat), SERVICE_host_stat_to_cost(host_stat));
#else
		// The cost of a cluster may decrease due to processing time,
		// because the data rate is measured using a running average.
		if(SERVICE_host_stat_to_cost(cluster->host_stat) != SERVICE_host_stat_to_cost(host_stat))
			C_DEBUG(GCMI_CLUSTERING_DEBUG_LEVEL, C_MSG_DEBUG, "GCMI/CLUST: Cluster with head %s reports inconsitent host statistics (%.1f != %.1f).\n", addr_to_str(cluster->host_stat->node->address), SERVICE_host_stat_to_cost(cluster->host_stat), SERVICE_host_stat_to_cost(host_stat));
#endif	// ARCH_NS2

		if(list_size(clusters) > 1) {
			C_ABORT_ON(list_empty(cluster->adjacent_clusters), "GCMI/CLUST: None of the %d other clusters is adjacent to cluster with head %s.\n", (list_size(clusters) - 1), addr_to_str(cluster->host_stat->node->address));
		}

		list_iterator_init(cluster->adjacent_clusters, &adjacent_clusters_iterator);
		while((adjacent_cluster = (cluster_t*) list_iterator_next(&adjacent_clusters_iterator)) != NULL) {
			C_ABORT_ON(!list_contains(clusters, adjacent_cluster), "GCMI/CLUST: Global cluster list does not contain adjacent cluster at %p.\n", adjacent_cluster);
		}

	}

	SERVICE_free_host_stat(&host_stat);

	for(node_id = 0; node_id < get_number_of_known_nodes(); node_id++) {
		if(node_id_to_cluster_map[node_id] != NULL) {
			C_ABORT_ON(!list_contains(clusters, node_id_to_cluster_map[node_id]), "GCMI/CLUST: Global cluster list does not contain cluster at %p pointed to by node ID %d (%s).\n", node_id_to_cluster_map[node_id], node_id, addr_to_str(node_id_to_address(node_id)));

			distance_to_cluster_head = get_distance_between_nodes(node_id_to_address(node_id), node_id_to_cluster_map[node_id]->host_stat->node->address);
			C_ABORT_ON(distance_to_cluster_head >= C_MAXIMAL_NETWORK_DIAMETER, "GCMI/CLUST: %s is not connected to its cluster head %s.\n", addr_to_str(node_id_to_address(node_id)), addr_to_str(node_id_to_cluster_map[node_id]->host_stat->node->address));

#ifndef C_SPI_OPTIMIZED_FOR_UNRELIABLE_LINKS
 			list_iterator_init(clusters, &clusters_iterator);
			while((cluster = (cluster_t*) list_iterator_next(&clusters_iterator)) != NULL) {
				float distance_to_alternative_cluster_head = get_distance_between_nodes(node_id_to_address(node_id), cluster->host_stat->node->address);
				C_ABORT_ON(distance_to_alternative_cluster_head < distance_to_cluster_head, "GCMI/CLUST: Distance of %s to cluster with head %s is smaller than distance to own cluster head %s (%.1f < %.1f).\n", addr_to_str(node_id_to_address(node_id)), addr_to_str(cluster->host_stat->node->address), addr_to_str(node_id_to_cluster_map[node_id]->host_stat->node->address), distance_to_alternative_cluster_head, distance_to_cluster_head);
			}
#endif	// C_SPI_OPTIMIZED_FOR_UNRELIABLE_LINKS
		}
	}
}
#endif	// C_ENABLE_CONSISTENCY_CHECKS



/**
 * @brief Compare the addresses of two node stats.
 *
 * @param item Node stat to check.
 * @param cmp_item Reference node stat.
 * @return <code>true</code> if the node stat has a smaller address than the
 * reference node stat, <code>false</code> otherwise.
 */
static bool node_stat_node_address_less_than(const void* item, const void* cmp_item) {
	SERVICE_node_stat_t* node_stat = (SERVICE_node_stat_t*) item;
	SERVICE_node_stat_t* cmp_node_stat = (SERVICE_node_stat_t*) cmp_item;
	return (node_stat->node->address < cmp_node_stat->node->address);
}

/**
 * @brief Print the node address of a node statistic item.
 *
 * @param debug_level Numeric debug level of the message.
 * @param msg_type Numeric type of the message.
 * @param item Pointer to node whose address is to be printed.
 */
static void print_node_stat_node_address(uint8_t debug_level, uint8_t msg_type, const void* item) {
	C_DEBUG(debug_level, msg_type, "%s", addr_to_str(((SERVICE_node_stat_t*) item)->node->address));
}

/**
 * @brief Compare the addresses of two cluster heads.
 *
 * @param item Cluster to check.
 * @param cmp_item Reference cluster.
 * @return <code>true</code> if the head of the cluster has a smaller address
 * than the head of the reference cluster, <code>false</code> otherwise.
 */
static bool cluster_head_address_less_than(const void* item, const void* cmp_item) {
	cluster_t* cluster = (cluster_t*) item;
	cluster_t* cmp_cluster = (cluster_t*) cmp_item;
	return (cluster->host_stat->node->address < cmp_cluster->host_stat->node->address);
}

/**
 * @brief Print the node address of a cluster head.
 *
 * @param debug_level Numeric debug level of the message.
 * @param msg_type Numeric type of the message.
 * @param item Pointer to cluser whose cluster head address is to be printed.
 */
static void print_cluster_head_address(uint8_t debug_level, uint8_t msg_type, const void* item) {
	C_DEBUG(debug_level, msg_type, "%s", addr_to_str(((cluster_t*) item)->host_stat->node->address));
}

/**
 * @brief Print all clusters.
 *
 * Time complexity:	O(clusters ^ 2
 *			  + clusters * (
 *			    avg_nodes_in_cluster ^ 2 + avg_nodes_in_cluster
 *			    + avg_adjacent_clusters ^ 2 + avg_adjacent_clusters))
 *			= O(clusters ^ 2
 *			    + clusters * (avg_nodes_in_cluster ^ 2 + avg_adjacent_clusters ^ 2))
 * Space complexity:	O(1)
 *
 * @param debug_level Numeric debug level of the message.
 * @param msg_type Numeric type of the message.
 * @param clusters Pointer to list of clusters.
 */
void print_clusters(uint8_t debug_level, uint8_t msg_type, list_t* clusters) {
	cluster_t* cluster = NULL;
	list_iterator_t clusters_iterator;
	uint16_t i = 1;

	list_sort(clusters, cluster_head_address_less_than);

	list_iterator_init(clusters, &clusters_iterator);
	while((cluster = (cluster_t*) list_iterator_next(&clusters_iterator)) != NULL) {
		C_DEBUG(debug_level, msg_type, "GCMI/CLUST: Cluster #%d: Head %s, %d node(s) ", i, addr_to_str(cluster->host_stat->node->address), list_size(cluster->node_stats));
		list_sort(cluster->node_stats, node_stat_node_address_less_than);
		list_print(debug_level, (msg_type | C_MSG_CONT), cluster->node_stats, print_node_stat_node_address);

		C_DEBUG(debug_level, (msg_type | C_MSG_CONT), ", adjacent cluster heads ");
		list_sort(cluster->adjacent_clusters, cluster_head_address_less_than);
		list_print(debug_level, (msg_type | C_MSG_CONT), cluster->adjacent_clusters, print_cluster_head_address);

		C_DEBUG(debug_level, (msg_type | C_MSG_CONT), ", total cost %.1f, average cost %.1f\n", SERVICE_host_stat_to_cost(cluster->host_stat), (SERVICE_host_stat_to_cost(cluster->host_stat) / list_size(cluster->node_stats)));
		i++;
	}
}



/**
 * @brief Allocate and initialize memory for cluster.
 *
 * Time complexity:	O(clusters + cluster_add_node())
 *			= O(clusters + nodes * avg_adjacent_clusters)
 *			= O(nodes * avg_adjacent_clusters)
 * Space complexity:	O(1)
 *
 * @param cluster Pointer to pointer where the cluster is to be allocated.
 * @param head_node_stat Node stat of the cluster head.
 * @param clusters Pointer to list of clusters.
 * @param node_id_to_cluster_map Array of pointers to clusters, indexed by node IDs.
 */
void cluster_init(cluster_t** cluster, SERVICE_node_stat_t* head_node_stat, list_t* clusters, cluster_t** node_id_to_cluster_map) {
	C_DEBUG(GCMI_CLUSTERING_DEBUG_LEVEL, C_MSG_INFO, "GCMI/CLUST: Initializing cluster with %s as head...\n", addr_to_str(head_node_stat->node->address));

	MALLOC(*cluster, cluster_t);
	list_add(clusters, *cluster);

	SERVICE_alloc_host_stat(&(*cluster)->host_stat, head_node_stat->node->address);
	list_alloc(&(*cluster)->node_stats);
	list_alloc(&(*cluster)->adjacent_clusters);

	cluster_add_node(*cluster, head_node_stat, clusters, node_id_to_cluster_map);
}

/**
 * @brief Free memory of a cluster.
 *
 * Time complexity:	O(adjacent_clusters + nodes_in_cluster)
 * Space complexity:	O(1)
 *
 * @param cluster Pointer to pointer to a cluster to be freed.
 */
static void cluster_free(cluster_t** cluster) {
	cluster_free_list(*cluster);
	*cluster = NULL;
}

/**
 * @brief Free memory of a cluster (list interface).
 *
 * Time complexity:	O(adjacent_clusters + nodes_in_cluster)
 * Space complexity:	O(1)
 *
 * @param p Pointer to host statistics structure to free.
 */
void cluster_free_list(void* p) {
	cluster_t* cluster = (cluster_t*) p;

	C_DEBUG(GCMI_CLUSTERING_DEBUG_LEVEL, C_MSG_INFO, "GCMI/CLUST: Freeing cluster with head %s...\n", addr_to_str(cluster->host_stat->node->address));

	SERVICE_free_host_stat(&cluster->host_stat);
	list_free(&cluster->node_stats);
	list_free(&cluster->adjacent_clusters);
	FREE(cluster);
}

/**
 * @brief Add node to a cluster.
 *
 * NOTE: This function does not update the cluster head. Clusters must be
 * consolidated after all nodes have been added.
 *
 * Time complexity:	O(nodes * avg_adjacent_clusters)
 * Space complexity:	O(1)
 *
 * @param cluster Pointer to cluster.
 * @param node_stat_to_add Pointer to node stat of node that is to be added.
 * @param clusters Pointer to list of clusters.
 * @param node_id_to_cluster_map Array of pointers to clusters, indexed by node IDs.
 * @return <code>true</code> if adjacency lists were changed as a side effect,
 * <code>false</code> otherwise.
 */
bool cluster_add_node(cluster_t* cluster, SERVICE_node_stat_t* node_stat_to_add, list_t* clusters, cluster_t** node_id_to_cluster_map) {
	uint16_t neighbor_node_id = C_UNKNOWN_NODE_ID;
	bool changed_adjacency_lists = false;
#ifdef C_ENABLE_CONSISTENCY_CHECKS
	float cost_before_adding_node = SERVICE_host_stat_to_cost(cluster->host_stat);
#endif	// C_ENABLE_CONSISTENCY_CHECKS

	C_ABORT_ON(node_id_to_cluster_map[node_address_to_id(node_stat_to_add->node->address)] != NULL, "GCMI/CLUST: Trying to add %s to cluster with head %s, but node is already in cluster with head %s...\n", addr_to_str(node_stat_to_add->node->address), addr_to_str(cluster->host_stat->node->address), addr_to_str(node_id_to_cluster_map[node_address_to_id(node_stat_to_add->node->address)]->host_stat->node->address));

	/// Add node to cluster data structure.

	C_DEBUG(GCMI_CLUSTERING_DEBUG_LEVEL, C_MSG_INFO, "GCMI/CLUST: Adding %s to cluster with head %s (without updating the cluster head)...\n", addr_to_str(node_stat_to_add->node->address), addr_to_str(cluster->host_stat->node->address));

	list_add(cluster->node_stats, node_stat_to_add);
	SERVICE_add_node_stat_to_host_stat(cluster->host_stat, node_stat_to_add);
	node_id_to_cluster_map[node_address_to_id(node_stat_to_add->node->address)] = cluster;

	/// Update adjacent clusters.

	C_DEBUG(GCMI_CLUSTERING_DEBUG_LEVEL, C_MSG_DEBUG, "GCMI/CLUST: Updating adjacent clusters of cluster with head %s after adding %s...\n", addr_to_str(cluster->host_stat->node->address), addr_to_str(node_stat_to_add->node->address));

	for(neighbor_node_id = 0; neighbor_node_id < get_number_of_known_nodes(); neighbor_node_id++) {
		// Check if neighboring node belongs to another cluster.
		if(
			exists_known_link(node_stat_to_add->node->address, node_id_to_address(neighbor_node_id))
			&& (node_id_to_cluster_map[neighbor_node_id] != NULL)
			&& (node_id_to_cluster_map[neighbor_node_id] != cluster)
		) {
			C_DEBUG(GCMI_CLUSTERING_DEBUG_LEVEL, C_MSG_VERBOSE, "GCMI/CLUST: Neighboring nodes %s and %s imply that clusters with heads %s and %s are adjacent.\n", addr_to_str(node_stat_to_add->node->address), addr_to_str(node_id_to_address(neighbor_node_id)), addr_to_str(cluster->host_stat->node->address), addr_to_str(node_id_to_cluster_map[neighbor_node_id]->host_stat->node->address));
#ifdef C_ENABLE_NAM_TRACE_OUTPUT
//			C_trace("l -t %f -s %d -d %d -S DOWN -c red", C_time(), node_stat_to_add->node->address, node_id_to_address(neighbor_node_id));
#endif	// C_ENABLE_NAM_TRACE_OUTPUT

			// Update adjacent cluster lists on both sides.
			if(!list_contains(cluster->adjacent_clusters, node_id_to_cluster_map[neighbor_node_id])) {
				C_DEBUG(GCMI_CLUSTERING_DEBUG_LEVEL, C_MSG_INFO, "GCMI/CLUST: Adding cluster with head %s to adjacency list of cluster with head %s...\n", addr_to_str(node_id_to_cluster_map[neighbor_node_id]->host_stat->node->address), addr_to_str(cluster->host_stat->node->address));
				list_add(cluster->adjacent_clusters, node_id_to_cluster_map[neighbor_node_id]);
				changed_adjacency_lists = true;
			}
			if(!list_contains(node_id_to_cluster_map[neighbor_node_id]->adjacent_clusters, cluster)) {
				C_DEBUG(GCMI_CLUSTERING_DEBUG_LEVEL, C_MSG_INFO, "GCMI/CLUST: Adding cluster with head %s to adjacency list of cluster with head %s...\n", addr_to_str(cluster->host_stat->node->address), addr_to_str(node_id_to_cluster_map[neighbor_node_id]->host_stat->node->address));
				list_add(node_id_to_cluster_map[neighbor_node_id]->adjacent_clusters, cluster);
				changed_adjacency_lists = true;
			}
		}
	}

#ifdef C_ENABLE_CONSISTENCY_CHECKS
#ifdef ARCH_NS2
	C_ABORT_ON(SERVICE_host_stat_to_cost(cluster->host_stat) < cost_before_adding_node, "GCMI/CLUST: Cost for cluster with head %s has decreased while adding %s (%.1f < %.1f)...\n", addr_to_str(cluster->host_stat->node->address), addr_to_str(node_stat_to_add->node->address), SERVICE_host_stat_to_cost(cluster->host_stat), cost_before_adding_node);
#else
	// The cost of a cluster may decrease due to processing time, because
	// the data rate is measured using a running average.
	if(SERVICE_host_stat_to_cost(cluster->host_stat) < cost_before_adding_node)
		C_DEBUG(GCMI_CLUSTERING_DEBUG_LEVEL, C_MSG_DEBUG, "GCMI/CLUST: Cost for cluster with head %s has decreased while adding %s (%.1f < %.1f)...\n", addr_to_str(cluster->host_stat->node->address), addr_to_str(node_stat_to_add->node->address), SERVICE_host_stat_to_cost(cluster->host_stat), cost_before_adding_node);
#endif	// ARCH_NS2
#endif	// C_ENABLE_CONSISTENCY_CHECKS

	return changed_adjacency_lists;
}

/**
 * @brief Remove a node from a cluster.
 *
 * NOTE: This function does not update the cluster head.
 *
 * Time complexity:	O(nodes_in_cluster
 *			  + nodes * adjacent_clusters
 *			  + adjacent_clusters * (
 *			    nodes_in_cluster * nodes
 *			    + adjacent_clusters))
 *			= O(adjacent_clusters * nodes_in_cluster * nodes)
 * Space complexity:	O(adjacent_clusters)
 *
 * @param cluster Pointer to cluster.
 * @param node_stat_to_remove Pointer to node stat of node that is to be removed.
 * @param clusters Pointer to list of clusters.
 * @param node_id_to_cluster_map Array of pointers to clusters, indexed by node IDs.
 * @return <code>true</code> if adjacency lists were changed as a side effect,
 * <code>false</code> otherwise.
 */
static bool cluster_remove_node(cluster_t* cluster, SERVICE_node_stat_t* node_stat_to_remove, list_t* clusters, cluster_t** node_id_to_cluster_map) {
	list_t* adjacent_clusters_to_be_removed = NULL;
	uint16_t neighbor_node_id = C_UNKNOWN_NODE_ID;
	cluster_t* adjacent_cluster = NULL;
	list_iterator_t adjacent_clusters_iterator;
	bool alternative_path_exists = false;
	SERVICE_node_stat_t* node_stat = NULL;
	list_iterator_t node_stats_iterator;
	bool changed_adjacency_lists = false;
#ifdef C_ENABLE_CONSISTENCY_CHECKS
	float cost_before_removing_node = SERVICE_host_stat_to_cost(cluster->host_stat);
#endif	// C_ENABLE_CONSISTENCY_CHECKS

	C_ABORT_ON(node_stat_to_remove->node->address == cluster->host_stat->node->address, "GCMI/CLUST: Trying to remove %s from cluster with head %s, but this is the cluster head...\n", addr_to_str(node_stat_to_remove->node->address), addr_to_str(cluster->host_stat->node->address));
	C_ABORT_ON(node_id_to_cluster_map[node_address_to_id(node_stat_to_remove->node->address)] != cluster, "GCMI/CLUST: Trying to remove %s from cluster with head %s, but node is supposed to be in cluster with head %s...\n", addr_to_str(node_stat_to_remove->node->address), addr_to_str(cluster->host_stat->node->address), addr_to_str(node_id_to_cluster_map[node_address_to_id(node_stat_to_remove->node->address)]->host_stat->node->address));

	/// Remove node from cluster data structure.

	C_DEBUG(GCMI_CLUSTERING_DEBUG_LEVEL, C_MSG_INFO, "GCMI/CLUST: Removing %s from cluster with head %s (without updating the cluster head)...\n", addr_to_str(node_stat_to_remove->node->address), addr_to_str(cluster->host_stat->node->address));

	list_remove(cluster->node_stats, node_stat_to_remove);
	SERVICE_subtract_node_stat_to_host_stat(cluster->host_stat, node_stat_to_remove);
	node_id_to_cluster_map[node_address_to_id(node_stat_to_remove->node->address)] = NULL;

	/// Update adjacent clusters.

	C_DEBUG(GCMI_CLUSTERING_DEBUG_LEVEL, C_MSG_DEBUG, "GCMI/CLUST: Updating adjacent clusters of cluster with head %s after removing %s...\n", addr_to_str(cluster->host_stat->node->address), addr_to_str(node_stat_to_remove->node->address));

	list_alloc(&adjacent_clusters_to_be_removed);

	// Select all adjacent clusters that that were connected via the
	// removed node.
	for(neighbor_node_id = 0; neighbor_node_id < get_number_of_known_nodes(); neighbor_node_id++) {
		// Check if neighboring node belongs to another cluster.
		if(
			exists_known_link(node_stat_to_remove->node->address, node_id_to_address(neighbor_node_id))
			&& (node_id_to_cluster_map[neighbor_node_id] != NULL)
			&& (node_id_to_cluster_map[neighbor_node_id] != cluster)
			&& !list_contains(adjacent_clusters_to_be_removed, node_id_to_cluster_map[neighbor_node_id])
		) {
			C_DEBUG(GCMI_CLUSTERING_DEBUG_LEVEL, C_MSG_VERBOSE, "GCMI/CLUST: Cluster with head %s may not be adjacent anymore to cluster with head %s...\n", addr_to_str(node_id_to_cluster_map[neighbor_node_id]->host_stat->node->address), addr_to_str(cluster->host_stat->node->address));
			list_add(adjacent_clusters_to_be_removed, node_id_to_cluster_map[neighbor_node_id]);
		}
	}

	// Iterate over all candidates for removal.
	list_iterator_init(adjacent_clusters_to_be_removed, &adjacent_clusters_iterator);
	while((adjacent_cluster = (cluster_t*) list_iterator_next(&adjacent_clusters_iterator)) != NULL) {

		alternative_path_exists = false;

		// Check whether there is another path that connects the
		// updated cluster with the adjacent cluster.
		list_iterator_init(cluster->node_stats, &node_stats_iterator);
		while((node_stat = (SERVICE_node_stat_t*) list_iterator_next(&node_stats_iterator)) != NULL) {
			// Check if neighboring node belongs to removal candidate.
			for(neighbor_node_id = 0; neighbor_node_id < get_number_of_known_nodes(); neighbor_node_id++) {
				if(
					exists_known_link(node_stat->node->address, node_id_to_address(neighbor_node_id))
					&& (node_id_to_cluster_map[neighbor_node_id] == adjacent_cluster)
				) {
					C_DEBUG(GCMI_CLUSTERING_DEBUG_LEVEL, C_MSG_VERBOSE, "GCMI/CLUST: Found an alternative path from cluster with head %s to cluster with head %s via %s...\n", addr_to_str(cluster->host_stat->node->address), addr_to_str(node_id_to_cluster_map[neighbor_node_id]->host_stat->node->address), addr_to_str(node_id_to_address(neighbor_node_id)));
					alternative_path_exists = true;
					break;
				}
			}
			if(alternative_path_exists)
				break;
		}

		// If not, set to cluster not to be adjacent anymore.
		if(!alternative_path_exists) {
			C_ABORT_ON(!list_contains(cluster->adjacent_clusters, adjacent_cluster), "GCMI/CLUST: Trying to remove adjacent cluster with head %s from list of cluster with head %s, but the cluster is not contained in the list.\n", addr_to_str(adjacent_cluster->host_stat->node->address), addr_to_str(cluster->host_stat->node->address));
			C_ABORT_ON(!list_contains(adjacent_cluster->adjacent_clusters, cluster), "GCMI/CLUST: Trying to remove adjacent cluster with head %s from list of cluster with head %s, but the cluster is not contained in the list.\n", addr_to_str(cluster->host_stat->node->address), addr_to_str(adjacent_cluster->host_stat->node->address));

			C_DEBUG(GCMI_CLUSTERING_DEBUG_LEVEL, C_MSG_INFO, "GCMI/CLUST: Removing clusters with heads %s and %s from each others adjacency lists...\n", addr_to_str(cluster->host_stat->node->address), addr_to_str(adjacent_cluster->host_stat->node->address));
			list_remove(cluster->adjacent_clusters, adjacent_cluster);
			list_remove(adjacent_cluster->adjacent_clusters, cluster);
			changed_adjacency_lists = true;
		}

	}

	list_free(&adjacent_clusters_to_be_removed);

#ifdef C_ENABLE_CONSISTENCY_CHECKS
	C_ABORT_ON(SERVICE_host_stat_to_cost(cluster->host_stat) > cost_before_removing_node, "GCMI/CLUST: Cost for cluster with head %s has increased while removing %s (%.1f < %.1f)...\n", addr_to_str(cluster->host_stat->node->address), addr_to_str(node_stat_to_remove->node->address), SERVICE_host_stat_to_cost(cluster->host_stat), cost_before_removing_node);
#endif	// C_ENABLE_CONSISTENCY_CHECKS

	return changed_adjacency_lists;
}

/**
 * @brief Update to cluster head of a cluster.
 *
 * This function updates the cluster head of a cluster to the most cost
 * efficient node.
 *
 * Time complexity:	O(nodes_in_cluster * GCSI_calculate_service_cost())
 *			O(nodes_in_cluster * nodes_in_cluster * 1)
 *			= O(nodes_in_cluster ^ 2)
 * Space complexity:	O(1)
 *
 * @param cluster Pointer to cluster whose head should be updated.
 * @return <code>true</code> if the cluster head has changed,
 * <code>false</code> otherwise.
 */
static bool cluster_update_head(cluster_t* cluster) {
	SERVICE_node_stat_t* optimal_node_stat = NULL;
	float optimal_service_cost = FLT_MAX;
	SERVICE_node_stat_t* node_stat = NULL;
	list_iterator_t node_stats_iterator;
	float host_candidate_cost = 0;

#ifdef C_SPI_OPTIMIZED_FOR_UNRELIABLE_LINKS
	// Never change the cluster head when running on unreliable links. The
	// intial cluster heads were selected with the probability of a
	// successful replication in mind.
	return false;
#endif	// C_SPI_OPTIMIZED_FOR_UNRELIABLE_LINKS

	C_DEBUG(GCMI_CLUSTERING_DEBUG_LEVEL, C_MSG_DEBUG, "GCMI/CLUST: Updating cluster head of cluster with head %s...\n", addr_to_str(cluster->host_stat->node->address));

	// Find the node with the smallest service cost, as given by the sum
	// over all nodes of the product of distance and service traffic.
	list_iterator_init(cluster->node_stats, &node_stats_iterator);
	while((node_stat = (SERVICE_node_stat_t*) list_iterator_next(&node_stats_iterator)) != NULL) {

		// Calculate the service cost for the current service host of
		// this loop iteration.
		host_candidate_cost = GCSI_calculate_service_cost(cluster->node_stats, node_stat->node->address);

		// Set optimal target node and service cost if the cost for the
		// current service host in this loop iteration is lower.
		if(
			(optimal_service_cost > host_candidate_cost)
			&& (!node_stat->refused_migration)	// Only consider nodes that have not refused a migration in the past.
		) {
			optimal_node_stat = node_stat;
			optimal_service_cost = host_candidate_cost;
		}
	}

	if(optimal_node_stat == NULL) {
		C_DEBUG(GCMI_CLUSTERING_DEBUG_LEVEL, C_MSG_ERROR, "GCMI/CLUST: All nodes in cluster with %s as head have previously refused to accept a service migration. Leaving cluster head unchanged...\n", addr_to_str(cluster->host_stat->node->address));
		return false;
	}

	if(cluster->host_stat->node->address != optimal_node_stat->node->address) {
		C_DEBUG(GCMI_CLUSTERING_DEBUG_LEVEL, C_MSG_INFO, "GCMI/CLUST: Cluster head changed from %s to %s, service cost is %.1f.\n", addr_to_str(cluster->host_stat->node->address), addr_to_str(optimal_node_stat->node->address), optimal_service_cost);
		// NOTE: This leaves the statistics in the host_stat structure
		// unchanged, just as it's supposed to be.
		cluster->host_stat->node = optimal_node_stat->node;
		return true;
	} else {
		C_DEBUG(GCMI_CLUSTERING_DEBUG_LEVEL, C_MSG_INFO, "GCMI/CLUST: Cluster head remains unchanged at %s, service cost is %.1f.\n", addr_to_str(cluster->host_stat->node->address), optimal_service_cost);
		return false;
	}
}

/**
 * @brief Conditionally move nodes from one cluster to an adjacent cluster.
 *
 * Move nodes from one cluster to another if the respective cluster head is
 * closer.
 *
 * Time complexity:	O(nodes_in_cluster * (
 *			  clusters
 *			  + cluster_remove_node()
 *			  + cluster_add_node()
 *			  + cluster_update_head()))
 *			= O(nodes_in_cluster * (
 *			    clusters
 *			    + (adjacent_clusters * nodes_in_cluster * nodes)
 *			    + (nodes * avg_adjacent_clusters)
 *			    + (nodes_in_cluster ^ 2)))
 *			= O(nodes_in_cluster ^ 2 * adjacent_clusters * nodes)
 * Space complexity:	O(nodes_in_cluster)
 *
 * @param cluster Pointer to cluster whose nodes may have to be moved.
 * @param clusters Pointer to list of clusters.
 * @param affected_clusters Pointer to list of affected clusters, i.e., whose
 * nodes might have to change clusters due to updated cluster heads.
 * @param node_id_to_cluster_map Array of pointers to clusters, indexed by node IDs.
 * @return <code>true</code> if a cluster head was changed as a side effect,
 * <code>false</code> otherwise.
 */
static bool cluster_conditionally_move_nodes_to_adjacent_cluster(cluster_t* cluster, list_t* clusters, list_t* affected_clusters, cluster_t** node_id_to_cluster_map) {
	list_t* node_stats_in_merged_cluster = NULL;
	SERVICE_node_stat_t* node_stat = NULL;
	list_iterator_t node_stats_iterator;
	cluster_t* candidate_cluster = NULL;
	list_iterator_t clusters_iterator;
	float distance_to_clostest_cluster_head = FLT_MAX;
	cluster_t* cluster_with_clostest_cluster_head = NULL;
	list_t* adjacent_clusters_before_moving_node = NULL;
	bool changed_adjacency_lists = false;
	bool changed_affected_clusters = false;

	C_DEBUG(GCMI_CLUSTERING_DEBUG_LEVEL, C_MSG_DEBUG, "GCMI/CLUST: Checking whether nodes need to be moved from cluster with head %s...\n", addr_to_str(cluster->host_stat->node->address));

#ifdef C_ENABLE_CONSISTENCY_CHECKS
	if(list_size(clusters) > 1) {
		cluster_t* cluster = NULL;
		list_iterator_init(clusters, &clusters_iterator);
		while((cluster = (cluster_t*) list_iterator_next(&clusters_iterator)) != NULL) {
			C_ABORT_ON(list_empty(cluster->adjacent_clusters), "GCMI/CLUST: Trying to move nodes between clusters, but none of the %d other clusters is adjacent to cluster with head %s.\n", (list_size(clusters) - 1), addr_to_str(cluster->host_stat->node->address));
		}
	}
#endif	// C_ENABLE_CONSISTENCY_CHECKS

	// Iterate over all nodes in the cluster.
	// NOTE: We have to iterate over a copy of the node stats list, because
	// the list may change as we move nodes between clusters.
	list_alloc(&node_stats_in_merged_cluster);
	list_append(node_stats_in_merged_cluster, cluster->node_stats);
	list_iterator_init(node_stats_in_merged_cluster, &node_stats_iterator);
	while((node_stat = (SERVICE_node_stat_t*) list_iterator_next(&node_stats_iterator)) != NULL) {

		// Find closest cluster head by checking the distances to
		// all other cluster heads.
		distance_to_clostest_cluster_head = get_distance_between_nodes(node_stat->node->address, cluster->host_stat->node->address);
		cluster_with_clostest_cluster_head = cluster;
		list_iterator_init(clusters, &clusters_iterator);
		while((candidate_cluster = (cluster_t*) list_iterator_next(&clusters_iterator)) != NULL) {
			if(get_distance_between_nodes(node_stat->node->address, candidate_cluster->host_stat->node->address) < distance_to_clostest_cluster_head) {
				C_DEBUG(GCMI_CLUSTERING_DEBUG_LEVEL, C_MSG_DEBUG, "GCMI/CLUST: %s is closer to cluster with head %s than to previously closest cluster with head %s... (%.1f < %.1f)\n", addr_to_str(node_stat->node->address), addr_to_str(candidate_cluster->host_stat->node->address), addr_to_str(cluster_with_clostest_cluster_head->host_stat->node->address), get_distance_between_nodes(node_stat->node->address, candidate_cluster->host_stat->node->address), distance_to_clostest_cluster_head);
				distance_to_clostest_cluster_head = get_distance_between_nodes(node_stat->node->address, candidate_cluster->host_stat->node->address);
				cluster_with_clostest_cluster_head = candidate_cluster;
			}
		}

		// If closest cluster head is not the same as the head of the
		// current cluster, move node to another cluster.
		if(cluster_with_clostest_cluster_head != cluster) {
			C_DEBUG(GCMI_CLUSTERING_DEBUG_LEVEL, C_MSG_INFO, "GCMI/CLUST: Moving %s from cluster with head %s to cluster with head %s...\n", addr_to_str(node_stat->node->address), addr_to_str(cluster->host_stat->node->address), addr_to_str(cluster_with_clostest_cluster_head->host_stat->node->address));

			// NOTE: We need a copy of the adjacency list because
			// it might change as the node is removed.
			list_alloc(&adjacent_clusters_before_moving_node);
			list_append(adjacent_clusters_before_moving_node, cluster->adjacent_clusters);
			changed_adjacency_lists = false;
			changed_adjacency_lists |= cluster_remove_node(cluster, node_stat, clusters, node_id_to_cluster_map);
			changed_adjacency_lists |= cluster_add_node(cluster_with_clostest_cluster_head, node_stat, clusters, node_id_to_cluster_map);
			if(changed_adjacency_lists) {
				C_DEBUG(GCMI_CLUSTERING_DEBUG_LEVEL, C_MSG_DEBUG, "GCMI/CLUST: Adding adjacent clusters ");
				list_print(GCMI_CLUSTERING_DEBUG_LEVEL, (C_MSG_DEBUG | C_MSG_CONT), adjacent_clusters_before_moving_node, print_cluster_head_address);
				C_DEBUG(GCMI_CLUSTERING_DEBUG_LEVEL, (C_MSG_DEBUG | C_MSG_CONT), " of cluster with head %s to affected clusters due to changing adjacency lists...\n", addr_to_str(cluster->host_stat->node->address));

				list_merge(affected_clusters, adjacent_clusters_before_moving_node);
				changed_affected_clusters = true;
			}
			list_free(&adjacent_clusters_before_moving_node);

			if(cluster_update_head(cluster_with_clostest_cluster_head)) {
				C_DEBUG(GCMI_CLUSTERING_DEBUG_LEVEL, C_MSG_DEBUG, "GCMI/CLUST: Adding adjacent clusters ");
				list_print(GCMI_CLUSTERING_DEBUG_LEVEL, (C_MSG_DEBUG | C_MSG_CONT), cluster_with_clostest_cluster_head->adjacent_clusters, print_cluster_head_address);
				C_DEBUG(GCMI_CLUSTERING_DEBUG_LEVEL, (C_MSG_DEBUG | C_MSG_CONT), " of cluster with head %s to affected clusters due to changing in cluster head...\n", addr_to_str(cluster_with_clostest_cluster_head->host_stat->node->address));

				list_merge(affected_clusters, cluster_with_clostest_cluster_head->adjacent_clusters);
				changed_affected_clusters = true;
			}
		}

#ifdef C_ENABLE_CONSISTENCY_CHECKS
		if(list_size(clusters) > 1) {
			cluster_t* cluster = NULL;
			list_iterator_init(clusters, &clusters_iterator);
			while((cluster = (cluster_t*) list_iterator_next(&clusters_iterator)) != NULL) {
				C_ABORT_ON(list_empty(cluster->adjacent_clusters), "GCMI/CLUST: Moved %s between clusters, but now none of the %d other clusters is adjacent to cluster with head %s.\n", addr_to_str(node_stat->node->address), (list_size(clusters) - 1), addr_to_str(cluster->host_stat->node->address));
			}
		}
#endif	// C_ENABLE_CONSISTENCY_CHECKS

	}
	list_free(&node_stats_in_merged_cluster);

	return changed_affected_clusters;
}

/**
 * @brief Consolidate clusters after a cluster has been modified.
 *
 * Time complexity:	O(cluster_update_head()
 *			  + adjacent_clusters
 *			  + cluster_head_changes * (
 *			    affected_clusters * (
 *			      cluster_conditionally_move_nodes_to_adjacent_cluster()
 *			      + cluster_update_head()))
 *			  + adjacent_clusters)
 *			= O(nodes_in_cluster ^ 2
 *			    + adjacent_clusters
 *			    + cluster_head_changes * (
 *			      affected_clusters * (
 *			        avg_nodes_in_cluster ^ 2 * avg_adjacent_clusters * nodes
 *			        + avg_nodes_in_cluster ^ 2))
 *			    + adjacent_clusters)
 *			= O(nodes_in_cluster ^ 2
 *			    + cluster_head_changes * affected_clusters
 *			      * avg_nodes_in_cluster ^ 2 * avg_adjacent_clusters * nodes)
 *			= O(cluster_head_changes * affected_clusters
 *			    * avg_nodes_in_cluster ^ 2 * avg_adjacent_clusters * nodes)
 *			  NOTE: It's not obvious how to estimate cluster_head_changes.
 *			  Empirical data suggests it's constant, hence we assume O(1).
 *			= O(affected_clusters
 *			    * avg_nodes_in_cluster ^ 2 * avg_adjacent_clusters * nodes)
 *			  NOTE: It's not obvious how to estimate affected_clusters.
 *			  It's slightly larger than O(avg_adjacent_clusters), but
 *			  empirical data suggest only by a constant factor.
 *			= O(avg_adjacent_clusters ^ 2 * avg_nodes_in_cluster ^ 2 * nodes)
 * Space complexity:	O(affected_clusters + avg_nodes_in_cluster)
 *			= O(adjacent_clusters + avg_nodes_in_cluster)
 *
 * @param changed_cluster Pointer to cluster whose nodes may have to be moved.
 * @param clusters Pointer to list of clusters.
 * @param node_id_to_cluster_map Array of pointers to clusters, indexed by node IDs.
 */
void cluster_consolidate(cluster_t* changed_cluster, list_t* clusters, cluster_t** node_id_to_cluster_map) {
	list_t* affected_clusters = NULL;
	list_iterator_t affected_clusters_iterator;
	bool changed_affected_clusters = false;
	cluster_t* cluster = NULL;

	C_DEBUG(GCMI_CLUSTERING_DEBUG_LEVEL, C_MSG_DEBUG, "GCMI/CLUST: Consolidating clusters after changing cluster with head %s...\n", addr_to_str(changed_cluster->host_stat->node->address));

	cluster_update_head(changed_cluster);

	list_alloc(&affected_clusters);
	list_add(affected_clusters, changed_cluster);
	list_append(affected_clusters, changed_cluster->adjacent_clusters);
	do {
		C_DEBUG(GCMI_CLUSTERING_DEBUG_LEVEL, C_MSG_DEBUG, "GCMI/CLUST: Affected clusters are ");
		list_print(GCMI_CLUSTERING_DEBUG_LEVEL, (C_MSG_DEBUG | C_MSG_CONT), affected_clusters, print_cluster_head_address);
		C_DEBUG(GCMI_CLUSTERING_DEBUG_LEVEL, (C_MSG_DEBUG | C_MSG_CONT), ".\n");
		changed_affected_clusters = false;

		list_iterator_init(affected_clusters, &affected_clusters_iterator);
		while((cluster = (cluster_t*) list_iterator_next(&affected_clusters_iterator)) != NULL) {

			if(cluster_conditionally_move_nodes_to_adjacent_cluster(cluster, clusters, affected_clusters, node_id_to_cluster_map)) {
				changed_affected_clusters = true;

				// Modifying affected_clusters invalidates the
				// iterator. And since we have to iterator over
				// all other clusters one more time anyway,
				// it's better to break now rather than storing
				// cluster->adjacent_clusters in another
				// to-be-merged list.
				break;
			}

			if(cluster_update_head(cluster)) {
				changed_affected_clusters = true;

				// Cluster head of adjacent cluster has
				// changed, hence we must check its adjacent
				// clusters, too.
				C_DEBUG(GCMI_CLUSTERING_DEBUG_LEVEL, C_MSG_DEBUG, "GCMI/CLUST: Adding adjacent clusters ");
				list_print(GCMI_CLUSTERING_DEBUG_LEVEL, (C_MSG_DEBUG | C_MSG_CONT), cluster->adjacent_clusters, print_cluster_head_address);
				C_DEBUG(GCMI_CLUSTERING_DEBUG_LEVEL, (C_MSG_DEBUG | C_MSG_CONT), " of cluster with head %s to affected clusters...\n", addr_to_str(cluster->host_stat->node->address));
				list_merge(affected_clusters, cluster->adjacent_clusters);

				// See above.
				break;
			}
		}

		if(changed_affected_clusters)
			C_DEBUG(GCMI_CLUSTERING_DEBUG_LEVEL, C_MSG_DEBUG, "GCMI/CLUST: Affected clusters have changed, re-running consolidation...\n");

	} while(changed_affected_clusters);
	list_free(&affected_clusters);

#ifdef C_ENABLE_CONSISTENCY_CHECKS
	check_clusters(clusters, node_id_to_cluster_map);
#endif	// C_ENABLE_CONSISTENCY_CHECKS

	C_DEBUG(GCMI_CLUSTERING_DEBUG_LEVEL, C_MSG_DEBUG, "GCMI/CLUST: Consolidation done.\n");
}

/**
 * @brief Merge two clusters.
 *
 * This function adds all nodes of one cluster (src_cluster) to another cluster
 * (dst_cluster). Previously existing nodes in the second cluster (dst_cluster)
 * remain unchanged.
 *
 * Time complexity:	O(clusters
 *			  + clusters * avg_adjacent_clusters
 *			  + nodes_in_cluster
 *			  + nodes_in_cluster * cluster_add_node()
 *			  + cluster_free()
 *			  + cluster_consolidate())
 *			= O(clusters
 *			    + clusters * avg_adjacent_clusters
 *			    + nodes_in_cluster
 *			    + nodes_in_cluster * nodes * avg_adjacent_clusters
 *			    + adjacent_clusters + nodes_in_cluster
 *			    + avg_adjacent_clusters ^ 2 * avg_nodes_in_cluster ^ 2 * nodes)
 *			= O(clusters * avg_adjacent_clusters
 *			    + nodes_in_cluster * nodes * avg_adjacent_clusters
 *			    + avg_adjacent_clusters ^ 2 * avg_nodes_in_cluster ^ 2 * nodes)
 *			= O(avg_adjacent_clusters ^ 2 * avg_nodes_in_cluster ^ 2 * nodes)
 * Space complexity:	O(adjacent_clusters + avg_nodes_in_cluster)
 *
 * @param dst_cluster Pointer to destination cluster.
 * @param src_cluster Pointer to source cluster.
 * @param clusters Pointer to list of clusters.
 * @param node_id_to_cluster_map Array of pointers to clusters, indexed by node IDs.
 */
void cluster_merge(cluster_t* dst_cluster, cluster_t* src_cluster, list_t* clusters, cluster_t** node_id_to_cluster_map) {
	cluster_t* cluster = NULL;
	list_iterator_t clusters_iterator;
	SERVICE_node_stat_t* node_stat = NULL;
	list_iterator_t node_stats_iterator;

	// TODO: Remove this after making sure that the check below doesn't trigger.
	C_DEBUG(GCMI_CLUSTERING_DEBUG_LEVEL, C_MSG_NOTE, "GCMI/CLUST: Merging cluster with head %s into cluster with head %s (checking clusters)...\n", addr_to_str(src_cluster->host_stat->node->address), addr_to_str(dst_cluster->host_stat->node->address));

#ifdef C_ENABLE_CONSISTENCY_CHECKS
	check_clusters(clusters, node_id_to_cluster_map);
#endif	// C_ENABLE_CONSISTENCY_CHECKS

	C_DEBUG(GCMI_CLUSTERING_DEBUG_LEVEL, C_MSG_NOTE, "GCMI/CLUST: Merging cluster with head %s into cluster with head %s...\n", addr_to_str(src_cluster->host_stat->node->address), addr_to_str(dst_cluster->host_stat->node->address));

	/// Move all nodes from source to destination cluster and free source
	/// cluster.

	// Remove source cluster from global data structures.
	list_remove(clusters, src_cluster);
	list_iterator_init(clusters, &clusters_iterator);
	while((cluster = (cluster_t*) list_iterator_next(&clusters_iterator)) != NULL) {
		if(list_contains(cluster->adjacent_clusters, src_cluster))
			list_remove(cluster->adjacent_clusters, src_cluster);
	}
	list_iterator_init(src_cluster->node_stats, &node_stats_iterator);
	while((node_stat = (SERVICE_node_stat_t*) list_iterator_next(&node_stats_iterator)) != NULL) {
		node_id_to_cluster_map[node_address_to_id(node_stat->node->address)] = NULL;
	}

	// Add all nodes from source cluster to destination cluster.
	list_iterator_init(src_cluster->node_stats, &node_stats_iterator);
	while((node_stat = (SERVICE_node_stat_t*) list_iterator_next(&node_stats_iterator)) != NULL) {
		cluster_add_node(dst_cluster, node_stat, clusters, node_id_to_cluster_map);
	}

	// Free source cluster.
	cluster_free(&src_cluster);

	// Now proceed to consolidate all affected clusters, starting with the
	// merged cluster and its adjacent clusters.
	cluster_consolidate(dst_cluster, clusters, node_id_to_cluster_map);

	// NOTE: This function call is left out of the complexity analysis of
	// this function.
	C_IF_DEBUG(GCMI_CLUSTERING_DEBUG_LEVEL, C_MSG_VERBOSE)
		print_clusters(GCMI_CLUSTERING_DEBUG_LEVEL, C_MSG_VERBOSE, clusters);

#ifdef C_ENABLE_CONSISTENCY_CHECKS
	check_clusters(clusters, node_id_to_cluster_map);
#endif	// C_ENABLE_CONSISTENCY_CHECKS
}

/**
 * @brief Generate a list of cluster head addresses.
 *
 * It is up to the caller to free the list and its elements.
 *
 * Time complexity:	O(clusters)
 * Space complexity:	O(1)
 *
 * @param clusters Pointer to list of clusters.
 * @param cluster_head_addresses Pointer to list of cluster head addresses (return value).
 */
void cluster_to_list_of_head_addresses(const list_t* clusters, list_t* cluster_head_addresses) {
	cluster_t* cluster = NULL;
	list_iterator_t clusters_iterator;
	C_addr_t* address = NULL;

	C_ABORT_ON(cluster_head_addresses == NULL, "GCMI/CLUST: Trying to generate list of cluster heads, but target list is at NULL.\n");
	C_ABORT_ON(!list_empty(cluster_head_addresses), "GCMI/CLUST: Trying to generate list of cluster head, but target list is not empty.\n");

	list_iterator_init(clusters, &clusters_iterator);
	while((cluster = (cluster_t*) list_iterator_next(&clusters_iterator)) != NULL) {
		MALLOC(address, C_addr_t);
		*address = cluster->host_stat->node->address;
		list_add(cluster_head_addresses, address);
		address = NULL;
	}
}

#endif	// C_SERVICE_PLACEMENT_EVALUATION
