/**
 * @file graph_cost_multiple_instances.c
 *
 * @brief Graph cost multiple instances.
 *
 * NOTE: The complexity analysis for each function is based on an earlier
 * version that explicitly calculated edge cost and distances of the network
 * graph. This is now done on demand by the mapping component. The analysis
 * remains valid.
 *
 * Copyright (c) 2007, 2008, 2009, 2010 Georg Wittenburg, Freie Universitaet
 * Berlin (wittenbu@inf.fu-berlin.de). All rights reserved.
 * Please refer to the COPYING file for detailed licensing information.
 *
 * @author Georg Wittenburg
 *
 * $Id: graph_cost_multiple_instances.c 3481 2010-04-27 21:38:53Z gwitten $
 */

#include <float.h>
#include <math.h>

#include "../../config.h"
#include "../../macros.h"
#include "../../types.h"
#include "../../vars.h"
#include "../../main.h"
#include "../../util/callback.h"
#include "../../util/debug.h"
#include "../../util/list.h"
#include "../../util/mapping.h"
#include "../../util/net.h"
#include "../../util/trigger.h"
#include "../../service/service.h"
#include "../placement.h"
#include "../placement_types.h"
#include "../placement_vars.h"
#include "../graph_cost_single_instance/graph_cost_single_instance.h"
#include "graph_cost_multiple_instances.h"
#include "gcmi_cost_calculation.h"
#include "gcmi_clustering.h"



#ifdef C_SERVICE_PLACEMENT_EVALUATION

#ifndef C_USE_NETWORK_MAPPING
	#error Trying to compile GCMI without network mapping support.
#endif	// C_USE_NETWORK_MAPPING

#ifndef C_ENABLE_SPI_OPTIMIZATIONS
	#warning Compiling GCMI without SPI optimizations...
#endif	// C_ENABLE_SPI_OPTIMIZATIONS



/**
 * @brief Debug level for this component.
 */
#define GCMI_DEBUG_LEVEL				C_MSG_INFO
// #define GCMI_DEBUG_LEVEL				C_MSG_VERBOSE

/**
 * @brief Minimal time interval between subsequent runs of the algorithm.
 */
#define GCMI_MINIMAL_RUN_INTERVAL			1.0

/**
 * @brief Maximal number of replications from a single host per adaptation.
 */
#define GCMI_MAXIMAL_REPLICATIONS_FROM_HOST		3

/**
 * @brief Hysteresis (multiplier) between estimated saving and adaptation cost.
 */
#define GCMI_ADAPTATION_HYSTERESIS_MULTIPLIER		1.05

/**
 * @brief Estimated data rate of service replication in B/s.
 */
#define GCMI_SERVICE_REPLICATION_DATARATE		5000.0



/**
 * @brief Interface to GCMI algorithm.
 */
PLACEMENT_algorithm_t graph_cost_multiple_instances_algorithm = {
	C_PLACEMENT_ALGORITHM_GRAPH_COST_MULTIPLE_INSTANCES,
	graph_cost_multiple_instances_start,
	graph_cost_multiple_instances_stop,
	graph_cost_multiple_instances_trigger,
	graph_cost_multiple_instances_get_data,
	graph_cost_multiple_instances_set_data,
	graph_cost_multiple_instances_update_data,
	NULL,
	NULL,
	graph_cost_multiple_instances_get_synchronization_hosts
};

/**
 * @brief GCMI algorithm data.
 */
typedef struct {
	C_time_t		last_run_time;			///< Time of the last run of the algorithm.
	trigger_t*		trigger;			///< Value-dependent event trigger.
	C_addr_t		coordinator_address;		///< Address of the coordinator node.
	list_t*			configuration;			///< Current service configuration, i.e. list of hosts.
	C_time_t		found_superior_configuration_time;	///< Point in time at which a superior configuration was found; or 0 if no superior configuration is known.
	bool			adaptation_in_progress;		///< Flag whether an adaptation is in progress.
	list_t*			inactive_hosts;			///< List of inactive hosts (except the coordinator), i.e. those from which no STATS have been recieved.
	callback_id_t		inactive_hosts_callback_id;	///< ID of callback until which all inactive hosts have to report in.
	float			previous_cost;			///< Cost as calculated during the previous iteration (debugging only).
} graph_cost_multiple_instances_data_t;



/**
 * @brief Timeout for inactive hosts to report in.
 *
 * For all hosts that have failed to report in before the timeout, we now
 * assume that their respective adaptation command has failed.
 *
 * @param parameter The service for the algorithm is running.
 */
static void inactive_hosts_timeout(callback_parameter_t parameter) {
	SERVICE_service_t* service = (SERVICE_service_t*) parameter;
	graph_cost_multiple_instances_data_t* placement_data = (graph_cost_multiple_instances_data_t*) service->placement_data;
	C_addr_t* inactive_host_address = NULL;

	C_ABORT_ON(list_empty(placement_data->inactive_hosts), "GCMI: Inactive host timeout has expired, but there are no inactive hosts to process.\n");
	placement_data->found_superior_configuration_time = 0.0;

	while(!list_empty(placement_data->inactive_hosts)) {
		inactive_host_address = (C_addr_t*) list_head(placement_data->inactive_hosts);
		C_DEBUG(GCMI_DEBUG_LEVEL, C_MSG_WARN, "GCMI: Host %s has not reported in for service \"%s\" after timeout, assuming adaptation failed and removing from current configuration...\n", addr_to_str(*inactive_host_address), service->id);
		list_remove(placement_data->configuration, inactive_host_address);
		list_remove(placement_data->inactive_hosts, inactive_host_address);
		FREE(inactive_host_address);
	}

	placement_data->inactive_hosts_callback_id = 0;
}

/**
 * @brief Process changed data of graph cost multiple instances algorithm.
 *
 * @param service Pointer to service for which to process data.
 * @param placement_data Pointer to placement data to process.
 */
static void process_placement_data(SERVICE_service_t* service, graph_cost_multiple_instances_data_t* placement_data) {
	C_addr_t* current_host_address = NULL;
	list_iterator_t current_hosts_iterator;

	placement_data->last_run_time = C_time();	// Reset to current time to gather some data before the next run.
	placement_data->found_superior_configuration_time = 0.0;

	if(placement_data->inactive_hosts_callback_id != 0) {
		callback_cancel(placement_data->inactive_hosts_callback_id);
		placement_data->inactive_hosts_callback_id = 0;
	}

#ifdef C_ENABLE_NAM_TRACE_OUTPUT
	C_trace("m -t %f -s %d -n %s -X", C_time(), C_LOCAL_ADDRESS, service->id);
#endif	// C_ENABLE_NAM_TRACE_OUTPUT

	if(placement_data->coordinator_address == C_LOCAL_ADDRESS) {

		// Set trigger parameters for service coordinator.
		trigger_reset(placement_data->trigger, 0, GCMI_COORDINATOR_TRIGGER_INITIAL_RANGE, GCMI_COORDINATOR_TRIGGER_DURATION);

		// Initialize list of inactive hosts.
		if(placement_data->inactive_hosts == NULL)
			list_alloc(&placement_data->inactive_hosts);
		else
			list_clear(placement_data->inactive_hosts);
		if(list_size(placement_data->configuration) > 1) {
			list_iterator_init(placement_data->configuration, &current_hosts_iterator);
			while((current_host_address = (C_addr_t*) list_iterator_next(&current_hosts_iterator)) != NULL) {
				if(*current_host_address != C_LOCAL_ADDRESS)
					list_add(placement_data->inactive_hosts, current_host_address);
			}
			// Start timeout until which inactive hosts have to
			// report in. The timeout is the sum of the time it
			// takes for the replications to complete and the time
			// for the new instance to report to the coordinator.
			// NOTE: For simplicity of the calculation, we assume
			// that a service replicates with 5kB/s.
			placement_data->inactive_hosts_callback_id = callback_add(inactive_hosts_timeout, service,
				(((float) GCMI_MAXIMAL_REPLICATIONS_FROM_HOST) * (((float) service->data_size) / GCMI_SERVICE_REPLICATION_DATARATE)
				+ 3.0 * (GCMI_HOST_TRIGGER_DURATION + GCMI_MINIMAL_RUN_INTERVAL))
			);
		}

		// Discard old stats. As hosts report in, more accurate data
		// will become available.
		SERVICE_reset_service_stats(service->id);

#ifdef C_ENABLE_NAM_TRACE_OUTPUT
		C_trace("m -t %f -s %d -n %s -h hexagon -c black", C_time(), C_LOCAL_ADDRESS, service->id);
#endif	// C_ENABLE_NAM_TRACE_OUTPUT

	} else {

		// Set trigger parameters for regular service instance.
		trigger_reset(placement_data->trigger, 0, GCMI_HOST_TRIGGER_INITIAL_RANGE, GCMI_HOST_TRIGGER_DURATION);

		// Free list of inactive hosts.
		if(placement_data->inactive_hosts != NULL)
			list_free(&placement_data->inactive_hosts);

#ifdef C_ENABLE_NAM_TRACE_OUTPUT
		C_trace("m -t %f -s %d -n %s -h circle -c black", C_time(), C_LOCAL_ADDRESS, service->id);
#endif	// C_ENABLE_NAM_TRACE_OUTPUT

	}

}

/**
 * @brief Initialize cluster data structures.
 *
 * Time complexity:	O(nodes * cluster_init())
 *			= O(nodes * nodes * avg_adjacent_clusters)
 *			= O(nodes ^ 2 * avg_adjacent_clusters)
 * Space complexity:	O(1)
 *
 * @param service Pointer to service.
 * @param current_host_addresses Pointer to list of service instance host addresses.
 * @param clusters Pointer to list of clusters.
 * @param node_id_to_cluster_map Array of pointers to clusters, indexed by node IDs.
 * @param number_of_initial_clusters Maximum number of clusters to create.
 */
static void init_clusters(const SERVICE_service_t* service, const list_t* current_host_addresses, list_t* clusters, cluster_t** node_id_to_cluster_map, unsigned int number_of_initial_clusters) {
	SERVICE_node_stat_t* node_stat = NULL;
	list_iterator_t note_stats_iterator;
	cluster_t* cluster = NULL;
#ifdef C_SPI_OPTIMIZED_FOR_UNRELIABLE_LINKS
	list_iterator_t current_host_addresses_iterator;
	C_addr_t* current_host_address = NULL;
	bool is_neighbor_of_current_host = false;
#endif	// C_SPI_OPTIMIZED_FOR_UNRELIABLE_LINKS

	C_ABORT_ON(number_of_initial_clusters == 0, "GCMI: Can't initialize 0 clusters.\n");

	list_iterator_init(service->node_stats, &note_stats_iterator);
	while(
		((node_stat = (SERVICE_node_stat_t*) list_iterator_next(&note_stats_iterator)) != NULL)
		&& (number_of_initial_clusters > 0)
	) {
		if(
			(node_stat->node->address == C_LOCAL_ADDRESS)
			|| (SERVICE_node_stat_to_cost(node_stat) > FLT_EPSILON)
		) {

			// Ignore nodes between which no path is known.
			if(!exists_known_path(node_stat->node->address, C_LOCAL_ADDRESS)) {
				C_DEBUG(GCMI_DEBUG_LEVEL, C_MSG_WARN, "GCMI: Trying to create a cluster for %s to which no path is known, skipping node...\n", addr_to_str(node_stat->node->address));
				continue;
			}

#ifdef C_SPI_OPTIMIZED_FOR_UNRELIABLE_LINKS
			// Skip nodes which are not direct neighbors of
			// existing hosts. Migrating to more distant nodes has
			// a lower probability of success.
			is_neighbor_of_current_host = false;
			list_iterator_init(current_host_addresses, &current_host_addresses_iterator);
			while((current_host_address = (C_addr_t*) list_iterator_next(&current_host_addresses_iterator)) != NULL) {
				if(
					(node_stat->node->address == *current_host_address)
					|| exists_known_link(node_stat->node->address, *current_host_address)
				) {
					is_neighbor_of_current_host = true;
					break;
				}
			}
			if(!is_neighbor_of_current_host)
				continue;
#endif	// C_SPI_OPTIMIZED_FOR_UNRELIABLE_LINKS

			C_DEBUG(GCMI_DEBUG_LEVEL, C_MSG_VERBOSE, "GCMI: Creating cluster for %s (%.1f service cost)...\n", addr_to_str(node_stat->node->address), SERVICE_node_stat_to_cost(node_stat));
			cluster_init(&cluster, node_stat, clusters, node_id_to_cluster_map);
			cluster = NULL;
			number_of_initial_clusters--;

		} else {
			C_DEBUG(GCMI_DEBUG_LEVEL, C_MSG_VERBOSE, "GCMI: Skipping %s while creating clusters (%.1f service cost)...\n", addr_to_str(node_stat->node->address), SERVICE_node_stat_to_cost(node_stat));
		}
	}
}

/**
 * @brief Add remaining nodes to clusters.
 *
 * The initial clustering consists only of a subset of nodes that actively
 * request a service. This function adds all the remaining known nodes to the
 * nearest cluster.
 *
 * Time complexity:	O(nodes * (
 *			  clusters + nodes
 *			  + cluster_add_node()))
 *			= O(nodes * (
 *			    clusters + nodes
 *			    + nodes * avg_adjacent_clusters))
 *			= O(nodes
 *			    * nodes * avg_adjacent_clusters)
 *			= O(nodes ^ 2 * node_degree)
 * Space complexity:	O(1)
 *
 * @param service Pointer to service.
 * @param clusters Pointer to list of clusters.
 * @param node_id_to_cluster_map Array of pointers to clusters, indexed by node IDs.
 */
static void add_remaining_nodes_to_clusters(SERVICE_service_t* service, list_t* clusters, cluster_t** node_id_to_cluster_map) {
	node_t* node = NULL;
	list_iterator_t C_known_nodes_iterator;
	cluster_t* cluster = NULL;
	list_iterator_t clusters_iterator;
	cluster_t* closest_cluster = NULL;
	float distance_to_closest_cluster_head = FLT_MAX;
	float distance_to_current_cluster_head = 0;
	SERVICE_node_stat_t* node_stat = NULL;
	SERVICE_node_stat_t cmp_node_stat;
	node_t cmp_node;
	memset(&cmp_node_stat, 0, sizeof(SERVICE_node_stat_t));
	memset(&cmp_node, 0, sizeof(node_t));
	cmp_node_stat.node = &cmp_node;

	list_iterator_init(C_known_nodes, &C_known_nodes_iterator);
	while((node = (node_t*) list_iterator_next(&C_known_nodes_iterator)) != NULL) {

		if(!exists_known_path(node->address, C_LOCAL_ADDRESS)) {
			C_DEBUG(GCMI_DEBUG_LEVEL, C_MSG_DEBUG, "GCMI: Skipping %s to which no path is known...\n", addr_to_str(node->address));
			continue;
		}
		if((cluster = node_id_to_cluster_map[node_address_to_id(node->address)]) != NULL) {
			C_DEBUG(GCMI_DEBUG_LEVEL, C_MSG_DEBUG, "GCMI: Skipping %s which is already contained in cluster with head %s...\n", addr_to_str(node->address), addr_to_str(cluster->host_stat->node->address));
			continue;
		}

		// Find the closest cluster for the node.
		closest_cluster = NULL;
		distance_to_closest_cluster_head = FLT_MAX;
		list_iterator_init(clusters, &clusters_iterator);
		while((cluster = (cluster_t*) list_iterator_next(&clusters_iterator)) != NULL) {
			distance_to_current_cluster_head = get_distance_between_nodes(node->address, cluster->host_stat->node->address);
			if(distance_to_current_cluster_head < distance_to_closest_cluster_head) {
				distance_to_closest_cluster_head = distance_to_current_cluster_head;
				closest_cluster = cluster;
			}
		}
		C_ABORT_ON(closest_cluster == NULL, "GCMI: No closest cluster found in list of %d cluster(s) while trying to add remaining nodes to clusters.\n", list_size(clusters));

		C_DEBUG(GCMI_DEBUG_LEVEL, C_MSG_DEBUG, "GCMI: Adding %s to cluster with %s as head...\n", addr_to_str(node->address), addr_to_str(closest_cluster->host_stat->node->address));
		cmp_node_stat.node->address = node->address;
		if((node_stat = (SERVICE_node_stat_t*) list_find(service->node_stats, &cmp_node_stat, SERVICE_matches_node_stat_address)) == NULL) {
			// Alloc and add node stats if not existant.
			SERVICE_alloc_node_stat(&node_stat, node->address);
			list_add(service->node_stats, node_stat);
		}
		cluster_add_node(closest_cluster, node_stat, clusters, node_id_to_cluster_map);
	}
}

/**
 * @brief Calculate the coordinator node for a given service configuration.
 *
 * Time complexity:	O(service_hosts ^ 2)
 * Space complexity:	O(1)
 *
 * @param service_configuration Pointer to list of host addresses.
 * @return Address of the best coordinator node for this service configuration.
 */
static C_addr_t calculate_coordinator(const list_t* service_configuration) {
	C_addr_t* service_host_address = NULL;
	list_iterator_t service_hosts_iterator;
	C_addr_t* service_host_address_2 = NULL;
	list_iterator_t service_hosts_iterator_2;
	C_addr_t coordinator_address = C_BROADCAST_ADDRESS;
	float coordinator_distance = FLT_MAX;
	float tmp_coordinator_distance = FLT_MAX;

	list_iterator_init(service_configuration, &service_hosts_iterator);
	while((service_host_address = (C_addr_t*) list_iterator_next(&service_hosts_iterator)) != NULL) {
		tmp_coordinator_distance = 0;

		list_iterator_init(service_configuration, &service_hosts_iterator_2);
		while((service_host_address_2 = (C_addr_t*) list_iterator_next(&service_hosts_iterator_2)) != NULL) {
			if(service_host_address_2 != service_host_address) {
				tmp_coordinator_distance += get_distance_between_nodes(*service_host_address_2, *service_host_address);
			}
		}
		if(tmp_coordinator_distance < coordinator_distance) {
			coordinator_distance = tmp_coordinator_distance;
			coordinator_address = *service_host_address;
		}
	}
	C_ABORT_ON(coordinator_address == C_BROADCAST_ADDRESS, "GCMI: Failed to calculate a new coordinator node.\n");

	return coordinator_address;
}

/**
 * @brief Find two clusters to merge.
 *
 * Time complexity:	O(clusters * avg_adjacent_clusters)
 *			= O(clusters * node_degree)
 * Space complexity:	O(1)
 *
 * @param clusters Pointer to list of clusters.
 * @param cluster_to_merge_A Pointer to first cluster to merge (return value).
 * @param cluster_to_merge_B Pointer to second cluster to merge (return value).
 * @return <code>true</code> if two clusters that are suitable for merging were
 * found, <code>false</code> otherwise.
 */
static bool find_clusters_to_merge(const list_t* clusters, cluster_t** cluster_to_merge_A, cluster_t** cluster_to_merge_B) {
	cluster_t* cluster = NULL;
	list_iterator_t clusters_iterator;
	cluster_t* adjacent_cluster = NULL;
	list_iterator_t adjacent_clusters_iterator;
	float recommended_cluster_pair_cost = FLT_MAX;
	float cluster_pair_cost = FLT_MAX;
	bool found_clusters_to_merge = false;

	list_iterator_init(clusters, &clusters_iterator);
	while((cluster = (cluster_t*) list_iterator_next(&clusters_iterator)) != NULL) {
		list_iterator_init(cluster->adjacent_clusters, &adjacent_clusters_iterator);
		while((adjacent_cluster = (cluster_t*) list_iterator_next(&adjacent_clusters_iterator)) != NULL) {
			cluster_pair_cost = SERVICE_host_stat_to_cost(cluster->host_stat) + SERVICE_host_stat_to_cost(adjacent_cluster->host_stat);
			if(cluster_pair_cost < recommended_cluster_pair_cost) {
				C_DEBUG(GCMI_DEBUG_LEVEL, C_MSG_VERBOSE, "GCMI: Adjacent clusters with %s and %s as heads have a combined client-level service cost of %.1f.\n", addr_to_str(cluster->host_stat->node->address), addr_to_str(adjacent_cluster->host_stat->node->address), cluster_pair_cost);
				*cluster_to_merge_A = cluster;
				*cluster_to_merge_B = adjacent_cluster;
				recommended_cluster_pair_cost = cluster_pair_cost;
				found_clusters_to_merge = true;
			}
		}
	}
	return found_clusters_to_merge;
}

/**
 * @brief Merge clusters.
 *
 * Starting with an initial clustering with one cluster per active node (and
 * all remaining nodes added to these clusters), this function iteratively
 * improves the clustering by greedily merging two clusters while there are
 * mergeable clusters available. At each step, the cost of the configuration is
 * calculated and the best configuration is stored.
 *
 * Time complexity:	O(cluster_to_list_of_head_addresses()
 *			  + calculate_service_cost()
 *			  + calculate_coordinator()
 *			  + merge_steps * (
 *			    nodes
 *			    + nodes
 *			    + nodes
 *			    + find_clusters_to_merge()
 *			    + cluster_merge()
 *			    + cluster_to_list_of_head_addresses()))
 *			= O(clusters
 *			    + nodes ^ 2
 *			    + service_hosts ^ 2
 *			    + merge_steps * (
 *			      nodes
 *			      + nodes
 *			      + nodes
 *			      + clusters * node_degree
 *			      + avg_adjacent_clusters ^ 2 * avg_nodes_in_cluster ^ 2 * nodes
 *			      + clusters))
 *			= O(nodes ^ 2
 *			    + merge_steps * (
 *			      nodes
 *			      + clusters * node_degree
 *			      + avg_adjacent_clusters ^ 2 * avg_nodes_in_cluster ^ 2 * nodes))
 *			= O(nodes ^ 2
 *			    + merge_steps *
 *			      node_degree ^ 2 * avg_nodes_in_cluster ^ 2 * nodes)
 *			= O(nodes ^ 2
 *			    + nodes *	// We start with one cluster per node.
 *			      node_degree ^ 2 * avg_nodes_in_cluster ^ 2 * nodes)
 *			= O(node_degree ^ 2 * avg_nodes_in_cluster ^ 2 * nodes ^ 2)
 * Space complexity:	O(nodes + node_degree + avg_nodes_in_cluster)
 *
 * @param service Pointer to service.
 * @param clusters Pointer to list of clusters.
 * @param node_id_to_cluster_map Array of pointers to clusters, indexed by node IDs.
 * @param recommended_service_configuration Pointer to list of target nodes, i.e. cluster heads of the recommended clustering (return value).
 * @param recommended_service_cost Pointer to global cost of the recommended clustering (return value).
 */
static void merge_clusters(SERVICE_service_t* service, list_t* clusters, cluster_t** node_id_to_cluster_map, list_t* recommended_service_configuration, float* recommended_service_cost) {
	float recommended_service_cost_tmp = FLT_MAX;
	list_t* recommended_service_configuration_tmp = NULL;
	cluster_t* cluster_to_merge_A = NULL;
	cluster_t* cluster_to_merge_B = NULL;
#ifdef GCMI_SNAPSHOT_MEASUREMENT
	unsigned int i = 0;
#endif	// GCMI_SNAPSHOT_MEASUREMENT
#ifdef GCMI_NETWORK_MAP_SNAPSHOT_MEASUREMENT
	unsigned int j = 0;
	unsigned int total_bidirectional_links = 0;
	unsigned int known_bidirectional_links = 0;
#endif	// GCMI_NETWORK_MAP_SNAPSHOT_MEASUREMENT

	// Initialize variables with target nodes and cost of the configuration
	// that was passed into this function.
	cluster_to_list_of_head_addresses(clusters, recommended_service_configuration);
	*recommended_service_cost = calculate_service_cost(service, recommended_service_configuration, C_LOCAL_ADDRESS);

#ifndef GCMI_SNAPSHOT_MEASUREMENT
	C_DEBUG(GCMI_DEBUG_LEVEL, C_MSG_DEBUG, "GCMI: Service cost of initial cluster configuration is %.1f with %d cluster(s).\n", *recommended_service_cost, list_size(recommended_service_configuration));
#else
	C_DEBUG(GCMI_DEBUG_LEVEL, (C_MSG_INFO | C_MSG_CONT), "Simulation complete.\n");
#endif	// GCMI_SNAPSHOT_MEASUREMENT
#ifdef GCMI_COST_CURVE_SNAPSHOT_MEASUREMENT
	if(list_size(recommended_service_configuration) != C_number_of_nodes_UNKNOWN_IN_ALGORITHMS) {
		if(get_number_of_known_nodes() == C_number_of_nodes_UNKNOWN_IN_ALGORITHMS) {
			C_DEBUG(GCMI_DEBUG_LEVEL, (C_MSG_ERROR | C_MSG_CONT), "Invalid measurement: No cluster for each node (%d != %d).\n\n", list_size(recommended_service_configuration), C_number_of_nodes_UNKNOWN_IN_ALGORITHMS);
		} else {
			C_DEBUG(GCMI_DEBUG_LEVEL, (C_MSG_ERROR | C_MSG_CONT), "Invalid measurement: No cluster for each node (%d != %d), possibly due to only %d node(s) being known to this node at the moment.\n\n", list_size(recommended_service_configuration), C_number_of_nodes_UNKNOWN_IN_ALGORITHMS, get_number_of_known_nodes());
		}
		fflush(NULL);
		abort();
	}
	C_DEBUG(GCMI_DEBUG_LEVEL, (C_MSG_INFO | C_MSG_CONT), "\n###\nservice_instances,cost\n");
	C_DEBUG(GCMI_DEBUG_LEVEL, (C_MSG_INFO | C_MSG_CONT), "%d,%f\n", list_size(recommended_service_configuration), *recommended_service_cost);
#endif	// GCMI_COST_CURVE_SNAPSHOT_MEASUREMENT

	list_alloc(&recommended_service_configuration_tmp);
	// Repeat while there are suitable clusters for merging.
	while(find_clusters_to_merge(clusters, &cluster_to_merge_A, &cluster_to_merge_B)) {

		C_DEBUG(GCMI_DEBUG_LEVEL, C_MSG_VERBOSE, "GCMI: Clusters with heads %s and %s selected for merging.\n", addr_to_str(cluster_to_merge_A->host_stat->node->address), addr_to_str(cluster_to_merge_B->host_stat->node->address));

		// Merge adjacent clusters with minimal cost.
		cluster_merge(cluster_to_merge_A, cluster_to_merge_B, clusters, node_id_to_cluster_map);
		C_IF_DEBUG(GCMI_DEBUG_LEVEL, C_MSG_VERBOSE)
			print_clusters(GCMI_DEBUG_LEVEL, C_MSG_VERBOSE, clusters);

		// Store current solution in temporary variables.
		cluster_to_list_of_head_addresses(clusters, recommended_service_configuration_tmp);
		recommended_service_cost_tmp = calculate_service_cost(service, recommended_service_configuration_tmp, calculate_coordinator(recommended_service_configuration_tmp));

#ifdef GCMI_COST_CURVE_SNAPSHOT_MEASUREMENT
		C_DEBUG(GCMI_DEBUG_LEVEL, (C_MSG_INFO | C_MSG_CONT), "%d,%f\n", list_size(recommended_service_configuration_tmp), recommended_service_cost_tmp);
#endif	// GCMI_COST_CURVE_SNAPSHOT_MEASUREMENT

		// Store results if temporary variables improve the best known
		// solution; continue to loop to avoid local minima.
		if(recommended_service_cost_tmp < *recommended_service_cost) {
#ifndef GCMI_SNAPSHOT_MEASUREMENT
			C_DEBUG(GCMI_DEBUG_LEVEL, C_MSG_DEBUG, "GCMI: Service cost of candidate cluster configuration is %.1f with %d cluster(s), using as new recommendation...\n", recommended_service_cost_tmp, list_size(recommended_service_configuration_tmp));
#endif	// GCMI_SNAPSHOT_MEASUREMENT
			list_clear_including_data(recommended_service_configuration, free);
			list_append(recommended_service_configuration, recommended_service_configuration_tmp);
			list_clear(recommended_service_configuration_tmp);
			*recommended_service_cost = recommended_service_cost_tmp;
		} else {
#ifndef GCMI_SNAPSHOT_MEASUREMENT
			C_DEBUG(GCMI_DEBUG_LEVEL, C_MSG_DEBUG, "GCMI: Service cost of candidate cluster configuration is %.1f with %d cluster(s), retaining old recommendation...\n", recommended_service_cost_tmp, list_size(recommended_service_configuration_tmp));
#endif	// GCMI_SNAPSHOT_MEASUREMENT
			list_clear_including_data(recommended_service_configuration_tmp, free);
		}
	}
	list_free(&recommended_service_configuration_tmp);

#ifdef GCMI_NETWORK_MAP_SNAPSHOT_MEASUREMENT
	total_bidirectional_links = (unsigned int) (pow((sqrt(C_number_of_nodes_UNKNOWN_IN_ALGORITHMS) - 1), 2) * 4 + 2 * (sqrt(C_number_of_nodes_UNKNOWN_IN_ALGORITHMS) - 1));
	for(i = 0; i < get_number_of_known_nodes(); i++) {
		for(j = 0; j < get_number_of_known_nodes(); j++) {
			if(
				(j < i)
				&& exists_directly_known_link(node_id_to_address(i), node_id_to_address(j))
			) {
				if(exists_directly_known_link(node_id_to_address(j), node_id_to_address(i))) {
					C_DEBUG(GCMI_DEBUG_LEVEL, C_MSG_VERBOSE, "GCMI: Couting known link between %s and %s...\n", addr_to_str(node_id_to_address(i)), addr_to_str(node_id_to_address(j)));
					known_bidirectional_links++;
#ifdef C_ENABLE_NAM_TRACE_OUTPUT
					C_trace("l -t %f -s %d -d %d -S UP -c black", C_time(), node_id_to_address(i), node_id_to_address(j));
#endif	// C_ENABLE_NAM_TRACE_OUTPUT
				} else {
					C_DEBUG(GCMI_DEBUG_LEVEL, (C_MSG_ERROR | C_MSG_CONT), "Invalid measurement: Found unidirectional link between %s and %s.\n\n", node_id_to_address(i), node_id_to_address(j));
					fflush(NULL);
					abort();
				}
			}
		}
	}
	C_DEBUG(GCMI_DEBUG_LEVEL, (C_MSG_INFO | C_MSG_CONT), "\n###\nknown_nodes,total_nodes,known_nodes_fraction,known_links,total_links,known_links_fraction\n");
	C_DEBUG(GCMI_DEBUG_LEVEL, (C_MSG_INFO | C_MSG_CONT), "%d,%d,%f,%d,%d,%f\n", get_number_of_known_nodes(), C_number_of_nodes_UNKNOWN_IN_ALGORITHMS, (((double) get_number_of_known_nodes()) / ((double) C_number_of_nodes_UNKNOWN_IN_ALGORITHMS)), known_bidirectional_links, total_bidirectional_links, (((double) known_bidirectional_links) / ((double) total_bidirectional_links)));
#endif	// GCMI_NETWORK_MAP_SNAPSHOT_MEASUREMENT
#ifdef GCMI_COST_CURVE_SNAPSHOT_MEASUREMENT
	C_DEBUG(GCMI_DEBUG_LEVEL, (C_MSG_INFO | C_MSG_CONT), "\n###\nrecommended_service_instances,recommended_cost\n");
	C_DEBUG(GCMI_DEBUG_LEVEL, (C_MSG_INFO | C_MSG_CONT), "%d,%f\n", list_size(recommended_service_configuration), *recommended_service_cost);
#endif	// GCMI_COST_CURVE_SNAPSHOT_MEASUREMENT
#ifndef GCMI_SNAPSHOT_MEASUREMENT
	C_DEBUG(GCMI_DEBUG_LEVEL, C_MSG_DEBUG, "GCMI: Service cost of recommended cluster configuration is %.1f.\n", *recommended_service_cost);
#else
	C_DEBUG(GCMI_DEBUG_LEVEL, (C_MSG_INFO | C_MSG_CONT), "\n");
#ifdef C_ENABLE_NAM_TRACE_OUTPUT
	for(i = 0; i < 4096; i++) C_trace("");	// Ugly hack to flush nam output stream.
#endif	// C_ENABLE_NAM_TRACE_OUTPUT
	fflush(NULL);
	exit(0);
#endif	// GCMI_SNAPSHOT_MEASUREMENT
}

/**
 * @brief Calculate recommended service configuration.
 *
 * Calculate the recommended service configuration based on the network topology
 * and service demand.
 *
 * NOTE: The complexity analysis below is based on an earlier version of this
 * function. The current version iteratares over increasing numbers of
 * initial clusters heads to reduce the cost of calls to merge_clusters()
 * while optaining equally valid results.
 *
 * Time complexity:	O(nodes
 *			  + init_clusters()
 *			  + add_remaining_nodes_to_clusters()
 *			  + cluster_consolidate()
 *			  + merge_clusters()
 *			  + clusters)
 *			= O(nodes
 *			  + nodes ^ 2 * avg_adjacent_clusters
 *			  + node_degree ^ 2 * avg_nodes_in_cluster ^ 2 * nodes
 *			  + nodes ^ 2 * node_degree
 *			  + node_degree ^ 2 * avg_nodes_in_cluster ^ 2 * nodes ^ 2
 *			  + clusters)
 *			= O(node_degree ^ 2 * avg_nodes_in_cluster ^ 2 * nodes ^ 2)
 * Space complexity:	O(nodes + node_degree + avg_nodes_in_cluster)
 *
 * @param service Pointer to service.
 * @param current_host_addresses Pointer to list of service instance host addresses.
 * @param recommended_service_configuration Pointer to list of target node addresses to be calculated.
 * @param recommended_service_cost Pointer to recommended service cost to be calculated.
 */
static void calculate_recommended_adaptation_target_nodes(SERVICE_service_t* service, const list_t* current_host_addresses, list_t* recommended_service_configuration, float* recommended_service_cost) {
	list_t* clusters = NULL;
	cluster_t** node_id_to_cluster_map = NULL;
	float recommended_service_cost_tmp = FLT_MAX;
	list_t* recommended_service_configuration_tmp = NULL;
	unsigned int number_of_initial_clusters = 0;
	bool try_with_more_initial_clusters = true;

	C_ABORT_ON(!list_empty(recommended_service_configuration), "GCMI: Trying to calculate recommended adaptation target nodes, but target node list is not empty.\n");
	*recommended_service_cost = FLT_MAX;

	list_alloc(&recommended_service_configuration_tmp);
	// Try with incrementing numbers of initial clusters to avoid
	// scalability problems with large networks.
	while(
		(number_of_initial_clusters < get_number_of_known_nodes())
		&& try_with_more_initial_clusters
	) {
#ifndef GCMI_SNAPSHOT_MEASUREMENT
		// Adjust the number of initial nodes for the clustering
		// depending on the fraction of synchronization traffic.
		if(service->sync_traffic_fraction < FLT_EPSILON) {
			number_of_initial_clusters += get_number_of_known_nodes();
		} else {
			// The constant of 0.025 gives us 25% of the known
			// nodes at a sync traffic fraction of 10%, which is
			// about what we need.
			number_of_initial_clusters += (get_number_of_known_nodes() * (0.025 / service->sync_traffic_fraction));
		}
		if(number_of_initial_clusters == 0)
			number_of_initial_clusters = 1;
		if(number_of_initial_clusters > get_number_of_known_nodes())
			number_of_initial_clusters = get_number_of_known_nodes();
#else
		// Do only one iteration if we're measuring the cost curve.
		number_of_initial_clusters = get_number_of_known_nodes();
#endif	// GCMI_SNAPSHOT_MEASUREMENT

		// Initialize clusters, one client per cluster.
		C_DEBUG(GCMI_DEBUG_LEVEL, C_MSG_DEBUG, "GCMI: Initializing clusters with client nodes...\n");
		list_alloc(&clusters);
		MALLOC_ARRAY(node_id_to_cluster_map, cluster_t*, get_number_of_known_nodes());
		init_clusters(service, current_host_addresses, clusters, node_id_to_cluster_map, number_of_initial_clusters);
		C_IF_DEBUG(GCMI_DEBUG_LEVEL, C_MSG_DEBUG)
			print_clusters(GCMI_DEBUG_LEVEL, C_MSG_DEBUG, clusters);

		if(list_empty(clusters)) {
			C_DEBUG(GCMI_DEBUG_LEVEL, C_MSG_DEBUG, "GCMI: Aborting due to lack of service demand...\n");
			try_with_more_initial_clusters = false;
			goto cleanup_after_calculating_target_nodes;
		}

		// Add remaining (non-client) nodes to nearest cluster.
		C_DEBUG(GCMI_DEBUG_LEVEL, C_MSG_DEBUG, "GCMI: Adding remaining nodes to clusters...\n");
		add_remaining_nodes_to_clusters(service, clusters, node_id_to_cluster_map);
		C_IF_DEBUG(GCMI_DEBUG_LEVEL, C_MSG_DEBUG)
			print_clusters(GCMI_DEBUG_LEVEL, C_MSG_DEBUG, clusters);

		// Consolidate cluster after adding all nodes.
		C_DEBUG(GCMI_DEBUG_LEVEL, C_MSG_DEBUG, "GCMI: Consolidating clusters...\n");
		cluster_consolidate((cluster_t*) list_head(clusters), clusters, node_id_to_cluster_map);

		// At this point all known, connected nodes are part of a cluster.
	
		// Merge adjacent clusters and retrieve recommended configuration.
		recommended_service_cost_tmp = FLT_MAX;
		merge_clusters(service, clusters, node_id_to_cluster_map, recommended_service_configuration_tmp, &recommended_service_cost_tmp);
		C_ABORT_ON(recommended_service_cost_tmp == FLT_MAX, "GCMI: Failed to calculate recommended service cost.\n");

		// At this point recommended_service_configuration_tmp contains the
		// recommended selection of target nodes given the initial number
		// of clusters, and recommended_service_cost_tmp the cost of that
		// selection.

		// Store results if temporary variables improve the prior
		// solution with less initial clusters.
		if(recommended_service_cost_tmp < *recommended_service_cost) {
			C_DEBUG(GCMI_DEBUG_LEVEL, C_MSG_INFO, "GCMI: Service cost of configuration based on max. %d initial clusters is %.1f with %d cluster(s), using as new recommendation...\n", number_of_initial_clusters, recommended_service_cost_tmp, list_size(recommended_service_configuration_tmp));
			list_clear_including_data(recommended_service_configuration, free);
			list_append(recommended_service_configuration, recommended_service_configuration_tmp);
			list_clear(recommended_service_configuration_tmp);
			*recommended_service_cost = recommended_service_cost_tmp;
		} else {
			C_DEBUG(GCMI_DEBUG_LEVEL, C_MSG_INFO, "GCMI: Service cost of configuration based on max. %d initial clusters is %.1f with %d cluster(s), returning recommendation...\n", number_of_initial_clusters, recommended_service_cost_tmp, list_size(recommended_service_configuration_tmp));
			list_clear_including_data(recommended_service_configuration_tmp, free);
			try_with_more_initial_clusters = false;
		}

cleanup_after_calculating_target_nodes:
		// Free data structures.
		list_free_including_data(&clusters, cluster_free_list);
		FREE(node_id_to_cluster_map);
	}
	list_free(&recommended_service_configuration_tmp);

	C_ABORT_ON(*recommended_service_cost < 0, "GCMI: Recommended service cost is negative (%.1f).\n", *recommended_service_cost);
}

/**
 * @brief Add a replication from the best source host.
 *
 * Time complexity:	O(current_service_hosts)
 * Space complexity:	O(1)
 *
 * @param service Pointer to service.
 * @param current_service_configuration Pointer to list of current host addresses.
 * @param target_host_address Address of the target host.
 * @param replications_per_source_host Pointer to array of replication counters for each host.
 * @param adaptation_commands Pointer to list of adaptation commands to be calculated.
 * @param estimated_adaptation_cost Pointer to estimated adaptation cost to be calculated.
 * @return <code>true</code> if a suitable replication command was added (or if
 * no replication is necessary), <code>false</code> otherwise.
 */
static bool add_replication_from_best_source(SERVICE_service_t* service, const list_t* current_service_configuration, C_addr_t target_host_address, uint8_t* replications_per_source_host, list_t* adaptation_commands, float* estimated_adaptation_cost) {
	C_addr_t best_source_host_address = C_BROADCAST_ADDRESS;
	float best_source_distance = FLT_MAX;
	float source_distance = FLT_MAX;
	C_addr_t* current_service_host_address = NULL;
	list_iterator_t current_service_hosts_iterator;
	PLACEMENT_command_t* command = NULL;

	list_iterator_init(current_service_configuration, &current_service_hosts_iterator);
	while((current_service_host_address = (C_addr_t*) list_iterator_next(&current_service_hosts_iterator)) != NULL) {
		if(replications_per_source_host[node_address_to_id(*current_service_host_address)] <= GCMI_MAXIMAL_REPLICATIONS_FROM_HOST) {
			source_distance = get_distance_between_nodes(*current_service_host_address, target_host_address);
			if(source_distance < best_source_distance) {
				best_source_distance = source_distance;
				best_source_host_address = *current_service_host_address;
			}
		}
	}
	if(best_source_host_address == C_BROADCAST_ADDRESS) {
		C_DEBUG(GCMI_DEBUG_LEVEL, C_MSG_INFO, "GCMI: No source found for recommended host %s, ignoring host for current adaptation...\n", addr_to_str(target_host_address));
		return false;
	}

	C_DEBUG(GCMI_DEBUG_LEVEL, C_MSG_DEBUG, "GCMI: Current host %s is the best source for recommended host %s (distance %.1f).\n", addr_to_str(best_source_host_address), addr_to_str(target_host_address), best_source_distance);

	// Add replication to adaptation plan if replication is necessary.
	if(best_source_host_address != target_host_address) {
		MALLOC(command, PLACEMENT_command_t);
		command->handler_address = best_source_host_address;
		command->id = PLACEMENT_COMMAND_REPLICATE;
		command->destination_address = target_host_address;
		list_add(adaptation_commands, command);

		replications_per_source_host[node_address_to_id(best_source_host_address)]++;
		*estimated_adaptation_cost += (((float) service->data_size) * best_source_distance);
	}
	return true;
}

/**
 * @brief Calculate adaptation commands and cost.
 *
 * Time complexity:	O(calculate_coordinator()
 *			  + recommended_service_hosts * add_replication_from_best_source()
 *			  + current_service_hosts * (recommended_service_hosts + commands)
 *			  + current_service_hosts * (recommended_service_hosts + commands))
 *			= O(recommended_service_hosts ^ 2
 *			  + recommended_service_hosts * current_service_hosts
 *			  + current_service_hosts_iterator * (recommended_service_hosts + commands)
 *			  + current_service_host_address * (recommended_service_hosts + commands))
 *			= O(recommended_service_hosts ^ 2
 *			  + current_service_hosts * (recommended_service_hosts + commands))
 *			= O(recommended_service_hosts ^ 2
 *			  + current_service_hosts * (recommended_service_hosts + commands))
 *			// recommended_service_hosts is of the magnitude as current_service_hosts.
 *			// commands is neglegible.
 *			= O(service_hosts ^ 2)
 * Space complexity:	O(commands)
 *
 * @param service Pointer to service.
 * @param amortization_time Time until which the adaptation traffic is to be amortized; divisor in calculation of adaptation cost.
 * @param current_service_configuration Pointer to list of current host addresses.
 * @param recommended_service_configuration Pointer to list of target host addresses.
 * @param adaptation_commands Pointer to list of adaptation commands to be calculated.
 * @param estimated_adaptation_cost Pointer to estimated adaptation cost to be calculated.
 */
static void calculate_adaptation_commands(SERVICE_service_t* service, C_time_t amortization_time, const list_t* current_service_configuration, const list_t* recommended_service_configuration, list_t* adaptation_commands, float* estimated_adaptation_cost) {
	C_addr_t new_coordinator_address = C_BROADCAST_ADDRESS;
	PLACEMENT_command_t* command = NULL;
	uint8_t* replications_per_source_host = NULL;
	C_addr_t* recommended_service_host_address = NULL;
	list_iterator_t recommended_service_hosts_iterator;
	C_addr_t* current_service_host_address = NULL;
	list_iterator_t current_service_hosts_iterator;
	PLACEMENT_command_t* last_replication_command = NULL;
	list_iterator_t commands_iterator;
	bool shutdown_service_instance = false;

	C_ABORT_ON(amortization_time < 0, "GCMI: Amortization time is negative.\n");
	C_ABORT_ON(fabs(amortization_time) < FLT_EPSILON, "GCMI: Amortization time is zero.\n");
	C_ABORT_ON(!list_empty(adaptation_commands), "GCMI: Adaptation plan command list at %p is not empty.\n", adaptation_commands);

	*estimated_adaptation_cost = 0;

	// Establish the new coordinator node.
	new_coordinator_address = calculate_coordinator(recommended_service_configuration);
	MALLOC(command, PLACEMENT_command_t);
	command->handler_address = C_BROADCAST_ADDRESS;
	command->id = PLACEMENT_COMMAND_COORDINATOR;
	command->destination_address = new_coordinator_address;
	list_add(adaptation_commands, command);
	command = NULL;

	// Find the best source host for all recommended target hosts.
	// Handle coordinator first to make sure that it is ready to receive
	// the stats packets that the other nodes will be sending.
	MALLOC_ARRAY(replications_per_source_host, uint8_t, get_number_of_known_nodes());
	memset(replications_per_source_host, 0, (sizeof(uint8_t) * get_number_of_known_nodes()));
	if(!add_replication_from_best_source(service, current_service_configuration, new_coordinator_address, replications_per_source_host, adaptation_commands, estimated_adaptation_cost))
		C_ABORT_ON(true, "GCMI: Failed to find suitable replication to new coordinator %s.\n", new_coordinator_address);
	list_iterator_init(recommended_service_configuration, &recommended_service_hosts_iterator);
	while((recommended_service_host_address = (C_addr_t*) list_iterator_next(&recommended_service_hosts_iterator)) != NULL) {
		if(*recommended_service_host_address != new_coordinator_address)
			add_replication_from_best_source(service, current_service_configuration, *recommended_service_host_address, replications_per_source_host, adaptation_commands, estimated_adaptation_cost);
	}
	FREE(replications_per_source_host);

	// Change last replication command for each source to a migration, if
	// the source node is not in the list of recommended nodes itself.
	list_iterator_init(current_service_configuration, &current_service_hosts_iterator);
	while((current_service_host_address = (C_addr_t*) list_iterator_next(&current_service_hosts_iterator)) != NULL) {

		if(list_find(recommended_service_configuration, current_service_host_address, matches_node_address) != NULL)
			continue;

		last_replication_command = NULL;
		list_iterator_init(adaptation_commands, &commands_iterator);
		while((command = (PLACEMENT_command_t*) list_iterator_next(&commands_iterator)) != NULL) {
			if(
				(command->handler_address == *current_service_host_address)
				&& (command->id == PLACEMENT_COMMAND_REPLICATE)
			) {
				last_replication_command = command;
			}
		}
		if(last_replication_command != NULL)
			last_replication_command->id = PLACEMENT_COMMAND_MIGRATE;
	}

	// Add shutdown commands for hosts of the current configuration that
	// aren't required anymore and have not already migrated their service
	// instances.
	list_iterator_init(current_service_configuration, &current_service_hosts_iterator);
	while((current_service_host_address = (C_addr_t*) list_iterator_next(&current_service_hosts_iterator)) != NULL) {

		if(list_find(recommended_service_configuration, current_service_host_address, matches_node_address) != NULL)
			continue;

		shutdown_service_instance = true;
		list_iterator_init(adaptation_commands, &commands_iterator);
		while((command = (PLACEMENT_command_t*) list_iterator_next(&commands_iterator)) != NULL) {
			if(
				(command->handler_address == *current_service_host_address)
				&& (command->id == PLACEMENT_COMMAND_MIGRATE)
			) {
				shutdown_service_instance = false;
				break;
			}
		}

		if(shutdown_service_instance) {
			MALLOC(command, PLACEMENT_command_t);
			command->handler_address = *current_service_host_address;
			command->id = PLACEMENT_COMMAND_DELAYED_SHUTDOWN;
			command->delay = (((float) service->data_size) / GCMI_SERVICE_REPLICATION_DATARATE);
			list_add(adaptation_commands, command);
			command = NULL;
		}
	}

	*estimated_adaptation_cost /= ((float) amortization_time);
	C_ABORT_ON(*estimated_adaptation_cost < 0, "GCMI: Estimated adaptation cost is negative (%.1f).\n", *estimated_adaptation_cost);
}

/**
 * @brief Service adaptation callback.
 *
 * Time complexity:	O(1)
 * Space complexity:	O(1)
 *
 * @param parameter Adaptation result.
 */
static void graph_cost_multiple_instances_callback(callback_parameter_t parameter) {
	PLACEMENT_adaptation_result_t* result = (PLACEMENT_adaptation_result_t*) parameter;
	SERVICE_service_t* service = (SERVICE_service_t*) result->parameter;
	graph_cost_multiple_instances_data_t* placement_data = NULL;

	if(SERVICE_exists_service(service)) {

		if(service->placement_data == NULL) {
			C_DEBUG(GCMI_DEBUG_LEVEL, C_MSG_NOTE, "GCMI: Adaptation of configuration of service \"%s\" has been completed (%d), but GCMI algorithm is not active anymore.\n", result->service_id, result->error_code);
			goto free_parameters;
		}
		placement_data = (graph_cost_multiple_instances_data_t*) service->placement_data;
		placement_data->found_superior_configuration_time = 0.0;
		placement_data->adaptation_in_progress = false;

		switch(result->error_code) {
		case C_ERROR_OK:
			C_DEBUG(GCMI_DEBUG_LEVEL, C_MSG_NOTE, "GCMI: Adaptation of configuration of service \"%s\" has been completed.\n", result->service_id);
			break;
		case C_ERROR_PARAMETER:
			C_DEBUG(GCMI_DEBUG_LEVEL, C_MSG_ERROR, "GCMI: Adaptation of configuration of service \"%s\" failed because of an invalid parameter, assuming that this node is still the coordinator...\n", result->service_id);
			placement_data->coordinator_address = C_LOCAL_ADDRESS;
			process_placement_data(service, placement_data);
			break;
		case C_ERROR_TIMEOUT:
			C_DEBUG(GCMI_DEBUG_LEVEL, C_MSG_WARN, "GCMI: Adaptation of configuration of service \"%s\" timed out, assuming that this node is still the coordinator...\n", result->service_id);
			placement_data->coordinator_address = C_LOCAL_ADDRESS;
			process_placement_data(service, placement_data);
			break;
		case C_ERROR_ABORTED:
			C_DEBUG(GCMI_DEBUG_LEVEL, C_MSG_WARN, "GCMI: Adaptation of configuration of service \"%s\" aborted, assuming that this node is still the coordinator...\n", result->service_id);
			placement_data->coordinator_address = C_LOCAL_ADDRESS;
			process_placement_data(service, placement_data);
			break;
		case C_ERROR_BUSY:
			C_DEBUG(GCMI_DEBUG_LEVEL, C_MSG_ERROR, "GCMI: Adaptation of configuration of service \"%s\" failed because system is busy, assuming that this node is still the coordinator...\n", result->service_id);
			placement_data->coordinator_address = C_LOCAL_ADDRESS;
			process_placement_data(service, placement_data);
			break;
		default:
			C_ABORT_ON(true, "GCMI: Adaptation callback with unknown error code 0x%X.\n", result->error_code);
		}

	} else {

		switch(result->error_code) {
		case C_ERROR_OK:
			C_DEBUG(GCMI_DEBUG_LEVEL, C_MSG_NOTE, "GCMI: Adaptation of configuration of service \"%s\" has been completed (service doesn't exists anymore on this node).\n", result->service_id);
			break;
		case C_ERROR_PARAMETER:
			C_DEBUG(GCMI_DEBUG_LEVEL, C_MSG_ERROR, "GCMI: Adaptation of configuration of service \"%s\" failed because of an invalid parameter, and the service doesn't exists anymore on this node...\n", result->service_id);
			break;
		case C_ERROR_TIMEOUT:
			C_DEBUG(GCMI_DEBUG_LEVEL, C_MSG_WARN, "GCMI: Adaptation of configuration of service \"%s\" timed out, and the service doesn't exists anymore on this node...\n", result->service_id);
			break;
		case C_ERROR_ABORTED:
			C_DEBUG(GCMI_DEBUG_LEVEL, C_MSG_ERROR, "GCMI: Adaptation of configuration of service \"%s\" aborted, and the service doesn't exists anymore on this node...\n", result->service_id);
			break;
		case C_ERROR_BUSY:
			C_DEBUG(GCMI_DEBUG_LEVEL, C_MSG_ERROR, "GCMI: Adaptation of configuration of service \"%s\" failed because system is busy, and the service doesn't exists anymore on this node...\n", result->service_id);
			break;
		default:
			C_ABORT_ON(true, "GCMI: Adaptation callback with unknown error code 0x%X.\n", result->error_code);
		}

	}

free_parameters:
	FREE(result);
}

/**
 * @brief Graph cost multiple instances algorithm.
 *
 * Time complexity:	O(calculate_edge_cost()
 *			  + floyd_warshall()
 *			  + calculate_service_cost()
 *			  + calculate_recommended_adaptation_target_nodes()
 *			  + calculate_adaptation())
 *			= O(nodes ^ 2
 *			  + nodes ^ 3
 *			  + nodes ^ 2
 *			  + node_degree ^ 2 * avg_nodes_in_cluster ^ 2 * nodes ^ 2
 *			  + service_hosts ^ 2)
 *			= O(nodes ^ 3
 *			  + node_degree ^ 2 * avg_nodes_in_cluster ^ 2 * nodes ^ 2
 *			  + service_hosts ^ 2)
 *			// service_hosts is less than nodes.
 *			= O(nodes ^ 3
 *			  + node_degree ^ 2 * avg_nodes_in_cluster ^ 2 * nodes ^ 2)
 *			// avg_nodes_in_cluster is less than nodes.
 *			= O(nodes ^ 3
 *			  + node_degree ^ 2 * nodes ^ 4)
 *			= O(nodes ^ 4 * node_degree ^ 2)
 * Space complexity:	O(nodes ^ 2)
 *
 * @param service Pointer to service for which to run this algorithm.
 */
static void graph_cost_multiple_instances_run(SERVICE_service_t* service) {
	graph_cost_multiple_instances_data_t* placement_data = ((graph_cost_multiple_instances_data_t*) service->placement_data);
	float previous_service_cost = placement_data->previous_cost;
	float current_service_cost = FLT_MAX;
	SERVICE_node_stat_t* node_stat = NULL;
	list_iterator_t node_stats_iterator;
	C_time_t next_trigger_duration = 0;
	list_t* recommended_service_configuration = NULL;
	float recommended_service_cost = FLT_MAX;
	C_time_t adaptation_amortization_time = 0;
	list_t* adaptation_commands = NULL;
	float estimated_savings = FLT_MIN;
	float estimated_adaptation_cost = FLT_MAX;
	PLACEMENT_adaptation_result_t* result = NULL;

	// Calculate current service cost, either for the host or the
	// coordinator.
	if(placement_data->coordinator_address != C_LOCAL_ADDRESS) {
		current_service_cost = GCSI_calculate_service_cost(service->node_stats, C_LOCAL_ADDRESS);
		C_DEBUG(GCMI_DEBUG_LEVEL, C_MSG_DEBUG, "GCMI: Current local cost of service \"%s\" is %.1f.\n", service->id, current_service_cost);
	} else {
		current_service_cost = calculate_service_cost(service, placement_data->configuration, placement_data->coordinator_address);
		C_DEBUG(GCMI_DEBUG_LEVEL, C_MSG_INFO, "GCMI: Current configuration of service \"%s\" is ", service->id);
		list_print(GCMI_DEBUG_LEVEL, (C_MSG_INFO | C_MSG_CONT), placement_data->configuration, print_node_address);
		C_DEBUG(GCMI_DEBUG_LEVEL, (C_MSG_INFO | C_MSG_CONT), " with %d host(s) and a cost of %.1f.\n", list_size(placement_data->configuration), current_service_cost);

		placement_data->previous_cost = current_service_cost;
	}

	// Abort if current cost has not changed considerably.
	if(!trigger_check(placement_data->trigger, current_service_cost))
		return;

	if(placement_data->coordinator_address != C_LOCAL_ADDRESS) {

		/// This node is not the coordinator.

		if(!placement_data->adaptation_in_progress) {

			// Transmit the statistics to the coordinator.
			C_DEBUG(GCMI_DEBUG_LEVEL, C_MSG_INFO, "GCMI: Sending local statistics of service \"%s\" with cost %.1f to coordinator %s...\n", service->id, current_service_cost, addr_to_str(placement_data->coordinator_address));
			SERVICE_send_statistics(service->id, placement_data->coordinator_address);

			// Reset node statistics after transmission and
			// re-calculate the service cost to properly reset
			// the trigger.
			// NOTE: We might lose some data if the packet is lost, but
			// that shouldn't be a problem.
			list_iterator_init(service->node_stats, &node_stats_iterator);
			while((node_stat = (SERVICE_node_stat_t*) list_iterator_next(&node_stats_iterator)) != NULL)
				SERVICE_reset_node_stat(node_stat, false);
			current_service_cost = GCSI_calculate_service_cost(service->node_stats, C_LOCAL_ADDRESS);
		}

		// Reset the trigger.
		next_trigger_duration = GCMI_HOST_TRIGGER_DURATION * (trigger_times_fired_at_end(placement_data->trigger) + 1);
		if(next_trigger_duration > GCMI_HOST_TRIGGER_MAX_DURATION)
			next_trigger_duration = GCMI_HOST_TRIGGER_MAX_DURATION;
		trigger_reset(placement_data->trigger, current_service_cost, (current_service_cost * GCMI_HOST_TRIGGER_RANGE_MULTIPLIER), next_trigger_duration);

		return;
	}

	/// This node is the coordinator.

	// Store current cost in statistics variable.
	C_PLACEMENT_current_service_cost = current_service_cost;

	// Reset the trigger.
	next_trigger_duration = GCMI_COORDINATOR_TRIGGER_DURATION * (trigger_times_fired_at_end(placement_data->trigger) + 1);
	if(next_trigger_duration > GCMI_COORDINATOR_TRIGGER_MAX_DURATION)
		next_trigger_duration = GCMI_COORDINATOR_TRIGGER_MAX_DURATION;
	trigger_reset(placement_data->trigger, current_service_cost, (current_service_cost * GCMI_COORDINATOR_TRIGGER_RANGE_MULTIPLIER), next_trigger_duration);

	// Abort if current cost is zero.
	if(fabs(current_service_cost) < FLT_EPSILON)
		return;

	// Calculate recommended service configuration and cost.
	list_alloc(&recommended_service_configuration);
	calculate_recommended_adaptation_target_nodes(service, placement_data->configuration, recommended_service_configuration, &recommended_service_cost);
	if(list_empty(recommended_service_configuration)) {
		C_DEBUG(GCMI_DEBUG_LEVEL, C_MSG_ERROR, "GCMI: No potential target for the migration of service \"%s\" could be established.\n", service->id);
		goto free_recommended_configuration;
	}
	list_sort(recommended_service_configuration, node_address_less_than);
	C_DEBUG(GCMI_DEBUG_LEVEL, C_MSG_INFO, "GCMI: Recommended configuration of service \"%s\" is ", service->id);
	list_print(GCMI_DEBUG_LEVEL, (C_MSG_INFO | C_MSG_CONT), recommended_service_configuration, print_node_address);
	C_DEBUG(GCMI_DEBUG_LEVEL, (C_MSG_INFO | C_MSG_CONT), " with %d host(s) and a cost of %.1f.\n", list_size(recommended_service_configuration), recommended_service_cost);

	// Store optimal service cost and number of hosts in statistics
	// variables.
	C_PLACEMENT_currently_optimal_service_cost = recommended_service_cost;
	C_PLACEMENT_currently_optimal_number_of_hosts = list_size(recommended_service_configuration);

	// Check whether the current configuration matches the calculated
	// recommended configuration.
	if(list_size(placement_data->configuration) == list_size(recommended_service_configuration)) {
		unsigned int i = 0;
		for(i = 0; i < list_size(placement_data->configuration); i++) {
			C_DEBUG(GCMI_DEBUG_LEVEL, C_MSG_VERBOSE, "GCMI: Current and recommended host #%d are %s and %s.\n", i, addr_to_str(*((C_addr_t*) list_get(placement_data->configuration, i))), addr_to_str(*((C_addr_t*) list_get(recommended_service_configuration, i))));
			if(*((C_addr_t*) list_get(placement_data->configuration, i)) != *((C_addr_t*) list_get(recommended_service_configuration, i)))
				goto current_and_recommended_configuration_differ;
		}

		// This happens rarely and is most propably due to rounding
		// errors when current and recommended configuration are sorted
		// differently.
		if(fabsf(current_service_cost - recommended_service_cost) > FLT_EPSILON)
			C_DEBUG(GCMI_DEBUG_LEVEL, C_MSG_DEBUG, "GCMI: Current configuration matches recommended configuration, but service costs differ (%.1f != %.1f)\n", current_service_cost, recommended_service_cost);

		C_DEBUG(GCMI_DEBUG_LEVEL, C_MSG_NOTE, "GCMI: Configuration of service \"%s\" with %d host(s) remains unchanged, recommended configuration matches current configuration.\n", service->id, list_size(placement_data->configuration));
		goto free_recommended_configuration;
	}
current_and_recommended_configuration_differ:

	// Abort if the the recommended configuration is more costly than the
	// current configuration.
	if(current_service_cost <= (recommended_service_cost * GCMI_ADAPTATION_HYSTERESIS_MULTIPLIER)) {
		C_DEBUG(GCMI_DEBUG_LEVEL, C_MSG_DEBUG, "GCMI: Current configuration of service \"%s\" is ", service->id);
		list_print(GCMI_DEBUG_LEVEL, (C_MSG_DEBUG | C_MSG_CONT), placement_data->configuration, print_node_address);
		C_DEBUG(GCMI_DEBUG_LEVEL, (C_MSG_DEBUG | C_MSG_CONT), " with %d host(s) and a cost of %.1f.\n", list_size(placement_data->configuration), current_service_cost);
		if(list_size(recommended_service_configuration) == list_size(placement_data->configuration)) {
			C_DEBUG(GCMI_DEBUG_LEVEL, C_MSG_DEBUG, "GCMI: Recommended service cost of service \"%s\" is more costly than current service cost (%.1f > %.1f).\n", service->id, recommended_service_cost, current_service_cost);
		} else {
			C_DEBUG(GCMI_DEBUG_LEVEL, C_MSG_DEBUG, "GCMI: Recommended service cost of service \"%s\" is more costly than current service cost (%.1f > %.1f) and number of service instances in recommended and current configuration differ (%d != %d).\n", service->id, recommended_service_cost, current_service_cost, list_size(recommended_service_configuration), list_size(placement_data->configuration));
		}
		C_DEBUG(GCMI_DEBUG_LEVEL, C_MSG_NOTE, "GCMI: Configuration of service \"%s\" with %d host(s) remains unchanged, recommended configuration with %d host(s) doesn't substantially reduce the cost (%.1f, %.1f\%).\n", service->id, list_size(placement_data->configuration), list_size(recommended_service_configuration), (recommended_service_cost - current_service_cost), ((recommended_service_cost - current_service_cost) / current_service_cost * 100));
		placement_data->found_superior_configuration_time = 0.0;
		goto free_recommended_configuration;
	}

	// If this is the first time that a superior configuration has been
	// found, keep a note of the time (required below) and abort for now.
	if(placement_data->found_superior_configuration_time == 0.0) {
		C_DEBUG(GCMI_DEBUG_LEVEL, C_MSG_NOTE, "GCMI: Configuration of service \"%s\" with %d host(s) remains unchanged, but superior configuration with %d host(s) has been found (%.1f, %.1f\%).\n", service->id, list_size(placement_data->configuration), list_size(recommended_service_configuration), (recommended_service_cost - current_service_cost), ((recommended_service_cost - current_service_cost) / current_service_cost * 100));
		placement_data->found_superior_configuration_time = C_time();
		goto free_recommended_configuration;
	}

	// Calculate adaptation plan and cost.
	adaptation_amortization_time = C_time() - placement_data->found_superior_configuration_time;
	list_alloc(&adaptation_commands);
	calculate_adaptation_commands(service, adaptation_amortization_time, placement_data->configuration, recommended_service_configuration, adaptation_commands, &estimated_adaptation_cost);
	C_DEBUG(GCMI_DEBUG_LEVEL, C_MSG_INFO, "GCMI: Adaptation plan for service \"%s\" is ", service->id);
	list_print(GCMI_DEBUG_LEVEL, (C_MSG_INFO | C_MSG_CONT), adaptation_commands, PLACEMENT_print_command);
	C_DEBUG(GCMI_DEBUG_LEVEL, (C_MSG_INFO | C_MSG_CONT), " with an estimated adaptation cost of %.1f to be amortized in %.1fs.\n", estimated_adaptation_cost, adaptation_amortization_time);

	// Store estimated adaptation cost in statistics variable.
	C_PLACEMENT_currently_estimated_adaptation_cost = estimated_adaptation_cost;

	// If it's worth the trouble, adapt the current service configuration.
	estimated_savings = current_service_cost - recommended_service_cost;
	if(estimated_savings <= estimated_adaptation_cost) {
		C_DEBUG(GCMI_DEBUG_LEVEL, C_MSG_NOTE, "GCMI: Configuration of service \"%s\" with %d host(s) remains unchanged, adaptation to configuration with %d host(s) is too expensive (%.1f savings <= %.1f cost).\n", service->id, list_size(placement_data->configuration), list_size(recommended_service_configuration), estimated_savings, estimated_adaptation_cost);
		goto free_adaptation_commands;
	}

	/// Adaptation of service configuration is required.

	// Sanity check for a suspicious increase in service cost. Usually,
	// this indicates a sudden change in this node's knowledge
	// about the network topology and thus hints at problems in the
	// network mapping component.
	if(
		(previous_service_cost != 0.0)
		&& ((current_service_cost - previous_service_cost) > 100000.0)
	) {
		C_DEBUG(GCMI_DEBUG_LEVEL, C_MSG_ERROR, "GCMI: Suspiciously high increase in cost of service \"%s\" (%.1f >> %.1f).\n", service->id, current_service_cost, previous_service_cost);
	}

	// Sanity check to avoid changing the configuration too much. If this
	// error occurs, it almost certainly indicates a bug somewhere else.
	if(
		(list_size(recommended_service_configuration) <= 5)
		&& (list_size(recommended_service_configuration) <= (list_size(placement_data->configuration) / 3))
	) {
		C_DEBUG(GCMI_DEBUG_LEVEL, C_MSG_ERROR, "GCMI: Suspiciously high reduction in the number of recommended instances for service \"%s\" (%d << %d).\n", service->id, list_size(recommended_service_configuration), list_size(placement_data->configuration));
	}

	C_DEBUG(GCMI_DEBUG_LEVEL, C_MSG_NOTE, "GCMI: Changing configuration of service \"%s\" from ", service->id);
	list_print(GCMI_DEBUG_LEVEL, (C_MSG_NOTE | C_MSG_CONT), placement_data->configuration, print_node_address);
	C_DEBUG(GCMI_DEBUG_LEVEL, (C_MSG_NOTE | C_MSG_CONT), " with %d host(s) to ", list_size(placement_data->configuration));
	list_print(GCMI_DEBUG_LEVEL, (C_MSG_NOTE | C_MSG_CONT), recommended_service_configuration, print_node_address);
	C_DEBUG(GCMI_DEBUG_LEVEL, (C_MSG_NOTE | C_MSG_CONT), " with %d host(s) (%.1f savings > %.1f cost)...\n", list_size(recommended_service_configuration), estimated_savings, estimated_adaptation_cost);

	placement_data->adaptation_in_progress = true;
	placement_data->previous_cost = 0.0;	// Invalidate previous cost calculation while adapting.
	MALLOC(result, PLACEMENT_adaptation_result_t);
	result->parameter = service;
	PLACEMENT_distribute_adaptation(service->id, adaptation_commands, placement_data->configuration, graph_cost_multiple_instances_callback, result);

free_adaptation_commands:
	list_free_including_data(&adaptation_commands, free);

free_recommended_configuration:
	list_free_including_data(&recommended_service_configuration, free);
}

/**
 * @brief Start graph cost multiple instances algorithm.
 *
 * Placement data is initialized assuming that this node is the only service
 * host and thus the coordinator.
 *
 * @param service Pointer to service for which to start this algorithm.
 */
void graph_cost_multiple_instances_start(SERVICE_service_t* service) {
	graph_cost_multiple_instances_data_t* placement_data = NULL;
	C_addr_t* host_address = NULL;

	C_ABORT_ON(service->placement_data != NULL, "GCMI: Trying to start algorithm for service \"%s\", but placement data is not NULL.\n", service->id);

	C_DEBUG(GCMI_DEBUG_LEVEL, C_MSG_DEBUG, "GCMI: Starting algorithm for service \"%s\"...\n", service->id);
	MALLOC(service->placement_data, graph_cost_multiple_instances_data_t);
	placement_data = (graph_cost_multiple_instances_data_t*) service->placement_data;
	placement_data->last_run_time = C_time();	// Initialize to current time to gather some data before the first run.

	// Set local node to be the coordinator.
	trigger_alloc(&placement_data->trigger, GCMI_COORDINATOR_TRIGGER_INITIAL_RANGE, GCMI_COORDINATOR_TRIGGER_DURATION);
	placement_data->coordinator_address = C_LOCAL_ADDRESS;

	// Set local node to be the only service instance.
	list_alloc(&placement_data->configuration);
	MALLOC(host_address, C_addr_t);
	*host_address = C_LOCAL_ADDRESS;
	list_add(placement_data->configuration, host_address);
	placement_data->found_superior_configuration_time = 0.0;
	placement_data->adaptation_in_progress = false;
	list_alloc(&placement_data->inactive_hosts);
	placement_data->inactive_hosts_callback_id = 0;
	placement_data->previous_cost = 0.0;

#ifdef C_ENABLE_NAM_TRACE_OUTPUT
	C_trace("m -t %f -s %d -n %s -X", C_time(), C_LOCAL_ADDRESS, service->id);
	C_trace("m -t %f -s %d -n %s -h hexagon -c black", C_time(), C_LOCAL_ADDRESS, service->id);
#endif	// C_ENABLE_NAM_TRACE_OUTPUT
}

/**
 * @brief Stop graph cost multiple instances algorithm.
 *
 * @param service Pointer to service for which to stop this algorithm.
 */
void graph_cost_multiple_instances_stop(SERVICE_service_t* service) {
	graph_cost_multiple_instances_data_t* placement_data = (graph_cost_multiple_instances_data_t*) service->placement_data;

	C_ABORT_ON(placement_data == NULL, "GCMI: Trying to stop algorithm for service \"%s\", but placement data is NULL.\n", service->id);

	if(placement_data->coordinator_address == C_LOCAL_ADDRESS) {
		if(placement_data->adaptation_in_progress)
			C_DEBUG(GCMI_DEBUG_LEVEL, C_MSG_WARN, "GCMI: Stopping algorithm and coordinator for service \"%s\", while an adaptation is in progress...\n", service->id);
		else
			C_DEBUG(GCMI_DEBUG_LEVEL, C_MSG_INFO, "GCMI: Stopping algorithm and coordinator for service \"%s\"...\n", service->id);
	} else {
		C_DEBUG(GCMI_DEBUG_LEVEL, C_MSG_DEBUG, "GCMI: Stopping algorithm for service \"%s\"...\n", service->id);
	}

#ifdef C_ENABLE_NAM_TRACE_OUTPUT
	if(placement_data->coordinator_address == C_LOCAL_ADDRESS) {
		C_trace("m -t %f -s %d -n %s -X", C_time(), C_LOCAL_ADDRESS, service->id);
		C_trace("m -t %f -s %d -n %s -h circle -c black", C_time(), C_LOCAL_ADDRESS, service->id);
	}
#endif	// C_ENABLE_NAM_TRACE_OUTPUT

	trigger_free(&placement_data->trigger);
	if(placement_data->configuration != NULL)
		list_free_including_data(&placement_data->configuration, free);
	if(placement_data->inactive_hosts != NULL)
		list_free(&placement_data->inactive_hosts);
	if(placement_data->inactive_hosts_callback_id != 0)
		callback_cancel(placement_data->inactive_hosts_callback_id);
	FREE(service->placement_data);
}

/**
 * @brief Trigger graph cost multiple instances algorithm.
 *
 * @param service Pointer to service for which to trigger this algorithm.
 * @param reason_id Numeric ID for the reason why the algorithm was triggered.
 * @param reason_parameter Pointer to optional, reason-specific parameter of
 * the reason.
 */
void graph_cost_multiple_instances_trigger(SERVICE_service_t* service, uint8_t reason_id, const void* reason_parameter) {
	graph_cost_multiple_instances_data_t* placement_data = (graph_cost_multiple_instances_data_t*) service->placement_data;
	PLACEMENT_trigger_reason_stats_parameter_t* trigger_reason_stats_parameter = (PLACEMENT_trigger_reason_stats_parameter_t*) reason_parameter;
	C_addr_t* host_address = NULL;

	C_ABORT_ON(placement_data == NULL, "GCMI: Trying to trigger algorithm for service \"%s\", but placement data is NULL.\n", service->id);

	if(reason_id == C_PLACEMENT_TRIGGER_REASON_STATS) {

		if(placement_data->coordinator_address != C_LOCAL_ADDRESS) {
			if(placement_data->adaptation_in_progress) {
				C_DEBUG(GCMI_DEBUG_LEVEL, C_MSG_INFO, "GCMI: Ignoring attempt to trigger algorithm for service \"%s\" due do received stats from %s. This node is not the coordinator anymore, but an adaptation is still in progress.\n", service->id, addr_to_str(trigger_reason_stats_parameter->source_address));
			} else {
				C_DEBUG(GCMI_DEBUG_LEVEL, C_MSG_WARN, "GCMI: Trying to trigger algorithm for service \"%s\" due do received stats from %s, but this node is not the coordinator. Updating coordinator on remote service instance to %s...\n", service->id, addr_to_str(trigger_reason_stats_parameter->source_address), addr_to_str(placement_data->coordinator_address));
				PLACEMENT_update_remote_coordinator(service->id, trigger_reason_stats_parameter->source_address, placement_data->coordinator_address);
			}
			return;
		}

		/// This is the coordinator node.

		if(list_find(placement_data->configuration, &trigger_reason_stats_parameter->source_address, matches_address) == NULL) {

			// We have received service statistics from a node that
			// is not currently part of the service configuration.
			// Hence, the current configuration has been incomplete
			// and we reset the timer for the calculations based on
			// it.
			placement_data->found_superior_configuration_time = 0.0;

			if(trigger_reason_stats_parameter->valid_instance) {
				C_DEBUG(GCMI_DEBUG_LEVEL, C_MSG_NOTE, "GCMI: Received stats for service \"%s\" from a %s which is not a current service host, but tagged as valid. Adding to current configuration...\n", service->id, addr_to_str(trigger_reason_stats_parameter->source_address));

				// Add source of stats to the configuration.
				MALLOC(host_address, C_addr_t);
				*host_address = trigger_reason_stats_parameter->source_address;
				list_add(placement_data->configuration, host_address);

				// Remove failed migration destination from the
				// configuration, but only if the coordinator
				// is still waiting for any inactive nodes to
				// report in.
				if(
					(trigger_reason_stats_parameter->last_failed_migration_destination != C_BROADCAST_ADDRESS)
					&& !list_empty(placement_data->inactive_hosts)
				) {
					host_address = (C_addr_t*) list_find(placement_data->inactive_hosts, &trigger_reason_stats_parameter->last_failed_migration_destination, matches_address);
					if(host_address != NULL) {
						C_DEBUG(GCMI_DEBUG_LEVEL, C_MSG_INFO, "GCMI: Also removing failed migration destination %s from configuration of service \"%s\"...\n", addr_to_str(trigger_reason_stats_parameter->last_failed_migration_destination), service->id);
						list_remove(placement_data->configuration, host_address);
						list_remove(placement_data->inactive_hosts, host_address);
						FREE(host_address);
	
						if(list_empty(placement_data->inactive_hosts)) {
							callback_cancel(placement_data->inactive_hosts_callback_id);
							placement_data->inactive_hosts_callback_id = 0;
						}
					} else {
						C_DEBUG(GCMI_DEBUG_LEVEL, C_MSG_WARN, "GCMI: Trying to remove supposedly failed migration destination %s from configuration of service \"%s\", but it is not part of the list of inactive hosts.\n", addr_to_str(trigger_reason_stats_parameter->last_failed_migration_destination), service->id);
					}
				}

			} else {
				C_DEBUG(GCMI_DEBUG_LEVEL, C_MSG_WARN, "GCMI: Trying to trigger algorithm for service \"%s\", but received stats from a %s which is not a current service host. Shutting down apparently left-over service instance...\n", service->id, addr_to_str(trigger_reason_stats_parameter->source_address));
				PLACEMENT_shutdown_remote_service_host(service->id, trigger_reason_stats_parameter->source_address);
				return;
			}
		}

		if((host_address = (C_addr_t*) list_find(placement_data->inactive_hosts, &trigger_reason_stats_parameter->source_address, matches_address)) != NULL) {
			C_DEBUG(GCMI_DEBUG_LEVEL, C_MSG_NOTE, "GCMI: Host %s reporting in for service \"%s\"...\n", addr_to_str(*host_address), service->id);
			list_remove(placement_data->inactive_hosts, host_address);
			// NOTE: Don't free the list element as it is also
			// contained in the configuration list.
			placement_data->found_superior_configuration_time = 0.0;

			if(list_empty(placement_data->inactive_hosts)) {
				callback_cancel(placement_data->inactive_hosts_callback_id);
				placement_data->inactive_hosts_callback_id = 0;
			}
		}
	}

	if(placement_data->coordinator_address == C_LOCAL_ADDRESS) {
		if(!list_empty(placement_data->inactive_hosts)) {
			C_DEBUG(GCMI_DEBUG_LEVEL, C_MSG_INFO, "GCMI: Current configuration of service \"%s\" is ", service->id);
			list_print(GCMI_DEBUG_LEVEL, (C_MSG_INFO | C_MSG_CONT), placement_data->configuration, print_node_address);
			C_DEBUG(GCMI_DEBUG_LEVEL, (C_MSG_INFO | C_MSG_CONT), " with %d host(s), waiting for %d host(s)...\n", list_size(placement_data->configuration), list_size(placement_data->inactive_hosts));
			return;
		} else if(placement_data->adaptation_in_progress) {
			C_DEBUG(GCMI_DEBUG_LEVEL, C_MSG_INFO, "GCMI: Current configuration of service \"%s\" is ", service->id);
			list_print(GCMI_DEBUG_LEVEL, (C_MSG_INFO | C_MSG_CONT), placement_data->configuration, print_node_address);
			C_DEBUG(GCMI_DEBUG_LEVEL, (C_MSG_INFO | C_MSG_CONT), " with %d host(s), adaptation is in progress...\n", list_size(placement_data->configuration));
			return;
		}
	}

	// Only run algorithm in certain minimal intervals or after receiving a
	// STATS packet.
	if(
		((C_time() - placement_data->last_run_time) > GCMI_MINIMAL_RUN_INTERVAL)
		|| (reason_id == C_PLACEMENT_TRIGGER_REASON_STATS)
	) {
		C_DEBUG(GCMI_DEBUG_LEVEL, C_MSG_DEBUG, "GCMI: Triggering algorithm for service \"%s\"...\n", service->id);
		graph_cost_multiple_instances_run(service);
		placement_data->last_run_time = C_time();
	}
}

/**
 * @brief Get data from graph cost multiple instances algorithm.
 *
 * @param service Pointer to service for which to get data.
 * @param data Pointer to pointer to data to return.
 * @return Size of the data in bytes.
 */
size_t graph_cost_multiple_instances_get_data(const SERVICE_service_t* service, char** data) {
	graph_cost_multiple_instances_data_t* placement_data = (graph_cost_multiple_instances_data_t*) service->placement_data;

	C_ABORT_ON(placement_data == NULL, "GCMI: Trying to get data from algorithm for service \"%s\", but placement data is NULL.\n", service->id);

	C_DEBUG(GCMI_DEBUG_LEVEL, C_MSG_DEBUG, "GCMI: Getting data of algorithm for service \"%s\"...\n", service->id);
	size_t size = sizeof(C_addr_t) + list_size(placement_data->configuration) * sizeof(C_addr_t);
	MALLOC_ARRAY(*data, char, size);

	memcpy(*data, &placement_data->coordinator_address, sizeof(C_addr_t));
	list_to_array(placement_data->configuration, (((char*) *data) + sizeof(C_addr_t)), sizeof(C_addr_t));
	return size;
}

/**
 * @brief Set data of graph cost multiple instances algorithm.
 *
 * @param service Pointer to service for which to set data.
 * @param data Pointer to data to set.
 * @param data_size Size of the data in bytes.
 */
void graph_cost_multiple_instances_set_data(SERVICE_service_t* service, const char* data, size_t data_size) {
	graph_cost_multiple_instances_data_t* placement_data = (graph_cost_multiple_instances_data_t*) service->placement_data;

	C_ABORT_ON(placement_data == NULL, "GCMI: Trying to set data of algorithm for service \"%s\", but placement data is NULL.\n", service->id);

	C_DEBUG(GCMI_DEBUG_LEVEL, C_MSG_DEBUG, "GCMI: Setting data for algorithm for service \"%s\"...\n", service->id);
	memcpy(&placement_data->coordinator_address, data, sizeof(C_addr_t));
	list_clear_including_data(placement_data->configuration, free);
	list_from_array(placement_data->configuration, (((char*) data) + sizeof(C_addr_t)), sizeof(C_addr_t), (data_size - sizeof(C_addr_t)));

	list_sort(placement_data->configuration, node_address_less_than);
	process_placement_data(service, placement_data);

	C_DEBUG(GCMI_DEBUG_LEVEL, C_MSG_DEBUG, "GCMI: Configuration of service \"%s\" set to ", service->id);
	list_print(GCMI_DEBUG_LEVEL, (C_MSG_DEBUG | C_MSG_CONT), placement_data->configuration, print_node_address);
	C_DEBUG(GCMI_DEBUG_LEVEL, (C_MSG_DEBUG | C_MSG_CONT), " with %s as coordinator.\n", addr_to_str(placement_data->coordinator_address));
}

/**
 * @brief Update data of graph cost multiple instances algorithm.
 *
 * @param service Pointer to service for which to update data.
 * @param adaptation Pointer to adaptation plan to use for the update.
 * @return <code>true</code> if the adaptation data is valid,
 * <code>false</code> otherwise.
 */
bool graph_cost_multiple_instances_update_data(SERVICE_service_t* service, const PLACEMENT_incoming_adaptation_t* adaptation) {
	graph_cost_multiple_instances_data_t* placement_data = (graph_cost_multiple_instances_data_t*) service->placement_data;
	PLACEMENT_command_t* command = NULL;
	list_iterator_t commands_iterator;
	C_addr_t* host_address = NULL;

	C_ABORT_ON(placement_data == NULL, "GCMI: Trying to update data of algorithm for service \"%s\", but placement data is NULL.\n", service->id);

	C_DEBUG(GCMI_DEBUG_LEVEL, C_MSG_DEBUG, "GCMI: Updating data of algorithm for service \"%s\"...\n", service->id);
	list_iterator_init(adaptation->commands, &commands_iterator);
	while((command = (PLACEMENT_command_t*) list_iterator_next(&commands_iterator)) != NULL) {
		switch(command->id) {
		case PLACEMENT_COMMAND_NOP:
			break;
		case PLACEMENT_COMMAND_COORDINATOR:
			placement_data->coordinator_address = command->destination_address;
			// Fall-through to make sure that the coordinator is
			// contained in the list of host addresses.
		case PLACEMENT_COMMAND_REPLICATE:
			if(list_find(placement_data->configuration, &command->destination_address, matches_node_address) == NULL) {
				MALLOC(host_address, C_addr_t);
				*host_address = command->destination_address;
				list_add(placement_data->configuration, host_address);
				host_address = NULL;
			}
			break;
		case PLACEMENT_COMMAND_MIGRATE:
			if(list_find(placement_data->configuration, &command->destination_address, matches_node_address) == NULL) {
				MALLOC(host_address, C_addr_t);
				*host_address = command->destination_address;
				list_add(placement_data->configuration, host_address);
				host_address = NULL;
			}
			if((host_address = (C_addr_t*) list_find(placement_data->configuration, &command->handler_address, matches_node_address)) != NULL) {
				list_remove(placement_data->configuration, host_address);
				FREE(host_address);
			}
			break;
		case PLACEMENT_COMMAND_SHUTDOWN:
		case PLACEMENT_COMMAND_DELAYED_SHUTDOWN:
			if(
				(placement_data->coordinator_address == C_LOCAL_ADDRESS)
				&& (adaptation->coordinator_address != C_LOCAL_ADDRESS)
			) {
				C_DEBUG(GCMI_DEBUG_LEVEL, C_MSG_WARN, "GCMI: Non-coordinator %s trying to shut down the coordinator for service \"%s\", ignoring request...\n", addr_to_str(adaptation->coordinator_address), service->id);
				return false;
			}
			if((host_address = (C_addr_t*) list_find(placement_data->configuration, &command->handler_address, matches_node_address)) != NULL) {
				list_remove(placement_data->configuration, host_address);
				FREE(host_address);
			}
			break;
		default:
			C_ABORT_ON(true, "GCMI: Invalid placement command (0x%X).\n", command->id);
		}
	}

	// Check whether coordinator is part of the configuration.
	if(list_find(placement_data->configuration, &placement_data->coordinator_address, matches_node_address) == NULL) {
		C_ABORT_ON(placement_data->coordinator_address != C_LOCAL_ADDRESS, "GCMI: Coordinator is neither part of the configuration nor set to the local node.\n");
		C_DEBUG(GCMI_DEBUG_LEVEL, C_MSG_WARN, "GCMI: Adding local node as coordinator to configuration of service \"%s\"...\n", service->id);
		MALLOC(host_address, C_addr_t);
		*host_address = placement_data->coordinator_address;
		list_add(placement_data->configuration, host_address);
		host_address = NULL;
	}

	list_sort(placement_data->configuration, node_address_less_than);
	process_placement_data(service, placement_data);

	C_DEBUG(GCMI_DEBUG_LEVEL, C_MSG_DEBUG, "GCMI: Configuration of service \"%s\" updated to ", service->id);
	list_print(GCMI_DEBUG_LEVEL, (C_MSG_DEBUG | C_MSG_CONT), placement_data->configuration, print_node_address);
	C_DEBUG(GCMI_DEBUG_LEVEL, (C_MSG_DEBUG | C_MSG_CONT), " with %s as coordinator.\n", addr_to_str(placement_data->coordinator_address));
	return true;
}

/**
 * @brief Get synchronization hosts from graph cost multiple instances algorithm.
 *
 * TODO: If an adaptation is in progress and some hosts are redirecting their
 * clients while continuing to serve read-only requests, those hosts should be
 * included in the list of hosts to synchronize.
 *
 * @param service Pointer to service for which to get the synchronization hosts.
 * @param inter_host_update Flag whether this update is an inter-host update.
 * @return Pointer to a list of nodes to which synchronize changes to (must be
 * freed by the caller), or NULL if no synchronization is necessary.
 */
list_t* graph_cost_multiple_instances_get_synchronization_hosts(const SERVICE_service_t* service, bool inter_host_update) {
	graph_cost_multiple_instances_data_t* placement_data = (graph_cost_multiple_instances_data_t*) service->placement_data;
	list_t* synchronization_hosts = NULL;
	C_addr_t* host_address = NULL;
	list_iterator_t hosts_iterator;

	C_ABORT_ON(placement_data == NULL, "GCMI: Trying to get synchronization hosts for service \"%s\", but placement data is NULL.\n", service->id);

	C_DEBUG(GCMI_DEBUG_LEVEL, C_MSG_DEBUG, "GCMI: Getting synchronization hosts for service \"%s\"...\n", service->id);
	list_alloc(&synchronization_hosts);
	if(placement_data->coordinator_address == C_LOCAL_ADDRESS) {
		list_iterator_init(placement_data->configuration, &hosts_iterator);
		while((host_address = (C_addr_t*) list_iterator_next(&hosts_iterator)) != NULL) {
			if(*host_address != placement_data->coordinator_address)
				list_add(synchronization_hosts, host_address);
		}
	} else if(!inter_host_update) {
		// Only synchronize regular updates on non-coordinator hosts.
		list_add(synchronization_hosts, &placement_data->coordinator_address);
	}

	if(inter_host_update)
		C_DEBUG(GCMI_DEBUG_LEVEL, C_MSG_DEBUG, "GCMI: Synchronization hosts for inter-host update of service \"%s\" are ", service->id);
	else
		C_DEBUG(GCMI_DEBUG_LEVEL, C_MSG_DEBUG, "GCMI: Synchronization hosts for update of service \"%s\" are ", service->id);
	list_print(GCMI_DEBUG_LEVEL, (C_MSG_DEBUG | C_MSG_CONT), synchronization_hosts, print_node_address);
	C_DEBUG(GCMI_DEBUG_LEVEL, (C_MSG_DEBUG | C_MSG_CONT), ".\n");

	return synchronization_hosts;
}

#else
	#warning Compiling without support for GCMI...
#endif	// C_SERVICE_PLACEMENT_EVALUATION
