/**
 * @file graph_cost_multiple_instances.h
 *
 * @brief Graph cost multiple instances (header).
 *
 * Copyright (c) 2007, 2008, 2009, 2010 Georg Wittenburg, Freie Universitaet
 * Berlin (wittenbu@inf.fu-berlin.de). All rights reserved.
 * Please refer to the COPYING file for detailed licensing information.
 *
 * @author Georg Wittenburg
 *
 * $Id: graph_cost_multiple_instances.h 3424 2010-03-18 14:45:09Z gwitten $
 */

#ifndef GRAPH_COST_MULTIPLE_INSTANCES_H
#define GRAPH_COST_MULTIPLE_INSTANCES_H

#include "../../types.h"
#include "../../service/service_types.h"
#include "../placement_types.h"

#ifdef C_SERVICE_PLACEMENT_EVALUATION



/**
 * @brief Change simulation to allow measurements of the cost metrics over time.
 */
//#define GCMI_COST_METRICS_MEASUREMENT

#ifdef GCMI_COST_METRICS_MEASUREMENT
#warning GCMI is configured for cost metric measurements.
#endif	// GCMI_COST_METRICS_MEASUREMENT

/**
 * @brief Change simulation to output raw data about a snapshot of the network
 * map.
 */
//#define GCMI_NETWORK_MAP_SNAPSHOT_MEASUREMENT

/**
 * @brief Change simulation to output raw data about a snapshot of the cost
 * curve.
 */
//#define GCMI_COST_CURVE_SNAPSHOT_MEASUREMENT

/**
 * @brief Change simulation to allow for output of raw data documenting a
 * single run of the GCMI main loop, i.e. a snapshot of the algorithm.
 */
#if (defined(GCMI_NETWORK_MAP_SNAPSHOT_MEASUREMENT) || defined(GCMI_COST_CURVE_SNAPSHOT_MEASUREMENT))
#define GCMI_SNAPSHOT_MEASUREMENT
#endif	// GCMI_*_MEASUREMENT

#ifdef GCMI_SNAPSHOT_MEASUREMENT
#warning GCMI is configured for snapshot measurements.
#endif	// GCMI_SNAPSHOT_MEASUREMENT



/**
 * @brief Initial trigger range for this algorithm on regular hosts.
 */
#define GCMI_HOST_TRIGGER_INITIAL_RANGE			100

/**
 * @brief Current cost multiplier of the trigger range for this algorithm on
 * regular hosts.
 *
 * This factor depends on a lot of factors such as cost metric, network
 * mobility, etc.
 */
#define GCMI_HOST_TRIGGER_RANGE_MULTIPLIER		0.1

/**
 * @brief Trigger duration for this algorithm on regular hosts in seconds.
 */
#ifndef GCMI_COST_METRICS_MEASUREMENT
	#define GCMI_HOST_TRIGGER_DURATION		20
#else
	#define GCMI_HOST_TRIGGER_DURATION		10
#endif	// GCMI_COST_METRICS_MEASUREMENT

/**
 * @brief Maximal trigger duration for this algorithm on regular hosts in
 * seconds.
 *
 * This variable controls how often aggregated statistics are sent to the
 * coordinator node. A crucial point about this value is that the data
 * structures at the coordinator need to take it into account, e.g., when
 * calculating a data rate (see DATARATE_STATS_BUCKETS).
 */
#define GCMI_HOST_TRIGGER_MAX_DURATION			(10 * GCMI_HOST_TRIGGER_DURATION)

/**
 * @brief Initial trigger range for this algorithm on the coordinator node.
 */
#define GCMI_COORDINATOR_TRIGGER_INITIAL_RANGE		1000

/**
 * @brief Current cost multiplier of the trigger range for this algorithm on
 * the coordinator node.
 *
 * This factor depends on a lot of factors such as cost metric, network
 * mobility, etc.
 */
#define GCMI_COORDINATOR_TRIGGER_RANGE_MULTIPLIER	0.1

/**
 * @brief Trigger duration for this algorithm on the coordinator node in
 * seconds.
 */
#ifndef GCMI_COST_METRICS_MEASUREMENT
	#define GCMI_COORDINATOR_TRIGGER_DURATION	(3 * GCMI_HOST_TRIGGER_DURATION)
#else
	#define GCMI_COORDINATOR_TRIGGER_DURATION	0.5
#endif	// GCMI_COST_METRICS_MEASUREMENT

/**
 * @brief Maximal trigger duration for this algorithm on the coordinator node
 * in seconds.
 */
#define GCMI_COORDINATOR_TRIGGER_MAX_DURATION		(10 * GCMI_COORDINATOR_TRIGGER_DURATION)



void graph_cost_multiple_instances_start(SERVICE_service_t* service);
void graph_cost_multiple_instances_stop(SERVICE_service_t* service);
void graph_cost_multiple_instances_trigger(SERVICE_service_t* service, uint8_t reason_id, const void* reason_parameter);
size_t graph_cost_multiple_instances_get_data(const SERVICE_service_t* service, char** data);
void graph_cost_multiple_instances_set_data(SERVICE_service_t* service, const char* data, size_t data_size);
bool graph_cost_multiple_instances_update_data(SERVICE_service_t* service, const PLACEMENT_incoming_adaptation_t* adaptation);
list_t* graph_cost_multiple_instances_get_synchronization_hosts(const SERVICE_service_t* service, bool inter_host_update);

#endif	// C_SERVICE_PLACEMENT_EVALUATION

#endif	// GRAPH_COST_MULTIPLE_INSTANCES_H
