/**
 * @file placement.c
 *
 * @brief Placement main file.
 *
 * Copyright (c) 2007, 2008, 2009, 2010 Georg Wittenburg, Freie Universitaet
 * Berlin (wittenbu@inf.fu-berlin.de). All rights reserved.
 * Please refer to the COPYING file for detailed licensing information.
 *
 * @author Georg Wittenburg
 *
 * $Id: placement.c 3483 2010-04-28 08:23:55Z gwitten $
 */

#include <string.h>

#include "../config.h"
#include "../macros.h"
#include "../types.h"
#include "../vars.h"
#include "../main.h"
#include "../util/callback.h"
#include "../util/debug.h"
#include "../util/list.h"
#include "../util/net.h"
#include "../util/trigger.h"
#include "../rudp-light/rudp-light.h"
#include "../slp-light/slp-light.h"
#include "../service/service.h"
#include "placement.h"
#include "placement_types.h"
#include "placement_vars.h"



#define PLACEMENT_PACKET_TYPE_ADAPTATION_REQUEST	0x0	///< PLACEMENT ADAPTATION REQUEST packet type.
#define PLACEMENT_PACKET_TYPE_ADAPTATION_COMPLETE	0x1	///< PLACEMENT ADAPTATION COMPLETE packet type.

/**
 * @brief Timeout in seconds after which an adaptation is expected to be
 * complete.
 */
#define PLACEMENT_ADAPTATION_TIMEOUT			180



extern PLACEMENT_algorithm_t simple_heuristic_algorithm;		///< Simple heuristic algorithm structure.
extern PLACEMENT_algorithm_t graph_cost_single_instance_algorithm;	///< SPi / GCSI algorithm structure.
extern PLACEMENT_algorithm_t graph_cost_multiple_instances_algorithm;	///< SPi / GCMI algorithm structure.
extern PLACEMENT_algorithm_t linkpull_migration_algorithm;		///< LinkPull migration algorithm structure.
extern PLACEMENT_algorithm_t peerpull_migration_algorithm;		///< PeerPull migration algorithm structure.
extern PLACEMENT_algorithm_t topocenter_migration_algorithm;		///< TopoCenter(1) algorithm structure.
extern PLACEMENT_algorithm_t tree_topology_migration_algorithm;		///< Tree-topology algorithm structure.
extern PLACEMENT_algorithm_t asg_simple_algorithm;			///< ASG / simple algorithm structure.
extern PLACEMENT_algorithm_t asg_event_flow_tree_algorithm;		///< ASG / EFT algorithm structure.



/*
 * Interface to main module.
 */

/**
 * @brief Initialize the placement module.
 */
void PLACEMENT_init() {
#ifdef C_SERVICE_PLACEMENT_EVALUATION
	C_PLACEMENT_currently_outgoing_adaptation = NULL;
	C_PLACEMENT_services_to_shutdown_after_delay = NULL;
	list_alloc(&C_PLACEMENT_services_to_shutdown_after_delay);
	C_PLACEMENT_current_service_cost = 0.0;
	C_PLACEMENT_currently_optimal_service_cost = 0.0;
	C_PLACEMENT_currently_optimal_number_of_hosts = 0;
	C_PLACEMENT_currently_estimated_adaptation_cost = 0.0;
	C_PLACEMENT_total_adaptations = 0;
	C_PLACEMENT_successful_adaptations = 0;
	C_PLACEMENT_failed_adaptations = 0;
#endif	// C_SERVICE_PLACEMENT_EVALUATION
}

/**
 * @brief Start the placement module.
 */
void PLACEMENT_start() {
#ifdef C_SERVICE_PLACEMENT_EVALUATION
#endif	// C_SERVICE_PLACEMENT_EVALUATION
}

/**
 * @brief Stop the placement module.
 */
void PLACEMENT_stop() {
#ifdef C_SERVICE_PLACEMENT_EVALUATION
	if(C_PLACEMENT_currently_outgoing_adaptation != NULL) {
		C_DEBUG(PLACEMENT_DEBUG_LEVEL, C_MSG_WARN, "PLACEMENT: Aborting outgoing adaptation...\n");
		if(C_PLACEMENT_currently_outgoing_adaptation->timeout_callback_id != 0)
			callback_cancel(C_PLACEMENT_currently_outgoing_adaptation->timeout_callback_id);
		C_PLACEMENT_currently_outgoing_adaptation->result->error_code = C_ERROR_ABORTED;
		C_PLACEMENT_currently_outgoing_adaptation->callback(C_PLACEMENT_currently_outgoing_adaptation->result);
		FREE(C_PLACEMENT_currently_outgoing_adaptation);
		C_PLACEMENT_failed_adaptations++;
	}
	list_free(&C_PLACEMENT_services_to_shutdown_after_delay);
#endif	// C_SERVICE_PLACEMENT_EVALUATION
}

/**
 * @brief Finalize the placement module.
 */
void PLACEMENT_finalize() {
#ifdef C_SERVICE_PLACEMENT_EVALUATION
	if(C_check_consistency_on_shutdown) {
		C_ABORT_ON(C_PLACEMENT_total_adaptations != (C_PLACEMENT_successful_adaptations + C_PLACEMENT_failed_adaptations), "PLACEMENT: Adaptation counters don't add up (%d != (%d + %d)).\n", C_PLACEMENT_total_adaptations, C_PLACEMENT_successful_adaptations, C_PLACEMENT_failed_adaptations);
	}
#endif	// C_SERVICE_PLACEMENT_EVALUATION
}



/*
 * Utility functions.
 */

#ifdef C_SERVICE_PLACEMENT_EVALUATION

/**
 * @brief Find the new coordinator in a set of commands.
 *
 * @param commands Pointer to list of adaptation commands.
 * @return Address of new coordinator, or C_BROADCAST_ADDRESS if no coordinator
 * was found.
 */
static C_addr_t get_coordinator_from_commands(const list_t* commands) {
	PLACEMENT_command_t* command = NULL;
	list_iterator_t commands_iterator;

	list_iterator_init(commands, &commands_iterator);
	while((command = (PLACEMENT_command_t*) list_iterator_next(&commands_iterator)) != NULL) {
		if(command->id == PLACEMENT_COMMAND_COORDINATOR) {
			return command->destination_address;
		}
	}
	return C_BROADCAST_ADDRESS;
}

/**
 * @brief Find the handler for the new coordinator in a set of commands.
 *
 * @param commands Pointer to list of adaptation commands.
 * @return Address of handler of new coordinator, or C_BROADCAST_ADDRESS if no
 * coordinator was found.
 */
static C_addr_t get_coordinator_handler_from_commands(const list_t* commands) {
	PLACEMENT_command_t* command = NULL;
	list_iterator_t commands_iterator;
	C_addr_t coordinator_address = get_coordinator_from_commands(commands);
	C_addr_t coordinator_handler_address = C_BROADCAST_ADDRESS;

	if(coordinator_address != C_BROADCAST_ADDRESS) {
		list_iterator_init(commands, &commands_iterator);
		while((command = (PLACEMENT_command_t*) list_iterator_next(&commands_iterator)) != NULL) {
			if(
				((command->id == PLACEMENT_COMMAND_REPLICATE)
				|| (command->id == PLACEMENT_COMMAND_MIGRATE))
				&& (command->destination_address == coordinator_address)
			) {
				C_ABORT_ON(coordinator_handler_address != C_BROADCAST_ADDRESS, "Found two handlers for the new coordinator (%s, %s).\n", addr_to_str(coordinator_handler_address), addr_to_str(command->handler_address));
				coordinator_handler_address = command->handler_address;
			}
		}
		// No node is responsible for the coordinator, hence the
		// coordinator remains on its current node.
		if(coordinator_handler_address == C_BROADCAST_ADDRESS)
			coordinator_handler_address = coordinator_address;
	}
	return coordinator_handler_address;
}

/**
 * @brief Notify the placement algorithm that the current node is to be used as
 * coordinator node.
 *
 * @param adaptation Pointer to adaptation plan to implement.
 */
static void use_current_node_as_coordinator(PLACEMENT_incoming_adaptation_t* adaptation) {
	SERVICE_service_t* service = NULL;
	PLACEMENT_incoming_adaptation_t* set_coordinator_adaptation = NULL;
	PLACEMENT_command_t set_coordinator_command = {C_BROADCAST_ADDRESS, PLACEMENT_COMMAND_COORDINATOR, C_LOCAL_ADDRESS};

	/// Update information for the ongoing adaptation.

	adaptation->coordinator_address = C_LOCAL_ADDRESS;

	/// Update information in the placement algorithm.

#ifdef C_ENABLE_CONSISTENCY_CHECKS
	SERVICE_check_service_id(adaptation->service_id);
#endif	// C_ENABLE_CONSISTENCY_CHECKS

	service = SERVICE_get_service(adaptation->service_id);
	if(service == NULL) {
		C_DEBUG(PLACEMENT_DEBUG_LEVEL, C_MSG_WARN, "PLACEMENT: Trying to use current node as coordinator for service \"%s\", but service is not active anymore.\n", adaptation->service_id);
		return;
	} else if(service->placement_algorithm == NULL) {
		C_DEBUG(PLACEMENT_DEBUG_LEVEL, C_MSG_WARN, "PLACEMENT: Trying to use current node as coordinator for service \"%s\", but algorithm is not active anymore.\n", adaptation->service_id);
		return;
	}

	if(service->placement_algorithm->update_data != NULL) {
		C_DEBUG(PLACEMENT_DEBUG_LEVEL, C_MSG_INFO, "PLACEMENT: Using current node as coordinator for service \"%s\"...\n", adaptation->service_id);

		MALLOC(set_coordinator_adaptation, PLACEMENT_incoming_adaptation_t);
		SERVICE_copy_service_id(set_coordinator_adaptation->service_id, adaptation->service_id);
		list_alloc(&set_coordinator_adaptation->commands);
		list_add(set_coordinator_adaptation->commands, &set_coordinator_command);
		set_coordinator_adaptation->coordinator_address = C_BROADCAST_ADDRESS;
		set_coordinator_adaptation->current_command_handles_coordinator = false;

		service->placement_algorithm->update_data(service, set_coordinator_adaptation);

		list_free(&set_coordinator_adaptation->commands);
		FREE(set_coordinator_adaptation);
	}
}



/*
 * Packet handling functions.
 */

/**
 * @brief Timeout callback for sending a packet via RUDP.
 *
 * @param parameter Transmission for which the timeout ocurred.
 */
static void handle_PLACEMENT_ADAPTATION_COMPLETE_timeout(callback_parameter_t parameter) {
	PLACEMENT_outgoing_adaptation_t* adaptation = (PLACEMENT_outgoing_adaptation_t*) parameter;

	if(C_PLACEMENT_currently_outgoing_adaptation == NULL) {
		C_DEBUG(PLACEMENT_DEBUG_LEVEL, C_MSG_WARN, "PLACEMENT: Ignoring timeout for previously completed adaptation...\n");
		return;
	}
	C_ABORT_ON(adaptation != C_PLACEMENT_currently_outgoing_adaptation, "PLACEMENT: Timeout parameter disagrees with global variable about outgoing adaptation (%p != %p).\n", adaptation, C_PLACEMENT_currently_outgoing_adaptation);

	C_DEBUG(PLACEMENT_DEBUG_LEVEL, C_MSG_WARN, "PLACEMENT: Timed out while waiting for ADAPTATION COMPLETE, adaptation of service \"%s\" failed.\n", C_PLACEMENT_currently_outgoing_adaptation->service_id);
	if(C_PLACEMENT_currently_outgoing_adaptation->callback != NULL) {
		C_ABORT_ON(C_PLACEMENT_currently_outgoing_adaptation->result == NULL, "PLACEMENT: Trying to call callback function, but result is NULL.\n"); 
		C_PLACEMENT_currently_outgoing_adaptation->result->error_code = C_ERROR_TIMEOUT;
		C_PLACEMENT_currently_outgoing_adaptation->callback(C_PLACEMENT_currently_outgoing_adaptation->result);
	}
	C_PLACEMENT_failed_adaptations++;
	FREE(C_PLACEMENT_currently_outgoing_adaptation);
}

/**
 * @brief RUDP callback for a PLACEMENT ADAPTATION REQUEST packet.
 *
 * @param parameter Pointer to RUDP transmission result.
 */
static void send_PLACEMENT_ADAPTATION_REQUEST_callback(callback_parameter_t parameter) {
	RUDP_transmission_result_t* result = (RUDP_transmission_result_t*) parameter;
	PLACEMENT_outgoing_adaptation_t* adaptation = NULL;

	if(result->parameter == NULL) {
		switch(result->error_code) {
		case C_ERROR_OK:
			C_DEBUG(PLACEMENT_DEBUG_LEVEL, C_MSG_DEBUG, "PLACEMENT: ADAPTATION REQUEST reached the destination node.\n");
			break;
		case C_ERROR_TIMEOUT:
			C_DEBUG(PLACEMENT_DEBUG_LEVEL, C_MSG_WARN, "PLACEMENT: ADAPTATION REQUEST timed out, adaptation command failed.\n");
			break;
		case C_ERROR_ABORTED:
			C_DEBUG(PLACEMENT_DEBUG_LEVEL, C_MSG_WARN, "PLACEMENT: Sending ADAPTATION REQUEST aborted, adaptation command failed.\n");
			break;
		case C_ERROR_BUSY:
			C_DEBUG(PLACEMENT_DEBUG_LEVEL, C_MSG_WARN, "PLACEMENT: ADAPTATION REQUEST hit full queue, adaptation command failed.\n");
			break;
		default:
			C_ABORT_ON(true, "PLACEMENT: Unhandled error while sending ADAPTATION REQUEST with RUDP (%d).\n", result->error_code);
		}
		FREE(result);
		return;
	}

	/// If this callback is being passed a parameter, then the related RUDP
	/// transmission was addressed to the handler of the new coordinator.
	/// Proceed to handle errors accordingly.

	adaptation = (PLACEMENT_outgoing_adaptation_t*) result->parameter;

	if(C_PLACEMENT_currently_outgoing_adaptation == NULL) {
		C_DEBUG(PLACEMENT_DEBUG_LEVEL, C_MSG_WARN, "PLACEMENT: Ignoring callback for previously completed adaptation...\n");
		FREE(result);
		return;
	}
	C_ABORT_ON(adaptation != C_PLACEMENT_currently_outgoing_adaptation, "PLACEMENT: Callback parameter disagrees with global variable about outgoing adaptation (%p != %p).\n", adaptation, C_PLACEMENT_currently_outgoing_adaptation);

	switch(result->error_code) {
	case C_ERROR_OK:
		C_DEBUG(PLACEMENT_DEBUG_LEVEL, C_MSG_DEBUG, "PLACEMENT: ADAPTATION REQUEST for service \"%s\" reached the coordinator handler.\n", adaptation->service_id);
		if(C_PLACEMENT_currently_outgoing_adaptation->coordinator_handler_address == C_PLACEMENT_currently_outgoing_adaptation->coordinator_address) {
			// The coordinator handler is identical with the
			// coordinator. Hence, receiving this ACK means that
			// the adaptation was successful.
			if(C_PLACEMENT_currently_outgoing_adaptation->callback != NULL) {
				C_ABORT_ON(C_PLACEMENT_currently_outgoing_adaptation->result == NULL, "PLACEMENT: Trying to call callback function, but result is NULL.\n"); 
				C_PLACEMENT_currently_outgoing_adaptation->result->error_code = C_ERROR_OK;
				C_PLACEMENT_currently_outgoing_adaptation->callback(C_PLACEMENT_currently_outgoing_adaptation->result);
			}
			C_PLACEMENT_successful_adaptations++;
		} else {
			// Wait for the coordinator handler to finish
			// replicating or migration to the new coordinator.
			adaptation->timeout_callback_id = callback_add(handle_PLACEMENT_ADAPTATION_COMPLETE_timeout, C_PLACEMENT_currently_outgoing_adaptation, PLACEMENT_ADAPTATION_TIMEOUT);
			FREE(result);	// Only free the RUDP transmission result, not
			return;		// the adaptation.
		}
		break;
	case C_ERROR_TIMEOUT:
	case C_ERROR_ABORTED:
	case C_ERROR_BUSY:
		C_DEBUG(PLACEMENT_DEBUG_LEVEL, C_MSG_WARN, "PLACEMENT: ADAPTATION REQUEST for coordinator handler failed (%d), adaptation of service \"%s\" failed.\n", result->error_code, C_PLACEMENT_currently_outgoing_adaptation->service_id);
		if(C_PLACEMENT_currently_outgoing_adaptation->callback != NULL) {
			C_ABORT_ON(C_PLACEMENT_currently_outgoing_adaptation->result == NULL, "PLACEMENT: Trying to call callback function, but result is NULL.\n"); 
			C_PLACEMENT_currently_outgoing_adaptation->result->error_code = result->error_code;
			C_PLACEMENT_currently_outgoing_adaptation->callback(C_PLACEMENT_currently_outgoing_adaptation->result);
		}
		C_PLACEMENT_failed_adaptations++;
		break;
	default:
		C_ABORT_ON(true, "PLACEMENT: Unhandled error while sending ADAPTATION REQUEST to coordinator handler with RUDP (%d).\n", result->error_code);
	}

	FREE(C_PLACEMENT_currently_outgoing_adaptation);
	FREE(result);
}

/**
 * @brief Send a PLACEMENT ADAPTATION REQUEST packet.
 *
 * @param destination_address Address of the destination node.
 * @param adaptation Pointer to adaptation plan to distribute.
 * @param commands Pointer to list of adaptation commands.
 * @param is_coordinator_handler Flag whether this adaptation request is for
 * the handler of the new coordinator.
 */
static void send_PLACEMENT_ADAPTATION_REQUEST(C_addr_t destination_address, PLACEMENT_outgoing_adaptation_t* adaptation, const list_t* commands, bool is_coordinator_handler) {
	PLACEMENT_ADAPTATION_REQUEST_packet_t request;
	RUDP_transmission_result_t* result = NULL;
	PLACEMENT_command_t* command_array = NULL;
	size_t command_array_size = list_size(commands) * sizeof(PLACEMENT_command_t);

	C_DEBUG(PLACEMENT_DEBUG_LEVEL, C_MSG_INFO, "PLACEMENT: Sending ADAPTATION REQUEST for service \"%s\" to %s...\n", adaptation->service_id, addr_to_str(destination_address));
	memset(&request, 0, sizeof(PLACEMENT_ADAPTATION_REQUEST_packet_t));
	SERVICE_copy_service_id(request.service_id, adaptation->service_id);
	MALLOC_ARRAY(command_array, PLACEMENT_command_t, list_size(commands));
	list_to_array(commands, command_array, sizeof(PLACEMENT_command_t));
	MALLOC(result, RUDP_transmission_result_t);
	if(is_coordinator_handler) {
		C_DEBUG(PLACEMENT_DEBUG_LEVEL, C_MSG_DEBUG, "PLACEMENT: ADAPTATION REQUEST is for the coordinator handler.\n");
		result->parameter = adaptation;
	} else {
		result->parameter = NULL;
	}
	RUDP_sendto2(destination_address, PACKET_PRIORITY_SIGNALLING_PLACEMENT, APP_TYPE_PLACEMENT, PLACEMENT_PACKET_TYPE_ADAPTATION_REQUEST, &request, sizeof(PLACEMENT_ADAPTATION_REQUEST_packet_t), command_array, command_array_size, send_PLACEMENT_ADAPTATION_REQUEST_callback, result);
	FREE(command_array);
}

/**
 * @brief RUDP callback for a PLACEMENT ADAPTATION COMPLETE packet.
 *
 * @param parameter Pointer to RUDP transmission result.
 */
static void send_PLACEMENT_ADAPTATION_COMPLETE_callback(callback_parameter_t parameter) {
	RUDP_transmission_result_t* result = (RUDP_transmission_result_t*) parameter;

	switch(result->error_code) {
	case C_ERROR_OK:
		C_DEBUG(PLACEMENT_DEBUG_LEVEL, C_MSG_DEBUG, "PLACEMENT: ADAPTATION COMPLETE reached the destination node.\n");
		break;
	case C_ERROR_TIMEOUT:
		C_DEBUG(PLACEMENT_DEBUG_LEVEL, C_MSG_WARN, "PLACEMENT: ADAPTATION COMPLETE timed out, adaptation command failed.\n");
		break;
	case C_ERROR_ABORTED:
		C_DEBUG(PLACEMENT_DEBUG_LEVEL, C_MSG_WARN, "PLACEMENT: Sending ADAPTATION COMPLETE aborted, adaptation command failed.\n");
		break;
	case C_ERROR_BUSY:
		C_DEBUG(PLACEMENT_DEBUG_LEVEL, C_MSG_WARN, "PLACEMENT: ADAPTATION COMPLETE hit full queue, adaptation command failed.\n");
		break;
	default:
		C_ABORT_ON(true, "PLACEMENT: Unhandled error while sending ADAPTATION COMPLETE with RUDP (%d).\n", result->error_code);
	}
	FREE(result);
}

/**
 * @brief Send a PLACEMENT ADAPTATION COMPLETE packet.
 *
 * @param destination_address Address of the destination node.
 * @param service_id ID of the service for which the adaptation was completed.
 */
static void send_PLACEMENT_ADAPTATION_COMPLETE(C_addr_t destination_address, const SERVICE_id_t* service_id) {
	PLACEMENT_ADAPTATION_COMPLETE_packet_t adaptation_complete;
	RUDP_transmission_result_t* result = NULL;

	C_DEBUG(PLACEMENT_DEBUG_LEVEL, C_MSG_INFO, "PLACEMENT: Sending ADAPTATION COMPLETE for service \"%s\" to %s...\n", service_id, addr_to_str(destination_address));
	memset(&adaptation_complete, 0, sizeof(PLACEMENT_ADAPTATION_COMPLETE_packet_t));
	SERVICE_copy_service_id(adaptation_complete.service_id, service_id);
	MALLOC(result, RUDP_transmission_result_t);
	RUDP_sendto(destination_address, PACKET_PRIORITY_SIGNALLING_PLACEMENT, APP_TYPE_PLACEMENT, PLACEMENT_PACKET_TYPE_ADAPTATION_COMPLETE, &adaptation_complete, sizeof(PLACEMENT_ADAPTATION_COMPLETE_packet_t), send_PLACEMENT_ADAPTATION_COMPLETE_callback, result);
}

static void implement_adaptation(PLACEMENT_incoming_adaptation_t* adaptation);

/**
 * @brief Replication callback function of adaptation plan implementation.
 *
 * @param parameter Replication result.
 */
static void implement_adaptation_replicate_service_callback(callback_parameter_t parameter) {
	SERVICE_replication_result_t* result = (SERVICE_replication_result_t*) parameter;
	PLACEMENT_incoming_adaptation_t* adaptation = (PLACEMENT_incoming_adaptation_t*) result->parameter;

	switch(result->error_code) {
	case C_ERROR_OK:
		C_DEBUG(PLACEMENT_DEBUG_LEVEL, C_MSG_NOTE, "PLACEMENT: Replication of service \"%s\" was successful, proceeding with next command in configuration adaptation...\n", adaptation->service_id);
		implement_adaptation(adaptation);
		break;
	case C_ERROR_ABORTED:
		if(adaptation->current_command_handles_coordinator) {
			C_DEBUG(PLACEMENT_DEBUG_LEVEL, C_MSG_NOTE, "PLACEMENT: Replication of service \"%s\" to new coordinator has been aborted, setting current node to be the coordinator and aborting configuration adaptation...\n", adaptation->service_id);
			use_current_node_as_coordinator(adaptation);
		} else {
			C_DEBUG(PLACEMENT_DEBUG_LEVEL, C_MSG_NOTE, "PLACEMENT: Replication of service \"%s\" has been aborted, aborting configuration adaptation...\n", adaptation->service_id);
		}
		list_free_including_data(&adaptation->commands, free);
		FREE(adaptation);
		break;
	case C_ERROR_PARAMETER:
	case C_ERROR_TIMEOUT:
	case C_ERROR_BUSY:
	case C_ERROR_IN_PROGRESS:
	case C_ERROR_COMM_ERROR:
	case C_ERROR_REMOTE_BUSY:
		if(adaptation->current_command_handles_coordinator) {
			C_DEBUG(PLACEMENT_DEBUG_LEVEL, C_MSG_WARN, "PLACEMENT: Replication of service \"%s\" to new coordinator failed (%d), setting current node to be the coordinator and proceeding with next command in configuration adaptation...\n", adaptation->service_id, result->error_code);
			use_current_node_as_coordinator(adaptation);
		} else {
			C_DEBUG(PLACEMENT_DEBUG_LEVEL, C_MSG_WARN, "PLACEMENT: Replication of service \"%s\" failed (%d), proceeding with next command in configuration adaptation...\n", adaptation->service_id, result->error_code);
		}
		implement_adaptation(adaptation);
		break;
	default:
		C_ABORT_ON(true, "PLACEMENT: Replication callback with unknown error code 0x%X.\n", result->error_code);
	}

	FREE(result);
}

/**
 * @brief Migration callback function of adaptation plan implementation.
 *
 * @param parameter Migration result.
 */
static void implement_adaptation_migrate_service_callback(callback_parameter_t parameter) {
	SERVICE_migration_result_t* result = (SERVICE_migration_result_t*) parameter;
	PLACEMENT_incoming_adaptation_t* adaptation = (PLACEMENT_incoming_adaptation_t*) result->parameter;

	switch(result->error_code) {
	case C_ERROR_OK:
		C_DEBUG(PLACEMENT_DEBUG_LEVEL, C_MSG_NOTE, "PLACEMENT: Migration of service \"%s\" was successful, proceeding with next command in configuration adaptation...\n", adaptation->service_id);
		implement_adaptation(adaptation);
		break;
	case C_ERROR_ABORTED:
		if(adaptation->current_command_handles_coordinator) {
			C_DEBUG(PLACEMENT_DEBUG_LEVEL, C_MSG_NOTE, "PLACEMENT: Migration of service \"%s\" to new coordinator has been aborted, setting current node to be the coordinator and aborting configuration adaptation...\n", adaptation->service_id);
			use_current_node_as_coordinator(adaptation);
		} else {
			C_DEBUG(PLACEMENT_DEBUG_LEVEL, C_MSG_NOTE, "PLACEMENT: Migration of service \"%s\" has been aborted, aborting configuration adaptation...\n", adaptation->service_id);
		}
		list_free_including_data(&adaptation->commands, free);
		FREE(adaptation);
		break;
	case C_ERROR_PARAMETER:
	case C_ERROR_TIMEOUT:
	case C_ERROR_BUSY:
	case C_ERROR_IN_PROGRESS:
	case C_ERROR_COMM_ERROR:
	case C_ERROR_REMOTE_BUSY:
		if(adaptation->current_command_handles_coordinator) {
			C_DEBUG(PLACEMENT_DEBUG_LEVEL, C_MSG_WARN, "PLACEMENT: Migration of service \"%s\" to new coordinator failed (%d), setting current node to be the coordinator and proceeding with next command in configuration adaptation...\n", adaptation->service_id, result->error_code);
			use_current_node_as_coordinator(adaptation);
		} else {
			C_DEBUG(PLACEMENT_DEBUG_LEVEL, C_MSG_WARN, "PLACEMENT: Migration of service \"%s\" failed (%d), retaining service instance and proceeding with next command in configuration adaptation...\n", adaptation->service_id, result->error_code);
			// The flag to retain this service instance is set
			// as part of the migration process.
		}
		implement_adaptation(adaptation);
		break;
	default:
		C_ABORT_ON(true, "PLACEMENT: Migration callback with unknown error code 0x%X.\n", result->error_code);
	}

	FREE(result);
}

/**
 * @brief Callback function to shutdown a service after a delay.
 *
 * @param parameter Pointer to service to shutdown.
 */
static void shutdown_service_after_delay_callback(callback_parameter_t parameter) {
	SERVICE_service_t* service = (SERVICE_service_t*) parameter;

	if(SERVICE_exists_service(service))
		SERVICE_shutdown_service(service->id);
}

/**
 * @brief Implement an adaptation plan.
 *
 * @param adaptation Pointer to adaptation plan to implement.
 */
static void implement_adaptation(PLACEMENT_incoming_adaptation_t* adaptation) {
	PLACEMENT_command_t* command = NULL;
	SERVICE_replication_result_t* replication_result = NULL;
	SERVICE_migration_result_t* migration_result = NULL;
	SERVICE_service_t* service = NULL;

	// Once the callback from either the replication or the migration of
	// the coordinator has occured, send a confirmation to the node that
	// requested this adaptation.
	if(adaptation->current_command_handles_coordinator)
		send_PLACEMENT_ADAPTATION_COMPLETE(adaptation->adaptation_source_address, adaptation->service_id);

	while(!list_empty(adaptation->commands)) {

		command = (PLACEMENT_command_t*) list_remove_head(adaptation->commands);
		adaptation->current_command_handles_coordinator = false;

		if((command->handler_address == C_LOCAL_ADDRESS) || (command->handler_address == C_BROADCAST_ADDRESS)) {

			if(SERVICE_get_service(adaptation->service_id) == NULL) {
				C_DEBUG(PLACEMENT_DEBUG_LEVEL, C_MSG_ERROR, "PLACEMENT: Trying to implement next step in adaptation plan for service \"%s\", but service is not active anymore.\n", adaptation->service_id);
				FREE(command);
				list_free_including_data(&adaptation->commands, free);
				FREE(adaptation);
				return;
			}

			// Make sure that the coordinator does not shut down.
			if(adaptation->coordinator_address == C_LOCAL_ADDRESS) {
				switch(command->id) {
				case PLACEMENT_COMMAND_NOP:
				case PLACEMENT_COMMAND_COORDINATOR:
				case PLACEMENT_COMMAND_REPLICATE:
					// Nothing to do.
					break;
				case PLACEMENT_COMMAND_MIGRATE:
					C_DEBUG(PLACEMENT_DEBUG_LEVEL, C_MSG_WARN, "PLACEMENT: This node is the coordinator, performing replication of service \"%s\" to %s instead of migration...\n", adaptation->service_id, addr_to_str(command->destination_address));
					command->id = PLACEMENT_COMMAND_REPLICATE;
					break;
				case PLACEMENT_COMMAND_SHUTDOWN:
				case PLACEMENT_COMMAND_DELAYED_SHUTDOWN:
					C_DEBUG(PLACEMENT_DEBUG_LEVEL, C_MSG_WARN, "PLACEMENT: This node is the coordinator, ignoring request for (delayed) shutdown of service \"%s\"...\n", adaptation->service_id);
					command->id = PLACEMENT_COMMAND_NOP;
					break;
				default:
					C_ABORT_ON(true, "PLACEMENT: Invalid placement command (0x%X).\n", command->id);
				}
			}

			switch(command->id) {
			case PLACEMENT_COMMAND_NOP:
				break;
			case PLACEMENT_COMMAND_COORDINATOR:
				// This is handled by the placement algorithm.
				break;
			case PLACEMENT_COMMAND_REPLICATE:
				C_DEBUG(PLACEMENT_DEBUG_LEVEL, C_MSG_INFO, "PLACEMENT: Replicating service \"%s\" to %s...\n", adaptation->service_id, addr_to_str(command->destination_address));

				if(command->destination_address == adaptation->coordinator_address)
					adaptation->current_command_handles_coordinator = true;
				MALLOC(replication_result, SERVICE_replication_result_t);
				replication_result->parameter = adaptation;
				SERVICE_replicate_service(adaptation->service_id, command->destination_address, implement_adaptation_replicate_service_callback, replication_result);
				FREE(command);
				return;		// Return and wait to be called again by callback.
			case PLACEMENT_COMMAND_MIGRATE:
				C_DEBUG(PLACEMENT_DEBUG_LEVEL, C_MSG_INFO, "PLACEMENT: Migrating service \"%s\" to %s...\n", adaptation->service_id, addr_to_str(command->destination_address));

				if(command->destination_address == adaptation->coordinator_address)
					adaptation->current_command_handles_coordinator = true;
				MALLOC(migration_result, SERVICE_migration_result_t);
				migration_result->parameter = adaptation;
				SERVICE_migrate_service(adaptation->service_id, command->destination_address, implement_adaptation_migrate_service_callback, migration_result);
				FREE(command);
				return;		// Return and wait to be called again by callback.
			case PLACEMENT_COMMAND_SHUTDOWN:
				C_DEBUG(PLACEMENT_DEBUG_LEVEL, C_MSG_INFO, "PLACEMENT: Shutting down service \"%s\"...\n", adaptation->service_id);
				SERVICE_shutdown_service(adaptation->service_id);
				break;
			case PLACEMENT_COMMAND_DELAYED_SHUTDOWN:
				service = SERVICE_get_service(adaptation->service_id);
				C_ABORT_ON(service == NULL, "PLACEMENT: Trying to shutdown service \"%s\" with a delay of %.1fs, but no service with this ID was found.\n", adaptation->service_id, command->delay);

				C_DEBUG(PLACEMENT_DEBUG_LEVEL, C_MSG_INFO, "PLACEMENT: Shutting down service \"%s\" in %.1fs...\n", adaptation->service_id, command->delay);
				list_add(C_PLACEMENT_services_to_shutdown_after_delay, service);
				callback_add(shutdown_service_after_delay_callback, service, command->delay);
				break;
			default:
				C_ABORT_ON(true, "PLACEMENT: Invalid placement command (0x%X).\n", command->id);
			}
		} else {
			C_DEBUG(PLACEMENT_DEBUG_LEVEL, C_MSG_DEBUG, "PLACEMENT: Ignoring command for %s...\n", addr_to_str(command->handler_address));
		}
		FREE(command);
	}

	C_DEBUG(PLACEMENT_DEBUG_LEVEL, C_MSG_NOTE, "PLACEMENT: Implementation of adaptation plan for service \"%s\" is complete.\n", adaptation->service_id);

	// Unlock, reset stats, and announce the service if it hasn't been shut
	// down by one of the commands.
	service = SERVICE_get_service(adaptation->service_id);
	if((service != NULL) && !list_contains(C_PLACEMENT_services_to_shutdown_after_delay, service)) {
		SERVICE_unlock_service(adaptation->service_id);

		// Only reset stats if this node is not the coordinator. A
		// reset on a running coordinator at an unspecified point in
		// time may lead to scewed data and incorrect placement
		// decisions.
		if(adaptation->coordinator_address != C_LOCAL_ADDRESS)
			SERVICE_reset_service_stats(adaptation->service_id);

		SLP_announce_service(adaptation->service_id, C_BROADCAST_ADDRESS);
	}

	list_free(&adaptation->commands);
	FREE(adaptation);
}

/**
 * @brief Handle an incoming PLACEMENT ADAPTATION REQUEST.
 *
 * @param header Pointer to the packet header.
 * @param request Pointer to the adaptation request.
 * @param command_array Pointer to array of commands.
 * @param command_array_size Size of command array in bytes.
 */
static void handle_PLACEMENT_ADAPTATION_REQUEST(const net_header_t* header, const PLACEMENT_ADAPTATION_REQUEST_packet_t* request, const PLACEMENT_command_t* command_array, size_t command_array_size) {
	PLACEMENT_incoming_adaptation_t* adaptation = NULL;
	SERVICE_service_t* service = NULL;
	
#ifdef C_ENABLE_CONSISTENCY_CHECKS
	SERVICE_check_service_id(request->service_id);
#endif	// C_ENABLE_CONSISTENCY_CHECKS

	service = SERVICE_get_running_service(request->service_id);
	if(service == NULL) {
		service = SERVICE_get_service(request->service_id);
		if(service == NULL)
			C_DEBUG(PLACEMENT_DEBUG_LEVEL, C_MSG_WARN, "PLACEMENT: Received ADAPTATION REQUEST for service \"%s\" from %s, but no such service is running on this node. Ignoring...\n", request->service_id, addr_to_str(header->source_address));
		else
			C_DEBUG(PLACEMENT_DEBUG_LEVEL, C_MSG_WARN, "PLACEMENT: Received ADAPTATION REQUEST for service \"%s\" from %s, but service is not running (0x%X). Ignoring...\n", request->service_id, addr_to_str(header->source_address), service->state);
		return;
	}

	C_DEBUG(PLACEMENT_DEBUG_LEVEL, C_MSG_DEBUG, "PLACEMENT: Received ADAPTATION REQUEST for service \"%s\" from %s.\n", request->service_id, addr_to_str(header->source_address));

	MALLOC(adaptation, PLACEMENT_incoming_adaptation_t);
	SERVICE_copy_service_id(adaptation->service_id, request->service_id);
	list_alloc(&adaptation->commands);
	list_from_array(adaptation->commands, command_array, sizeof(PLACEMENT_command_t), command_array_size);
	adaptation->coordinator_address = get_coordinator_from_commands(adaptation->commands);
	adaptation->adaptation_source_address = header->source_address;
	adaptation->current_command_handles_coordinator = false;

	if(
		(service->placement_algorithm != NULL)
		&& (service->placement_algorithm->update_data != NULL)
	) {
		C_DEBUG(PLACEMENT_DEBUG_LEVEL, C_MSG_INFO, "PLACEMENT: Updating algorithm data for service \"%s\"...\n", adaptation->service_id);
		if(!service->placement_algorithm->update_data(service, adaptation)) {
			C_DEBUG(PLACEMENT_DEBUG_LEVEL, C_MSG_WARN, "PLACEMENT: Algorithm data for service \"%s\" is invalid, aborting processing of ADAPTATION REQUEST from %s...\n", adaptation->service_id, addr_to_str(header->source_address));
			list_free_including_data(&adaptation->commands, free);
			FREE(adaptation);
			return;
		}
	}

	C_DEBUG(PLACEMENT_DEBUG_LEVEL, C_MSG_NOTE, "PLACEMENT: Starting implementation of adaptation plan for service \"%s\"...\n", adaptation->service_id);
	SERVICE_lock_service(adaptation->service_id);
	implement_adaptation(adaptation);
}

/**
 * @brief Handle an incoming PLACEMENT ADAPTATION COMPLETE.
 *
 * @param header Pointer to the packet header.
 * @param adaptation_complete Pointer to the adaptation complete struct.
 */
static void handle_PLACEMENT_ADAPTATION_COMPLETE(const net_header_t* header, const PLACEMENT_ADAPTATION_COMPLETE_packet_t* adaptation_complete) {
	if(C_PLACEMENT_currently_outgoing_adaptation == NULL) {
		C_DEBUG(PLACEMENT_DEBUG_LEVEL, C_MSG_WARN, "PLACEMENT: Ignoring ADAPTATION COMPLETE for previously completed adaptation...\n");
		return;
	}
	C_ABORT_ON(!SERVICE_same_service_id(adaptation_complete->service_id, C_PLACEMENT_currently_outgoing_adaptation->service_id), "PLACEMENT: ADAPTATION COMPLETE parameter disagrees with global variable about outgoing adaptation (%s != %s).\n", adaptation_complete->service_id, C_PLACEMENT_currently_outgoing_adaptation->service_id);

	// Don't cancel the callback if we received the ADAPTATION_COMPLETE
	// before the RUDP ACK for the ADAPTATION_REQUEST.
	if(C_PLACEMENT_currently_outgoing_adaptation->timeout_callback_id != 0)
		callback_cancel(C_PLACEMENT_currently_outgoing_adaptation->timeout_callback_id);

	if(C_PLACEMENT_currently_outgoing_adaptation->callback != NULL) {
		C_ABORT_ON(C_PLACEMENT_currently_outgoing_adaptation->result == NULL, "PLACEMENT: Trying to call callback function, but result is NULL.\n"); 
		C_PLACEMENT_currently_outgoing_adaptation->result->error_code = C_ERROR_OK;
		C_PLACEMENT_currently_outgoing_adaptation->callback(C_PLACEMENT_currently_outgoing_adaptation->result);
	}
	C_PLACEMENT_successful_adaptations++;
	FREE(C_PLACEMENT_currently_outgoing_adaptation);
}

#endif	// C_SERVICE_PLACEMENT_EVALUATION

/**
 * @brief Placement packet classifier.
 *
 * @param header Pointer to header of an incoming packet.
 * @param payload Pointer to payload of an incoming packet.
 */
void PLACEMENT_recv(const net_header_t* header, const void* payload) {
#ifdef C_SERVICE_PLACEMENT_EVALUATION
	PLACEMENT_command_t* command_array = NULL;
	size_t command_array_size = 0;

	C_ABORT_ON(header->app != APP_TYPE_PLACEMENT, "PLACEMENT: Received non-PLACEMENT packet.\n");

	switch(header->type) {
	case PLACEMENT_PACKET_TYPE_ADAPTATION_REQUEST:
		C_ABORT_ON(header->payload_length < sizeof(PLACEMENT_ADAPTATION_REQUEST_packet_t), "PLACEMENT: Received ADAPTATION REQUEST packet with wrong payload length (%d < %d).\n", header->payload_length, sizeof(PLACEMENT_ADAPTATION_REQUEST_packet_t));
		command_array = ((PLACEMENT_command_t*) (((char*) payload) + sizeof(PLACEMENT_ADAPTATION_REQUEST_packet_t)));
		command_array_size = header->payload_length - sizeof(PLACEMENT_ADAPTATION_REQUEST_packet_t);
		handle_PLACEMENT_ADAPTATION_REQUEST(header, (PLACEMENT_ADAPTATION_REQUEST_packet_t*) payload, command_array, command_array_size);
		break;
	case PLACEMENT_PACKET_TYPE_ADAPTATION_COMPLETE:
		C_ABORT_ON(header->payload_length != sizeof(PLACEMENT_ADAPTATION_COMPLETE_packet_t), "PLACEMENT: Received ADAPTATION COMPLETE packet with wrong payload length (%d != %d).\n", header->payload_length, sizeof(PLACEMENT_ADAPTATION_COMPLETE_packet_t));
		handle_PLACEMENT_ADAPTATION_COMPLETE(header, (PLACEMENT_ADAPTATION_COMPLETE_packet_t*) payload);
		break;
	default:
		C_ABORT_ON(true, "PLACEMENT: Received packet of unknown type 0x%X from %s.\n", header->type, addr_to_str(header->source_address));
	}
#else
	C_DEBUG(PLACEMENT_DEBUG_LEVEL, C_MSG_ERROR, "PLACEMENT: Received a placement packet, but framework was built without service placement support.\n");
#endif	// C_SERVICE_PLACEMENT_EVALUATION
}



/*
 * Interface provided by this module.
 */

/**
 * @brief Start the selected service placement algorithm.
 *
 * This function is called once when the service is started.
 *
 * @param service Pointer to service for which the placement algorithm is to be
 * started.
 * @param algorithm_id ID of service placement algorithm.
 */
void PLACEMENT_start_algorithm(SERVICE_service_t* service, uint8_t algorithm_id) {
#ifdef C_SERVICE_PLACEMENT_EVALUATION
	C_ABORT_ON(service == NULL, "PLACEMENT: Trying to start placement algorithm for NULL service.\n");
	C_ABORT_ON(service->placement_algorithm != NULL, "PLACEMENT: Trying to start placement algorithm for service \"%s\", but an algorithm is already active.\n", service->id);

	// Reset stats before starting the placement algorithm.
	C_PLACEMENT_current_service_cost = 0.0;
	C_PLACEMENT_currently_optimal_service_cost = 0.0;
	C_PLACEMENT_currently_optimal_number_of_hosts = 0;
	C_PLACEMENT_currently_estimated_adaptation_cost = 0.0;

	switch(algorithm_id) {
	case C_PLACEMENT_ALGORITHM_NONE:
		service->placement_algorithm = NULL;
		return;
	case C_PLACEMENT_ALGORITHM_SIMPLE_HEURISTIC:
		service->placement_algorithm = &simple_heuristic_algorithm;
		break;
	case C_PLACEMENT_ALGORITHM_GRAPH_COST_SINGLE_INSTANCE:
		service->placement_algorithm = &graph_cost_single_instance_algorithm;
		break;
	case C_PLACEMENT_ALGORITHM_GRAPH_COST_MULTIPLE_INSTANCES:
		service->placement_algorithm = &graph_cost_multiple_instances_algorithm;
		break;
	case C_PLACEMENT_ALGORITHM_LINKPULL_MIGRATION:
		service->placement_algorithm = &linkpull_migration_algorithm;
		break;
	case C_PLACEMENT_ALGORITHM_PEERPULL_MIGRATION:
		service->placement_algorithm = &peerpull_migration_algorithm;
		break;
	case C_PLACEMENT_ALGORITHM_TOPOCENTER_MIGRATION:
		service->placement_algorithm = &topocenter_migration_algorithm;
		break;
	case C_PLACEMENT_ALGORITHM_TREE_TOPOLOGY_MIGRATION:
		service->placement_algorithm = &tree_topology_migration_algorithm;
		break;
	case C_PLACEMENT_ALGORITHM_ASG_SIMPLE:
		service->placement_algorithm = &asg_simple_algorithm;
		break;
	case C_PLACEMENT_ALGORITHM_ASG_EVENT_FLOW_TREE:
		service->placement_algorithm = &asg_event_flow_tree_algorithm;
		break;
	default:
		C_ABORT_ON(true, "PLACEMENT: Invalid placement algorithm (0x%X) while starting service \"%s\".\n", service->placement_algorithm, service->id);
	}

	if(service->placement_algorithm->start) {
		C_DEBUG(PLACEMENT_DEBUG_LEVEL, C_MSG_DEBUG, "PLACEMENT: Calling start function for placement algorithm on service \"%s\"...\n", service->id);
		service->placement_algorithm->start(service);
	} else {
		C_DEBUG(PLACEMENT_DEBUG_LEVEL, C_MSG_DEBUG, "PLACEMENT: Trying to call start function for placement algorithm on service \"%s\", but no function given.\n", service->id);
	}
#else
	C_DEBUG(PLACEMENT_DEBUG_LEVEL, C_MSG_WARN, "PLACEMENT: Trying to start algorithm, but framework was built without service placement support.\n");
#endif	// C_SERVICE_PLACEMENT_EVALUATION
}

/**
 * @brief Stop the selected service placement algorithm.
 *
 * This function is called once when the service is stopped.
 *
 * @param service Pointer to service for which the placement algorithm is to be
 * stopped.
 */
void PLACEMENT_stop_algorithm(SERVICE_service_t* service) {
#ifdef C_SERVICE_PLACEMENT_EVALUATION
	C_ABORT_ON(service == NULL, "PLACEMENT: Trying to stop placement algorithm for NULL service.\n");

	if(service->placement_algorithm == NULL) {
		C_DEBUG(PLACEMENT_DEBUG_LEVEL, C_MSG_DEBUG, "PLACEMENT: Trying to stop placement algorithm for service \"%s\", but no algorithm is currently active.\n", service->id);
		return;
	}

	if(service->placement_algorithm->stop) {
		C_DEBUG(PLACEMENT_DEBUG_LEVEL, C_MSG_DEBUG, "PLACEMENT: Calling stop function for placement algorithm on service \"%s\"...\n", service->id);
		service->placement_algorithm->stop(service);
	} else {
		C_DEBUG(PLACEMENT_DEBUG_LEVEL, C_MSG_DEBUG, "PLACEMENT: Trying to call stop function for placement algorithm on service \"%s\", but no function given.\n", service->id);
	}
	service->placement_algorithm = NULL;
#else
	C_DEBUG(PLACEMENT_DEBUG_LEVEL, C_MSG_WARN, "PLACEMENT: Trying to stop algorithm, but framework was built without service placement support.\n");
#endif	// C_SERVICE_PLACEMENT_EVALUATION
}

/**
 * @brief Trigger the selected service placement algorithm.
 *
 * This function is called every time the service processes a request.
 *
 * @param service Pointer to service for which the placement algorithm is to be
 * triggered.
 * @param reason_id Numeric ID for the reason why the algorithm was triggered.
 * @param reason_parameter Pointer to optional, reason-specific parameter of
 * the reason.
 */
void PLACEMENT_trigger_algorithm(SERVICE_service_t* service, uint8_t reason_id, const void* reason_parameter) {
#ifdef C_SERVICE_PLACEMENT_EVALUATION
	C_ABORT_ON(service == NULL, "PLACEMENT: Trying to trigger placement algorithm for NULL service.\n");

	if(service->placement_algorithm == NULL) {
		C_DEBUG(PLACEMENT_DEBUG_LEVEL, C_MSG_VERBOSE, "PLACEMENT: Trying to trigger placement algorithm for service \"%s\", but no algorithm is currently active.\n", service->id);
		return;
	}

	if(service->placement_algorithm->trigger) {
		C_DEBUG(PLACEMENT_DEBUG_LEVEL, C_MSG_VERBOSE, "PLACEMENT: Calling trigger function for placement algorithm on service \"%s\"...\n", service->id);
		service->placement_algorithm->trigger(service, reason_id, reason_parameter);
	} else {
		C_DEBUG(PLACEMENT_DEBUG_LEVEL, C_MSG_VERBOSE, "PLACEMENT: Trying to call trigger function for placement algorithm on service \"%s\", but no function given.\n", service->id);
	}
#else
	C_DEBUG(PLACEMENT_DEBUG_LEVEL, C_MSG_WARN, "PLACEMENT: Trying to trigger algorithm, but framework was built without service placement support.\n");
#endif	// C_SERVICE_PLACEMENT_EVALUATION
}

/**
 * @brief Get data from the selected service placement algorithm.
 *
 * @param service Pointer to service for which to get the algorithm data.
 * @param data Pointer to pointer to data to return.
 * @return Size of the data in bytes.
 */
size_t PLACEMENT_get_algorithm_data(const SERVICE_service_t* service, char** data) {
#ifdef C_SERVICE_PLACEMENT_EVALUATION
	C_ABORT_ON(service == NULL, "PLACEMENT: Trying to get placement algorithm data for NULL service.\n");

	if(service->placement_algorithm == NULL) {
		C_DEBUG(PLACEMENT_DEBUG_LEVEL, C_MSG_DEBUG, "PLACEMENT: Trying to get data from placement algorithm for service \"%s\", but no algorithm is currently active.\n", service->id);
		return 0;
	}

	if(service->placement_algorithm->get_data) {
		C_DEBUG(PLACEMENT_DEBUG_LEVEL, C_MSG_DEBUG, "PLACEMENT: Getting data from placement algorithm on service \"%s\"...\n", service->id);
		return service->placement_algorithm->get_data(service, data);
	} else {
		C_DEBUG(PLACEMENT_DEBUG_LEVEL, C_MSG_DEBUG, "PLACEMENT: Trying to get data from placement algorithm on service \"%s\", but no function given.\n", service->id);
		return 0;
	}
#else
	C_DEBUG(PLACEMENT_DEBUG_LEVEL, C_MSG_WARN, "PLACEMENT: Trying to get algorithm data, but framework was built without service placement support.\n");
	return 0;
#endif	// C_SERVICE_PLACEMENT_EVALUATION
}

/**
 * @brief Set data of the selected service placement algorithm.
 *
 * @param service Pointer to service for which to set the algorithm data.
 * @param data Pointer to data to set.
 * @param data_size Size of the data in bytes.
 */
void PLACEMENT_set_algorithm_data(SERVICE_service_t* service, const char* data, size_t data_size) {
#ifdef C_SERVICE_PLACEMENT_EVALUATION
	C_ABORT_ON(service == NULL, "PLACEMENT: Trying to set placement algorithm data for NULL service.\n");

	if(service->placement_algorithm == NULL) {
		C_DEBUG(PLACEMENT_DEBUG_LEVEL, C_MSG_DEBUG, "PLACEMENT: Trying to set data from placement algorithm for service \"%s\", but no algorithm is currently active.\n", service->id);
		return;
	}

	if(service->placement_algorithm->set_data) {
		C_DEBUG(PLACEMENT_DEBUG_LEVEL, C_MSG_DEBUG, "PLACEMENT: Setting data of placement algorithm on service \"%s\"...\n", service->id);
		service->placement_algorithm->set_data(service, data, data_size);
	} else {
		C_DEBUG(PLACEMENT_DEBUG_LEVEL, C_MSG_DEBUG, "PLACEMENT: Trying to set data of placement algorithm on service \"%s\", but no function given.\n", service->id);
	}
#else
	C_DEBUG(PLACEMENT_DEBUG_LEVEL, C_MSG_WARN, "PLACEMENT: Trying to set algorithm data, but framework was built without service placement support.\n");
#endif	// C_SERVICE_PLACEMENT_EVALUATION
}

/**
 * @brief Update data of the selected service placement algorithm.
 *
 * @param service Pointer to service for which to update the algorithm data.
 * @param adaptation Pointer to adaptation plan to use for the update.
 * @return <code>true</code> if the adaptation data is valid,
 * <code>false</code> otherwise.
 */
bool PLACEMENT_update_algorithm_data(SERVICE_service_t* service, const PLACEMENT_incoming_adaptation_t* adaptation) {
#ifdef C_SERVICE_PLACEMENT_EVALUATION
	C_ABORT_ON(service == NULL, "PLACEMENT: Trying to update placement algorithm data for NULL service.\n");

	if(service->placement_algorithm == NULL) {
		C_DEBUG(PLACEMENT_DEBUG_LEVEL, C_MSG_DEBUG, "PLACEMENT: Trying to update data of placement algorithm for service \"%s\", but no algorithm is currently active.\n", service->id);
		return true;
	}

	if(service->placement_algorithm->update_data) {
		C_DEBUG(PLACEMENT_DEBUG_LEVEL, C_MSG_DEBUG, "PLACEMENT: Updating data of placement algorithm on service \"%s\"...\n", service->id);
		return service->placement_algorithm->update_data(service, adaptation);
	} else {
		C_DEBUG(PLACEMENT_DEBUG_LEVEL, C_MSG_DEBUG, "PLACEMENT: Trying to update data of placement algorithm on service \"%s\", but no function given.\n", service->id);
		return true;
	}
#else
	C_DEBUG(PLACEMENT_DEBUG_LEVEL, C_MSG_WARN, "PLACEMENT: Trying to update algorithm data, but framework was built without service placement support.\n");
	return false;
#endif	// C_SERVICE_PLACEMENT_EVALUATION
}

/**
 * @brief Report known service instance to selected service placement algorithm.
 *
 * @param service Pointer to service for which report a known service instance.
 * @param host_address Address of the host of the known service instance.
 */
void PLACEMENT_report_known_service_instance(SERVICE_service_t* service, C_addr_t host_address) {
#ifdef C_SERVICE_PLACEMENT_EVALUATION
	C_ABORT_ON(service == NULL, "PLACEMENT: Trying to report known service instance for NULL service.\n");
	if(service->placement_algorithm == NULL) {
		C_DEBUG(PLACEMENT_DEBUG_LEVEL, C_MSG_VERBOSE, "PLACEMENT: Trying to report known service instance of service \"%s\", but no algorithm is currently active.\n", service->id);
		return;
	}

	if(service->placement_algorithm->report_known_service_instance) {
		C_DEBUG(PLACEMENT_DEBUG_LEVEL, C_MSG_VERBOSE, "PLACEMENT: Reporting known service instance of service \"%s\" at %s...\n", service->id, addr_to_str(host_address));
		service->placement_algorithm->report_known_service_instance(service, host_address);
	} else {
		C_DEBUG(PLACEMENT_DEBUG_LEVEL, C_MSG_VERBOSE, "PLACEMENT: Trying to report known service instance of service \"%s\" at %s, but no function given.\n", service->id, addr_to_str(host_address));
	}
#else
	C_DEBUG(PLACEMENT_DEBUG_LEVEL, C_MSG_WARN, "PLACEMENT: Trying to report known service instance, but framework was built without service placement support.\n");
#endif	// C_SERVICE_PLACEMENT_EVALUATION
}

/**
 * @brief Report invalid service instance to selected service placement algorithm.
 *
 * @param service Pointer to service for which report a known service instance.
 * @param host_address Address of the host of the invalid service instance.
 */
void PLACEMENT_report_invalid_service_instance(SERVICE_service_t* service, C_addr_t host_address) {
#ifdef C_SERVICE_PLACEMENT_EVALUATION
	C_ABORT_ON(service == NULL, "PLACEMENT: Trying to report invalid service instance for NULL service.\n");
	if(service->placement_algorithm == NULL) {
		C_DEBUG(PLACEMENT_DEBUG_LEVEL, C_MSG_VERBOSE, "PLACEMENT: Trying to report invalid service instance of service \"%s\", but no algorithm is currently active.\n", service->id);
		return;
	}

	if(service->placement_algorithm->report_invalid_service_instance) {
		C_DEBUG(PLACEMENT_DEBUG_LEVEL, C_MSG_VERBOSE, "PLACEMENT: Reporting invalid service instance of service \"%s\" at %s...\n", service->id, addr_to_str(host_address));
		service->placement_algorithm->report_invalid_service_instance(service, host_address);
	} else {
		C_DEBUG(PLACEMENT_DEBUG_LEVEL, C_MSG_VERBOSE, "PLACEMENT: Trying to report invalid service instance of service \"%s\" at %s, but no function given.\n", service->id, addr_to_str(host_address));
	}
#else
	C_DEBUG(PLACEMENT_DEBUG_LEVEL, C_MSG_WARN, "PLACEMENT: Trying to report invalid service instance, but framework was built without service placement support.\n");
#endif	// C_SERVICE_PLACEMENT_EVALUATION
}

/**
 * @brief Get synchronization hosts from selected service placement algorithm.
 *
 * @param service Pointer to service for which to get the synchronization hosts.
 * @param inter_host_update Flag whether this update is an inter-host update.
 * @return Pointer to a list of nodes to which synchronize changes to (must be
 * freed by the caller), or NULL if no synchronization is necessary.
 */
list_t* PLACEMENT_get_synchronization_hosts(const SERVICE_service_t* service, bool inter_host_update) {
#ifdef C_SERVICE_PLACEMENT_EVALUATION
	C_ABORT_ON(service == NULL, "PLACEMENT: Trying to get synchronization hosts for NULL service.\n");

	if(service->placement_algorithm == NULL) {
		C_DEBUG(PLACEMENT_DEBUG_LEVEL, C_MSG_VERBOSE, "PLACEMENT: Trying to get synchronization hosts for service \"%s\", but no algorithm is currently active.\n", service->id);
		return NULL;
	}

	if(service->placement_algorithm->get_synchronization_hosts) {
		C_DEBUG(PLACEMENT_DEBUG_LEVEL, C_MSG_VERBOSE, "PLACEMENT: Getting synchronization hosts for service \"%s\"...\n", service->id);
		return service->placement_algorithm->get_synchronization_hosts(service, inter_host_update);
	} else {
		C_DEBUG(PLACEMENT_DEBUG_LEVEL, C_MSG_VERBOSE, "PLACEMENT: Trying to get synchronization hosts for service \"%s\", but no function given.\n", service->id);
		return NULL;
	}
#else
	C_DEBUG(PLACEMENT_DEBUG_LEVEL, C_MSG_WARN, "PLACEMENT: Trying to get synchronization hosts, but framework was built without service placement support.\n");
	return NULL;
#endif	// C_SERVICE_PLACEMENT_EVALUATION
}

/**
 * @brief Distribute an adaptation plan.
 *
 * @param service_id ID of the service whose configuration is to be adapted.
 * @param commands Pointer to list of adaptation commands.
 * @param destination_nodes Pointer to list of destination node addresses.
 * @param callback Callback function.
 * @param result Pointer to replication result.
 */
void PLACEMENT_distribute_adaptation(const SERVICE_id_t* service_id, const list_t* commands, const list_t* destination_nodes, callback_function_t callback, PLACEMENT_adaptation_result_t* result) {
#ifdef C_SERVICE_PLACEMENT_EVALUATION
	PLACEMENT_outgoing_adaptation_t* adaptation = NULL;
	C_addr_t coordinator_address = C_BROADCAST_ADDRESS;
	C_addr_t coordinator_handler_address = C_BROADCAST_ADDRESS;
	SERVICE_id_t service_id_tmp[SERVICE_ID_LENGTH];
	list_t* destination_nodes_tmp = NULL;
	C_addr_t* destination_node_address = NULL;
	list_iterator_t destination_nodes_iterator;
	C_addr_t* destination_node_address_tmp = NULL;
	bool callback_registered = false;

#ifdef C_ENABLE_CONSISTENCY_CHECKS
	SERVICE_check_service_id(service_id);
#endif	// C_ENABLE_CONSISTENCY_CHECKS
	C_ABORT_ON(commands == NULL, "PLACEMENT: Trying to distribute adaptation plan with NULL commands.\n");
	C_ABORT_ON(destination_nodes == NULL, "PLACEMENT: Trying to distribute adaptation plan to NULL destination nodes.\n");
	C_ABORT_ON(((callback == NULL) && (result != NULL)) || ((callback != NULL) && (result == NULL)), "PLACEMENT: Trying to distribute adaptation plan, but callback parameters are inconsistent (%p, %p).\n", callback, result);

	if(C_PLACEMENT_currently_outgoing_adaptation != NULL) {
		C_DEBUG(PLACEMENT_DEBUG_LEVEL, C_MSG_WARN, "PLACEMENT: Trying to distribute adaptation plan while another adaptation is still in progress (%p).\n", C_PLACEMENT_currently_outgoing_adaptation);
		result->error_code = C_ERROR_BUSY;
		callback(result);
		return;
	}

	coordinator_address = get_coordinator_from_commands(commands);
	if(coordinator_address == C_BROADCAST_ADDRESS) {
		C_DEBUG(PLACEMENT_DEBUG_LEVEL, C_MSG_WARN, "PLACEMENT: Adaptation plan for service \"%s\" does not specify a coordinator node.\n", service_id);
		result->error_code = C_ERROR_PARAMETER;
		callback(result);
		return;
	}

	C_DEBUG(PLACEMENT_DEBUG_LEVEL, C_MSG_NOTE, "PLACEMENT: Distributing adaptation plan for service \"%s\" ", service_id);
	list_print(PLACEMENT_DEBUG_LEVEL, (C_MSG_NOTE | C_MSG_CONT), commands, PLACEMENT_print_command);
	C_DEBUG(PLACEMENT_DEBUG_LEVEL, (C_MSG_NOTE | C_MSG_CONT), " to node(s) ");
	list_print(PLACEMENT_DEBUG_LEVEL, (C_MSG_NOTE | C_MSG_CONT), destination_nodes, print_node_address);
	C_DEBUG(PLACEMENT_DEBUG_LEVEL, (C_MSG_NOTE | C_MSG_CONT), ".\n");
	C_PLACEMENT_total_adaptations++;

	// Reset coordinator-specific statistics variables, if the coordinator
	// is about to change.
	if(coordinator_address != C_LOCAL_ADDRESS) {
		C_PLACEMENT_current_service_cost = 0.0;
		C_PLACEMENT_currently_optimal_service_cost = 0.0;
		C_PLACEMENT_currently_optimal_number_of_hosts = 0;
		C_PLACEMENT_currently_estimated_adaptation_cost = 0.0;
	}

	// Establish which node (if any) is responsible for the new coordinator
	// (if any).
	coordinator_handler_address = get_coordinator_handler_from_commands(commands);
	C_ABORT_ON(coordinator_handler_address == C_BROADCAST_ADDRESS, "PLACEMENT: Failed to extract coordinator handler from adaptation plan for service \"%s\".\n", service_id);

	// If the adaptation plan affects the local node, the service ID and
	// the list of destination nodes (which are both merely a pointer to
	// the current configuration) will change. Hence, we need to make a
	// temporary copy.
	SERVICE_copy_service_id(service_id_tmp, service_id);

	// NOTE: Make sure that the handler for the new coordinator is the
	// first node in the adaptation process, as otherwise the error
	// handling doesn't work.
	C_DEBUG(PLACEMENT_DEBUG_LEVEL, C_MSG_DEBUG, "PLACEMENT: Starting adaptation with %s, which is the handler for the new coordinator.\n", addr_to_str(coordinator_handler_address));
	list_alloc(&destination_nodes_tmp);
	MALLOC(destination_node_address_tmp, C_addr_t);
	*destination_node_address_tmp = coordinator_handler_address;
	list_add(destination_nodes_tmp, destination_node_address_tmp);
	destination_node_address_tmp = NULL;
	list_iterator_init(destination_nodes, &destination_nodes_iterator);
	while((destination_node_address = (C_addr_t*) list_iterator_next(&destination_nodes_iterator)) != NULL) {
		if(*destination_node_address != coordinator_handler_address) {
			MALLOC(destination_node_address_tmp, C_addr_t);
			*destination_node_address_tmp = *destination_node_address;
			list_add(destination_nodes_tmp, destination_node_address_tmp);
			destination_node_address_tmp = NULL;
		}
	}

	// Implement the adaptation plan for all destination nodes.
	list_iterator_init(destination_nodes_tmp, &destination_nodes_iterator);
	while((destination_node_address = (C_addr_t*) list_iterator_next(&destination_nodes_iterator)) != NULL) {

		C_DEBUG(PLACEMENT_DEBUG_LEVEL, C_MSG_DEBUG, "PLACEMENT: Preparing adaptation of service \"%s\" on %s...\n", service_id_tmp, addr_to_str(*destination_node_address));

		MALLOC(adaptation, PLACEMENT_outgoing_adaptation_t);
		SERVICE_copy_service_id(adaptation->service_id, service_id_tmp);
		adaptation->coordinator_address = coordinator_address;
		adaptation->coordinator_handler_address = coordinator_handler_address;
		adaptation->timeout_callback_id = 0;

		if(*destination_node_address == coordinator_handler_address) {

			C_ABORT_ON(callback_registered, "PLACEMENT: Trying to register callback twice while distributing adaptation plan.\n");
			callback_registered = true;

			C_PLACEMENT_currently_outgoing_adaptation = adaptation;
			adaptation->callback = callback;
			adaptation->result = result;
			if(adaptation->result != NULL)
				SERVICE_copy_service_id(adaptation->result->service_id, service_id_tmp);
			send_PLACEMENT_ADAPTATION_REQUEST(*destination_node_address, adaptation, commands, true);
			adaptation = NULL;
			// The the adaptation structure will be freed later on.

		} else {

			adaptation->callback = NULL;
			adaptation->result = NULL;
			send_PLACEMENT_ADAPTATION_REQUEST(*destination_node_address, adaptation, commands, false);
			FREE(adaptation);

		}
	}

	list_free_including_data(&destination_nodes_tmp, free);

	if((callback != NULL) && !callback_registered) {
		C_DEBUG(PLACEMENT_DEBUG_LEVEL, C_MSG_WARN, "PLACEMENT: Caller tried to register callback at %p, but no callback was registered given the commands.\n", callback);
		result->error_code = C_ERROR_PARAMETER;
		callback(result);
		C_PLACEMENT_currently_outgoing_adaptation = NULL;
	}
#else
	C_DEBUG(PLACEMENT_DEBUG_LEVEL, C_MSG_ERROR, "PLACEMENT: Trying to start adaptation, but framework was built without service placement support.\n");
#endif	// C_SERVICE_PLACEMENT_EVALUATION
}

/**
 * @brief Update a the coordinator of a remote service host.
 *
 * @param service_id ID of the service which is to be updated.
 * @param destination_node_address Address of the destination node.
 * @param coordinator_address Address of the new coordinator node.
 */
void PLACEMENT_update_remote_coordinator(const SERVICE_id_t* service_id, C_addr_t destination_node_address, C_addr_t coordinator_address) {
#ifdef C_SERVICE_PLACEMENT_EVALUATION
	PLACEMENT_outgoing_adaptation_t* set_coordinator_adaptation = NULL;
	list_t* set_coordinator_commands = NULL;
	PLACEMENT_command_t set_coordinator_command = {destination_node_address, PLACEMENT_COMMAND_COORDINATOR, coordinator_address};

#ifdef C_ENABLE_CONSISTENCY_CHECKS
	SERVICE_check_service_id(service_id);
#endif	// C_ENABLE_CONSISTENCY_CHECKS
	C_ABORT_ON(destination_node_address == C_BROADCAST_ADDRESS, "PLACEMENT: Trying to set coordinator of service \"%s\", but request is addressed to the broadcast address.\n", service_id);

	C_DEBUG(PLACEMENT_DEBUG_LEVEL, C_MSG_NOTE, "PLACEMENT: Setting coordinator of service \"%s\" on %s to %s...\n", service_id, addr_to_str(destination_node_address), addr_to_str(coordinator_address));

	MALLOC(set_coordinator_adaptation, PLACEMENT_outgoing_adaptation_t);
	SERVICE_copy_service_id(set_coordinator_adaptation->service_id, service_id);
	list_alloc(&set_coordinator_commands);
	list_add(set_coordinator_commands, &set_coordinator_command);
	set_coordinator_adaptation->coordinator_address = C_BROADCAST_ADDRESS;
	set_coordinator_adaptation->coordinator_handler_address = C_BROADCAST_ADDRESS;
	set_coordinator_adaptation->timeout_callback_id = 0;
	set_coordinator_adaptation->callback = NULL;
	set_coordinator_adaptation->result = NULL;
	send_PLACEMENT_ADAPTATION_REQUEST(destination_node_address, set_coordinator_adaptation, set_coordinator_commands, false);
	list_free(&set_coordinator_commands);
	FREE(set_coordinator_adaptation);
#else
	C_DEBUG(PLACEMENT_DEBUG_LEVEL, C_MSG_ERROR, "PLACEMENT: Trying to set coordinator on remote service host, but framework was built without service placement support.\n");
#endif	// C_SERVICE_PLACEMENT_EVALUATION
}

/**
 * @brief Shutdown a remote service host.
 *
 * @param service_id ID of the service which is to be shut down.
 * @param destination_node_address Address of the destination node.
 */
void PLACEMENT_shutdown_remote_service_host(const SERVICE_id_t* service_id, C_addr_t destination_node_address) {
#ifdef C_SERVICE_PLACEMENT_EVALUATION
	PLACEMENT_outgoing_adaptation_t* shutdown_adaptation = NULL;
	list_t* shutdown_commands = NULL;
	PLACEMENT_command_t shutdown_command = {destination_node_address, PLACEMENT_COMMAND_SHUTDOWN, C_BROADCAST_ADDRESS};

#ifdef C_ENABLE_CONSISTENCY_CHECKS
	SERVICE_check_service_id(service_id);
#endif	// C_ENABLE_CONSISTENCY_CHECKS
	C_ABORT_ON(destination_node_address == C_BROADCAST_ADDRESS, "PLACEMENT: Trying to shutdown service \"%s\", but request is addressed to the broadcast address.\n", service_id);

	C_DEBUG(PLACEMENT_DEBUG_LEVEL, C_MSG_NOTE, "PLACEMENT: Shutting down service \"%s\" on %s...\n", service_id, addr_to_str(destination_node_address));

	MALLOC(shutdown_adaptation, PLACEMENT_outgoing_adaptation_t);
	SERVICE_copy_service_id(shutdown_adaptation->service_id, service_id);
	list_alloc(&shutdown_commands);
	list_add(shutdown_commands, &shutdown_command);
	shutdown_adaptation->coordinator_address = C_BROADCAST_ADDRESS;
	shutdown_adaptation->coordinator_handler_address = C_BROADCAST_ADDRESS;
	shutdown_adaptation->timeout_callback_id = 0;
	shutdown_adaptation->callback = NULL;
	shutdown_adaptation->result = NULL;
	send_PLACEMENT_ADAPTATION_REQUEST(destination_node_address, shutdown_adaptation, shutdown_commands, false);
	list_free(&shutdown_commands);
	FREE(shutdown_adaptation);
#else
	C_DEBUG(PLACEMENT_DEBUG_LEVEL, C_MSG_ERROR, "PLACEMENT: Trying to shutdown remote service host, but framework was built without service placement support.\n");
#endif	// C_SERVICE_PLACEMENT_EVALUATION
}

/**
 * @brief Print a placement command.
 *
 * @param debug_level Numeric debug level of the message.
 * @param msg_type Numeric type of the message.
 * @param item Pointer to placement command to print.
 */
void PLACEMENT_print_command(uint8_t debug_level, uint8_t msg_type, const void* item) {
	PLACEMENT_command_t* command = (PLACEMENT_command_t*) item;

	switch(command->id) {
	case PLACEMENT_COMMAND_NOP:
		C_DEBUG(debug_level, msg_type, "NOP: %s", addr_to_str(command->handler_address));
		break;
	case PLACEMENT_COMMAND_COORDINATOR:
		C_DEBUG(debug_level, msg_type, "CRD: %s", addr_to_str(command->destination_address));
		break;
	case PLACEMENT_COMMAND_REPLICATE:
		C_DEBUG(debug_level, msg_type, "REP: %s -> %s", addr_to_str(command->handler_address), addr_to_str(command->destination_address));
		break;
	case PLACEMENT_COMMAND_MIGRATE:
		C_DEBUG(debug_level, msg_type, "MIG: %s -> %s", addr_to_str(command->handler_address), addr_to_str(command->destination_address));
		break;
	case PLACEMENT_COMMAND_SHUTDOWN:
		C_DEBUG(debug_level, msg_type, "STD: %s", addr_to_str(command->handler_address));
		break;
	case PLACEMENT_COMMAND_DELAYED_SHUTDOWN:
		C_DEBUG(debug_level, msg_type, "STD: %s (%.1fs)", addr_to_str(command->handler_address), command->delay);
		break;
	default:
		C_ABORT_ON(true, "PLACEMENT: Invalid placement command (0x%X).\n", command->id);
	}
}



/*
 * Interface provided by this module for direct commands.
 */

/**
 * @brief Placement command function.
 *
 * @param argc Number of command parameters.
 * @param argv Array of strings used as command parameters.
 * @return <code>true</code> if the command was executed successfully,
 * <code>false</code> otherwise.
 */
bool PLACEMENT_command(int argc, const char*const* argv) {
	return false;
}
