/**
 * @file topocenter_migration.c
 *
 * @brief TopoCenter(1) migration algorithm.
 *
 * This is an implementation of the TopoCenter(1) service placement algorithm
 * proposed in:
 * Hongzhou Liu, Tom Roeder, Kevin Walsh, Rimon Barr, and Emin Guen Sirer.
 * Design and Implementation of a Single System Image Operating System for Ad
 * Hoc Networks. In Proceedings of the Third International Conference on Mobile
 * Systems, Applications, and Services (MobiSys '05), Seattle, Washington, USA,
 * June 2005.
 *
 * At the end of an epoch, this algorithm computes a partial network view based
 * on information from the routing protocol, and migrates the service to the
 * node that minimizes the sum of migration cost (proportional to the shortest
 * path) and estimated future communication cost (based on the packet
 * statistics of the last epoch).
 *
 * Copyright (c) 2007, 2008, 2009, 2010 Georg Wittenburg, Freie Universitaet
 * Berlin (wittenbu@inf.fu-berlin.de). All rights reserved.
 * Please refer to the COPYING file for detailed licensing information.
 *
 * @author Georg Wittenburg
 *
 * $Id: topocenter_migration.c 3408 2010-03-15 18:52:15Z gwitten $
 */

#include <float.h>

#include "../../config.h"
#include "../../macros.h"
#include "../../types.h"
#include "../../vars.h"
#include "../../main.h"
#include "../../util/debug.h"
#include "../../util/list.h"
#include "../../util/callback.h"
#include "../../util/mapping.h"
#include "../../util/net.h"
#include "../../service/service.h"
#include "../placement.h"
#include "../placement_types.h"
#include "../placement_vars.h"
#include "../placement_types.h"
#include "topocenter_migration.h"



#ifdef C_SERVICE_PLACEMENT_EVALUATION



/**
 * @brief Interface to TopoCenter(1) algorithm.
 */
PLACEMENT_algorithm_t topocenter_migration_algorithm = {
	C_PLACEMENT_ALGORITHM_TOPOCENTER_MIGRATION,
	topocenter_migration_start,
	topocenter_migration_stop,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL
};

/**
 * @brief TopoCenter(1) algorithm data.
 */
typedef struct {
	SERVICE_id_t*	service_id;	///< Service ID.
	callback_id_t	callback_id;	///< ID of the periodic callback.
} topocenter_migration_data_t;

/**
 * @brief Epoch of the TopoCenter(1) algorithm in seconds.
 */
#define PLACEMENT_TOPOCENTER_EPOCH	60



static void topocenter_migration_run(callback_parameter_t parameter);

/**
 * @brief Service migration callback.
 *
 * @param parameter Migration result.
 */
static void topocenter_migration_callback(callback_parameter_t parameter) {
	SERVICE_migration_result_t* result = (SERVICE_migration_result_t*) parameter;
	SERVICE_id_t* service_id = (SERVICE_id_t*) result->parameter;
	SERVICE_service_t* service = SERVICE_get_service(service_id);
	topocenter_migration_data_t* placement_data = NULL;

	if(result->error_code != C_ERROR_OK) {
		C_ABORT_ON(service == NULL, "PLACEMENT: Trying to call migration callback of TopoCenter(1) algorithm for service \"%s\", but service is NULL.\n", service_id);
		if(service->placement_data == NULL) {
			C_DEBUG(PLACEMENT_DEBUG_LEVEL, C_MSG_WARN, "PLACEMENT: Migration of service \"%s\" failed (%d), but TopoCenter(1) algorithm is not active anymore.\n", service_id, result->error_code);
			goto free_parameters;
		}
		placement_data = (topocenter_migration_data_t*) service->placement_data;
	}

	switch(result->error_code) {
	case C_ERROR_OK:
		C_DEBUG(PLACEMENT_DEBUG_LEVEL, C_MSG_NOTE, "PLACEMENT: Migration of service \"%s\" was successful.\n", service_id);
		// There is no need to unlock the service since no instance
		// remains on this node.
		break;
	case C_ERROR_PARAMETER:
		C_DEBUG(PLACEMENT_DEBUG_LEVEL, C_MSG_WARN, "PLACEMENT: Migration of service \"%s\" failed because the service was found running on the destination node. Shutting down service on this node...\n", service_id);
		SERVICE_unlock_service(service_id);
		SERVICE_shutdown_service(service_id);
		break;
	case C_ERROR_TIMEOUT:
		C_DEBUG(PLACEMENT_DEBUG_LEVEL, C_MSG_WARN, "PLACEMENT: Migration of service \"%s\" timed out.\n", service_id);
		SERVICE_unlock_service(service_id);
		placement_data->callback_id = callback_add(topocenter_migration_run, service, PLACEMENT_TOPOCENTER_EPOCH);
		break;
	case C_ERROR_ABORTED:
		C_DEBUG(PLACEMENT_DEBUG_LEVEL, C_MSG_WARN, "PLACEMENT: Migration of service \"%s\" has been aborted.\n", service_id);
		SERVICE_unlock_service(service_id);
		// Don't re-add the callback if aborted by the user.
		break;
	case C_ERROR_BUSY:
		C_DEBUG(PLACEMENT_DEBUG_LEVEL, C_MSG_ERROR, "PLACEMENT: Migration of service \"%s\" failed because the local system is busy.\n", service_id);
		SERVICE_unlock_service(service_id);
		placement_data->callback_id = callback_add(topocenter_migration_run, service, PLACEMENT_TOPOCENTER_EPOCH);
		break;
	case C_ERROR_IN_PROGRESS:
		C_DEBUG(PLACEMENT_DEBUG_LEVEL, C_MSG_ERROR, "PLACEMENT: Migration of service \"%s\" is already in progress.\n", service_id);
		SERVICE_unlock_service(service_id);
		placement_data->callback_id = callback_add(topocenter_migration_run, service, PLACEMENT_TOPOCENTER_EPOCH);
		break;
	case C_ERROR_COMM_ERROR:
		C_DEBUG(PLACEMENT_DEBUG_LEVEL, C_MSG_WARN, "PLACEMENT: Migration of service \"%s\" failed due to a communication error.\n", service_id);
		SERVICE_unlock_service(service_id);
		placement_data->callback_id = callback_add(topocenter_migration_run, service, PLACEMENT_TOPOCENTER_EPOCH);
		break;
	case C_ERROR_REMOTE_BUSY:
		C_DEBUG(PLACEMENT_DEBUG_LEVEL, C_MSG_ERROR, "PLACEMENT: Migration of service \"%s\" failed because the remote system refused to host the service.\n", service_id);
		SERVICE_unlock_service(service_id);
		placement_data->callback_id = callback_add(topocenter_migration_run, service, PLACEMENT_TOPOCENTER_EPOCH);
		break;
	default:
		C_ABORT_ON(true, "PLACEMENT: Migration callback with unknown error code 0x%X.\n", result->error_code);
	}

free_parameters:
	FREE(service_id);
	FREE(result);
}

/**
 * @brief TopoCenter(1) migration algorithm.
 *
 * @param parameter Pointer to service for which to run this algorithm.
 */
static void topocenter_migration_run(callback_parameter_t parameter) {
	SERVICE_id_t* service_id = (SERVICE_id_t*) parameter;
	SERVICE_service_t* service = SERVICE_get_running_service(service_id);
	topocenter_migration_data_t* placement_data = NULL;
	SERVICE_node_stat_t* node_stat = NULL;
	list_iterator_t node_stats_iterator;
	float lowest_migration_and_communication_cost = FLT_MAX;
	node_t* node_with_lowest_migration_and_communication_cost = NULL;

	if(service == NULL) {
		C_DEBUG(PLACEMENT_DEBUG_LEVEL, C_MSG_ERROR, "PLACEMENT: Trying to run TopoCenter(1) migration algorithm for service \"%s\", but service is not running.\n", service->id);
		return;
	}

	C_ABORT_ON(service->placement_data == NULL, "PLACEMENT: Trying to run TopoCenter(1) algorithm for service \"%s\", but placement data is NULL.\n", service->id);
	placement_data = (topocenter_migration_data_t*) service->placement_data;
	placement_data->callback_id = 0;

	if(service->host_stat->request_count == 0) {
		C_DEBUG(PLACEMENT_DEBUG_LEVEL, C_MSG_WARN, "PLACEMENT: Trying to run TopoCenter(1) migration algorithm for service \"%s\", but no service requests have been received.\n", service->id);
		placement_data->callback_id = callback_add(topocenter_migration_run, service, PLACEMENT_TOPOCENTER_EPOCH);
		return;
	}

	// Look for node that minimizes the sum of migration cost and estimated
	// future communication cost.
	list_iterator_init(service->node_stats, &node_stats_iterator);
	while((node_stat = (SERVICE_node_stat_t*) list_iterator_next(&node_stats_iterator)) != NULL) {
		float migration_cost = (((float) service->data_size) * get_distance_between_nodes(C_LOCAL_ADDRESS, node_stat->node->address));
		float communication_cost = 0.0;
		SERVICE_node_stat_t* node_stat2 = NULL;
		list_iterator_t node_stats_iterator2;
		list_iterator_init(service->node_stats, &node_stats_iterator2);
		while((node_stat2 = (SERVICE_node_stat_t*) list_iterator_next(&node_stats_iterator2)) != NULL)
			communication_cost += (((float) (node_stat2->traffic)) * get_distance_between_nodes(node_stat2->node->address, node_stat->node->address));
		if((migration_cost + communication_cost) < lowest_migration_and_communication_cost) {
			lowest_migration_and_communication_cost = (migration_cost + communication_cost);
			node_with_lowest_migration_and_communication_cost = node_stat->node;
		}
	}
	C_ABORT_ON(node_with_lowest_migration_and_communication_cost == NULL, "PLACEMENT: %d service request(s) for service \"%s\" but no node with lowest cost.\n", service->host_stat->request_count, service->id);

	if(node_with_lowest_migration_and_communication_cost->address == C_LOCAL_ADDRESS) {
		C_DEBUG(PLACEMENT_DEBUG_LEVEL, C_MSG_NOTE, "PLACEMENT: Service \"%s\" is placed optimally on the current node (%.1f cost after %d total requests).\n", service->id, lowest_migration_and_communication_cost, service->host_stat->request_count);
		placement_data->callback_id = callback_add(topocenter_migration_run, service, PLACEMENT_TOPOCENTER_EPOCH);
	} else {
		SERVICE_migration_result_t* result = NULL;

		C_DEBUG(PLACEMENT_DEBUG_LEVEL, C_MSG_NOTE, "PLACEMENT: Migrating service \"%s\" to %s (%.1f cost after %d total requests)...\n", service->id, addr_to_str(node_with_lowest_migration_and_communication_cost->address), lowest_migration_and_communication_cost, service->host_stat->request_count);

		SERVICE_lock_service(service->id);
		MALLOC(result, SERVICE_migration_result_t);
		SERVICE_alloc_service_id((SERVICE_id_t**) &result->parameter, service->id);
		SERVICE_migrate_service(service->id, node_with_lowest_migration_and_communication_cost->address, topocenter_migration_callback, result);
	}
}

/**
 * @brief Start TopoCenter(1) migration algorithm.
 *
 * @param service Pointer to service for which to start this algorithm.
 */
void topocenter_migration_start(SERVICE_service_t* service) {
	topocenter_migration_data_t* placement_data = NULL;

	C_ABORT_ON(service->placement_data != NULL, "PLACEMENT: Trying to start TopoCenter(1) algorithm for service \"%s\", but placement data is not NULL.\n", service->id);

	C_DEBUG(PLACEMENT_DEBUG_LEVEL, C_MSG_INFO, "PLACEMENT: Starting TopoCenter(1) algorithm for service \"%s\"...\n", service->id);
	MALLOC(service->placement_data, topocenter_migration_data_t);
	placement_data = (topocenter_migration_data_t*) service->placement_data;
	SERVICE_alloc_service_id(&placement_data->service_id, service->id);
	placement_data->callback_id = callback_add(topocenter_migration_run, placement_data->service_id, PLACEMENT_TOPOCENTER_EPOCH);
}

/**
 * @brief Stop TopoCenter(1) migration algorithm.
 *
 * @param service Pointer to service for which to stop this algorithm.
 */
void topocenter_migration_stop(SERVICE_service_t* service) {
	topocenter_migration_data_t* placement_data = (topocenter_migration_data_t*) service->placement_data;

	C_ABORT_ON(placement_data == NULL, "PLACEMENT: Trying to stop TopoCenter(1) algorithm for service \"%s\", but placement data is NULL.\n", service->id);

	C_DEBUG(PLACEMENT_DEBUG_LEVEL, C_MSG_INFO, "PLACEMENT: Stopping TopoCenter(1) algorithm for service \"%s\"...\n", service->id);
	FREE(placement_data->service_id);
	if(placement_data->callback_id != 0)
		callback_cancel(placement_data->callback_id);
	FREE(service->placement_data);
}

#else
	#warning Compiling without support for TopoCenter migration...
#endif	// C_SERVICE_PLACEMENT_EVALUATION
