/**
 * @file rudp-light.c
 *
 * @brief Reliable User Datagram Protocol (RUDP) - minimal implemenation.
 *
 * Copyright (c) 2007, 2008, 2009, 2010 Georg Wittenburg, Freie Universitaet
 * Berlin (wittenbu@inf.fu-berlin.de). All rights reserved.
 * Please refer to the COPYING file for detailed licensing information.
 *
 * @author Georg Wittenburg
 *
 * $Id: rudp-light.c 3368 2010-02-28 16:10:09Z gwitten $
 */

#include <math.h>
#include <stdlib.h>
#include <string.h>

#include "../config.h"
#include "../macros.h"
#include "../types.h"
#include "../vars.h"
#include "../main.h"
#include "../util/debug.h"
#include "../util/list.h"
#include "../util/callback.h"
#include "../util/net.h"
#include "rudp-light.h"
#include "rudp-light_types.h"
#include "rudp-light_vars.h"



/**
 * @brief Maximal transmission queue length.
 */
#define RUDP_MAX_QUEUED_PACKETS	512



/*
 * Interface to main module.
 */

/**
 * @brief Initialize the RUDP module.
 */
void RUDP_init() {
	C_RUDP_transmission_queue = NULL;
	list_alloc(&C_RUDP_transmission_queue);
	memset(C_RUDP_received_packet_cache, 0, C_RUDP_RECEIVED_PACKET_CACHE_SIZE * sizeof(packet_fingerprint_t));
	C_RUDP_total_transmissions = 0;
	C_RUDP_successful_transmissions = 0;
	C_RUDP_failed_transmissions = 0;
}

/**
 * @brief Start the RUDP module.
 */
void RUDP_start() {
}

/**
 * @brief Stop the RUDP module.
 */
void RUDP_stop() {
	RUDP_transmission_t* transmission = NULL;
	while(!list_empty(C_RUDP_transmission_queue)) {
		transmission = (RUDP_transmission_t*) list_head(C_RUDP_transmission_queue);

		C_DEBUG(RUDP_DEBUG_LEVEL, C_MSG_INFO, "RUDP: Aborting transmission of DATA packet to %s with sequence number %d...\n", addr_to_str(transmission->packet.header->destination_address), transmission->packet.header->seqno);
		callback_cancel(transmission->callback_id);
		if(transmission->function != NULL) {
			transmission->result->error_code = C_ERROR_ABORTED;
			transmission->function(transmission->result);
		}
		list_remove(C_RUDP_transmission_queue, transmission);
		free_header(&transmission->packet.header);
		if(transmission->packet.payload != NULL)
			FREE(transmission->packet.payload);
		FREE(transmission);

		C_RUDP_failed_transmissions++;
	}
}

/**
 * @brief Finalize the RUDP module.
 */
void RUDP_finalize() {
	if(C_check_consistency_on_shutdown) {
		C_ABORT_ON(!list_empty(C_RUDP_transmission_queue), "RUDP: There are still %d packet(s) in the transmission queue.\n", list_size(C_RUDP_transmission_queue));
		C_ABORT_ON(C_RUDP_total_transmissions != (C_RUDP_successful_transmissions + C_RUDP_failed_transmissions), "RUDP: Transmit counters don't add up (%d != (%d + %d)).\n", C_RUDP_total_transmissions, C_RUDP_successful_transmissions, C_RUDP_failed_transmissions);
	}
	list_free(&C_RUDP_transmission_queue);
}



/*
 * Packet handling functions.
 */

/**
 * @brief Send a RUDP DATA packet.
 *
 * @param orig_header Pointer to header of the packet to be transmitted.
 * @param orig_payload Pointer to payload of the packet to be transmitted.
 */
static void send_RUDP_DATA(const net_header_t* orig_header, const void* orig_payload) {
	RUDP_DATA_packet_t data;
	net_header_t* new_header = NULL;

	C_DEBUG(RUDP_DEBUG_LEVEL, C_MSG_INFO, "RUDP: Sending DATA packet #%d to %s...\n", orig_header->seqno, addr_to_str(orig_header->destination_address));

	// Store original application ID and type ID in RUDP data header.
	memset(&data, 0, sizeof(RUDP_DATA_packet_t));
	data.orig_app = orig_header->app;
	data.orig_type = orig_header->type;

	alloc_header_from_header(&new_header, orig_header);
	new_header->app = APP_TYPE_RUDP;
	new_header->type = RUDP_PACKET_TYPE_DATA;
	if(new_header->payload_length == 0) {
		new_header->payload_length = sizeof(RUDP_DATA_packet_t);
		C_route(new_header, &data);
	} else {
		void* rudp_payload = NULL;
		MALLOC_JOIN(rudp_payload, char, &data, sizeof(RUDP_DATA_packet_t), orig_payload, new_header->payload_length);
		new_header->payload_length += sizeof(RUDP_DATA_packet_t);
		C_route(new_header, rudp_payload);
		FREE(rudp_payload);
	}
	free_header(&new_header);
}

/**
 * @brief Send a RUDP ACK packet.
 *
 * @param to Address of the destination node.
 * @param seqno Sequence number of the packet to be acknowledged.
 */
static void send_RUDP_ACK(C_addr_t to, seqno_t seqno) {
	RUDP_ACK_packet_t ack;

	C_DEBUG(RUDP_DEBUG_LEVEL, C_MSG_INFO, "RUDP: Sending ACK for DATA packet #%d to %s...\n", seqno, addr_to_str(to));
	memset(&ack, 0, sizeof(RUDP_ACK_packet_t));
	ack.seqno = seqno;
	C_sendto(to, PACKET_PRIORITY_SIGNALLING_TRANSPORT, APP_TYPE_RUDP, RUDP_PACKET_TYPE_ACK, &ack, sizeof(RUDP_ACK_packet_t));
}

/**
 * @brief Handle incoming RUDP DATA packet.
 *
 * @param rudp_header Pointer to header of incoming RUDP DATA packet.
 * @param rudp_payload Pointer to payload of incoming RUDP DATA packet.
 */
static void handle_RUDP_DATA(const net_header_t* rudp_header, const void* rudp_payload) {
	net_header_t* orig_header = NULL;
	RUDP_DATA_packet_t* data = (RUDP_DATA_packet_t*) rudp_payload;

	C_ABORT_ON(rudp_header->payload_length < sizeof(RUDP_DATA_packet_t), "RUDP: Received DATA packet with wrong payload length (%d < %d).\n", rudp_header->payload_length, sizeof(RUDP_DATA_packet_t));

	if(check_for_duplicate(C_RUDP_received_packet_cache, C_RUDP_RECEIVED_PACKET_CACHE_SIZE, rudp_header)) {
		C_DEBUG(RUDP_DEBUG_LEVEL, C_MSG_DEBUG, "RUDP: Dropping duplicate DATA packet #%d from %s...\n", rudp_header->seqno, addr_to_str(rudp_header->source_address));
		send_RUDP_ACK(rudp_header->source_address, rudp_header->seqno);
		return;
	}

	// Restore application ID and type ID in original header, adjust
	// payload length.
	alloc_header_from_header(&orig_header, rudp_header);
	orig_header->app = data->orig_app;
	orig_header->type = data->orig_type;
	orig_header->payload_length -= sizeof(RUDP_DATA_packet_t);

	C_DEBUG(RUDP_DEBUG_LEVEL, C_MSG_INFO, "RUDP: Received DATA packet #%d from %s...\n", orig_header->seqno, addr_to_str(orig_header->source_address));
	send_RUDP_ACK(orig_header->source_address, orig_header->seqno);

	if(orig_header->payload_length == 0)
		C_handle_packet(orig_header, NULL);
	else
		C_handle_packet(orig_header, ((char*) rudp_payload) + sizeof(RUDP_DATA_packet_t));
	free_header(&orig_header);
}

/**
 * @brief Check whether destination address and sequence number of two
 * transmission match.
 *
 * @param item Transmission to check.
 * @param cmp_item Reference transmission.
 * @return <code>true</code> if destination address and sequence number of two
 * transmissions are equal, <code>false</code> otherwise.
 */
static bool matches_transmission_net_header_to_and_seqno(const void* item, const void* cmp_item) {
	RUDP_transmission_t* transmission = (RUDP_transmission_t*) item;
	RUDP_transmission_t* cmp_transmission = (RUDP_transmission_t*) cmp_item;
	return (
		(transmission->packet.header->destination_address == cmp_transmission->packet.header->destination_address)
		&& (transmission->packet.header->seqno == cmp_transmission->packet.header->seqno)
	);
}

/**
 * @brief Handle incoming RUDP ACK packet.
 *
 * @param header Pointer to header of incoming RUDP ACK packet.
 * @param ack Pointer to ACK information of incoming RUDP ACK packet.
 */
static void handle_RUDP_ACK(const net_header_t* header, const RUDP_ACK_packet_t* ack) {
	RUDP_transmission_t cmp_transmission;
	net_header_t cmp_header;
	RUDP_transmission_t* transmission = NULL;

	memset(&cmp_transmission, 0, sizeof(RUDP_transmission_t));
	memset(&cmp_header, 0, sizeof(net_header_t));
	cmp_transmission.packet.header = &cmp_header;
	cmp_transmission.packet.header->destination_address = header->source_address;
	cmp_transmission.packet.header->seqno = ack->seqno;
	transmission = (RUDP_transmission_t*) list_find(C_RUDP_transmission_queue, &cmp_transmission, matches_transmission_net_header_to_and_seqno);
	if(transmission == NULL) {
		C_DEBUG(RUDP_DEBUG_LEVEL, C_MSG_DEBUG, "RUDP: Ignoring ACK from %s for packet with sequence number %d that is not in queue...\n", addr_to_str(header->source_address), ack->seqno);
		return;
	}

	C_DEBUG(RUDP_DEBUG_LEVEL, C_MSG_INFO, "RUDP: Received ACK from %s for packet with sequence number %d.\n", addr_to_str(header->source_address), ack->seqno);
	if(transmission->function != NULL) {
		transmission->result->error_code = C_ERROR_OK;
		transmission->function(transmission->result);
	}
	transmission->result = NULL;
	list_remove(C_RUDP_transmission_queue, transmission);
	callback_cancel(transmission->callback_id);
	if(transmission->packet.payload != NULL)
		FREE(transmission->packet.payload);
	free_header(&transmission->packet.header);
	FREE(transmission);

	C_RUDP_successful_transmissions++;
}

/**
 * @brief RUDP packet classifier.
 *
 * @param header Pointer to header of an incoming packet.
 * @param payload Pointer to payload of an incoming packet.
 */
void RUDP_recv(const net_header_t* header, const void* payload) {
	C_ABORT_ON(header->app != APP_TYPE_RUDP, "RUDP: Received non-RUDP packet.\n");

	switch(header->type) {
	case RUDP_PACKET_TYPE_DATA:
		handle_RUDP_DATA(header, payload);
		break;
	case RUDP_PACKET_TYPE_ACK:
		C_ABORT_ON(header->payload_length != sizeof(RUDP_ACK_packet_t), "RUDP: Received ACK packet with wrong payload length (%d != %d).\n", header->payload_length, sizeof(RUDP_ACK_packet_t));
		handle_RUDP_ACK(header, (RUDP_ACK_packet_t*) payload);
		break;
	default:
		C_ABORT_ON(true, "RUDP: Received packet of unknown type 0x%X from %s.\n", header->type, addr_to_str(header->source_address));
	}
}



/*
 * Interface provided by this module.
 */

/**
 * @brief Timeout callback for sending a packet via RUDP.
 *
 * @param parameter Transmission for which the timeout ocurred.
 */
static void RUDP_send_timeout(callback_parameter_t parameter) {
	RUDP_transmission_t* transmission = (RUDP_transmission_t*) parameter;

	// Retry if the packet hasn't been removed from the queue.
	if(transmission->retries > 0) {
		int retry = C_RUDP_PACKET_RETRIES - (--transmission->retries);
		C_time_t backoff_time = C_RUDP_MIN_BACKOFF_TIME + C_random(C_RUDP_MIN_BACKOFF_TIME * pow(2, retry));
		if(backoff_time > C_RUDP_MAX_BACKOFF_TIME)
			backoff_time = C_RUDP_MAX_BACKOFF_TIME;
		C_DEBUG(RUDP_DEBUG_LEVEL, C_MSG_DEBUG, "RUDP: Timeout for acknowledgement of DATA packet to %s with sequence number %d has expired (retry #%d). Resending packet and rescheduling by %.2fs...\n", addr_to_str(transmission->packet.header->destination_address), transmission->packet.header->seqno, retry, backoff_time);
		transmission->callback_id = callback_add(RUDP_send_timeout, transmission, backoff_time);
		send_RUDP_DATA(transmission->packet.header, transmission->packet.payload);

	// Remove from queue and drop packet if no retries are left.
	} else {
		float pdr = 0.0;
		if(C_RUDP_total_transmissions > 0)
			pdr = ((float) C_RUDP_successful_transmissions) / ((float) C_RUDP_total_transmissions);
		C_DEBUG(RUDP_DEBUG_LEVEL, C_MSG_WARN, "RUDP: Timeout for ACK of DATA packet to %s with sequence number %d has expired. No retries left, dropping packet. PDR is %.2f%%.\n", addr_to_str(transmission->packet.header->destination_address), transmission->packet.header->seqno, (pdr * 100.0));
		if(transmission->function != NULL) {
			transmission->result->error_code = C_ERROR_TIMEOUT;
			transmission->function(transmission->result);
		}
		list_remove(C_RUDP_transmission_queue, transmission);
		free_header(&transmission->packet.header);
		if(transmission->packet.payload != NULL)
			FREE(transmission->packet.payload);
		FREE(transmission);

		C_RUDP_failed_transmissions++;
	}
}

/**
 * @brief Send a packet via RUDP.
 *
 * @param header Pointer to the header of the packet to send.
 * @param payload Pointer to the payload of the packet to send.
 * @param function Callback function.
 * @param result Result of the transmission.
 */
void RUDP_send(net_header_t* header, const void* payload, callback_function_t function, RUDP_transmission_result_t* result) {
	RUDP_transmission_t* transmission = NULL;

	if((function != NULL) && (result == NULL)) {
		C_DEBUG(RUDP_DEBUG_LEVEL, C_MSG_ERROR, "RUDP: Trying to send packet with callback, but no memory for the results has been allocated.\n");
		return;
	} else if(header == NULL) {
		C_DEBUG(RUDP_DEBUG_LEVEL, C_MSG_ERROR, "RUDP: Trying to send packet without header.\n");
		if(function != NULL) {
			result->error_code = C_ERROR_PARAMETER;
			function(result);
		}
		return;
	} else if((header->payload_length > 0) && (payload == NULL)) {
		C_DEBUG(RUDP_DEBUG_LEVEL, C_MSG_ERROR, "RUDP: Trying to send non-empty packet without payload.\n");
		if(function != NULL) {
			result->error_code = C_ERROR_PARAMETER;
			function(result);
		}
		return;
	} else if(header->destination_address == C_BROADCAST_ADDRESS) {
		C_DEBUG(RUDP_DEBUG_LEVEL, C_MSG_ERROR, "RUDP: Broadcasts are not supported, ignoring packet...\n");
		if(function != NULL) {
			result->error_code = C_ERROR_PARAMETER;
			function(result);
		}
		return;
	}

	C_RUDP_total_transmissions++;

	if(list_size(C_RUDP_transmission_queue) >= RUDP_MAX_QUEUED_PACKETS) {
		C_DEBUG(RUDP_DEBUG_LEVEL, C_MSG_WARN, "RUDP: Queue full, dropping packet...\n");
		if(function != NULL) {
			result->error_code = C_ERROR_BUSY;
			function(result);
		}
		C_RUDP_failed_transmissions++;
		return;
	}

	C_DEBUG(RUDP_DEBUG_LEVEL, C_MSG_NOTE, "RUDP: Reliably sending packet #%d from %s to %s...\n", header->seqno, addr_to_str(header->source_address), addr_to_str(header->destination_address));
	MALLOC(transmission, RUDP_transmission_t);
	alloc_header_from_header(&transmission->packet.header, header);
	if(header->payload_length > 0) {
		void* p = NULL;
		MALLOC_ARRAY(p, char, header->payload_length);
		memcpy(p, payload, header->payload_length);
		transmission->packet.payload = p;
	} else {
		transmission->packet.payload = NULL;
	}
	transmission->retries = C_RUDP_PACKET_RETRIES;
	if(function == NULL) {
		transmission->function = NULL;
		transmission->result = NULL;
	} else {
		transmission->function = function;
		transmission->result = result;
		transmission->result->error_code = C_ERROR_NOT_SET;
	}
	list_add(C_RUDP_transmission_queue, transmission);

	transmission->callback_id = callback_add(RUDP_send_timeout, transmission, C_NETWORK_TRAVERSAL_TIME);
	send_RUDP_DATA(transmission->packet.header, transmission->packet.payload);
}


/**
 * @brief Send a packet via RUDP with application parameters and payload.
 *
 * @param to Address of destination node.
 * @param priority Priority of the packet.
 * @param app Application identifier.
 * @param type Packet type identifier.
 * @param payload Pointer to payload.
 * @param payload_length Length of payload in bytes.
 * @param function Callback function.
 * @param result Result of the transmission.
 */
void RUDP_sendto(C_addr_t to, uint8_t priority, uint8_t app, uint8_t type, const void* payload, size_t payload_length, callback_function_t function, RUDP_transmission_result_t* result) {
	RUDP_sendto2(to, priority, app, type, payload, payload_length, NULL, 0, function, result);
}

/**
 * @brief Send a packet via RUDP with application parameters and two payloads.
 *
 * @param to Address of destination node.
 * @param priority Priority of the packet.
 * @param app Application identifier.
 * @param type Packet type identifier.
 * @param payload1 Pointer to first payload.
 * @param payload_length1 Length of first payload in bytes.
 * @param payload2 Pointer to second payload.
 * @param payload_length2 Length of second payload in bytes.
 * @param function Callback function.
 * @param result Result of the transmission.
 */
void RUDP_sendto2(C_addr_t to, uint8_t priority, uint8_t app, uint8_t type, const void* payload1, size_t payload_length1, const void* payload2, size_t payload_length2, callback_function_t function, RUDP_transmission_result_t* result) {
	net_header_t* header = NULL;

	C_ABORT_ON(((payload1 == NULL) && (payload_length1 != 0)) || ((payload1 != NULL) && (payload_length1 == 0)) || ((payload2 == NULL) && (payload_length2 != 0)) || ((payload2 != NULL) && (payload_length2 == 0)), "Trying to assemble packet for sending to %s via RUDP, but payload parameters are inconsistent.\n", addr_to_str(to));

	C_DEBUG(RUDP_DEBUG_LEVEL, C_MSG_VERBOSE, "Assembling packet for sending to %s via RUDP...\n", addr_to_str(to));
	alloc_header(&header, to, priority, app, type, (payload_length1 + payload_length2));
	if(header->payload_length == 0) {
		RUDP_send(header, NULL, function, result);
	} else {
		void* payload = NULL;
		MALLOC_JOIN(payload, char, payload1, payload_length1, payload2, payload_length2);
		RUDP_send(header, payload, function, result);
		FREE(payload);
	}
	free_header(&header);
}

/**
 * @brief Return the MTU size supported by RUDP.
 *
 * @return Maximal Transmission Unit (MTU) size in bytes.
 */
size_t RUDP_max_payload_size() {
	return C_max_net_payload_size() - sizeof(net_header_t);
}



/*
 * Interface provided by this module for direct commands.
 */

/**
 * @brief RUDP command function.
 *
 * @param argc Number of command parameters.
 * @param argv Array of strings used as command parameters.
 * @return <code>true</code> if the command was executed successfully,
 * <code>false</code> otherwise.
 */
bool RUDP_command(int argc, const char*const* argv) {
	return false;
}
