/**
 * @file service_client.c
 *
 * @brief Service - Client.
 *
 * Copyright (c) 2007, 2008, 2009, 2010 Georg Wittenburg, Freie Universitaet
 * Berlin (wittenbu@inf.fu-berlin.de). All rights reserved.
 * Please refer to the COPYING file for detailed licensing information.
 *
 * @author Georg Wittenburg
 *
 * $Id: service_client.c 3427 2010-03-18 18:35:15Z gwitten $
 */

#include <math.h>

#include "../config.h"
#include "../macros.h"
#include "../types.h"
#include "../vars.h"
#include "../main.h"
#include "../util/debug.h"
#include "../util/callback.h"
#include "../util/net.h"
#include "../rudp-light/rudp-light.h"
#include "../slp-light/slp-light.h"
#include "../placement/placement.h"
#include "service.h"
#include "service_types.h"
#include "service_vars.h"
#include "service_client.h"
#include "service_management.h"
#include "service_util.h"



/**
 * @brief Debug level for the client component.
 */
#define SERVICE_CLIENT_DEBUG_LEVEL	C_MSG_WARN

/**
 * @brief Number of service instances to locate before selecting the best
 * server.
 */
#define CLIENT_MIN_SERVICE_INSTANCES	1

/**
 * @brief Maximal number of data/stream requests/updates to schedule for retry.
 */
#define CLIENT_MAX_REQUEST_SCHEDULED_FOR_RETRY	100

#ifdef C_ENABLE_NAM_TRACE_OUTPUT
/**
 * @brief Timeout for resetting the color of a node after a request in seconds.
 */
#define REQUEST_INDICATION_TIMEOUT	0.1

/**
 * @brief Timeout for resetting the color of a node after a period of activity
 * in seconds.
 */
#define ACTIVITY_INDICATION_TIMEOUT		10.0
#endif	// C_ENABLE_NAM_TRACE_OUTPUT



/*
 * Packet handling functions.
 */

/**
 * @brief Send SERVICE DATA REQUEST packet.
 *
 * @param to Address of destination node.
 * @param service_id ID of service of which data is requested.
 * @param offset Offset in service data.
 * @param size Request size in bytes.
 */
static void send_SERVICE_CLIENT_DATA_REQUEST(C_addr_t to, const SERVICE_id_t* service_id, uint16_t offset, size_t size) {
	SERVICE_CLIENT_DATA_REQUEST_packet_t request;

	C_DEBUG(SERVICE_CLIENT_DEBUG_LEVEL, C_MSG_DEBUG, "SERVICE: Sending CLIENT DATA REQUEST for %d byte(s) at 0x%X of service \"%s\" to %s...\n", size, offset, service_id, addr_to_str(to));
	memset(&request, 0, sizeof(SERVICE_CLIENT_DATA_REQUEST_packet_t));
	copy_service_id(request.service_id, service_id);
	request.offset = offset;
	request.size = size;
	request.start_time = C_time();
	C_sendto(to, PACKET_PRIORITY_DATA, APP_TYPE_SERVICE, SERVICE_PACKET_TYPE_CLIENT_DATA_REQUEST, &request, sizeof(SERVICE_CLIENT_DATA_REQUEST_packet_t));

	C_SERVICE_transmitted_bytes += sizeof(SERVICE_CLIENT_DATA_REQUEST_packet_t);
	add_bytes_to_datarate_stat(C_SERVICE_transmitted_datarate_stat, sizeof(SERVICE_CLIENT_DATA_REQUEST_packet_t));
}

#ifdef SERVICE_CLIENT_USE_RUDP_FOR_UPDATES
/**
 * @brief RUDP callback for SERVICE DATA UPDATE packets.
 *
 * @param parameter RUDP transmission result.
 */
static void send_SERVICE_CLIENT_DATA_UPDATE_callback(callback_parameter_t parameter) {
	RUDP_transmission_result_t* result = (RUDP_transmission_result_t*) parameter;

	if(result->error_code == C_ERROR_OK) {
		C_DEBUG(SERVICE_CLIENT_DEBUG_LEVEL, C_MSG_DEBUG, "SERVICE: CLIENT DATA UPDATE was successful.\n");
		C_SERVICE_successful_data_updates++;
		C_SERVICE_successful_requests++;
	} else {
		C_DEBUG(SERVICE_CLIENT_DEBUG_LEVEL, C_MSG_DEBUG, "SERVICE: CLIENT DATA UPDATE failed (%d).\n", result->error_code);
		C_SERVICE_failed_data_updates++;
		C_SERVICE_failed_requests++;
	}

	FREE(result);
}
#endif	// SERVICE_CLIENT_USE_RUDP_FOR_UPDATES

/**
 * @brief Send SERVICE DATA UPDATE packet.
 *
 * @param to Address of destination node.
 * @param service_id ID of service of which data is to be updated.
 * @param offset Offset in service data.
 * @param size Update size in bytes.
 * @param data Pointer to new service data.
 * @param inter_host_update Flag whether this update is an inter-host update.
 */
void send_SERVICE_CLIENT_DATA_UPDATE(C_addr_t to, const SERVICE_id_t* service_id, uint16_t offset, size_t size, const char* data, bool inter_host_update) {
	SERVICE_CLIENT_DATA_UPDATE_packet_t update;
#ifdef SERVICE_CLIENT_USE_RUDP_FOR_UPDATES
	RUDP_transmission_result_t* result = NULL;
#endif	// SERVICE_CLIENT_USE_RUDP_FOR_UPDATES

	C_DEBUG(SERVICE_CLIENT_DEBUG_LEVEL, C_MSG_DEBUG, "SERVICE: Sending CLIENT DATA UPDATE for %d byte(s) at 0x%X of service \"%s\" (0x%X...) to %s...\n", size, offset, service_id, (uint8_t) *data, addr_to_str(to));
	memset(&update, 0, sizeof(SERVICE_CLIENT_DATA_UPDATE_packet_t));
	copy_service_id(update.service_id, service_id);
	update.offset = offset;
	update.inter_host_update = inter_host_update;

#ifdef SERVICE_CLIENT_USE_RUDP_FOR_UPDATES
	MALLOC(result, RUDP_transmission_result_t);
	RUDP_sendto2(to, PACKET_PRIORITY_DATA, APP_TYPE_SERVICE, SERVICE_PACKET_TYPE_CLIENT_DATA_UPDATE, &update, sizeof(SERVICE_CLIENT_DATA_UPDATE_packet_t), data, size, send_SERVICE_CLIENT_DATA_UPDATE_callback, result);
#else
	C_sendto2(to, PACKET_PRIORITY_DATA, APP_TYPE_SERVICE, SERVICE_PACKET_TYPE_CLIENT_DATA_UPDATE, &update, sizeof(SERVICE_CLIENT_DATA_UPDATE_packet_t), data, size);
#endif	// SERVICE_CLIENT_USE_RUDP_FOR_UPDATES

	if(inter_host_update) {
		C_SERVICE_sync_overhead_bytes += sizeof(SERVICE_CLIENT_DATA_UPDATE_packet_t) + size;
	} else {
		C_SERVICE_transmitted_bytes += sizeof(SERVICE_CLIENT_DATA_UPDATE_packet_t) + size;
		add_bytes_to_datarate_stat(C_SERVICE_transmitted_datarate_stat, (sizeof(SERVICE_CLIENT_DATA_UPDATE_packet_t) + size));
	}
}

/**
 * @brief RUDP callback for SERVICE STREAM REQUEST packets.
 *
 * @param parameter RUDP transmission result.
 */
static void send_SERVICE_CLIENT_STREAM_REQUEST_callback(callback_parameter_t parameter) {
	RUDP_transmission_result_t* result = (RUDP_transmission_result_t*) parameter;

	if(result->error_code == C_ERROR_OK) {
		C_DEBUG(SERVICE_CLIENT_DEBUG_LEVEL, C_MSG_DEBUG, "SERVICE: CLIENT STREAM REQUEST was successful.\n");
		C_SERVICE_successful_stream_requests++;
		C_SERVICE_successful_requests++;
	} else {
		C_DEBUG(SERVICE_CLIENT_DEBUG_LEVEL, C_MSG_DEBUG, "SERVICE: CLIENT STREAM REQUEST failed (%d).\n", result->error_code);
		C_SERVICE_failed_stream_requests++;
		C_SERVICE_failed_requests++;
	}

	FREE(result);
}

/**
 * @brief Send SERVICE STREAM REQUEST packet.
 *
 * @param to Address of destination node.
 * @param service_id ID of service of which data is requested.
 * @param offset Offset in service data.
 * @param size Request size in bytes.
 * @param interval Interval between streamed packets in seconds.
 * @param duration Time the stream is to be active in seconds.
 */
static void send_SERVICE_CLIENT_STREAM_REQUEST(C_addr_t to, const SERVICE_id_t* service_id, uint16_t offset, size_t size, C_time_t interval, C_time_t duration) {
	SERVICE_CLIENT_STREAM_REQUEST_packet_t request;
	RUDP_transmission_result_t* result = NULL;

	C_DEBUG(SERVICE_CLIENT_DEBUG_LEVEL, C_MSG_DEBUG, "SERVICE: Sending CLIENT STREAM REQUEST for %d byte(s) at 0x%X of service \"%s\" at interval %.3fs for a duration of %.2fs to %s...\n", size, offset, service_id, interval, duration, addr_to_str(to));
	memset(&request, 0, sizeof(SERVICE_CLIENT_STREAM_REQUEST_packet_t));
	copy_service_id(request.service_id, service_id);
	request.offset = offset;
	request.size = size;
	request.interval = interval;
	request.duration = duration;
	MALLOC(result, RUDP_transmission_result_t);
	RUDP_sendto(to, PACKET_PRIORITY_DATA, APP_TYPE_SERVICE, SERVICE_PACKET_TYPE_CLIENT_STREAM_REQUEST, &request, sizeof(SERVICE_CLIENT_STREAM_REQUEST_packet_t), send_SERVICE_CLIENT_STREAM_REQUEST_callback, result);

	C_SERVICE_transmitted_bytes += sizeof(SERVICE_CLIENT_STREAM_REQUEST_packet_t);
	add_bytes_to_datarate_stat(C_SERVICE_transmitted_datarate_stat, sizeof(SERVICE_CLIENT_STREAM_REQUEST_packet_t));
}

/**
 * @brief Handle incoming SERVICE DATA ITEM packet.
 *
 * Possibly redirect future requests to this service if server provided
 * corresponding information
 *
 * @param header Header of the incoming packet.
 * @param data_item SERVER DATA ITEM header.
 * @param data Requested data.
 * @param data_size Size of requested data.
 */
void handle_SERVICE_SERVER_DATA_ITEM(const net_header_t* header, const SERVICE_SERVER_DATA_ITEM_packet_t* data_item, const char* data, size_t data_size) {
	C_time_t delay = C_time() - data_item->start_time;
	float recall = 0.0;

	if(data_item->redirect_addr == C_BROADCAST_ADDRESS) {
		SLP_cache_update_with_metrics(data_item->service_id, header->source_address, header->hop_count, delay);
#ifdef SERVICE_MIGRATION_USE_REDIRECTION
	} else {
		SERVICE_service_t* service = SERVICE_get_service(data_item->service_id);

		C_DEBUG(SERVICE_CLIENT_DEBUG_LEVEL, C_MSG_INFO, "SERVICE: Redirecting future request for service \"%s\" to %s.\n", data_item->service_id, addr_to_str(data_item->redirect_addr));
		SLP_cache_remove(data_item->service_id, header->source_address);
		SLP_cache_update(data_item->service_id, data_item->redirect_addr);
		if(service != NULL) {
			PLACEMENT_report_invalid_service_instance(service, header->source_address);
			PLACEMENT_report_known_service_instance(service, data_item->redirect_addr);
		}
#endif	// SERVICE_MIGRATION_USE_REDIRECTION
	}

	C_SERVICE_successful_data_requests++;
	C_SERVICE_total_successful_data_request_delay += delay;
	C_SERVICE_successful_requests++;
	C_SERVICE_received_bytes += sizeof(SERVICE_SERVER_DATA_ITEM_packet_t) + data_size;
	add_bytes_to_datarate_stat(C_SERVICE_received_datarate_stat, (sizeof(SERVICE_SERVER_DATA_ITEM_packet_t) + data_size));

	if(C_SERVICE_total_data_requests > 0)
		recall = ((float) C_SERVICE_successful_data_requests) / ((float) C_SERVICE_total_data_requests);
	C_DEBUG(SERVICE_CLIENT_DEBUG_LEVEL, C_MSG_INFO, "SERVICE: Received %d byte(s) of SERVER DATA ITEM for service \"%s\" (0x%X...) from %s after %.2fs. Recall is %.2f%% (%d / %d).\n", data_size, data_item->service_id, (uint8_t) *data, addr_to_str(header->source_address), delay, (recall * 100.0), C_SERVICE_successful_data_requests, C_SERVICE_total_data_requests);
}

/**
 * @brief Handle incoming SERVICE STREAM ITEM packet.
 *
 * Possibly redirect future requests to this service if server provided
 * corresponding information
 *
 * @param header Header of the incoming packet.
 * @param stream_item SERVER DATA ITEM header.
 * @param data Requested data.
 * @param data_size Size of requested data.
 */
void handle_SERVICE_SERVER_STREAM_ITEM(const net_header_t* header, const SERVICE_SERVER_STREAM_ITEM_packet_t* stream_item, const char* data, size_t data_size) {
	if(stream_item->redirect_addr == C_BROADCAST_ADDRESS) {
		SLP_cache_update(stream_item->service_id, header->source_address);
#ifdef SERVICE_MIGRATION_USE_REDIRECTION
	} else {
		SERVICE_service_t* service = SERVICE_get_service(stream_item->service_id);

		C_DEBUG(SERVICE_CLIENT_DEBUG_LEVEL, C_MSG_INFO, "SERVICE: Redirecting future request to service \"%s\" to %s.\n", stream_item->service_id, addr_to_str(stream_item->redirect_addr));
		SLP_cache_remove(stream_item->service_id, header->source_address);
		SLP_cache_update(stream_item->service_id, stream_item->redirect_addr);
		if(service != NULL) {
			PLACEMENT_report_invalid_service_instance(service, header->source_address);
			PLACEMENT_report_known_service_instance(service, stream_item->redirect_addr);
		}
#endif	// SERVICE_MIGRATION_USE_REDIRECTION
	}

	C_SERVICE_received_stream_packets++;
	C_SERVICE_received_bytes += sizeof(SERVICE_SERVER_STREAM_ITEM_packet_t) + data_size;
	add_bytes_to_datarate_stat(C_SERVICE_received_datarate_stat, (sizeof(SERVICE_SERVER_STREAM_ITEM_packet_t) + data_size));

	C_DEBUG(SERVICE_CLIENT_DEBUG_LEVEL, C_MSG_DEBUG, "SERVICE: Received %d byte(s) of SERVER STREAM ITEM (0x%X...) from %s.\n", header->payload_length, (uint8_t) *data, addr_to_str(header->source_address));
}

/**
 * @brief Handle incoming SERVICE NOT AVAILABLE packet.
 *
 * Possibly redirect future requests to this service if server provided
 * corresponding information
 *
 * @param header Header of incoming packet.
 * @param not_available Data structure with redirection information.
 */
void handle_SERVICE_SERVER_NOT_AVAILABLE(const net_header_t* header, const SERVICE_SERVER_NOT_AVAILABLE_packet_t* not_available) {
	SERVICE_service_t* service = SERVICE_get_service(not_available->service_id);

	C_DEBUG(SERVICE_CLIENT_DEBUG_LEVEL, C_MSG_DEBUG, "SERVICE: Received of SERVER NOT AVAILABLE for service \"%s\" from %s.\n", not_available->service_id, addr_to_str(header->source_address));

#ifdef SERVICE_MIGRATION_USE_REDIRECTION
	if(not_available->redirect_addr != C_BROADCAST_ADDRESS) {
		C_DEBUG(SERVICE_CLIENT_DEBUG_LEVEL, C_MSG_INFO, "SERVICE: Redirecting future requests to service \"%s\" to %s...\n", not_available->service_id, addr_to_str(not_available->redirect_addr));
		SLP_cache_remove(not_available->service_id, header->source_address);
		SLP_cache_update(not_available->service_id, not_available->redirect_addr);
		if(service != NULL) {
			PLACEMENT_report_invalid_service_instance(service, header->source_address);
			PLACEMENT_report_known_service_instance(service, not_available->redirect_addr);
		}
	} else {
#endif	// SERVICE_MIGRATION_USE_REDIRECTION
		C_DEBUG(SERVICE_CLIENT_DEBUG_LEVEL, C_MSG_INFO, "SERVICE: Invalidating service cache entry for \"%s\" at %s...\n", not_available->service_id, addr_to_str(header->source_address));
		SLP_cache_remove(not_available->service_id, header->source_address);
		if(service != NULL)
			PLACEMENT_report_invalid_service_instance(service, header->source_address);
#ifdef SERVICE_MIGRATION_USE_REDIRECTION
	}
#endif	// SERVICE_MIGRATION_USE_REDIRECTION
}



/*
 * Control function.
 */

#ifdef C_ENABLE_NAM_TRACE_OUTPUT
/**
 * @brief Changed color timeout after indicating a pending request.
 *
 * @param parameter Not used.
 */
static void indicate_request_timeout(callback_parameter_t parameter) {
	if(C_SERVICE_indicate_activity_callback_id == 0) {
		C_ABORT_ON(true, "Illegal state while indicating node activity. Pending request is complete, but node is inactive.\n");
	} else {
		// Pending request is complete while node is active.
		C_trace("n -t %f -s %d -S COLOR -c green3 -o red", C_time(), C_LOCAL_ADDRESS);
	}
	C_SERVICE_indicate_request_callback_id = 0;
}

/**
 * @brief Changed color timeout after indicating node activity.
 *
 * @param parameter Not used.
 */
static void indicate_activity_timeout(callback_parameter_t parameter) {
	if(C_SERVICE_indicate_request_callback_id == 0) {
		// Node becomes inactive while no request is pending.
		C_trace("n -t %f -s %d -S COLOR -c black -o green3", C_time(), C_LOCAL_ADDRESS);
	} else {
		C_ABORT_ON(true, "Illegal state while indicating node activity. Node becomes inactive while a request is pending.\n");
	}
	C_SERVICE_indicate_activity_callback_id = 0;
}

/**
 * @brief Change color to indicate node activity.
 *
 * Color names are listed in /etc/X11/rgb.txt.
 */
static void indicate_activity() {
	if((C_SERVICE_indicate_request_callback_id == 0) && (C_SERVICE_indicate_activity_callback_id == 0)) {
		// Node becomes active with a pending request after a period of inactivity.
		C_trace("n -t %f -s %d -S COLOR -c red -o black", C_time(), C_LOCAL_ADDRESS);
		C_SERVICE_indicate_request_callback_id = callback_add(indicate_request_timeout, NULL, REQUEST_INDICATION_TIMEOUT);
		C_SERVICE_indicate_activity_callback_id = callback_add(indicate_activity_timeout, NULL, ACTIVITY_INDICATION_TIMEOUT);
	} else if((C_SERVICE_indicate_request_callback_id != 0) && (C_SERVICE_indicate_activity_callback_id == 0)) {
		C_ABORT_ON(true, "Illegal state while indicating node activity. A request is pending, but node is not active.\n");
	} else if((C_SERVICE_indicate_request_callback_id == 0) && (C_SERVICE_indicate_activity_callback_id != 0)) {
		// Node is active and a request becomes pending.
		C_trace("n -t %f -s %d -S COLOR -c red -o green3", C_time(), C_LOCAL_ADDRESS);
		callback_cancel(C_SERVICE_indicate_activity_callback_id);
		C_SERVICE_indicate_request_callback_id = callback_add(indicate_request_timeout, NULL, REQUEST_INDICATION_TIMEOUT);
		C_SERVICE_indicate_activity_callback_id = callback_add(indicate_activity_timeout, NULL, ACTIVITY_INDICATION_TIMEOUT);
	} else if((C_SERVICE_indicate_request_callback_id != 0) && (C_SERVICE_indicate_activity_callback_id != 0)) {
		// Node is active and a request is pending.
		callback_cancel(C_SERVICE_indicate_request_callback_id);
		callback_cancel(C_SERVICE_indicate_activity_callback_id);
		C_SERVICE_indicate_request_callback_id = callback_add(indicate_request_timeout, NULL, REQUEST_INDICATION_TIMEOUT);
		C_SERVICE_indicate_activity_callback_id = callback_add(indicate_activity_timeout, NULL, ACTIVITY_INDICATION_TIMEOUT);
	} else {
		C_ABORT_ON(true, "Impossible state while indicating node activity.\n");
	}
}
#endif	// C_ENABLE_NAM_TRACE_OUTPUT

/**
 * @brief Data request callback function for SLP queries.
 *
 * Proceed with data request if query was succesful, schedule retry if query is
 * in progress, abort otherwise.
 *
 * @param parameter SLP query result
 */
static void request_data_callback(callback_parameter_t parameter) {
	SLP_query_result_t* slp_query_result = (SLP_query_result_t*) parameter;
	SERVICE_data_request_t* request = (SERVICE_data_request_t*) slp_query_result->parameter;

	switch(slp_query_result->error_code) {
	case C_ERROR_OK:
		C_DEBUG(SERVICE_CLIENT_DEBUG_LEVEL, C_MSG_INFO, "SERVICE: Using %s as server for service \"%s\"...\n", addr_to_str(slp_query_result->server), request->service_id);
		send_SERVICE_CLIENT_DATA_REQUEST(slp_query_result->server, request->service_id, request->offset, request->size);
		goto cleanup;
	case C_ERROR_PARAMETER:
		C_DEBUG(SERVICE_CLIENT_DEBUG_LEVEL, C_MSG_ERROR, "SERVICE: Wrong parameter passed to SLP, dropping data request...\n");
		break;
	case C_ERROR_TIMEOUT:
#ifdef C_SERVICE_PLACEMENT_OPTIMIZED_FOR_SPI_EVALUATION
		C_DEBUG(SERVICE_CLIENT_DEBUG_LEVEL, C_MSG_WARN, "SERVICE: Service \"%s\" is not reachable, dropping data request...\n", request->service_id);
#else
		C_DEBUG(SERVICE_CLIENT_DEBUG_LEVEL, C_MSG_DEBUG, "SERVICE: Service \"%s\" is not reachable, dropping data request...\n", request->service_id);
#endif	// C_SERVICE_PLACEMENT_OPTIMIZED_FOR_SPI_EVALUATION
		break;
	case C_ERROR_ABORTED:
		C_DEBUG(SERVICE_CLIENT_DEBUG_LEVEL, C_MSG_WARN, "SERVICE: Data request for service \"%s\" aborted...\n", request->service_id);
		break;
	case C_ERROR_BUSY:
		C_DEBUG(SERVICE_CLIENT_DEBUG_LEVEL, C_MSG_WARN, "SERVICE: SLP is busy, dropping data request...\n");
		break;
	default:
		C_ABORT_ON(true, "SERVICE: request_data() callback with error code %d.\n", slp_query_result->error_code);
	}

	C_SERVICE_failed_data_requests++;
	C_SERVICE_failed_requests++;

cleanup:
	FREE(request);
	FREE(slp_query_result);
}

/**
 * @brief Request specific service data of a service.
 *
 * Perform SLP query for service in case it is not contained in the service
 * cache.
 *
 * @param service_id ID of the service from which to request data.
 * @param offset Offset in service data.
 * @param size Size of requested service data.
 */
void request_data(const SERVICE_id_t* service_id, uint16_t offset, size_t size) {
	SLP_query_result_t* slp_query_result = NULL;
	SERVICE_data_request_t* request = NULL;

	C_DEBUG(SERVICE_CLIENT_DEBUG_LEVEL, C_MSG_NOTE, "SERVICE: Requesting %d byte(s) at 0x%X of service \"%s\"...\n", size, offset, service_id);
#ifdef C_ENABLE_NAM_TRACE_OUTPUT
	indicate_activity();
#endif	// C_ENABLE_NAM_TRACE_OUTPUT
	C_SERVICE_total_data_requests++;
	C_SERVICE_total_requests++;

	MALLOC(slp_query_result, SLP_query_result_t);
	MALLOC(request, SERVICE_data_request_t);
	copy_service_id(request->service_id, service_id);
	request->offset = offset;
	request->size = size;
	slp_query_result->parameter = request;
	SLP_locate_service(service_id, request_data_callback, slp_query_result);
}

/**
 * @brief Data update callback function for SLP queries.
 *
 * Proceed with data update if query was succesful, schedule retry if query is
 * in progress, abort otherwise.
 *
 * @param parameter SLP query result
 */
static void update_data_callback(callback_parameter_t parameter) {
	SLP_query_result_t* slp_query_result = (SLP_query_result_t*) parameter;
	SERVICE_data_update_t* update = (SERVICE_data_update_t*) slp_query_result->parameter;

	switch(slp_query_result->error_code) {
	case C_ERROR_OK:
		C_DEBUG(SERVICE_CLIENT_DEBUG_LEVEL, C_MSG_INFO, "SERVICE: Using %s as server for service \"%s\"...\n", addr_to_str(slp_query_result->server), update->service_id);
		send_SERVICE_CLIENT_DATA_UPDATE(slp_query_result->server, update->service_id, update->offset, update->size, update->data, false);
		break;
	case C_ERROR_PARAMETER:
		C_DEBUG(SERVICE_CLIENT_DEBUG_LEVEL, C_MSG_ERROR, "SERVICE: Wrong parameter passed to SLP, dropping data request...\n");
		C_SERVICE_failed_data_updates++;
		C_SERVICE_failed_requests++;
		break;
	case C_ERROR_TIMEOUT:
		C_DEBUG(SERVICE_CLIENT_DEBUG_LEVEL, C_MSG_INFO, "SERVICE: Service \"%s\" is not reachable, dropping data update...\n", update->service_id);
		C_SERVICE_failed_data_updates++;
		C_SERVICE_failed_requests++;
		break;
	case C_ERROR_ABORTED:
		C_DEBUG(SERVICE_CLIENT_DEBUG_LEVEL, C_MSG_WARN, "SERVICE: Data update for service \"%s\" aborted...\n", update->service_id);
		C_SERVICE_failed_data_updates++;
		C_SERVICE_failed_requests++;
		break;
	case C_ERROR_BUSY:
		C_DEBUG(SERVICE_CLIENT_DEBUG_LEVEL, C_MSG_WARN, "SERVICE: SLP is busy, dropping data request...\n");
		C_SERVICE_failed_data_updates++;
		C_SERVICE_failed_requests++;
		break;
	default:
		C_ABORT_ON(true, "SERVICE: update_data() callback with error code %d.\n", slp_query_result->error_code);
	}

	FREE(update->data);
	FREE(update);
	FREE(slp_query_result);
}

/**
 * @brief Update specific service data of a service.
 *
 * Perform SLP query for service in case it is not contained in the service
 * cache.
 *
 * @param service_id ID of the service of which to update data.
 * @param offset Offset in service data.
 * @param size Size of service data to update.
 * @param data Pointer to new service data (which is freed internally).
 */
void update_data(const SERVICE_id_t* service_id, uint16_t offset, size_t size, char* data) {
	SLP_query_result_t* slp_query_result = NULL;
	SERVICE_data_update_t* update = NULL;

	C_ABORT_ON((size != 0) && (data == NULL), "SERVICE: Trying to update %d byte(s) at 0x%X of service \"%s\", but data pointer is NULL...\n", size, offset, service_id);

	C_DEBUG(SERVICE_CLIENT_DEBUG_LEVEL, C_MSG_NOTE, "SERVICE: Updating %d byte(s) at 0x%X of service \"%s\"...\n", size, offset, service_id);
#ifdef C_ENABLE_NAM_TRACE_OUTPUT
	indicate_activity();
#endif	// C_ENABLE_NAM_TRACE_OUTPUT
	C_SERVICE_total_data_updates++;
	C_SERVICE_total_requests++;

	MALLOC(slp_query_result, SLP_query_result_t);
	MALLOC(update, SERVICE_data_update_t);
	copy_service_id(update->service_id, service_id);
	update->offset = offset;
	update->size = size;
	MALLOC_ARRAY(update->data, char, size);
	memcpy(update->data, data, size);
	slp_query_result->parameter = update;
	SLP_locate_service(service_id, update_data_callback, slp_query_result);
}

/**
 * @brief Stream request callback function for SLP queries.
 *
 * Proceed with stream request if query was succesful, schedule retry if query
 * is in progress, abort otherwise.
 *
 * @param parameter SLP query result
 */
static void request_stream_callback(callback_parameter_t parameter) {
	SLP_query_result_t* slp_query_result = (SLP_query_result_t*) parameter;
	SERVICE_stream_request_t* request = (SERVICE_stream_request_t*) slp_query_result->parameter;

	switch(slp_query_result->error_code) {
	case C_ERROR_OK:
		C_DEBUG(SERVICE_CLIENT_DEBUG_LEVEL, C_MSG_INFO, "SERVICE: Using %s as server for service \"%s\"...\n", addr_to_str(slp_query_result->server), request->service_id);
		send_SERVICE_CLIENT_STREAM_REQUEST(slp_query_result->server, request->service_id, request->offset, request->size, request->interval, request->duration);
		break;
	case C_ERROR_PARAMETER:
		C_DEBUG(SERVICE_CLIENT_DEBUG_LEVEL, C_MSG_ERROR, "SERVICE: Wrong parameter passed to SLP, dropping data request...\n");
		C_SERVICE_failed_stream_requests++;
		C_SERVICE_failed_requests++;
		break;
	case C_ERROR_TIMEOUT:
		C_DEBUG(SERVICE_CLIENT_DEBUG_LEVEL, C_MSG_INFO, "SERVICE: Service \"%s\" is not reachable, dropping stream request...\n", request->service_id);
		C_SERVICE_failed_stream_requests++;
		C_SERVICE_failed_requests++;
		break;
	case C_ERROR_ABORTED:
		C_DEBUG(SERVICE_CLIENT_DEBUG_LEVEL, C_MSG_WARN, "SERVICE: Stream request for service \"%s\" aborted...\n", request->service_id);
		C_SERVICE_failed_stream_requests++;
		C_SERVICE_failed_requests++;
		break;
	case C_ERROR_BUSY:
		C_DEBUG(SERVICE_CLIENT_DEBUG_LEVEL, C_MSG_WARN, "SERVICE: SLP is busy, dropping data request...\n");
		C_SERVICE_failed_stream_requests++;
		C_SERVICE_failed_requests++;
		break;
	default:
		C_ABORT_ON(true, "SERVICE: request_stream() callback with error code %d.\n", slp_query_result->error_code);
	}

	FREE(request);
	FREE(slp_query_result);
}

/**
 * @brief Request a stream of specific service data of a service.
 *
 * Perform SLP query for service in case it is not contained in the service
 * cache.
 *
 * @param service_id ID of the service from which to request data.
 * @param offset Offset in service data.
 * @param size Size of requested service data.
 * @param interval Interval at which to send stream packets in seconds.
 * @param duration Time during which the stream is to be active in seconds.
 */
void request_stream(const SERVICE_id_t* service_id, uint16_t offset, size_t size, C_time_t interval, C_time_t duration) {
	SLP_query_result_t* slp_query_result = NULL;
	SERVICE_stream_request_t* request = NULL;

	C_DEBUG(SERVICE_CLIENT_DEBUG_LEVEL, C_MSG_NOTE, "SERVICE: Requesting a stream of %d byte(s) at 0x%X of service \"%s\" at interval %.2fs for a duration of %.2fs...\n", size, offset, service_id, interval, duration);
#ifdef C_ENABLE_NAM_TRACE_OUTPUT
	indicate_activity();
#endif	// C_ENABLE_NAM_TRACE_OUTPUT
	C_SERVICE_total_stream_requests++;
	C_SERVICE_total_requests++;

	// NOTE: With the current implementation, we cannot tell different
	// streams apart. As far as counting packets is concerned, this
	// shouldn't be a problem because RUDP allows us to tell apart whether
	// we have lost single packets or a stream request was never received.
	C_SERVICE_expected_stream_packets = ((uint32_t) (duration / interval)) + 1;

	MALLOC(slp_query_result, SLP_query_result_t);
	MALLOC(request, SERVICE_stream_request_t);
	copy_service_id(request->service_id, service_id);
	request->offset = offset;
	request->size = size;
	request->interval = interval;
	request->duration = duration;
	slp_query_result->parameter = request;
	SLP_locate_service(service_id, request_stream_callback, slp_query_result);
}
