/**
 * @file service_replication.c
 *
 * @brief Service replication.
 *
 * The replication protocol is as follows: Node A initiates replication to node
 * B by sending a SERVICE REPLICATION REQUEST packet via RUDP. If node B is
 * capable of host the service, it sends SERVICE REPLICATION ACK packets to
 * node A requesting segments of the service data of the service to be
 * replicated. Node A responds to these ACKs by sending the SERVICE REPLICATION
 * DATA packets with the requested segments. Once all service data segments
 * have been received by node B, it sends a SERVICE REPLICATION ACK with
 * segment size set to 0. Node A can then either proceed to shutdown the
 * service in case it was migrating it.
 *
 * Service migration works two phases: In the first phase the service is
 * replicated to the destination node, in the second phase the old host
 * redirects service request to the new host. The second phase terminates
 * after a timeout and the service is shutdown.
 *
 * Copyright (c) 2007, 2008, 2009, 2010 Georg Wittenburg, Freie Universitaet
 * Berlin (wittenbu@inf.fu-berlin.de). All rights reserved.
 * Please refer to the COPYING file for detailed licensing information.
 *
 * @author Georg Wittenburg
 *
 * $Id: service_replication.c 3386 2010-03-04 17:44:41Z gwitten $
 */

#include "../config.h"
#include "../macros.h"
#include "../types.h"
#include "../vars.h"
#include "../main.h"
#include "../util/debug.h"
#include "../util/callback.h"
#include "../util/net.h"
#include "../rudp-light/rudp-light.h"
#include "../slp-light/slp-light.h"
#include "../placement/placement.h"
#include "service.h"
#include "service_types.h"
#include "service_vars.h"
#include "service_replication.h"
#include "service_management.h"
#include "service_util.h"



/**
 * @brief Debug level for the replication component.
 */
#define SERVICE_REPLICATION_DEBUG_LEVEL	C_MSG_NOTE

/**
 * @brief Maximal segment size (in bytes) when configured to run on unreliable
 * links.
 */
#define SERVICE_REPLICATION_SEGMENT_SIZE_FOR_UNRELIABLE_LINKS	512

/**
 * @brief Timeout for service replication in seconds.
 */
#define SERVICE_REPLICATION_TIMEOUT	(2 * (C_PATH_DISCOVERY_TIME + (C_NETWORK_TRAVERSAL_TIME * service->data_size / RUDP_max_payload_size())))

/**
 * @brief Timeout for service redirection in seconds.
 *
 * Since redirection information is piggy-backed on replies to service request,
 * we should allow for the clients to make at least one requests. Otherwise,
 * the clients have to broadcast SLP queries to locate the new host.
 */
#ifdef SERVICE_MIGRATION_USE_REDIRECTION
	#define SERVICE_REDIRECTION_TIMEOUT	(2 * C_NETWORK_TRAVERSAL_TIME)
#endif	// SERVICE_MIGRATION_USE_REDIRECTION



/*
 * Utility functions.
 */

/**
 * @brief Allocate and initialize an outgoing replication structure.
 *
 * @param outgoing_replication Pointer to pointer to outgoing replication structure to allocate.
 * @param service Pointer to the service to replicate.
 * @param destination_address Address of the destination node.
 * @param is_migration Flag whether this replication is part of a migration.
 * @param callback Callback function.
 * @param result Pointer to replication result.
 */
static void alloc_outgoing_replication(SERVICE_outgoing_replication_t** outgoing_replication, SERVICE_service_t* service, C_addr_t destination_address, bool is_migration, callback_function_t callback, SERVICE_replication_result_t* result) {
	MALLOC(*outgoing_replication, SERVICE_outgoing_replication_t);
	(*outgoing_replication)->service = service;
	(*outgoing_replication)->destination_address = destination_address;
	(*outgoing_replication)->is_migration = is_migration;
	(*outgoing_replication)->callback = callback;
	(*outgoing_replication)->result = result;
	(*outgoing_replication)->result->error_code = C_ERROR_NOT_SET;
}

/**
 * @brief Free a outgoing replication structure.
 *
 * @param outgoing_replication Pointer to pointer to outgoing replication
 * structure to free.
 */
static void free_outgoing_replication(SERVICE_outgoing_replication_t** outgoing_replication) {
	FREE(*outgoing_replication);
}

/**
 * @brief Allocate and initialize an incoming replication structure.
 *
 * @param incoming_replication Pointer to pointer to incoming replication structure to allocate.
 * @param service Pointer to the service to replicate.
 * @param source_address Address of the source node.
 */
static void alloc_incoming_replication(SERVICE_incoming_replication_t** incoming_replication, SERVICE_service_t* service, C_addr_t source_address) {
#ifdef C_SPI_OPTIMIZED_FOR_UNRELIABLE_LINKS
	size_t max_segment_size = SERVICE_REPLICATION_SEGMENT_SIZE_FOR_UNRELIABLE_LINKS - sizeof(SERVICE_REPLICATION_DATA_packet_t);
#else
	size_t max_segment_size = RUDP_max_payload_size() - sizeof(SERVICE_REPLICATION_DATA_packet_t);
#endif	// C_SPI_OPTIMIZED_FOR_UNRELIABLE_LINKS
	size_t offset = 0;
	SERVICE_segment_t* segment = NULL;

	MALLOC(*incoming_replication, SERVICE_incoming_replication_t);
	(*incoming_replication)->service = service;
	(*incoming_replication)->source_address = source_address;
	list_alloc(&(*incoming_replication)->missing_segments);
	(*incoming_replication)->callback_id = 0;
	(*incoming_replication)->placement_algorithm_id = C_PLACEMENT_ALGORITHM_NONE;
	(*incoming_replication)->placement_algorithm_data = NULL;
	(*incoming_replication)->placement_algorithm_data_size = 0;

	// Split the service into segments for transmission and tag these
	// segments as missing.
	for(offset = 0x0000; offset < service->data_size; offset += max_segment_size) {
		MALLOC(segment, SERVICE_segment_t);
		segment->offset = offset;
		segment->size = service->data_size - offset;
		if(segment->size > max_segment_size)
			segment->size = max_segment_size;
		C_DEBUG(SERVICE_REPLICATION_DEBUG_LEVEL, C_MSG_DEBUG, "SERVICE: Adding segment with offset 0x%X and size %d to missing segments list of service \"%s\"...\n", segment->offset, segment->size, service->id);
		list_add((*incoming_replication)->missing_segments, segment);
		segment = NULL;
	}
}

/**
 * @brief Free a incoming replication structure.
 *
 * @param incoming_replication Pointer to pointer to incoming replication
 * structure to free.
 */
static void free_incoming_replication(SERVICE_incoming_replication_t** incoming_replication) {
	list_free_including_data(&(*incoming_replication)->missing_segments, free);
	if((*incoming_replication)->placement_algorithm_data != NULL)
		FREE((*incoming_replication)->placement_algorithm_data);
	FREE(*incoming_replication);
}

/**
 * @brief Abort an outgoing replication of a service.
 *
 * @param outgoing_replication Pointer to the outgoing replication.
 * @param error_code Error code to use in the result of the callback.
 */
void abort_outgoing_replication(SERVICE_outgoing_replication_t* outgoing_replication, C_errno_t error_code) {
	C_ABORT_ON(outgoing_replication == NULL, "SERVICE: Trying to abort outgoing replication at NULL.\n");
	C_ABORT_ON(!exists_service(outgoing_replication->service), "SERVICE: Trying to abort outgoing replication, but service at %p does not exist.\n", outgoing_replication->service);
	C_ABORT_ON(outgoing_replication->service->state != SERVICE_STATE_REPLICATING, "SERVICE: Trying to abort outgoing replication of service \"%s\", but service is not replicating (%d).\n", outgoing_replication->service->id, outgoing_replication->service->state);

	C_DEBUG(SERVICE_REPLICATION_DEBUG_LEVEL, C_MSG_INFO, "SERVICE: Aborting outgoing replication of service \"%s\" to %s...\n", outgoing_replication->service->id, addr_to_str(outgoing_replication->destination_address));

	outgoing_replication->service->state = SERVICE_STATE_LOCKED;
	outgoing_replication->service->outgoing_replication = NULL;
	if(outgoing_replication->callback_id != 0)
		callback_cancel(outgoing_replication->callback_id);
	outgoing_replication->result->error_code = error_code;
	outgoing_replication->callback(outgoing_replication->result);
	list_remove(C_SERVICE_outgoing_replications, outgoing_replication);
	free_outgoing_replication(&outgoing_replication);
}

/**
 * @brief Abort an incoming replication of a service.
 *
 * @param incoming_replication Pointer to the incoming replication.
 */
void abort_incoming_replication(SERVICE_incoming_replication_t* incoming_replication) {
	C_ABORT_ON(incoming_replication == NULL, "SERVICE: Trying to abort incoming replication at NULL.\n");
	C_ABORT_ON(!exists_service(incoming_replication->service), "SERVICE: Trying to abort incoming replication, but service at %p does not exist.\n", incoming_replication->service);
	C_ABORT_ON(incoming_replication->service->state != SERVICE_STATE_INIT, "SERVICE: Trying to abort incoming replication of service \"%s\", but service is not initializing (%d).\n", incoming_replication->service->id, incoming_replication->service->state);

	C_DEBUG(SERVICE_REPLICATION_DEBUG_LEVEL, C_MSG_INFO, "SERVICE: Aborting incoming replication of service \"%s\" from %s...\n", incoming_replication->service->id, addr_to_str(incoming_replication->source_address));

	list_remove(C_SERVICE_incoming_replications, incoming_replication);
	if(incoming_replication->callback_id != 0)
		callback_cancel(incoming_replication->callback_id);
	free_incoming_replication(&incoming_replication);
}



/*
 * Packet handling functions.
 */

/**
 * @brief RUDP callback for a SERVICE REPLICATION REQUEST packet.
 *
 * @param parameter Pointer to RUDP transmission result.
 */
static void send_SERVICE_SERVER_REPLICATION_REQUEST_callback(callback_parameter_t parameter) {
	RUDP_transmission_result_t* result = (RUDP_transmission_result_t*) parameter;
	SERVICE_outgoing_replication_t* replication = (SERVICE_outgoing_replication_t*) result->parameter;

	if(list_contains(C_SERVICE_outgoing_replications, replication)){

		C_ABORT_ON(!exists_service(replication->service), "SERVICE: Trying to handle REPLICATION REQUEST callback, but service at %p does not exist.\n", replication->service);
		C_ABORT_ON(replication->service->state != SERVICE_STATE_REPLICATING, "SERVICE: Trying to handle REPLICATION REQUEST callback for service \"%s\", but service is not replicating (%d).\n", replication->service->id, replication->service->state);

		switch(result->error_code) {
		case C_ERROR_OK:
			// Nothing to do, segments will be requested by new server.
			break;
		case C_ERROR_TIMEOUT:
			C_DEBUG(SERVICE_REPLICATION_DEBUG_LEVEL, C_MSG_WARN, "SERVICE: REPLICATION REQUEST timed out, replication of service \"%s\" failed.\n", replication->service->id);
			abort_outgoing_replication(replication, C_ERROR_COMM_ERROR);
			C_SERVICE_failed_replications++;
			break;
		case C_ERROR_ABORTED:
			C_DEBUG(SERVICE_REPLICATION_DEBUG_LEVEL, C_MSG_WARN, "SERVICE: Sending REPLICATION REQUEST aborted, replication of service \"%s\" failed.\n", replication->service->id);
			abort_outgoing_replication(replication, C_ERROR_ABORTED);
			C_SERVICE_failed_replications++;
			break;
		case C_ERROR_BUSY:
			C_DEBUG(SERVICE_REPLICATION_DEBUG_LEVEL, C_MSG_WARN, "SERVICE: REPLICATION REQUEST hit full queue, replication of service \"%s\" failed.\n", replication->service->id);
			abort_outgoing_replication(replication, C_ERROR_BUSY);
			C_SERVICE_failed_replications++;
			break;
		default:
			C_ABORT_ON(true, "SERVICE: Unhandled error while sending REPLICATION REQUEST with RUDP (%d).\n", result->error_code);
		}
	} else {
		C_DEBUG(SERVICE_REPLICATION_DEBUG_LEVEL, C_MSG_DEBUG, "SERVICE: Callback on REPLICATION REQUEST for previously cancelled replication.\n");
	}
	FREE(result);
}

/**
 * @brief Send a SERVICE REPLICATION REQUEST packet.
 *
 * @param replication Pointer to outgoing replication information.
 */
static void send_SERVICE_SERVER_REPLICATION_REQUEST(SERVICE_outgoing_replication_t* replication) {
	SERVICE_REPLICATION_REQUEST_packet_t request;
	char* placement_algorithm_data = NULL;
	size_t placement_algorithm_data_size = 0;
	RUDP_transmission_result_t* result = NULL;

	C_DEBUG(SERVICE_REPLICATION_DEBUG_LEVEL, C_MSG_INFO, "SERVICE: Sending REPLICATION REQUEST for %d byte(s) of service \"%s\" to %s...\n", replication->service->data_size, replication->service->id, addr_to_str(replication->destination_address));
	memset(&request, 0, sizeof(SERVICE_REPLICATION_REQUEST_packet_t));
	copy_service_id(request.service_id, replication->service->id);
	request.service_data_size = replication->service->data_size;
	request.is_migration = replication->is_migration;
	request.sync_traffic_fraction = replication->service->sync_traffic_fraction;
	if(replication->service->placement_algorithm != NULL)
		request.placement_algorithm_id = replication->service->placement_algorithm->id;
	else
		request.placement_algorithm_id = C_PLACEMENT_ALGORITHM_NONE;
	placement_algorithm_data_size = PLACEMENT_get_algorithm_data(replication->service, &placement_algorithm_data);
	MALLOC(result, RUDP_transmission_result_t);
	result->parameter = replication;
	RUDP_sendto2(replication->destination_address, PACKET_PRIORITY_SIGNALLING_SERVICE, APP_TYPE_SERVICE, SERVICE_PACKET_TYPE_REPLICATION_REQUEST, &request, sizeof(SERVICE_REPLICATION_REQUEST_packet_t), placement_algorithm_data, placement_algorithm_data_size, send_SERVICE_SERVER_REPLICATION_REQUEST_callback, result);
	FREE(placement_algorithm_data);
}

/**
 * @brief RUDP callback for a SERVICE REPLICATION DATA packet.
 *
 * @param parameter Pointer to RUDP transmission result.
 */
static void send_SERVICE_REPLICATION_DATA_callback(callback_parameter_t parameter) {
	RUDP_transmission_result_t* result = (RUDP_transmission_result_t*) parameter;
	SERVICE_outgoing_replication_t* replication = (SERVICE_outgoing_replication_t*) result->parameter;

	if(list_contains(C_SERVICE_outgoing_replications, replication)){

		C_ABORT_ON(!exists_service(replication->service), "SERVICE: Trying to handle REPLICATION DATA callback, but service at %p does not exist.\n", replication->service);
		C_ABORT_ON(replication->service->state != SERVICE_STATE_REPLICATING, "SERVICE: Trying to handle REPLICATION DATA callback for service \"%s\", but service is not replicating (%d).\n", replication->service->id, replication->service->state);

		switch(result->error_code) {
		case C_ERROR_OK:
			// Nothing to do, more segments will be requested by new server.
			break;
		case C_ERROR_TIMEOUT:
			C_DEBUG(SERVICE_REPLICATION_DEBUG_LEVEL, C_MSG_WARN, "SERVICE: REPLICATION DATA timed out, replication of service \"%s\" failed.\n", replication->service->id);
			abort_outgoing_replication(replication, C_ERROR_COMM_ERROR);
			C_SERVICE_failed_replications++;
			break;
		case C_ERROR_ABORTED:
			C_DEBUG(SERVICE_REPLICATION_DEBUG_LEVEL, C_MSG_WARN, "SERVICE: Sending REPLICATION DATA aborted, replication of service \"%s\" failed.\n", replication->service->id);
			abort_outgoing_replication(replication, C_ERROR_ABORTED);
			C_SERVICE_failed_replications++;
			break;
		case C_ERROR_BUSY:
			C_DEBUG(SERVICE_REPLICATION_DEBUG_LEVEL, C_MSG_WARN, "SERVICE: REPLICATION DATA hit full queue, replication of service \"%s\" failed.\n", replication->service->id);
			abort_outgoing_replication(replication, C_ERROR_BUSY);
			C_SERVICE_failed_replications++;
			break;
		default:
			C_ABORT_ON(true, "SERVICE: Unhandled error while sending REPLICATION DATA with RUDP (%d).\n", result->error_code);
		}
	} else {
		C_DEBUG(SERVICE_REPLICATION_DEBUG_LEVEL, C_MSG_DEBUG, "SERVICE: Callback on REPLICATION DATA for previously cancelled replication.\n");
	}
	FREE(result);
}

/**
 * @brief Send a SERVICE REPLICATION DATA packet.
 *
 * @param replication Pointer to outgoing replication information.
 * @param segment Pointer to segment to send.
 */
static void send_SERVICE_REPLICATION_DATA(SERVICE_outgoing_replication_t* replication, const SERVICE_segment_t* segment) {
	SERVICE_service_t* service = replication->service;
	SERVICE_REPLICATION_DATA_packet_t data_header;
	RUDP_transmission_result_t* result = NULL;

	C_DEBUG(SERVICE_REPLICATION_DEBUG_LEVEL, C_MSG_INFO, "SERVICE: Sending REPLICATION DATA for %d byte(s) at offset 0x%X of service \"%s\" to %s...\n", segment->size, segment->offset, service->id, addr_to_str(replication->destination_address));
	memset(&data_header, 0, sizeof(SERVICE_REPLICATION_DATA_packet_t));
	copy_service_id(data_header.service_id, service->id);
	data_header.segment_offset = segment->offset;
	MALLOC(result, RUDP_transmission_result_t);
	result->parameter = replication;
	RUDP_sendto2(replication->destination_address, PACKET_PRIORITY_SIGNALLING_SERVICE, APP_TYPE_SERVICE, SERVICE_PACKET_TYPE_REPLICATION_DATA, &data_header, sizeof(SERVICE_REPLICATION_DATA_packet_t), (service->data + segment->offset), segment->size, send_SERVICE_REPLICATION_DATA_callback, result);
}

static void send_SERVICE_REPLICATION_NACK(C_addr_t to, const char* service_id, C_errno_t error_code);

/**
 * @brief RUDP callback for a SERVICE REPLICATION ACK packet.
 *
 * @param parameter Pointer to RUDP transmission result.
 */
static void send_SERVICE_REPLICATION_ACK_callback(callback_parameter_t parameter) {
	RUDP_transmission_result_t* result = (RUDP_transmission_result_t*) parameter;
	SERVICE_incoming_replication_t* replication = (SERVICE_incoming_replication_t*) result->parameter;
	SERVICE_service_t* service = NULL;

	if(list_contains(C_SERVICE_incoming_replications, replication)) {
		service = replication->service;
		C_ABORT_ON(service == NULL, "SERVICE: Service is NULL in callback on REPLICATION ACK for incoming replication from %s.\n", addr_to_str(replication->source_address));

		switch(result->error_code) {
		case C_ERROR_OK:
			// Nothing to do, next segment will be sent by old server.
			break;
		case C_ERROR_TIMEOUT:
			C_DEBUG(SERVICE_REPLICATION_DEBUG_LEVEL, C_MSG_WARN, "SERVICE: REPLICATION ACK timed out, replication of service \"%s\" failed.\n", replication->service->id);
			send_SERVICE_REPLICATION_NACK(replication->source_address, replication->service->id, C_ERROR_TIMEOUT);
			abort_incoming_replication(replication);
			free_service(&service);
			break;
		case C_ERROR_ABORTED:
			C_DEBUG(SERVICE_REPLICATION_DEBUG_LEVEL, C_MSG_WARN, "SERVICE: Sending REPLICATION ACK aborted, replication of service \"%s\" failed.\n", replication->service->id);
			abort_incoming_replication(replication);
			free_service(&service);
			break;
		case C_ERROR_BUSY:
			C_DEBUG(SERVICE_REPLICATION_DEBUG_LEVEL, C_MSG_WARN, "SERVICE: REPLICATION ACK hit full queue, replication of service \"%s\" failed.\n", replication->service->id);
			abort_incoming_replication(replication);
			free_service(&service);
			break;
		default:
			C_ABORT_ON(true, "SERVICE: Unhandled error while sending REPLICATION ACK with RUDP (%d).\n", result->error_code);
		}
	} else {
		C_DEBUG(SERVICE_REPLICATION_DEBUG_LEVEL, C_MSG_DEBUG, "SERVICE: Callback on REPLICATION ACK for previously completed replication.\n");
	}
	FREE(result);
}

/**
 * @brief Send a SERVICE REPLICATION ACK packet.
 *
 * @param replication Pointer to incoming replication information.
 * @param segment Pointer to segment to request.
 */
static void send_SERVICE_REPLICATION_ACK(SERVICE_incoming_replication_t* replication, const SERVICE_segment_t* segment) {
	SERVICE_service_t* service = replication->service;
	SERVICE_REPLICATION_ACK_packet_t ack;
	RUDP_transmission_result_t* result = NULL;

	C_DEBUG(SERVICE_REPLICATION_DEBUG_LEVEL, C_MSG_INFO, "SERVICE: Sending REPLICATION ACK requesting %d byte(s) at offset 0x%X of service \"%s\" to %s...\n", segment->size, segment->offset, service->id, addr_to_str(replication->source_address));
	memset(&ack, 0, sizeof(SERVICE_REPLICATION_ACK_packet_t));
	copy_service_id(ack.service_id, service->id);
	memcpy(&ack.next_segment, segment, sizeof(SERVICE_segment_t));
	MALLOC(result, RUDP_transmission_result_t);
	result->parameter = replication;
	RUDP_sendto(replication->source_address, PACKET_PRIORITY_SIGNALLING_SERVICE, APP_TYPE_SERVICE, SERVICE_PACKET_TYPE_REPLICATION_ACK, &ack, sizeof(SERVICE_REPLICATION_ACK_packet_t), send_SERVICE_REPLICATION_ACK_callback, result);
}

/**
 * @brief Send a SERVICE REPLICATION NACK packet.
 *
 * @param to Address of destination node.
 * @param service_id ID of the service for which the replication has failed.
 * @param error_code Error code.
 */
static void send_SERVICE_REPLICATION_NACK(C_addr_t to, const SERVICE_id_t* service_id, C_errno_t error_code) {
	SERVICE_REPLICATION_NACK_packet_t nack;

	C_DEBUG(SERVICE_REPLICATION_DEBUG_LEVEL, C_MSG_INFO, "SERVICE: Sending REPLICATION NACK for service \"%s\" to %s...\n", service_id, addr_to_str(to));
	memset(&nack, 0, sizeof(SERVICE_REPLICATION_NACK_packet_t));
	copy_service_id(nack.service_id, service_id);
	nack.error_code = error_code;
	RUDP_sendto(to, PACKET_PRIORITY_SIGNALLING_SERVICE, APP_TYPE_SERVICE, SERVICE_PACKET_TYPE_REPLICATION_NACK, &nack, sizeof(SERVICE_REPLICATION_NACK_packet_t), NULL, NULL);
}

/**
 * @brief Timeout for handled incoming SERVICE REPLICATION REQUEST packet.
 *
 * If the service is does not exist anymore or is still initializing then we
 * abort the replication and free the resources.
 *
 * @param parameter Incoming replication information.
 */
static void handle_SERVICE_REPLICATION_REQUEST_timeout(callback_parameter_t parameter) {
	SERVICE_incoming_replication_t* replication = (SERVICE_incoming_replication_t*) parameter;
	SERVICE_service_t* service = replication->service;

	C_ABORT_ON(!exists_service(service), "SERVICE: Trying to handle timed out REPLICATION REQUEST, but service at %p does not exist.\n", service);
	C_ABORT_ON(service->state != SERVICE_STATE_INIT, "SERVICE: Trying to handle timed out REPLICATION REQUEST for service \"%s\", but service is not initializing (%d).\n", service->id, service->state);

	C_DEBUG(SERVICE_REPLICATION_DEBUG_LEVEL, C_MSG_WARN, "SERVICE: Replication of service \"%s\" from %s timed out, freeing resources...\n", service->id, addr_to_str(service->incoming_replication->source_address));
	replication->callback_id = 0;
	abort_incoming_replication(replication);
	free_service(&service);
}

/**
 * @brief Handle incoming SERVICE REPLICATION REQUEST packet.
 *
 * If this node can host the service then setup a list of missing segments and
 * start requesting them by sending a first SERVICE REPLICATION ACK.
 *
 * @param header Header of incoming packet.
 * @param request Replication request information.
 * @param placement_algorithm_data Pointer to data of the placement algorithm.
 * @param placement_algorithm_data_size Size of the data for the placement algorithm in bytes.
 */
void handle_SERVICE_REPLICATION_REQUEST(const net_header_t* header, const SERVICE_REPLICATION_REQUEST_packet_t* request, void* placement_algorithm_data, size_t placement_algorithm_data_size) {
	SERVICE_incoming_replication_t* replication = NULL;
	SERVICE_service_t* service = NULL;

	if(list_size(C_SERVICE_hosted_services) >= SERVICE_MAX_HOSTED_SERVICES) {
		C_DEBUG(SERVICE_REPLICATION_DEBUG_LEVEL, C_MSG_ERROR, "SERVICE: Received REPLICATION REQUEST for service \"%s\" from %s, but no this node already hosts %d services.\n", request->service_id, addr_to_str(header->source_address), SERVICE_MAX_HOSTED_SERVICES);
		send_SERVICE_REPLICATION_NACK(header->source_address, request->service_id, C_ERROR_BUSY);
		return;
	}
	if((service = get_service(request->service_id)) != NULL) {
		// Accept the replication only if the existing service is not
		// active, i.e., initializing or stopped.
		if(service->state == SERVICE_STATE_INIT) {
			if(service->incoming_replication != NULL) {
				C_DEBUG(SERVICE_REPLICATION_DEBUG_LEVEL, C_MSG_WARN, "SERVICE: Aborting presumably failed incoming replication of service \"%s\" from %s in order to accept new replication from %s...\n", service->id, addr_to_str(service->incoming_replication->source_address), addr_to_str(header->source_address));
				abort_incoming_replication(service->incoming_replication);
			}
			free_service(&service);
		} else if(service->state == SERVICE_STATE_STOPPED) {
			C_DEBUG(SERVICE_REPLICATION_DEBUG_LEVEL, C_MSG_WARN, "SERVICE: Freeing stopped service \"%s\" in order to accept new replication...\n", service->id);
			free_service(&service);
		} else {
			C_DEBUG(SERVICE_REPLICATION_DEBUG_LEVEL, C_MSG_WARN, "SERVICE: Received REPLICATION REQUEST for service \"%s\" from %s, but service is already running on this node.\n", request->service_id, addr_to_str(header->source_address));
			send_SERVICE_REPLICATION_NACK(header->source_address, request->service_id, C_ERROR_PARAMETER);
			return;
		}
	}

	C_DEBUG(SERVICE_REPLICATION_DEBUG_LEVEL, C_MSG_INFO, "SERVICE: Received REPLICATION REQUEST for service \"%s\" from %s...\n", request->service_id, addr_to_str(header->source_address));
	alloc_service(&service, request->service_id, request->service_data_size, request->sync_traffic_fraction);
	if(request->is_migration)
		service->replace_addr = header->source_address;
	alloc_incoming_replication(&replication, service, header->source_address);
	replication->placement_algorithm_id = request->placement_algorithm_id;
	if(placement_algorithm_data_size > 0) {
		MALLOC_ARRAY(replication->placement_algorithm_data, char, placement_algorithm_data_size);
		memcpy(replication->placement_algorithm_data, placement_algorithm_data, placement_algorithm_data_size);
		replication->placement_algorithm_data_size = placement_algorithm_data_size;
	}
	service->incoming_replication = replication;
	list_add(C_SERVICE_incoming_replications, replication);

	replication->callback_id = callback_add(handle_SERVICE_REPLICATION_REQUEST_timeout, replication, SERVICE_REPLICATION_TIMEOUT);
	send_SERVICE_REPLICATION_ACK(replication, (SERVICE_segment_t*) list_head(replication->missing_segments));
}

/**
 * @brief Check whether two service data segments match.
 *
 * @param item Segment to check.
 * @param cmp_item Reference segment.
 * @return <code>true</code> if the two segments match, <code>false</code>
 * otherwise.
 */
static bool matches_segment(const void* item, const void* cmp_item) {
	SERVICE_segment_t* segment = (SERVICE_segment_t*) item;
	SERVICE_segment_t* cmp_segment = (SERVICE_segment_t*) cmp_item;
	return (
		(segment->offset == cmp_segment->offset)
		&& (segment->size == cmp_segment->size)
	);
}

/**
 * @brief Handle incoming SERVICE REPLICATION DATA packet.
 *
 * @param header Header of the incoming packet.
 * @param data_header REPLICATION DATA header.
 * @param segment_data Segement data.
 * @param segment_size Size of segment data.
 */
void handle_SERVICE_REPLICATION_DATA(const net_header_t* header, const SERVICE_REPLICATION_DATA_packet_t* data_header, const char* segment_data, size_t segment_size) {
	SERVICE_segment_t segment;
	SERVICE_service_t* service = NULL;
	SERVICE_incoming_replication_t* replication = NULL;
	SERVICE_segment_t* missing_segment = NULL;

	segment.size = segment_size;
	segment.offset = data_header->segment_offset;
	service = get_service(data_header->service_id);
	if(service == NULL) {	// May occur if replication timed out, but source didn't receive NACK.
		C_DEBUG(SERVICE_REPLICATION_DEBUG_LEVEL, C_MSG_WARN, "SERVICE: Received REPLICATION DATA packet for %d byte(s) at offset 0x%X of service \"%s\" from %s, but service is unknown...\n", segment.size, segment.offset, data_header->service_id, addr_to_str(header->source_address));
		return;
	} else if(service->state != SERVICE_STATE_INIT) {	// May occur on retransmitted REPLICATION DATA packets.
		C_DEBUG(SERVICE_REPLICATION_DEBUG_LEVEL, C_MSG_WARN, "SERVICE: Received REPLICATION DATA packet for %d byte(s) at offset 0x%X of service \"%s\" from %s, but service is not initializing...\n", segment.size, segment.offset, data_header->service_id, addr_to_str(header->source_address));
		return;
	} else if(header->source_address != service->incoming_replication->source_address) {
		C_DEBUG(SERVICE_REPLICATION_DEBUG_LEVEL, C_MSG_WARN, "SERVICE: Received REPLICATION DATA packet for %d byte(s) at offset 0x%X of service \"%s\" from %s, but no incoming replication is active...\n", segment.size, segment.offset, data_header->service_id, addr_to_str(header->source_address));
		return;
	} else if(header->source_address != service->incoming_replication->source_address) {
		C_DEBUG(SERVICE_REPLICATION_DEBUG_LEVEL, C_MSG_ERROR, "SERVICE: Received REPLICATION DATA packet for %d byte(s) at offset 0x%X of service \"%s\" from %s, but service replication was initialized by %s...\n", segment.size, segment.offset, data_header->service_id, addr_to_str(header->source_address), addr_to_str(service->incoming_replication->source_address));
		return;
	}

	replication = service->incoming_replication;
	missing_segment = (SERVICE_segment_t*) list_find(replication->missing_segments, &segment, matches_segment);
	if(missing_segment != NULL) {
		C_DEBUG(SERVICE_REPLICATION_DEBUG_LEVEL, C_MSG_INFO, "SERVICE: Received REPLICATION DATA packet for %d byte(s) at offset 0x%X of service \"%s\" from %s...\n", segment.size, segment.offset, data_header->service_id, addr_to_str(header->source_address));
		update_service_data(service, segment.offset, segment.size, segment_data);
		list_remove(replication->missing_segments, missing_segment);
		FREE(missing_segment);
	} else {
		C_DEBUG(SERVICE_REPLICATION_DEBUG_LEVEL, C_MSG_WARN, "SERVICE: Received REPLICATION DATA packet for %d byte(s) at offset 0x%X of service \"%s\" from %s, but segment was not missing...\n", segment.size, segment.offset, data_header->service_id, addr_to_str(header->source_address));
	}

	if(!list_empty(replication->missing_segments)) {
		send_SERVICE_REPLICATION_ACK(replication, (SERVICE_segment_t*) list_head(replication->missing_segments));
	} else {
		// All segments have been transfered, send ACK with segment.size = 0.
		SERVICE_segment_t segment;
		memset(&segment, 0, sizeof(SERVICE_segment_t));
		send_SERVICE_REPLICATION_ACK(replication, &segment);
		service->incoming_replication = NULL;
		list_remove(C_SERVICE_incoming_replications, replication);
		callback_cancel(replication->callback_id);
		start_service(service, replication->placement_algorithm_id);
		PLACEMENT_set_algorithm_data(service, replication->placement_algorithm_data, replication->placement_algorithm_data_size);
		free_incoming_replication(&replication);
		C_DEBUG(SERVICE_REPLICATION_DEBUG_LEVEL, C_MSG_INFO, "SERVICE: Replication of service \"%s\" successful.\n", service->id);
	}
}

/**
 * @brief Handle incoming SERVICE REPLICATION ACK packet.
 *
 * @param header Header of incoming packet.
 * @param ack Acknowledgement information either requesting another segment of
 * service data or confirming the successful replication.
 */
void handle_SERVICE_REPLICATION_ACK(const net_header_t* header, const SERVICE_REPLICATION_ACK_packet_t* ack) {
	SERVICE_outgoing_replication_t* replication = NULL;

	SERVICE_service_t* service = get_service(ack->service_id);
	if(service == NULL) {	// May occur on delayed REPLICATION ACKs.
		C_DEBUG(SERVICE_REPLICATION_DEBUG_LEVEL, C_MSG_WARN, "SERVICE: Received REPLICATION ACK requesting %d byte(s) offset 0x%X of service \"%s\" from %s, but service is unknown...\n", ack->next_segment.size, ack->next_segment.offset, ack->service_id, addr_to_str(header->source_address));
		return;
	} else if(service->state != SERVICE_STATE_REPLICATING) {	// May occur on delayed REPLICATION ACKs.
		C_DEBUG(SERVICE_REPLICATION_DEBUG_LEVEL, C_MSG_WARN, "SERVICE: Received REPLICATION ACK requesting %d byte(s) offset 0x%X of service \"%s\" from %s, but service is not replicating...\n", ack->next_segment.size, ack->next_segment.offset, ack->service_id, addr_to_str(header->source_address));
		return;
	} else if((ack->next_segment.offset + ack->next_segment.size) > service->data_size) {
		C_DEBUG(SERVICE_REPLICATION_DEBUG_LEVEL, C_MSG_ERROR, "SERVICE: Received REPLICATION ACK requesting %d byte(s) offset 0x%X of service \"%s\" from %s, but service data size is only %d...\n", ack->next_segment.size, ack->next_segment.offset, ack->service_id, addr_to_str(header->source_address), service->data_size);
		return;
	} else if(header->source_address != service->outgoing_replication->destination_address) {	// May occur on delayed REPLICATION ACKs.
		C_DEBUG(SERVICE_REPLICATION_DEBUG_LEVEL, C_MSG_WARN, "SERVICE: Received REPLICATION ACK requesting %d byte(s) offset 0x%X of service \"%s\" from %s, but service replication was initialized to %s...\n", ack->next_segment.size, ack->next_segment.offset, ack->service_id, addr_to_str(header->source_address), addr_to_str(service->outgoing_replication->destination_address));
		return;
	}

	replication = service->outgoing_replication;
	C_DEBUG(SERVICE_REPLICATION_DEBUG_LEVEL, C_MSG_INFO, "SERVICE: Received REPLICATION ACK requesting %d byte(s) offset 0x%X of service \"%s\" from %s...\n", ack->next_segment.size, ack->next_segment.offset, ack->service_id, addr_to_str(header->source_address));
	if(ack->next_segment.size != 0) {
		send_SERVICE_REPLICATION_DATA(replication, &ack->next_segment);
	} else {
		// All segments have been transfered.
		C_DEBUG(SERVICE_REPLICATION_DEBUG_LEVEL, C_MSG_INFO, "SERVICE: Replication of service \"%s\" to %s is complete.\n", service->id, addr_to_str(header->source_address));
		service->state = SERVICE_STATE_LOCKED;
		service->outgoing_replication = NULL;
		replication->result->error_code = C_ERROR_OK;
		replication->callback(replication->result);
		list_remove(C_SERVICE_outgoing_replications, replication);
		callback_cancel(replication->callback_id);
		free_outgoing_replication(&replication);
		C_SERVICE_successful_replications++;
	}
}

/**
 * @brief Handle incoming SERVICE REPLICATION NACK packet.
 *
 * @param header Header of the incoming packet.
 * @param nack Information about why the replication failed.
 */
void handle_SERVICE_REPLICATION_NACK(const net_header_t* header, const SERVICE_REPLICATION_NACK_packet_t* nack) {
	SERVICE_service_t* service = get_service(nack->service_id);

	if(service == NULL) {
		C_DEBUG(SERVICE_REPLICATION_DEBUG_LEVEL, C_MSG_WARN, "SERVICE: Received REPLICATION NACK for service \"%s\" from %s, but that service is not hosted on this node...\n", nack->service_id, addr_to_str(header->source_address));
		return;
	} else if(service->state != SERVICE_STATE_REPLICATING) {
		C_DEBUG(SERVICE_REPLICATION_DEBUG_LEVEL, C_MSG_WARN, "SERVICE: Received REPLICATION NACK for service \"%s\" from %s, but that service is not replicating...\n", nack->service_id, addr_to_str(header->source_address));
		return;
	}
	C_ABORT_ON(service->outgoing_replication == NULL, "SERVICE: Received REPLICATION NACK for service \"%s\" from %s, but the outgoing replication of that service is NULL.\n", nack->service_id, addr_to_str(header->source_address));

	switch(nack->error_code) {
	case C_ERROR_PARAMETER:
		C_DEBUG(SERVICE_REPLICATION_DEBUG_LEVEL, C_MSG_WARN, "SERVICE: Received REPLICATION NACK for service \"%s\" from %s reporting that the service is already running on the remote node.\n", service->id, addr_to_str(header->source_address));
		service->outgoing_replication->result->error_code = C_ERROR_PARAMETER;
		break;
	case C_ERROR_TIMEOUT:
		C_DEBUG(SERVICE_REPLICATION_DEBUG_LEVEL, C_MSG_WARN, "SERVICE: Received REPLICATION NACK for service \"%s\" from %s reporting that the replication has timed out.\n", service->id, addr_to_str(header->source_address));
		service->outgoing_replication->result->error_code = C_ERROR_COMM_ERROR;
		break;
	case C_ERROR_BUSY:
		C_DEBUG(SERVICE_REPLICATION_DEBUG_LEVEL, C_MSG_WARN, "SERVICE: Received REPLICATION NACK for service \"%s\" from %s reporting that the remote node has refused to host the service.\n", service->id, addr_to_str(header->source_address));
		service->outgoing_replication->result->error_code = C_ERROR_REMOTE_BUSY;
		break;
	default:
		C_ABORT_ON(true, "SERVICE: Received REPLICATION NACK for service \"%s\" from %s with illegal error code (%d).\n", service->id, addr_to_str(header->source_address), nack->error_code);
	}

	abort_outgoing_replication(service->outgoing_replication, service->outgoing_replication->result->error_code);
	C_SERVICE_failed_replications++;
}



/*
 * Control function.
 */

/**
 * @brief Timeout function for service replication.
 *
 * @param parameter Pointer to outgoing replication information.
 */
static void replicate_service_timeout(callback_parameter_t parameter) {
	SERVICE_outgoing_replication_t* replication = (SERVICE_outgoing_replication_t*) parameter;
	SERVICE_service_t* service = replication->service;

	C_ABORT_ON(!exists_service(service), "SERVICE: Replication has timed out, but service at %p does not exist.\n", service);
	C_ABORT_ON(service->state != SERVICE_STATE_REPLICATING, "SERVICE: Replication of service \"%s\" has timed out, but service is not replicating (%d).\n", service->id, service->state);

	C_DEBUG(SERVICE_REPLICATION_DEBUG_LEVEL, C_MSG_WARN, "SERVICE: Replication of service \"%s\" timed out.\n", service->id);
	replication->callback_id = 0;
	abort_outgoing_replication(replication, C_ERROR_TIMEOUT);
	C_SERVICE_failed_replications++;
}

/**
 * @brief Replicate a service from this node to another node.
 *
 * @param service Pointer to the service to replicate.
 * @param destination_address Address of the destination node.
 * @param is_migration Flag whether this replication is part of a migration.
 * @param callback Callback function.
 * @param result Pointer to replication result.
 */
void replicate_service(SERVICE_service_t* service, C_addr_t destination_address, bool is_migration, callback_function_t callback, SERVICE_replication_result_t* result) {
	SERVICE_outgoing_replication_t* replication = NULL;

	if(service->state == SERVICE_STATE_REPLICATING) {
		C_DEBUG(SERVICE_REPLICATION_DEBUG_LEVEL, C_MSG_ERROR, "SERVICE: Trying to replicate service \"%s\" to %s, but service is already replicating (%d).\n", service->id, addr_to_str(destination_address), service->state);
		result->error_code = C_ERROR_IN_PROGRESS;
		callback(result);
		return;
	} else if(service->state != SERVICE_STATE_LOCKED) {
		C_DEBUG(SERVICE_REPLICATION_DEBUG_LEVEL, C_MSG_ERROR, "SERVICE: Trying to replicate service \"%s\" to %s, but service is not locked (%d).\n", service->id, addr_to_str(destination_address), service->state);
		result->error_code = C_ERROR_PARAMETER;
		callback(result);
		return;
	}

	C_DEBUG(SERVICE_REPLICATION_DEBUG_LEVEL, C_MSG_NOTE, "SERVICE: Replicating service \"%s\" to %s...\n", service->id, addr_to_str(destination_address));
	service->state = SERVICE_STATE_REPLICATING;
	C_SERVICE_total_replications++;

	alloc_outgoing_replication(&replication, service, destination_address, is_migration, callback, result);
	service->outgoing_replication = replication;
	list_add(C_SERVICE_outgoing_replications, replication);

	replication->callback_id = callback_add(replicate_service_timeout, replication, SERVICE_REPLICATION_TIMEOUT);
	send_SERVICE_SERVER_REPLICATION_REQUEST(replication);
}

#ifdef SERVICE_MIGRATION_USE_REDIRECTION
/**
 * @brief Redirect client requests to a service to another node.
 *
 * @param service Pointer to the service for which requests are to be
 * redirected.
 * @param to Address of the alternative server.
 */
static void redirect_clients(SERVICE_service_t* service, C_addr_t to) {
	C_ABORT_ON(service->state != SERVICE_STATE_LOCKED, "SERVICE: Trying to redirect clients of service \"%s\" to %s, but service is not locked (%d).\n", service->id, addr_to_str(to), service->state);

	// Update local SLP cache entry.
	SLP_cache_remove(service->id, C_LOCAL_ADDRESS);
	SLP_cache_update(service->id, to);

	service->state = SERVICE_STATE_REDIRECTING;
	service->redirect_addr = to;

	// TODO: Migrate active streams (possibly cancelling their current callbacks).
}

/**
 * @brief Service migration callback after the redirection phase.
 *
 * @param parameter Service migration information.
 */
static void migrate_service_redirection_callback(callback_parameter_t parameter) {
	SERVICE_migration_t* migration = (SERVICE_migration_t*) parameter;

	if(exists_service(migration->service)) {
		C_DEBUG(SERVICE_REPLICATION_DEBUG_LEVEL, C_MSG_NOTE, "SERVICE: Redirection of clients of service \"%s\" complete, stopping service...\n", migration->service->id);
		migration->service->redirect_addr = C_BROADCAST_ADDRESS;
		stop_service(migration->service);
		free_service(&migration->service);
	} else {
		C_DEBUG(SERVICE_REPLICATION_DEBUG_LEVEL, C_MSG_WARN, "SERVICE: Redirection of clients of service complete, but service has been stopped before.\n", migration->service->id);
	}

	migration->result->error_code = C_ERROR_OK;
	migration->callback(migration->result);
	C_SERVICE_successful_migrations++;
	FREE(migration);
}
#endif	// SERVICE_MIGRATION_USE_REDIRECTION

/**
 * @brief Service migration callback for the replication phase.
 *
 * @param parameter Service replication result.
 */
static void migrate_service_replication_callback(callback_parameter_t parameter) {
	SERVICE_replication_result_t* result = (SERVICE_replication_result_t*) parameter;
	SERVICE_migration_t* migration = (SERVICE_migration_t*) result->parameter;

	switch(result->error_code) {
	case C_ERROR_OK:
#ifdef SERVICE_MIGRATION_USE_REDIRECTION
		C_DEBUG(SERVICE_REPLICATION_DEBUG_LEVEL, C_MSG_NOTE, "SERVICE: Replication of service \"%s\" successful, proceeding with redirection of clients...\n", migration->service->id);
		redirect_clients(migration->service, migration->destination);
		callback_add(migrate_service_redirection_callback, migration, SERVICE_REDIRECTION_TIMEOUT);
#else
		C_DEBUG(SERVICE_REPLICATION_DEBUG_LEVEL, C_MSG_NOTE, "SERVICE: Replication of service \"%s\" complete, stopping service...\n", migration->service->id);
		migration->service->redirect_addr = C_BROADCAST_ADDRESS;
		stop_service(migration->service);
		free_service(&migration->service);

		migration->result->error_code = C_ERROR_OK;
		migration->callback(migration->result);
		C_SERVICE_successful_migrations++;
		FREE(migration);
#endif	// SERVICE_MIGRATION_USE_REDIRECTION
		break;
	case C_ERROR_ABORTED:
		C_DEBUG(SERVICE_REPLICATION_DEBUG_LEVEL, C_MSG_NOTE, "SERVICE: Replication of service \"%s\" has been aborted, aborting service migration...\n", migration->service->id);
		migration->result->error_code = C_ERROR_ABORTED;
		migration->callback(migration->result);
		C_SERVICE_total_migrations--;
		FREE(migration);
		break;
	case C_ERROR_PARAMETER:
	case C_ERROR_TIMEOUT:
	case C_ERROR_BUSY:
	case C_ERROR_IN_PROGRESS:
	case C_ERROR_COMM_ERROR:
	case C_ERROR_REMOTE_BUSY:
		C_DEBUG(SERVICE_REPLICATION_DEBUG_LEVEL, C_MSG_WARN, "SERVICE: Replication of service \"%s\" failed (%d), aborting service migration.\n", migration->service->id, result->error_code);
		migration->service->last_failed_migration_destination = migration->destination;
		migration->result->error_code = result->error_code;
		migration->callback(migration->result);
		C_SERVICE_failed_migrations++;
		FREE(migration);
		break;
	default:
		C_ABORT_ON(true, "SERVICE: Replication callback with unknown error code 0x%X.\n", result->error_code);
	}
	FREE(result);
}

/**
 * @brief Migrate a service from this node to another node.
 *
 * @param service Pointer to the service to migrate.
 * @param destination_address Address of the destination node.
 * @param callback Callback function.
 * @param result Pointer to migration result.
 */
void migrate_service(SERVICE_service_t* service, C_addr_t destination_address, callback_function_t callback, SERVICE_migration_result_t* result) {
	SERVICE_migration_t* migration = NULL;
	SERVICE_replication_result_t* replication_result = NULL;

	if(service->state == SERVICE_STATE_REPLICATING) {
		C_DEBUG(SERVICE_REPLICATION_DEBUG_LEVEL, C_MSG_ERROR, "SERVICE: Trying to migrate service \"%s\" to %s, but service is not running (%d).\n", service->id, addr_to_str(destination_address), service->state);
		result->error_code = C_ERROR_IN_PROGRESS;
		callback(result);
		return;
	} else if(service->state != SERVICE_STATE_LOCKED) {
		C_DEBUG(SERVICE_REPLICATION_DEBUG_LEVEL, C_MSG_ERROR, "SERVICE: Trying to migrate service \"%s\" to %s, but service is not locked (%d).\n", service->id, service->state, addr_to_str(destination_address));
		result->error_code = C_ERROR_PARAMETER;
		callback(result);
		return;
	}

	C_DEBUG(SERVICE_REPLICATION_DEBUG_LEVEL, C_MSG_NOTE, "SERVICE: Migrating service \"%s\" to %s...\n", service->id, addr_to_str(destination_address));
	C_SERVICE_total_migrations++;

	// Reset the migration failed flag.
	service->last_failed_migration_destination = C_BROADCAST_ADDRESS;

	MALLOC(migration, SERVICE_migration_t);
	migration->service = service;
	migration->destination = destination_address;
	migration->callback = callback;
	migration->result = result;
	migration->result->error_code = C_ERROR_NOT_SET;

	MALLOC(replication_result, SERVICE_replication_result_t);
	replication_result->parameter = migration;
	replicate_service(service, destination_address, true, migrate_service_replication_callback, replication_result);
}
