/**
 * @file service_types.h
 *
 * @brief Service (types).
 *
 * Copyright (c) 2007, 2008, 2009, 2010 Georg Wittenburg, Freie Universitaet
 * Berlin (wittenbu@inf.fu-berlin.de). All rights reserved.
 * Please refer to the COPYING file for detailed licensing information.
 *
 * @author Georg Wittenburg
 *
 * $Id: service_types.h 3475 2010-04-26 15:07:35Z gwitten $
 */

#ifndef SERVICE_TYPES_H
#define SERVICE_TYPES_H

#include "../types.h"



/**
 * @brief Type of the service ID.
 */
typedef char	SERVICE_id_t;

/**
 * @brief Length of the service ID string.
 */
#define SERVICE_ID_LENGTH	8



/*
 * Service statistics types.
 */

/**
 * @brief Service host statistic item.
 */
typedef struct {
	node_t*			node;			///< Node to which these statistics belong.
	uint32_t		request_count;		///< Number of service requests (data, update or stream).
	uint32_t		traffic;		///< Traffic of service requests to this host in bytes.
	datarate_stat_t*	datarate_stat;		///< Data rate of service requests to this host in B/s.
} SERVICE_host_stat_t;

/**
 * @brief Client node statistic item.
 */
typedef struct {
	node_t*			node;				///< Node to which these statistics belong.
	uint32_t		request_count;			///< Number of service requests from this node (data, update or stream).
	uint32_t		traffic;			///< Traffic of service requests from this node in bytes.
	datarate_stat_t*	datarate_stat;			///< Data rate of service requests from this node in B/s.
	uint32_t		last_hop_request_count;		///< Number of service requests this node was the last routing hop for.
	uint32_t		last_hop_traffic;		///< Traffic of service requests this node was the last routing hop for in bytes.
	datarate_stat_t*	last_hop_datarate_stat;		///< Data rate of service requests this node was the last routing hop for in B/s.
	uint32_t		last_hop_hop_count_running_sum;	///< Running sum of the hop count of request this node was the last routing hop for (used by ASG).
	uint32_t 		requests_forwarded;		///< Number of requests this node has forwarded (used by ASG).
	C_addr_t 		last_hop_of_last_request;	///< Address of the node that was the last routing hop of the last request from this node (used by ASG).
	bool			refused_migration;		///< Flag whether this node has refused a migration in the past.
} SERVICE_node_stat_t;

/**
 * @brief Client node statistic item (optimized for transmission).
 */
typedef struct {
	C_addr_t	node_address;		///< Node address.
	uint32_t	request_count;		///< Number of service requests from this node (data, update or stream).
	uint32_t	traffic;		///< Traffic of service requests from this node in bytes.
	uint32_t	datarate;		///< Data rate of service requests from this node in B/s.
} SERVICE_condensed_node_stat_t;



/*
 * General service types.
 */

struct SERVICE_outgoing_replication;
struct SERVICE_incoming_replication;
struct PLACEMENT_algorithm;

/**
 * @brief Service data structure.
 */
typedef struct {
	SERVICE_id_t				id[SERVICE_ID_LENGTH];		///< Service ID.
	uint8_t					state;				///< Current state.
	char*					data;				///< Pointer to service data.
	size_t					data_size;			///< Size of service data in bytes.
	float					sync_traffic_fraction;		///< Fraction of data traffic that is required for synchronization between service instances.
	C_time_t				start_time;			///< Time at which this service instance was started.
	SERVICE_host_stat_t*			host_stat;			///< Host statistics for a service.
	list_t*					node_stats;			///< List of node statistics for a service.
	struct PLACEMENT_algorithm*		placement_algorithm;		///< Placement algorithm and data.
	void*					placement_data;			///< Pointer to placement algorithm-specific data.
	struct SERVICE_outgoing_replication*	outgoing_replication;		///< Pointer to outgoing replication.
	struct SERVICE_incoming_replication*	incoming_replication;		///< Pointer to incoming replication.
	C_addr_t				redirect_addr;			///< Address of node to redirect requests to.
	C_addr_t				replace_addr;			///< Address of node that this host replaces.
	C_addr_t				last_failed_migration_destination;	///< Address of the node to which the last migration attempt has failed, or broadcast if not applicable.
	uint8_t					failed_stat_transmissions;	///< Number of transmissions of STATS to the coordinator that have failed.
} SERVICE_service_t;

// Include after typedefs to work around a circular dependency between headers.
#include "../placement/placement_types.h"

/**
 * @brief Data request data structure.
 */
typedef struct {
	SERVICE_id_t				service_id[SERVICE_ID_LENGTH];	///< Service ID.
	uint16_t				offset;				///< Offset in service data.
	size_t					size;				///< Request size in bytes.
} SERVICE_data_request_t;

/**
 * @brief Data update data structure.
 */
typedef struct {
	SERVICE_id_t				service_id[SERVICE_ID_LENGTH];	///< Service ID.
	uint16_t				offset;				///< Offset in service data.
	size_t					size;				///< Update size in bytes.
	char*					data;				///< New service data.
} SERVICE_data_update_t;

/**
 * @brief Stream request data structure.
 */
typedef struct {
	SERVICE_id_t				service_id[SERVICE_ID_LENGTH];	///< Service ID.
	uint16_t				offset;				///< Offset in service data.
	size_t					size;				///< Request size in bytes.
	C_time_t				interval;			///< Request interval in seconds.
	C_time_t				duration;			///< Request duration in seconds.
} SERVICE_stream_request_t;

/**
 * @brief Active stream being served.
 */
typedef struct {
	C_addr_t				to;				///< Address of the destination node.
	SERVICE_id_t				service_id[SERVICE_ID_LENGTH];	///< Service ID.
	uint16_t				offset;				///< Offset in service data.
	size_t					size;				///< Request size in bytes.
	C_time_t				interval;			///< Request interval in seconds.
	C_time_t				end_time;			///< Request duration in seconds.
} SERVICE_stream_t;

/**
 * @brief Segment of service data.
 */
typedef struct {
	uint16_t				offset;				///< Segment offset.
	size_t					size;				///< Segment size.
} SERVICE_segment_t;

/**
 * @brief Service replication result.
 */
typedef struct {
	C_errno_t				error_code;			///< Error code.
	callback_parameter_t			parameter;			///< Parameter for callback function.
} SERVICE_replication_result_t;

/**
 * @brief Outgoing service replication.
 */
typedef struct SERVICE_outgoing_replication {
	SERVICE_service_t*			service;			///< Pointer to service.
	C_addr_t				destination_address;		///< Address of destination node.
	bool					is_migration;			///< Flag whether this replication is part of a migration.
	callback_id_t				callback_id;			///< ID of the timeout callback.
	callback_function_t			callback;			///< Callback function.
	SERVICE_replication_result_t*		result;				///< Pointer to replication result.
} SERVICE_outgoing_replication_t;

/**
 * @brief Incoming service replication.
 */
typedef struct SERVICE_incoming_replication {
	SERVICE_service_t*			service;			///< Pointer to service.
	C_addr_t				source_address;			///< Address of source node.
	list_t*					missing_segments;		///< List of missing service data segments.
	callback_id_t				callback_id;			///< ID of the timeout callback.
	uint8_t					placement_algorithm_id;		///< ID of service placement algorithm.
	char*					placement_algorithm_data;	///< Pointer to data of the placement algorithm.
	size_t					placement_algorithm_data_size;	///< Size of the data for the placement algorithm in bytes.
} SERVICE_incoming_replication_t;

/**
 * @brief Service migration result.
 */
typedef struct {
	C_errno_t				error_code;			///< Error code.
	callback_parameter_t			parameter;			///< Parameter for callback function.
} SERVICE_migration_result_t;

/**
 * @brief Service migration.
 */
typedef struct {
	SERVICE_service_t*			service;			///< Pointer to service.
	C_addr_t				destination;			///< Address of destination node.
	callback_function_t			callback;			///< Callback function.
	SERVICE_migration_result_t*		result;				///< Pointer to migration result.
} SERVICE_migration_t;



/*
 * Packet types.
 */

/**
 * @brief SERVICE DATA REQUEST packet.
 */
typedef struct {
	SERVICE_id_t				service_id[SERVICE_ID_LENGTH];	///< Service ID.
	uint16_t				offset;				///< Offset in service data.
	size_t					size;				///< Request size in bytes.
	C_time_t				start_time;			///< Start time of request.
} SERVICE_CLIENT_DATA_REQUEST_packet_t;

/**
 * @brief SERVICE DATA UPDATE packet header.
 */
typedef struct {
	SERVICE_id_t				service_id[SERVICE_ID_LENGTH];	///< Service ID.
	uint16_t				offset;				///< Offset in service data.
	bool					inter_host_update;		///< Flag whether this update is an inter-host update.
} SERVICE_CLIENT_DATA_UPDATE_packet_t;

/**
 * @brief SERVICE STREAM REQUEST packet.
 */
typedef SERVICE_stream_request_t SERVICE_CLIENT_STREAM_REQUEST_packet_t;

/**
 * @brief SERVICE DATA ITEM packet.
 */
typedef struct {
	SERVICE_id_t				service_id[SERVICE_ID_LENGTH];	///< Service ID.
	C_time_t				start_time;			///< Start time of request.
	C_addr_t				redirect_addr;			///< Address of new server, if service is redirecting.
} SERVICE_SERVER_DATA_ITEM_packet_t;

/**
 * @brief SERVICE STREAM ITEM packet.
 */
typedef struct {
	SERVICE_id_t				service_id[SERVICE_ID_LENGTH];	///< Service ID.
	C_addr_t				redirect_addr;			///< Address of new server, if service is redirecting.
} SERVICE_SERVER_STREAM_ITEM_packet_t;

/**
 * @brief SERVICE SERVER NOT AVAILABLE packet.
 */
typedef struct {
	SERVICE_id_t				service_id[SERVICE_ID_LENGTH];	///< Service ID.
	C_addr_t				redirect_addr;			///< Address of new server, if service is redirecting.
} SERVICE_SERVER_NOT_AVAILABLE_packet_t;

/**
 * @brief SERVICE STATS packet.
 */
typedef struct {
	SERVICE_id_t				service_id[SERVICE_ID_LENGTH];	///< Service ID.
	size_t					node_stat_array_size;		///< Size of the node stat array in bytes.
	uint16_t				number_of_nodes;		///< Number of known nodes on the source node.
	bool					valid_instance;			///< Flag whether this service instance should (in doubt) be considered valid by the coordinator node.
	C_addr_t				last_failed_migration_destination;	///< Address of the node to which the last migration attempt has failed, or broadcast if not applicable.
} SERVICE_STATS_packet_t;

/**
 * @brief SERVICE REPLICATION REQUEST packet.
 */
typedef struct {
	SERVICE_id_t				service_id[SERVICE_ID_LENGTH];	///< Service ID.
	size_t					service_data_size;		///< Size of service data in bytes.
	bool					is_migration;			///< Flag whether this replication is part of a migration.
	float					sync_traffic_fraction;		///< Fraction of data traffic that is required for synchronization between service instances.
	uint8_t					placement_algorithm_id;		///< ID of service placement algorithm.
} SERVICE_REPLICATION_REQUEST_packet_t;

/**
 * @brief SERVICE REPLICATION DATA packet.
 */
typedef struct {
	SERVICE_id_t				service_id[SERVICE_ID_LENGTH];	///< Service ID.
	uint16_t				segment_offset;			///< Offset of service data segment.
} SERVICE_REPLICATION_DATA_packet_t;

/**
 * @brief SERVICE REPLICATION ACK packet.
 */
typedef struct {
	SERVICE_id_t				service_id[SERVICE_ID_LENGTH];	///< Service ID.
	SERVICE_segment_t			next_segment;			///< Next segment to transmit.
} SERVICE_REPLICATION_ACK_packet_t;

/**
 * @brief SERVICE REPLICATION NACK packet.
 */
typedef struct {
	SERVICE_id_t				service_id[SERVICE_ID_LENGTH];	///< Service ID.
	C_errno_t				error_code;			///< Error code.
} SERVICE_REPLICATION_NACK_packet_t;

#endif	// SERVICE_TYPES_H
