/**
 * @file slp_cache.c
 *
 * @brief SLP caching.
 *
 * Copyright (c) 2007, 2008, 2009, 2010 Georg Wittenburg, Freie Universitaet
 * Berlin (wittenbu@inf.fu-berlin.de). All rights reserved.
 * Please refer to the COPYING file for detailed licensing information.
 *
 * @author Georg Wittenburg
 *
 * $Id: slp_cache.c 3237 2010-01-24 15:36:41Z wittenbu $
 */

#include "../config.h"
#include "../macros.h"
#include "../types.h"
#include "../vars.h"
#include "../main.h"
#include "../util/debug.h"
#include "../util/list.h"
#include "../util/callback.h"
#include "../util/net.h"
#include "../service/service.h"
#include "slp-light.h"
#include "slp-light_types.h"
#include "slp-light_vars.h"
#include "slp_cache.h"



/**
 * @brief Size of cache services cache.
 */
#define SLP_SERVER_CACHE_SIZE		(C_MAX_SERVICES * C_MAXIMAL_NUMBER_OF_NODES)

/**
 * @brief Maximal age of a cache entry in seconds.
 */
#define SLP_MAX_CACHE_ENTRY_AGE		60



/**
 * @brief Check whether one service instances is better then another.
 *
 * @param service_instance_A First service instance to compare.
 * @param service_instance_B Second service instance to compare.
 * @return <code>true</code> if the first service instances is better than the
 * second, <code>false</code> otherwise.
 */
static bool service_instance_is_better(const SLP_service_instance_t* service_instance_A, const SLP_service_instance_t* service_instance_B) {
	bool is_better = false;

	C_ABORT_ON(service_instance_A == NULL, "SLP: First service instance to compare is NULL.\n");

	if(service_instance_B == NULL)
		return true;

	is_better = (
		(service_instance_A->hop_count < service_instance_B->hop_count)
		|| (
			(service_instance_A->hop_count == service_instance_B->hop_count)
			&& (service_instance_A->delay < service_instance_B->delay)
		)
	);

	if(is_better)
		C_DEBUG(SLP_DEBUG_LEVEL, C_MSG_VERBOSE, "SLP: Service instance A at %s (%d, %.2fs) is better than service instance B at %s (%d, %.2fs).\n", addr_to_str(service_instance_A->server), service_instance_A->hop_count, service_instance_A->delay, addr_to_str(service_instance_B->server), service_instance_B->hop_count, service_instance_B->delay);
	else
		C_DEBUG(SLP_DEBUG_LEVEL, C_MSG_VERBOSE, "SLP: Service instance B at %s (%d, %.2fs) is better than service instance A at %s (%d, %.2fs).\n", addr_to_str(service_instance_B->server), service_instance_B->hop_count, service_instance_B->delay, addr_to_str(service_instance_A->server), service_instance_A->hop_count, service_instance_A->delay);

	return is_better;
}

/**
 * @brief Check whether two service instances have the same service ID.
 *
 * @param item Service instance to check.
 * @param cmp_item Reference service instance.
 * @return <code>true</code> if the two service instances have the same service
 * ID, <code>false</code> otherwise.
 */
bool matches_service_instance_service_id(const void* item, const void* cmp_item) {
	SLP_service_instance_t* service_instance = (SLP_service_instance_t*) item;
	SLP_service_instance_t* cmp_service_instance = (SLP_service_instance_t*) cmp_item;
	return SERVICE_same_service_id(service_instance->service_id, cmp_service_instance->service_id);
}

/**
 * @brief Check whether two service instances have the same service ID and
 * server.
 *
 * @param item Service instance to check.
 * @param cmp_item Reference service instance.
 * @return <code>true</code> if the two service instances have the same service
 * ID and server, <code>false</code> otherwise.
 */
static bool matches_service_instance_service_id_and_server(const void* item, const void* cmp_item) {
	SLP_service_instance_t* service_instance = (SLP_service_instance_t*) item;
	SLP_service_instance_t* cmp_service_instance = (SLP_service_instance_t*) cmp_item;
	return (
		SERVICE_same_service_id(service_instance->service_id, cmp_service_instance->service_id)
		&& (service_instance->server == cmp_service_instance->server)
	);
}

/**
 * @brief Allocate and initialize a new service instance.
 *
 * @param service_instance Pointer to pointer of the service instance to be
 * allocated and initialized.
 * @param service_id Service ID for the new service instance.
 * @param server Server address for the new service instance.
 */
void alloc_service_instance(SLP_service_instance_t** service_instance, const SERVICE_id_t* service_id, C_addr_t server) {
	MALLOC(*service_instance, SLP_service_instance_t);
	SERVICE_copy_service_id((*service_instance)->service_id, service_id);
	(*service_instance)->server = server;
	(*service_instance)->hop_count = C_MAXIMAL_NETWORK_DIAMETER;
	(*service_instance)->delay = (float) C_NETWORK_TRAVERSAL_TIME;
	(*service_instance)->last_confirmed = C_time();
	(*service_instance)->delayed_removal_callback_id = 0;
}

/**
 * @brief Print the service ID and server of a service instance.
 *
 * @param debug_level Numeric debug level of the message.
 * @param msg_type Numeric type of the message.
 * @param item Pointer to service instance whose ID and server are to be
 * printed.
 */
static void print_service_instance_id_and_server(uint8_t debug_level, uint8_t msg_type, const void* item) {
	C_DEBUG(debug_level, msg_type, "(%s @ %s)", ((SLP_service_instance_t*) item)->service_id, addr_to_str(((SLP_service_instance_t*) item)->server));
}



/**
 * @brief Lookup a service in the cache.
 *
 * @param service_id ID of the service to lookup.
 * @param server Pointer to the server address if C_ERROR_OK is returned,
 * Pointer to C_BROADCAST_ADDRESS otherwise.
 * @return C_ERROR_NOT_FOUND if the service in not in the cache,
 * C_ERROR_TIMEOUT if the service is currently not reachable, or C_ERROR_OK
 * if a valid cache service_instance has been found.
 */
uint8_t cache_lookup(const SERVICE_id_t* service_id, C_addr_t* server) {
	SLP_service_instance_t* best_service_instance = NULL;
	SLP_service_instance_t* service_instance = NULL;
	list_iterator_t service_instance_iterator;
	list_t* stale_service_instances = NULL;
	unsigned int matching_service_instances = 0;

	C_DEBUG(SLP_DEBUG_LEVEL, C_MSG_DEBUG, "SLP: Looking up cached service instance for service \"%s\"...\n", service_id);
	list_alloc(&stale_service_instances);

	// Find best service instance.
	list_iterator_init(C_SLP_service_cache, &service_instance_iterator);
	while((service_instance = (SLP_service_instance_t*) list_iterator_next(&service_instance_iterator)) != NULL) {

		if((C_time() - service_instance->last_confirmed) > SLP_MAX_CACHE_ENTRY_AGE) {
			list_add(stale_service_instances, service_instance);
			continue;
		}

		if(SERVICE_same_service_id(service_instance->service_id, service_id)) {
			matching_service_instances++;
			if(service_instance_is_better(service_instance, best_service_instance))
				best_service_instance = service_instance;
		}
	}

	// Remove stale service instances.
	list_iterator_init(stale_service_instances, &service_instance_iterator);
	while((service_instance = (SLP_service_instance_t*) list_iterator_next(&service_instance_iterator)) != NULL) {
		C_DEBUG(SLP_DEBUG_LEVEL, C_MSG_INFO, "SLP: Purging stale entry for service \"%s\" at %s from cache...\n", service_instance->service_id, addr_to_str(service_instance->server));
		list_remove(C_SLP_service_cache, service_instance);
	}
	list_free_including_data(&stale_service_instances, free);

	if(best_service_instance == NULL) {
		*server = C_BROADCAST_ADDRESS;
		return C_ERROR_NOT_FOUND;
	} else {
		C_ABORT_ON(best_service_instance->server == C_BROADCAST_ADDRESS, "SLP: Best service instance for service \"%s\" is pointing to the broadcast address.\n", service_id);
		C_DEBUG(SLP_DEBUG_LEVEL, C_MSG_DEBUG, "SLP: The best cached service instance for service \"%s\" out of %d is %s.\n", service_id, matching_service_instances, addr_to_str(best_service_instance->server));
		*server = best_service_instance->server;
		return C_ERROR_OK;
	}
}

/**
 * @brief Update the service instance in the service cache with a server
 * address and new metrics.
 *
 * Adds a new service instance of it doesn't exist.
 *
 * @param service_id Service ID of the cached service.
 * @param server Address of the server for this service.
 * @param hop_count Distance to the server in hops.
 * @param delay Time it took the reply to arrive.
 * @param authorative Flag whether the information is authorative.
 */
void cache_update(const SERVICE_id_t* service_id, C_addr_t server, uint8_t hop_count, C_time_t delay, bool authorative) {
	SLP_service_instance_t cmp_service_instance;
	SLP_service_instance_t* service_instance = NULL;

	if(list_size(C_SLP_service_cache) >= SLP_SERVER_CACHE_SIZE) {
		C_DEBUG(SLP_DEBUG_LEVEL, C_MSG_ERROR, "SLP: Server cache is full: ");
		list_print(SLP_DEBUG_LEVEL, (C_MSG_ERROR | C_MSG_CONT), C_SLP_service_cache, print_service_instance_id_and_server);
		C_DEBUG(SLP_DEBUG_LEVEL, (C_MSG_ERROR | C_MSG_CONT), "\n");
		return;
	}

	C_DEBUG(SLP_DEBUG_LEVEL, C_MSG_DEBUG, "SLP: Setting cached service instance for service \"%s\" to %s...\n", service_id, addr_to_str(server));
	memset(&cmp_service_instance, 0, sizeof(SLP_service_instance_t));
	SERVICE_copy_service_id(cmp_service_instance.service_id, service_id);
	cmp_service_instance.server = server;
	service_instance = (SLP_service_instance_t*) list_find(C_SLP_service_cache, &cmp_service_instance, matches_service_instance_service_id_and_server);
	if(service_instance == NULL) {
		alloc_service_instance(&service_instance, service_id, server);
		list_add(C_SLP_service_cache, service_instance);
		service_instance->hop_count = hop_count;
		service_instance->delay = delay;
		service_instance->last_confirmed = C_time();
	} else if(authorative) {
		service_instance->hop_count = hop_count;
		service_instance->delay = delay;
		service_instance->last_confirmed = C_time();
	}
}

/**
 * @brief Remove a cache entry.
 *
 * If no corresponding entry is in the cache, no action is performed.
 *
 * @param service_id Service ID of the cached service to be removed.
 * @param server Address of the server for this service.
 */
void cache_remove(const SERVICE_id_t* service_id, C_addr_t server) {
	SLP_service_instance_t cmp_service_instance;
	SLP_service_instance_t* service_instance = NULL;

	SERVICE_copy_service_id(cmp_service_instance.service_id, service_id);
	cmp_service_instance.server = server;
	service_instance = (SLP_service_instance_t*) list_find(C_SLP_service_cache, &cmp_service_instance, matches_service_instance_service_id_and_server);
	if(service_instance != NULL) {
		C_DEBUG(SLP_DEBUG_LEVEL, C_MSG_DEBUG, "SLP: Removing cached service instance for service \"%s\" to %s...\n", service_instance->service_id, addr_to_str(service_instance->server));
		list_remove(C_SLP_service_cache, service_instance);
		FREE(service_instance);
	}
}

/**
 * @brief Check whether the cache contains a certain entry.
 *
 * @param service_id Service ID of the cached service.
 * @param server Address of the server for this service.
 */
bool cache_contains(const SERVICE_id_t* service_id, C_addr_t server) {
	SLP_service_instance_t cmp_service_instance;
	SLP_service_instance_t* service_instance = NULL;

	SERVICE_copy_service_id(cmp_service_instance.service_id, service_id);
	cmp_service_instance.server = server;
	service_instance = (SLP_service_instance_t*) list_find(C_SLP_service_cache, &cmp_service_instance, matches_service_instance_service_id_and_server);
	return (service_instance != NULL);
}
