/**
 * @file types.h
 *
 * @brief Types and macro definitions.
 *
 * Copyright (c) 2007, 2008, 2009, 2010 Georg Wittenburg, Freie Universitaet
 * Berlin (wittenbu@inf.fu-berlin.de). All rights reserved.
 * Please refer to the COPYING file for detailed licensing information.
 *
 * @author Georg Wittenburg
 *
 * $Id: types.h 3566 2010-06-09 12:24:20Z gwitten $
 */

#ifndef TYPES_H
#define TYPES_H



#ifdef ARCH_NONE
	// Used when generating dependencies.
#elif defined(ARCH_LINUX)
	#include "arch/linux/types.h"
#elif defined(ARCH_WIN32)
	#include "arch/win32/types.h"
#elif defined(ARCH_NS2)
	// Nothing to do.
#else
	#error No architecture defined (available architectures are: ARCH_LINUX, ARCH_WIN32 and ARCH_NS2).
#endif	// ARCH_*



#include "config.h"



/**
 * @brief Item of a single-linked list.
 */
typedef struct _list_item {
	void*			data;		///< Pointer to data of this list item.
	struct _list_item*	next;		///< Pointer to next list item.
} _list_item_t;

/**
 * @brief Single-linked list with some extra fields for efficiency.
 */
typedef struct {
	_list_item_t*		first_item;	///< Pointer to first item of the list.
	_list_item_t*		last_item;	///< Pointer to last item of the list.
	uint16_t 		size;		///< Number of items in the list.
} list_t;

/**
 * @brief Basic list iterator (no consistency checks, etc.).
 */
typedef struct {
	const list_t*		list;		///< Pointer to the list.
	_list_item_t*		current_item;	///< Pointer to current item of the list.
#ifdef C_ENABLE_CONSISTENCY_CHECKS
	uint16_t 		list_size;	///< Number of items in the list when the ierator was initialized.
#endif	// C_ENABLE_CONSISTENCY_CHECKS
} list_iterator_t;

/**
 * @brief Callback ID type.
 */
typedef uint32_t callback_id_t;

/**
 * @brief Callback parameter type.
 */
typedef void* callback_parameter_t;

/**
 * @brief Callback function type.
 */
typedef void(*callback_function_t)(callback_parameter_t);

/**
 * @brief Callback consisting of function, parameter and time.
 */
typedef struct {
	callback_id_t		id;		///< Unique ID of this callback.
	callback_function_t	function;	///< Callback function.
	callback_parameter_t	parameter;	///< Callback parameter.
	C_time_t		time;		///< Time at which the callback function will be called.
} callback_t;

/**
 * @brief Callback error code function type.
 */
typedef uint8_t C_errno_t;

/**
 * @brief Value-dependent event trigger.
 */
typedef struct {
	float			max_range;	///< Maximum range (in both diretions) of the hysteresis at start time.
	C_time_t		max_duration;	///< Maximum duration of the trigger; range reaches 0 if exceeded.
	float			start_value;	///< Start value to center the hysteresis around at start time.
	C_time_t		start_time;	///< Start time at which the range of the hysteresis is maximal.
	uint16_t		triggered_at_end_count;	///< Number of times that the trigger triggered at the end of its duration.
} trigger_t;



/**
 * @brief Time-to-live (TTL) type.
 */
typedef uint8_t ttl_t;

/**
 * @brief Sequence number type.
 */
typedef uint16_t seqno_t;

/**
 * @brief Layer 3 (NET) packet header.
 *
 * NOTE: Some of these fields are changed as the packet is in transit. If the
 * packet contains an encapsulated packet (e.g. in DYMO or RUDP), the fields of
 * the encapsulated packet need to be updated accordingly.
 */
typedef struct {
	C_addr_t	source_address;		///< Source address.
	C_addr_t	destination_address;	///< Destination address.
	C_addr_t	last_hop_address;	///< Address of the last node that forwarded this packet.
	C_addr_t	next_hop_address;	///< Address of the next node that should handle this packet.
	ttl_t		ttl;			///< Time-to-live (TTL).
	uint8_t		priority;		///< Priority (used in QoS-aware queueing).
	seqno_t		seqno;			///< Sequence number.
	uint8_t		app;			///< Application identifier.
	uint8_t		type;			///< Application-specific packet type identifier.
	uint8_t		hop_count;		///< Number of times this packet was forwarded.
	uint8_t		neighbor_count;		///< Number of neighbors of the source node.
	size_t		payload_length;		///< Size of the payload in bytes.
#ifdef C_USE_NETWORK_MAPPING
	C_addr_t*	hops;			///< Addresses of the hops on the route.
	C_addr_t*	neighbors;		///< Addresses of 1-hop neighbors of the source node.
#endif	// C_USE_NETWORK_MAPPING
} net_header_t;

/**
 * @brief Packet consisting of header and payload.
 */
typedef struct {
	net_header_t*	header;		///< Pointer to packet header.
	void*		payload;	///< Pointer to packet payload.
} C_packet_t;

/**
 * @brief Packet fingerprint.
 *
 * This data structure is used for checking for duplicates in recently arrived
 * packets.
 */
typedef struct {
	C_time_t	time;			///< Time of arrival.
	C_addr_t	from;			///< Source address.
	seqno_t		seqno;			///< Sequence number.
} packet_fingerprint_t;

/**
 * @brief Basic information about a node in the network.
 */
typedef struct {
	C_addr_t	address;			///< Node address.
	C_time_t	creation_time;			///< Time at which this node struct was created.
	uint8_t		minimal_hop_count;		///< Minimal hop count with which a packet from this node has been received.
#ifdef C_USE_INTERNAL_LINK_MONITORING
	uint8_t		unanswered_packets;		///< Number of packets to which no corresponding packet has been received.
	callback_id_t	expected_packet_callback_id;	///< ID of the expected packet timeout callback. A packet MUST be recieved before this timeout expires, otherwise the link is considered broken.
	callback_id_t	delayed_send_callback_id;	///< ID of the delayed reply timeout callback. Sending may be delayed for a certain amount of time. If this timeout expires, an artificial packet is sent just to keep the link alive.
#endif	// C_USE_INTERNAL_LINK_MONITORING
#ifdef C_USE_ETX_LINK_METRIC
	uint8_t		received_beacons;		///< Number of received BEACON packets during the LAST time interval.
	uint8_t		received_beacons_counter;	///< Number of received BEACON packets during the CURRENT time interval.
	uint8_t		reported_beacons;		///< Number of packets this node reported to have received from the local node.
	C_time_t	last_beacon_purge_time;		///< Time when the beacon counters where last purged.
#endif	// C_USE_ETX_LINK_METRIC
} node_t;

/**
 * @brief Variables to calculate data rates.
 */
typedef struct {
	uint32_t*	bytes_per_second_buckets;	///< Array of byte-counting buckets for each past second.
	uint16_t	valid_buckets;			///< Number of buckets (starting from index 0) that contain valid data.
	C_time_t	time_of_last_purge;		///< Last time at which the array was purged.
	uint32_t	added_datarate;			///< Data rate to add when calculating this one.
	C_time_t	added_datarate_time;		///< Time at which additional data rate was added.
	uint32_t	cached_datarate;		///< Cached value of the data rate described by this struct.
	bool		cached_datarate_is_dirty;	///< Flag whether the datarate field needs to be recalculated.
	list_t*		added_datarate_stats;		///< List of data rate statistics to add when calculating this one.
	bool		may_aggregate_other_stats;	///< Flag whether this struct may aggregate the data from other structs.
} datarate_stat_t;

#endif	// TYPES_H
