/**
 * @file list.c
 *
 * @brief List functions.
 *
 * Copyright (c) 2007, 2008, 2009, 2010 Georg Wittenburg, Freie Universitaet
 * Berlin (wittenbu@inf.fu-berlin.de). All rights reserved.
 * Please refer to the COPYING file for detailed licensing information.
 *
 * @author Georg Wittenburg
 *
 * $Id: list.c 3177 2010-01-05 16:43:45Z wittenbu $
 */

#include "../config.h"
#include "../macros.h"
#include "../types.h"
#include "../vars.h"
#include "../main.h"
#include "debug.h"
#include "list.h"



/**
 * @brief Debug level for this component.
 */
#define LIST_DEBUG_LEVEL	C_MSG_WARN

/**
 * @brief Debug level for this component.
 */
#ifdef C_ENABLE_CONSISTENCY_CHECKS
	// List consistency hasn't been a problem since 10/09, hence it should
	// be safe to disable these checks.
	//#define C_ENABLE_LIST_CONSISTENCY_CHECKS
#endif	// C_ENABLE_CONSISTENCY_CHECKS



#ifdef C_ENABLE_LIST_CONSISTENCY_CHECKS
/**
 * @brief Check the consistency of a list data structure.
 *
 * @param list Pointer to a list to check.
 */
static void list_check_consistency(const list_t* list) {
	_list_item_t* item = NULL;
	_list_item_t* last_item = NULL;
	unsigned int index = 0;

	C_ABORT_ON(list == NULL, "List consistency check failed; NULL list\n");

	item = last_item = list->first_item;
	while(item != NULL) {
		C_ABORT_ON(item->data == NULL, "List consistency check of list at %p failed; NULL data item for item #%d at %p.\n", list, index, item);
		last_item = item;
		item = item->next;
		index++;
	}
	C_ABORT_ON(last_item != list->last_item, "List consistency check of list at %p failed; last item is at %p, expected %p\n", list, item, list->last_item);
	C_ABORT_ON(index != list->size, "List consistency check of list at %p failed; size mismatch (%d != %d)\n", list, index, list->size);
}
#endif	// C_ENABLE_LIST_CONSISTENCY_CHECKS

/**
 * @brief Allocate and initialize a new list.
 *
 * @param list Pointer to a pointer to a (uninitialized) list.
 */
void list_alloc(list_t** list) {
	C_ABORT_ON(*list != NULL, "Trying to allocate on pointer that is not NULL...\n");
	MALLOC(*list, list_t);
	(*list)->first_item = NULL;
	(*list)->last_item = NULL;
	(*list)->size = 0;
	C_DEBUG(LIST_DEBUG_LEVEL, C_MSG_VERBOSE, "New list allocated at %p.\n", *list);
}

/**
 * @brief Free a list and a its elements and set the list pointer to NULL.
 *
 * Use this function to free a list whose data elements are still being pointed
 * to elsewhere.
 *
 * @param list Pointer to pointer to the list to free.
 */
void list_free(list_t** list) {
	_list_item_t* item;

	C_ABORT_ON(*list == NULL, "Trying to free list at NULL...\n");
#ifdef C_ENABLE_LIST_CONSISTENCY_CHECKS
	list_check_consistency(*list);
#endif	// C_ENABLE_LIST_CONSISTENCY_CHECKS

	C_DEBUG(LIST_DEBUG_LEVEL, C_MSG_VERBOSE, "Freeing list at %p...\n", *list);
	item = (*list)->first_item;
	while(item != NULL) {
		_list_item_t* tmp = item->next;
		FREE(item);
		item = tmp;
	}
	FREE(*list);
}

/**
 * @brief Free a list and a its elements including the data items and set the
 * list pointer to NULL.
 *
 * Use this function to free a list whose data elements are NOT used ANYWHERE
 * else.
 *
 * @param list Pointer to pointer to the list to free.
 * @param free_data Pointer to function that properly frees the data items.
 */
void list_free_including_data(list_t** list, void(*free_data)(void*)) {
	_list_item_t* item;

	C_ABORT_ON(*list == NULL, "Trying to free list at NULL...\n");
#ifdef C_ENABLE_LIST_CONSISTENCY_CHECKS
	list_check_consistency(*list);
#endif	// C_ENABLE_LIST_CONSISTENCY_CHECKS

	C_DEBUG(LIST_DEBUG_LEVEL, C_MSG_VERBOSE, "Freeing list at %p...\n", *list);
	item = (*list)->first_item;
	while(item != NULL) {
		_list_item_t* tmp = item->next;
		free_data(item->data);
		FREE(item);
		item = tmp;
	}
	FREE(*list);
}

/**
 * @brief Add a new data item to a list.
 *
 * @param list Pointer to the list.
 * @param data Pointer to the data item.
 */
void list_add(list_t* list, void* data) {
	_list_item_t* item = NULL;

	C_ABORT_ON(list == NULL, "Trying to add item to list at NULL...\n");
	C_ABORT_ON(data == NULL, "Trying to add NULL data item to list at %p...\n", list);
#ifdef C_ENABLE_LIST_CONSISTENCY_CHECKS
	list_check_consistency(list);
#endif	// C_ENABLE_LIST_CONSISTENCY_CHECKS

	C_DEBUG(LIST_DEBUG_LEVEL, C_MSG_VERBOSE, "Adding new data item at %p to list at %p; %d item(s) in list...\n", data, list, list->size + 1);
	MALLOC(item, _list_item_t);
	item->data = data;
	item->next = NULL;
	if(list->size == 0)
		list->first_item = item;
	else
		list->last_item->next = item;
	list->last_item = item;
	list->size++;

#ifdef C_ENABLE_LIST_CONSISTENCY_CHECKS
	list_check_consistency(list);
#endif	// C_ENABLE_LIST_CONSISTENCY_CHECKS
}

/**
 * @brief Insert a new data item at a specific location of a list.
 *
 * The list is traversed from beginning to end and for each data item the
 * boolean parameter function is called. The new data item is inserted into
 * the list before the first element for which the parameter function becomes
 * <code>true</code>. If the parameter function is <code>false</code> for all
 * elements of the list, the new data item is inserted at the end of the list.
 *
 * @param list Pointer to the list.
 * @param data Pointer to the data item.
 * @param matches Function used for finding the correct location.
 */
void list_insert_where(list_t* list, void* data, bool(*matches)(const void*, const void*)) {
	_list_item_t* item = NULL;

	C_ABORT_ON(list == NULL, "Trying to insert data item into list at NULL...\n");
	C_ABORT_ON(data == NULL, "Trying to insert NULL data item to list at %p...\n", list);
	C_ABORT_ON(matches == NULL, "Trying to insert data item into list at %p with NULL function...\n");
#ifdef C_ENABLE_LIST_CONSISTENCY_CHECKS
	list_check_consistency(list);
#endif	// C_ENABLE_LIST_CONSISTENCY_CHECKS

	C_DEBUG(LIST_DEBUG_LEVEL, C_MSG_VERBOSE, "Inserting new data item at %p into list at %p; %d item(s) in list...\n", data, list, list->size + 1);
	MALLOC(item, _list_item_t);
	item->data = data;
	if(list->size == 0) {
		item->next = NULL;
		list->first_item = item;
		list->last_item = item;
	} else if(list->size == 1) {
		if(matches(list->first_item->data, data)) {
			item->next = list->first_item;
			list->first_item = item;
		} else {
			item->next = NULL;
			list->first_item->next = item;
			list->last_item = item;
		}
	} else if(matches(list->first_item->data, data)) {
		item->next = list->first_item;
		list->first_item = item;
	} else if(!matches(list->last_item->data, data)) {
		item->next = NULL;
		list->last_item->next = item;
		list->last_item = item;
	} else {
		_list_item_t* current_item = list->first_item;
		_list_item_t* previous_item = NULL;
		while(!matches(current_item->data, data)) {
			previous_item = current_item;
			current_item = current_item->next;
		}
		previous_item->next = item;
		item->next = current_item;
	}
	list->size++;

#ifdef C_ENABLE_LIST_CONSISTENCY_CHECKS
	list_check_consistency(list);
#endif	// C_ENABLE_LIST_CONSISTENCY_CHECKS
}

/**
 * @brief Append data from a list to another list.
 *
 * NOTE: New memory is only allocated for the list data structures. The actual
 * data contained in the source list is not replicated, i.e. the two lists
 * point to the same data structures.
 *
 * @param dst_list Pointer to the destination list.
 * @param src_list Pointer to the source list.
 */
void list_append(list_t* dst_list, const list_t* src_list) {
	_list_item_t* current_src_item = NULL;
	_list_item_t* tmp_item = NULL;

	C_ABORT_ON(dst_list == NULL, "Trying to append list into list at NULL...\n");
	C_ABORT_ON(src_list == NULL, "Trying to append list from list at NULL...\n");
#ifdef C_ENABLE_LIST_CONSISTENCY_CHECKS
	list_check_consistency(dst_list);
	list_check_consistency(src_list);
#endif	// C_ENABLE_LIST_CONSISTENCY_CHECKS

	current_src_item = src_list->first_item;
	while(current_src_item != NULL) {
		MALLOC(tmp_item, _list_item_t);
		tmp_item->data = current_src_item->data;
		tmp_item->next = NULL;

		if(dst_list->size == 0)
			dst_list->first_item = tmp_item;
		else
			dst_list->last_item->next = tmp_item;
		dst_list->last_item = tmp_item;
		dst_list->size++;

		tmp_item = NULL;
		current_src_item = current_src_item->next;
	}

#ifdef C_ENABLE_LIST_CONSISTENCY_CHECKS
	list_check_consistency(dst_list);
	list_check_consistency(src_list);
#endif	// C_ENABLE_LIST_CONSISTENCY_CHECKS
}

/**
 * @brief Merge data from two lists into one list.
 *
 * This function appends all items from a list SRC, which are not contained (by
 * comparison of pointers) in a list DST, to the list DST.
 *
 * NOTE: New memory is only allocated for the list data structures. The actual
 * data contained in the source list is not replicated, i.e. the two lists
 * point to the same data structures.
 *
 * @param dst_list Pointer to the destination list.
 * @param src_list Pointer to the source list.
 */
void list_merge(list_t* dst_list, const list_t* src_list) {
	_list_item_t* current_src_item = NULL;
	bool item_contained_in_dst_list = false;
	_list_item_t* current_dst_item = NULL;
	_list_item_t* tmp_item = NULL;

	C_ABORT_ON(dst_list == NULL, "Trying to merge list into list at NULL...\n");
	C_ABORT_ON(src_list == NULL, "Trying to merge list from list at NULL...\n");
#ifdef C_ENABLE_LIST_CONSISTENCY_CHECKS
	list_check_consistency(dst_list);
	list_check_consistency(src_list);
#endif	// C_ENABLE_LIST_CONSISTENCY_CHECKS

	current_src_item = src_list->first_item;
	while(current_src_item != NULL) {
		item_contained_in_dst_list = false;

		current_dst_item = dst_list->first_item;
		while(current_dst_item != NULL) {
			if(current_dst_item->data == current_src_item->data) {
				item_contained_in_dst_list = true;
				break;
			}
			current_dst_item = current_dst_item->next;
		}

		if(!item_contained_in_dst_list) {
			MALLOC(tmp_item, _list_item_t);
			tmp_item->data = current_src_item->data;
			tmp_item->next = NULL;

			if(dst_list->size == 0)
				dst_list->first_item = tmp_item;
			else
				dst_list->last_item->next = tmp_item;
			dst_list->last_item = tmp_item;
			dst_list->size++;

			tmp_item = NULL;
		}

		current_src_item = current_src_item->next;
	}

#ifdef C_ENABLE_LIST_CONSISTENCY_CHECKS
	list_check_consistency(dst_list);
	list_check_consistency(src_list);
#endif	// C_ENABLE_LIST_CONSISTENCY_CHECKS
}

/**
 * @brief Remove a data item from the list.
 *
 * @param list Pointer to the list.
 * @param data Pointer to the data item.
 */
void list_remove(list_t* list, const void* data) {
	C_ABORT_ON(list == NULL, "Trying to remove data item from list at NULL...\n");
	C_ABORT_ON(data == NULL, "Trying to remove NULL data item from list at %p...\n", list);
	C_ABORT_ON(list->size == 0, "Trying to remove data item from empty list...\n", list);
#ifdef C_ENABLE_LIST_CONSISTENCY_CHECKS
	list_check_consistency(list);
#endif	// C_ENABLE_LIST_CONSISTENCY_CHECKS

	if(list->size == 1) {
		if(list->first_item->data == data) {
			C_DEBUG(LIST_DEBUG_LEVEL, C_MSG_VERBOSE, "Removing data item at %p from list at %p; %d item(s) in list...\n", data, list, list->size - 1);
			FREE(list->first_item);
			list->first_item = NULL;
			list->last_item = NULL;
			list->size = 0;
#ifdef C_ENABLE_LIST_CONSISTENCY_CHECKS
			list_check_consistency(list);
#endif	// C_ENABLE_LIST_CONSISTENCY_CHECKS
			return;
		}
	} else if(list->first_item->data == data) {
		_list_item_t* tmp;
		C_DEBUG(LIST_DEBUG_LEVEL, C_MSG_VERBOSE, "Removing data item at %p from list at %p; %d item(s) in list...\n", data, list, list->size - 1);
		tmp = list->first_item;
		list->first_item = list->first_item->next;
		FREE(tmp);
		list->size--;
#ifdef C_ENABLE_LIST_CONSISTENCY_CHECKS
		list_check_consistency(list);
#endif	// C_ENABLE_LIST_CONSISTENCY_CHECKS
		return;
	} else if(list->last_item->data == data) {
		_list_item_t* item;
		C_DEBUG(LIST_DEBUG_LEVEL, C_MSG_VERBOSE, "Removing data item at %p from list at %p; %d item(s) in list...\n", data, list, list->size - 1);
		item = list->first_item;
		while(item->next != list->last_item)
			item = item->next;
		FREE(list->last_item);
		list->last_item = item;
		list->last_item->next = NULL;
		list->size--;
#ifdef C_ENABLE_LIST_CONSISTENCY_CHECKS
		list_check_consistency(list);
#endif	// C_ENABLE_LIST_CONSISTENCY_CHECKS
		return;
	} else {
		_list_item_t* current_item = list->first_item;
		_list_item_t* previous_item = NULL;
		while(current_item != NULL) {
			if(current_item->data == data) {
				C_DEBUG(LIST_DEBUG_LEVEL, C_MSG_VERBOSE, "Removing data item at %p from list at %p; %d item(s) in list...\n", data, list, list->size - 1);
				previous_item->next = current_item->next;
				FREE(current_item);
				list->size--;
#ifdef C_ENABLE_LIST_CONSISTENCY_CHECKS
				list_check_consistency(list);
#endif	// C_ENABLE_LIST_CONSISTENCY_CHECKS
				return;
			}
			previous_item = current_item;
			current_item = current_item->next;
		}
	}
	C_DEBUG(LIST_DEBUG_LEVEL, C_MSG_ERROR, "Trying to remove data item at %p that is not list at %p...\n", data, list);

#ifdef C_ENABLE_LIST_CONSISTENCY_CHECKS
	list_check_consistency(list);
#endif	// C_ENABLE_LIST_CONSISTENCY_CHECKS
}

/**
 * @brief Remove the first data item from the list.
 *
 * @param list Pointer to the list.
 * @return Pointer to removed data item.
 */
void* list_remove_head(list_t* list) {
	void* data = NULL;
	_list_item_t* tmp = NULL;

	C_ABORT_ON(list == NULL, "Trying to remove first data item from list at NULL...\n");
	C_ABORT_ON(list->size == 0, "Trying to remove first data item from empty list...\n", list);
#ifdef C_ENABLE_LIST_CONSISTENCY_CHECKS
	list_check_consistency(list);
#endif	// C_ENABLE_LIST_CONSISTENCY_CHECKS

	data = list->first_item->data;
	C_DEBUG(LIST_DEBUG_LEVEL, C_MSG_VERBOSE, "Removing first data item at %p from list at %p; %d item(s) in list...\n", data, list, list->size - 1);
	if(list->size == 1) {
		FREE(list->first_item);
		list->first_item = NULL;
		list->last_item = NULL;
	} else {
		tmp = list->first_item;
		list->first_item = list->first_item->next;
		FREE(tmp);
	}
	list->size--;

#ifdef C_ENABLE_LIST_CONSISTENCY_CHECKS
	list_check_consistency(list);
#endif	// C_ENABLE_LIST_CONSISTENCY_CHECKS
	return data;
}

/**
 * @brief Remove all elements from a list (without freeing the data items).
 *
 * @param list Pointer to the list.
 */
void list_clear(list_t* list) {
	_list_item_t* item;

	C_ABORT_ON(list == NULL, "Trying to clear list at NULL...\n");
#ifdef C_ENABLE_LIST_CONSISTENCY_CHECKS
	list_check_consistency(list);
#endif	// C_ENABLE_LIST_CONSISTENCY_CHECKS

	C_DEBUG(LIST_DEBUG_LEVEL, C_MSG_VERBOSE, "Clearing list at %p with %d item(s)...\n", list, list->size);
	item = list->first_item;
	while(item != NULL) {
		_list_item_t* tmp = item->next;
		FREE(item);
		item = tmp;
	}
	list->first_item = NULL;
	list->last_item = NULL;
	list->size = 0;

#ifdef C_ENABLE_LIST_CONSISTENCY_CHECKS
	list_check_consistency(list);
#endif	// C_ENABLE_LIST_CONSISTENCY_CHECKS
}

/**
 * @brief Remove all elements from a list (freeing the data items).
 *
 * @param list Pointer to the list.
 * @param free_data Pointer to function that properly frees the data items.
 */
void list_clear_including_data(list_t* list, void(*free_data)(void*)) {
	_list_item_t* item;

	C_ABORT_ON(list == NULL, "Trying to clear list at NULL...\n");
#ifdef C_ENABLE_LIST_CONSISTENCY_CHECKS
	list_check_consistency(list);
#endif	// C_ENABLE_LIST_CONSISTENCY_CHECKS

	C_DEBUG(LIST_DEBUG_LEVEL, C_MSG_VERBOSE, "Clearing list at %p with %d item(s)...\n", list, list->size);
	item = list->first_item;
	while(item != NULL) {
		_list_item_t* tmp = item->next;
		free_data(item->data);
		FREE(item);
		item = tmp;
	}
	list->first_item = NULL;
	list->last_item = NULL;
	list->size = 0;

#ifdef C_ENABLE_LIST_CONSISTENCY_CHECKS
	list_check_consistency(list);
#endif	// C_ENABLE_LIST_CONSISTENCY_CHECKS
}

/**
 * @brief Check whether a list contains a specific data item.
 *
 * @param list Pointer to the list.
 * @param data Pointer to the data item.
 * @return <code>true</code> if the data item is contained in the list,
 * <code>false</code> otherwise.
 */
bool list_contains(const list_t* list, const void* data) {
	_list_item_t* current_item;

	C_ABORT_ON(list == NULL, "Trying to check for data item at %p of list at NULL...\n", data);
	C_ABORT_ON(data == NULL, "Trying to check for data item at NULL of list at %p...\n", list);
#ifdef C_ENABLE_LIST_CONSISTENCY_CHECKS
	list_check_consistency(list);
#endif	// C_ENABLE_LIST_CONSISTENCY_CHECKS

	C_DEBUG(LIST_DEBUG_LEVEL, C_MSG_VERBOSE, "Checking for data item at %p in list at %p...\n", data, list);

	current_item = list->first_item;
	while(current_item != NULL) {
		if(current_item->data == data) {
			C_DEBUG(LIST_DEBUG_LEVEL, C_MSG_VERBOSE, "Found data item at %p in list at %p...\n", current_item->data, list);
			return true;
		}
		current_item = current_item->next;
	}
	return false;
}

/**
 * @brief Get a data item at a specific index of a list.
 *
 * The first item of the list has index 0.
 *
 * @param list Pointer to the list.
 * @param index Index of the data item to retrieve.
 * @return Pointer to the data item at the specified index of the list.
 */
void* list_get(const list_t* list, unsigned int index) {
	_list_item_t* item;

	C_ABORT_ON(list == NULL, "Trying to retrieve data item #%d of list at NULL...\n", index);
	C_ABORT_ON(index >= list_size(list), "Trying to retrieve data item #%d of list at %p, but list has only %d elements...\n", index, list, list_size(list));
#ifdef C_ENABLE_LIST_CONSISTENCY_CHECKS
	list_check_consistency(list);
#endif	// C_ENABLE_LIST_CONSISTENCY_CHECKS

	C_DEBUG(LIST_DEBUG_LEVEL, C_MSG_VERBOSE, "Retrieving data item #%d of list at %p...\n", index, list);

	item = list->first_item;
	while(index-- > 0) {
		item = item->next;
	}
	return item->data;
}

/**
 * @brief Get the size of a list.
 *
 * @param list Pointer to the list.
 * @return Size of the list.
 */
unsigned int list_size(const list_t* list) {
	C_ABORT_ON(list == NULL, "Trying to retrieve size of list at NULL...\n");
#ifdef C_ENABLE_LIST_CONSISTENCY_CHECKS
	list_check_consistency(list);
#endif	// C_ENABLE_LIST_CONSISTENCY_CHECKS

	C_DEBUG(LIST_DEBUG_LEVEL, C_MSG_VERBOSE, "List at %p has %d data items.\n", list, list->size);
	return list->size;
}

/**
 * @brief Check whether a list is empty.
 *
 * @param list Pointer to the list.
 * @return <code>true</code> if the list doesn't contain any elements,
 * <code>false</code> otherwise.
 */
bool list_empty(const list_t* list) {
	C_ABORT_ON(list == NULL, "Trying to check whether list at NULL is empty...\n");
#ifdef C_ENABLE_LIST_CONSISTENCY_CHECKS
	list_check_consistency(list);
#endif	// C_ENABLE_LIST_CONSISTENCY_CHECKS

	if(list->size == 0) {
		C_DEBUG(LIST_DEBUG_LEVEL, C_MSG_VERBOSE, "List at %p is empty.\n", list);
		return true;
	} else {
		C_DEBUG(LIST_DEBUG_LEVEL, C_MSG_VERBOSE, "List at %p is not empty.\n", list);
		return false;
	}
}

/**
 * @brief Return the first data item of a list.
 *
 * @param list Pointer to the list.
 * @return Pointer to the first data item of the list.
 */
void* list_head(const list_t* list) {
	C_ABORT_ON(list == NULL, "Trying to retrieve first data item of list at NULL...\n");
	C_ABORT_ON(list_empty(list), "Trying to retrieve first data item of empty list at %p...\n", list);
#ifdef C_ENABLE_LIST_CONSISTENCY_CHECKS
	list_check_consistency(list);
#endif	// C_ENABLE_LIST_CONSISTENCY_CHECKS

	C_DEBUG(LIST_DEBUG_LEVEL, C_MSG_VERBOSE, "Returning first data item %p in list at %p...\n", list->first_item->data, list);
	return list->first_item->data;
}

/**
 * @brief Return the last data item of a list.
 *
 * @param list Pointer to the list.
 * @return Pointer to the last data item of the list.
 */
void* list_tail(const list_t* list) {
	C_ABORT_ON(list == NULL, "Trying to retrieve last data item of list at NULL...\n");
	C_ABORT_ON(list_empty(list), "Trying to retrieve last data item of empty list at %p...\n", list);
#ifdef C_ENABLE_LIST_CONSISTENCY_CHECKS
	list_check_consistency(list);
#endif	// C_ENABLE_LIST_CONSISTENCY_CHECKS

	C_DEBUG(LIST_DEBUG_LEVEL, C_MSG_VERBOSE, "Returning last data item %p in list at %p...\n", list->last_item->data, list);
	return list->last_item->data;
}

/**
 * @brief Find a specific data item in a list.
 *
 * The list is traversed from beginning to end and for each data item the
 * boolean parameter function is called. A pointer to the first data item for
 * which the function becomes <code>true</code> is returned. If the parameter
 * function is <code>false</code> for all data items of the list, this function
 * returns NULL.
 *
 * @param list Pointer to the list.
 * @param data Pointer to the data item.
 * @param matches Function used for finding a matching data item.
 * @return Pointer to a matching data item or NULL if no item matches.
 */
void* list_find(const list_t* list, const void* data, bool(*matches)(const void*, const void*)) {
	_list_item_t* current_item;

	C_ABORT_ON(list == NULL, "Trying to find data item in list at NULL...\n");
	C_ABORT_ON(matches == NULL, "Trying to find data item in list at %p with NULL function...\n");
#ifdef C_ENABLE_LIST_CONSISTENCY_CHECKS
	list_check_consistency(list);
#endif	// C_ENABLE_LIST_CONSISTENCY_CHECKS

	current_item = list->first_item;
	while(current_item != NULL) {
		if(matches(current_item->data, data)) {
			C_DEBUG(LIST_DEBUG_LEVEL, C_MSG_VERBOSE, "Found data item at %p in list at %p...\n", current_item->data, list);
			return current_item->data;
		}
		current_item = current_item->next;
	}
	C_DEBUG(LIST_DEBUG_LEVEL, C_MSG_VERBOSE, "No matching data item found in list at %p...\n", list);
	return NULL;
}

/**
 * @brief Print a list.
 *
 * @param debug_level Numeric debug level of the message.
 * @param msg_type Numeric type of the message.
 * @param list Pointer to the list.
 * @param print_data Function used for printing data items.
 */
void list_print(uint8_t debug_level, uint8_t msg_type, const list_t* list, void(*print_data)(uint8_t, uint8_t, const void*)) {
	C_ABORT_ON(list == NULL, "Trying to print list at NULL...\n");
	C_ABORT_ON(print_data == NULL, "Trying to print list at %p with NULL function...\n");
#ifdef C_ENABLE_LIST_CONSISTENCY_CHECKS
	list_check_consistency(list);
#endif	// C_ENABLE_LIST_CONSISTENCY_CHECKS

	C_DEBUG(debug_level, msg_type, "[");
	if(list->size != 0) {
		_list_item_t* current_item = list->first_item;
		print_data(debug_level, msg_type, current_item->data);
		current_item = current_item->next;
		while(current_item != NULL) {
			C_DEBUG(debug_level, msg_type, ", ");
			print_data(debug_level, msg_type, current_item->data);
			current_item = current_item->next;
		}
	}
	C_DEBUG(debug_level, msg_type, "]");
}

/**
 * @brief Apply a function to all data items in a list.
 *
 * @param list Pointer to the list.
 * @param function Function to apply on data items.
 */
void list_map(const list_t* list, void(*function)(const void*)) {
	_list_item_t* current_item;

	C_ABORT_ON(list == NULL, "Trying to apply function to list at NULL...\n");
	C_ABORT_ON(function == NULL, "Trying to apply NULL function to list at %p...\n");
#ifdef C_ENABLE_LIST_CONSISTENCY_CHECKS
	list_check_consistency(list);
#endif	// C_ENABLE_LIST_CONSISTENCY_CHECKS

	current_item = list->first_item;
	while(current_item != NULL) {
		function(current_item->data);
		current_item = current_item->next;
	}
}

/**
 * @brief Sort a list using a given function for comparisson.
 *
 * The algorithm used for sorting is BubbleSort, because the list
 * implementation is single-linked.
 *
 * Time complexity:	O(list_size ^ 2)
 * Space complexity:	O(1)
 *
 * @param list Pointer to the list.
 * @param less_than Function to use for comparing data items.
 */
void list_sort(list_t* list, bool(*less_than)(const void*, const void*)) {
	unsigned int i = 0;
	unsigned int j = 0;
	_list_item_t* current_item = NULL;
	_list_item_t* next_item = NULL;
	void* tmp_data = NULL;

	C_ABORT_ON(list == NULL, "Trying to sort list at NULL...\n");
	C_ABORT_ON(less_than == NULL, "Trying to sort list at %p with NULL function...\n");
#ifdef C_ENABLE_LIST_CONSISTENCY_CHECKS
	list_check_consistency(list);
#endif	// C_ENABLE_LIST_CONSISTENCY_CHECKS

	C_DEBUG(LIST_DEBUG_LEVEL, C_MSG_DEBUG, "Sorting list at %p...\n", list);

	if(list->size == 0)
		return;

	for(i = list->size - 1; i > 0; i--) {
		current_item = list->first_item;
		for(j = 0; j < i; j++) {
			next_item = current_item->next;
			if(less_than(next_item->data, current_item->data)) {
				tmp_data = current_item->data;
				current_item->data = next_item->data;
				next_item->data = tmp_data;
			}
			current_item = current_item->next;
		}
	}

#ifdef C_ENABLE_LIST_CONSISTENCY_CHECKS
	list_check_consistency(list);
#endif	// C_ENABLE_LIST_CONSISTENCY_CHECKS
}

/**
 * @brief Initialize an iterator for a list.
 *
 * @param list Pointer to the list.
 * @param iterator Pointer to the uninitialized iterator.
 */
void list_iterator_init(const list_t* list, list_iterator_t* iterator) {
	C_ABORT_ON(list == NULL, "Trying to initialize iterator for list at NULL...\n");
	C_ABORT_ON(iterator == NULL, "Trying to initialize iterator at NULL...\n");
#ifdef C_ENABLE_LIST_CONSISTENCY_CHECKS
	list_check_consistency(list);
#endif	// C_ENABLE_LIST_CONSISTENCY_CHECKS

	iterator->list = list;
	iterator->current_item = NULL;
#ifdef C_ENABLE_CONSISTENCY_CHECKS
	iterator->list_size = list->size;
#endif	// C_ENABLE_CONSISTENCY_CHECKS
}

/**
 * @brief Move iterator to the next element of the list and return the
 * corresponding data item.
 *
 * A side effect of this implementation is that an iterator that has reached
 * the end of the list can be reused for another iteration.
 *
 * @param iterator Pointer to the iterator.
 * @return Pointer to the next data item or NULL if the end of the list has
 * been reached.
 */
void* list_iterator_next(list_iterator_t* iterator) {
	C_ABORT_ON(iterator == NULL, "Trying to access next element on iterator at NULL...\n");
#ifdef C_ENABLE_CONSISTENCY_CHECKS
	C_ABORT_ON(iterator->list_size != iterator->list->size, "Trying to access next element on iterator at %p, but the size of the list at %p has changed from %d to %d.\n", iterator, iterator->list, iterator->list_size, iterator->list->size);
#endif	// C_ENABLE_CONSISTENCY_CHECKS
#ifdef C_ENABLE_LIST_CONSISTENCY_CHECKS
	list_check_consistency(iterator->list);
#endif	// C_ENABLE_LIST_CONSISTENCY_CHECKS

	// Set iterator to first element, if available.
	if(iterator->current_item == NULL) {
		if(iterator->list->size > 0) {
			iterator->current_item = iterator->list->first_item;
			return iterator->current_item->data;
		} else {
			return NULL;
		}

	// Set iterator to NULL upon reaching the end of the list.
	} else if(iterator->current_item == iterator->list->last_item) {
		iterator->current_item = NULL;
		return NULL;

	// Move iterator to the next element.
	} else {
		iterator->current_item = iterator->current_item->next;
		return iterator->current_item->data;
	}
}

/**
 * @brief Convert a list into an array.
 *
 * @param list Pointer to the list.
 * @param array Pointer to the array.
 * @param data_item_size Size of the data items of the list in bytes.
 */
void list_to_array(const list_t* list, void* array, size_t data_item_size) {
	_list_item_t* current_item;
	char* position = (char*) array;

	C_ABORT_ON(list == NULL, "Trying to convert list at NULL into array...\n");
	C_ABORT_ON(array == NULL, "Trying to convert list at %p into array at NULL...\n");
	C_ABORT_ON(data_item_size == 0, "Trying to convert list at %p into array, but data item size is 0...\n");
#ifdef C_ENABLE_LIST_CONSISTENCY_CHECKS
	list_check_consistency(list);
#endif	// C_ENABLE_LIST_CONSISTENCY_CHECKS

	current_item = list->first_item;
	while(current_item != NULL) {
		memcpy(position, current_item->data, data_item_size);
		position += data_item_size;
		current_item = current_item->next;
	}
}

/**
 * @brief Fill a list with the content of an array.
 *
 * @param list Pointer to the list.
 * @param array Pointer to the array.
 * @param data_item_size Size of the data items of the list in bytes.
 * @param array_size Size of the array in bytes.
 */
void list_from_array(list_t* list, const void* array, size_t data_item_size, size_t array_size) {
	char* position = NULL;

	C_ABORT_ON(list == NULL, "Trying to fill list at NULL from array...\n");
	C_ABORT_ON(array == NULL, "Trying to fill list at %p from array at NULL...\n");
	C_ABORT_ON(data_item_size == 0, "Trying to fill list at %p from array, but data item size is 0...\n");
	C_ABORT_ON((array_size % data_item_size) != 0, "Trying to fill list at %p from array, but array size is not a multiple of data_item_size (%d != %d)...\n", array_size, data_item_size);
	C_ABORT_ON(list->size != 0, "Trying to fill list at %p from array, but data list is not empty...\n");
#ifdef C_ENABLE_LIST_CONSISTENCY_CHECKS
	list_check_consistency(list);
#endif	// C_ENABLE_LIST_CONSISTENCY_CHECKS

	for(position = (char*) array; position < (((char*) array) + array_size); position += data_item_size) {
		char* p = NULL;
		MALLOC_ARRAY(p, char, data_item_size);
		memcpy(p, position, data_item_size);
		list_add(list, p);
	}
}
