/**
 * @file net.c
 *
 * @brief Network utility functions.
 *
 * Copyright (c) 2007, 2008, 2009, 2010 Georg Wittenburg, Freie Universitaet
 * Berlin (wittenbu@inf.fu-berlin.de). All rights reserved.
 * Please refer to the COPYING file for detailed licensing information.
 *
 * @author Georg Wittenburg
 *
 * $Id: net.c 3577 2010-06-10 10:14:45Z gwitten $
 */

#include <float.h>
#include <stdlib.h>
#include <string.h>

#ifdef ARCH_LINUX
#include <arpa/inet.h>
#include <netinet/in.h>
#include <sys/socket.h>
#elif defined(ARCH_WIN32)
#include <winsock2.h>
#endif	// ARCH_*

#include "../config.h"
#include "../macros.h"
#include "../types.h"
#include "../vars.h"
#include "../main.h"
#include "debug.h"
#include "list.h"
#include "callback.h"
#include "net.h"
#include "mapping.h"
#include "../icmp-light/icmp-light.h"



/**
 * @brief Debug level for this component.
 */
#define NET_DEBUG_LEVEL	C_MSG_WARN

/**
 * @brief Convert a numeric node address to a string.
 *
 * This function may be called several times in function parameters, so we have
 * to be careful to use different buffers every time.
 *
 * @param address Numeric node address.
 * @return String representation of the node address.
 */
char* addr_to_str(C_addr_t address) {
	int index = ++C_addr_to_str_buffer_index % C_ADDR_TO_STR_BUFFER_SIZE;
#ifdef ARCH_NS2
	if(address != C_BROADCAST_ADDRESS)
		snprintf(C_addr_to_str_buffer[index], C_ADDR_TO_STR_BUFFER_STR_LEN, "node #%d", address);
	else
		snprintf(C_addr_to_str_buffer[index], C_ADDR_TO_STR_BUFFER_STR_LEN, "BROADCAST");
#else
	struct in_addr tmp_addr;
	char* str = NULL;
	tmp_addr.s_addr = address;
	str = inet_ntoa(tmp_addr);
#ifdef ARCH_WIN32
	strncpy_s(C_addr_to_str_buffer[index], C_ADDR_TO_STR_BUFFER_STR_LEN, str, strlen(str));
#else
	strncpy(C_addr_to_str_buffer[index], str, C_ADDR_TO_STR_BUFFER_STR_LEN);
#endif	// ARCH_WIN32
#endif	// ARCH_*
	return C_addr_to_str_buffer[index];
}

/**
 * @brief Check a numeric node address for consistency.
 *
 * @param address Numeric node address.
 * @return <code>true</code> if the address is valid, <code>false</code>
 * otherwise.
 */
bool is_valid_address(C_addr_t address) {
#ifdef C_ENABLE_CONSISTENCY_CHECKS
#ifndef ARCH_NS2
	struct in_addr tmp_addr;
	tmp_addr.s_addr = address;
        unsigned int n1, n2, n3, n4;
	C_ABORT_ON(sscanf(inet_ntoa(tmp_addr), "%u.%u.%u.%u", &n1, &n2, &n3, &n4) != 4, "Invalid address %s (%d, 0x%X).\n", addr_to_str(address), address, address);
#endif	// ARCH_NS2

	if(
		(address == C_BROADCAST_ADDRESS)
#ifdef ARCH_NS2
		|| (address < C_MAXIMAL_NUMBER_OF_NODES)
#else
		|| ((n1 == 10) && (n2 < 256) && (n3 < 256) && (n4 < 256))	// 10.0.0.0/8 (private address range)
		|| ((n1 == 172) && (n2 >= 16) && (n2 < 32) && (n3 < 256) && (n4 < 256))	// 172.16.0.0/12 (private address range)
		|| ((n1 == 192) && (n2 == 168) && (n3 < 256) && (n4 < 256))	// 192.168.0.0/16 (private address range)
		|| ((n1 == 160) && (n2 == 45) && (n3 < 256) && (n4 < 256))	// Freie Universitaet Berlin
#endif	// ARCH_NS2
	) {
		return true;
	}
	return false;
#endif	// C_ENABLE_CONSISTENCY_CHECKS
}

/**
 * @brief Check whether two addresses match.
 *
 * @param item Address to check.
 * @param cmp_item Reference address.
 * @return <code>true</code> if the two addresses are equal, <code>false</code>
 * otherwise.
 */
bool matches_address(const void* item, const void* cmp_item) {
	C_addr_t* address = (C_addr_t*) item;
	C_addr_t* cmp_address = (C_addr_t*) cmp_item;
	return (*address == *cmp_address);
}



/**
 * @brief Allocate and initialize a packet header.
 *
 * @param header Pointer to pointer to packet header to allocate.
 * @param destination_address Address of destination node.
 * @param priority Priority of the packet.
 * @param app Application identifier.
 * @param type Packet type identifier.
 * @param payload_length Length of the payload in bytes.
 */
void alloc_header(net_header_t** header, C_addr_t destination_address, uint8_t priority, uint8_t app, uint8_t type, size_t payload_length) {
#ifdef C_USE_NETWORK_MAPPING
	unsigned int i = 0;
	node_t* node = NULL;
	list_iterator_t C_known_nodes_iterator;
#endif	// C_USE_NETWORK_MAPPING

	C_ABORT_ON(!is_valid_address(destination_address), "Trying to allocate packet header with invalid address (%s).\n", addr_to_str(destination_address));
	C_ABORT_ON((priority < PACKET_PRIORITY_MIN) || (priority > PACKET_PRIORITY_MAX), "Trying to allocate packet header with invalid packet priority (0x%x).\n", priority);

	MALLOC(*header, net_header_t);
	(*header)->source_address = C_LOCAL_ADDRESS;
	(*header)->destination_address = destination_address;
	(*header)->last_hop_address = C_LOCAL_ADDRESS;
	(*header)->next_hop_address = C_BROADCAST_ADDRESS;
	(*header)->ttl = C_DEFAULT_TTL;
	(*header)->priority = priority;
	(*header)->seqno = C_sequence_number++;
	(*header)->app = app;
	(*header)->type = type;
	(*header)->hop_count = 0;
	(*header)->neighbor_count = 0;
	(*header)->payload_length = payload_length;

#ifdef C_USE_NETWORK_MAPPING
	(*header)->hops = NULL;
	(*header)->neighbors = NULL;

	list_iterator_init(C_known_nodes, &C_known_nodes_iterator);
	while((node = (node_t*) list_iterator_next(&C_known_nodes_iterator)) != NULL) {
#ifdef C_ENABLE_CONSISTENCY_CHECKS
		C_ABORT_ON(!is_valid_address(node->address), "Found invalid address in list of known nodes while allocating packet header (%s).\n", addr_to_str(node->address));
#endif	// C_ENABLE_CONSISTENCY_CHECKS
		if(exists_known_link(C_LOCAL_ADDRESS, node->address)) {
			if((*header)->neighbors == NULL)
				MALLOC_ARRAY((*header)->neighbors, C_addr_t, C_MAXIMAL_NODE_DEGREE);
			C_DEBUG(NET_DEBUG_LEVEL, C_MSG_VERBOSE, "%s is a direct neighbor of this node.\n", addr_to_str(node->address));
			(*header)->neighbors[(*header)->neighbor_count] = node->address;
			(*header)->neighbor_count++;
			if((*header)->neighbor_count >= C_MAXIMAL_NODE_DEGREE) {
				C_DEBUG(NET_DEBUG_LEVEL, C_MSG_ERROR, "Number of neighbors exceeds maximal expected number of neighbors (%d).\n", C_MAXIMAL_NODE_DEGREE);
				break;
			}
		}
	}

#ifdef C_ENABLE_CONSISTENCY_CHECKS
	for(i = 0; i < (*header)->neighbor_count; i++)
		C_ABORT_ON(!is_valid_address((*header)->neighbors[i]), "Allocated packet header with invalid address for neighbor #%d/%d (%s).\n", (i + 1), (*header)->neighbor_count, addr_to_str((*header)->neighbors[i]));
#endif	// C_ENABLE_CONSISTENCY_CHECKS
#endif	// C_USE_NETWORK_MAPPING

	C_DEBUG(NET_DEBUG_LEVEL, C_MSG_DEBUG, "Packet #%d to %s initialized for %d byte(s) of payload.\n", (*header)->seqno, addr_to_str((*header)->destination_address), payload_length);
}

/**
 * @brief Allocate and initialize a packet header based on an existing header.
 *
 * @param new_header Pointer to pointer to packet header to allocate.
 * @param orig_header Pointer to original packet header.
 */
void alloc_header_from_header(net_header_t** new_header, const net_header_t* orig_header) {
	C_ABORT_ON(!is_valid_address(orig_header->source_address), "Trying to allocate packet header with invalid source address (%s).\n", addr_to_str(orig_header->source_address));
	C_ABORT_ON(!is_valid_address(orig_header->destination_address), "Trying to allocate packet header with invalid destination address (%s).\n", addr_to_str(orig_header->destination_address));
	C_ABORT_ON(!is_valid_address(orig_header->last_hop_address), "Trying to allocate packet header with invalid last hop address (%s).\n", addr_to_str(orig_header->last_hop_address));
	C_ABORT_ON(!is_valid_address(orig_header->next_hop_address), "Trying to allocate packet header with invalid next hop address (%s).\n", addr_to_str(orig_header->next_hop_address));

	MALLOC(*new_header, net_header_t);
	memcpy(*new_header, orig_header, sizeof(net_header_t));
#ifdef C_USE_NETWORK_MAPPING
	(*new_header)->hops = NULL;
	(*new_header)->neighbors = NULL;
	if(orig_header->hop_count != 0) {
		MALLOC_ARRAY((*new_header)->hops, C_addr_t, orig_header->hop_count);
		memcpy((*new_header)->hops, orig_header->hops, (orig_header->hop_count * sizeof(C_addr_t)));
	}
	if(orig_header->neighbor_count != 0) {
		MALLOC_ARRAY((*new_header)->neighbors, C_addr_t, orig_header->neighbor_count);
		memcpy((*new_header)->neighbors, orig_header->neighbors, (orig_header->neighbor_count * sizeof(C_addr_t)));
	}
#endif	// C_USE_NETWORK_MAPPING
}

/**
 * @brief Free a packet header.
 *
 * @param header Pointer to pointer to packet header to free.
 */
void free_header(net_header_t** header) {
#ifdef C_USE_NETWORK_MAPPING
	C_ABORT_ON((((*header)->hop_count == 0) && ((*header)->hops != NULL)) || (((*header)->hop_count != 0) && ((*header)->hops == NULL)), "Hop data is inconsistent for packet #%d from %s to %s.\n", (*header)->seqno, addr_to_str((*header)->source_address), addr_to_str((*header)->destination_address));
	C_ABORT_ON((((*header)->neighbor_count == 0) && ((*header)->neighbors != NULL)) || (((*header)->neighbor_count != 0) && ((*header)->neighbors == NULL)), "Neighbor data is inconsistent for packet #%d from %s to %s.\n", (*header)->seqno, addr_to_str((*header)->source_address), addr_to_str((*header)->destination_address));
	if((*header)->hops != NULL) {
		FREE((*header)->hops);
	}
	if((*header)->neighbors != NULL) {
		FREE((*header)->neighbors);
	}
#endif	// C_USE_NETWORK_MAPPING
	FREE(*header);
}

/**
 * @brief Send a packet with application parameters and payload.
 *
 * The packet is routed to the destination node.
 *
 * @param destination_address Address of destination node.
 * @param priority Priority of the packet.
 * @param app Application identifier.
 * @param type Packet type identifier.
 * @param payload Pointer to payload.
 * @param payload_length Length of payload in bytes.
 */
void C_sendto(C_addr_t destination_address, uint8_t priority, uint8_t app, uint8_t type, const void* payload, size_t payload_length) {
	net_header_t* header = NULL;

	C_ABORT_ON(((payload == NULL) && (payload_length != 0)) || ((payload != NULL) && (payload_length == 0)), "Trying to assemble packet for sending to %s, but payload parameters are inconsistent.\n", addr_to_str(destination_address));

	C_DEBUG(NET_DEBUG_LEVEL, C_MSG_VERBOSE, "Assembling packet for sending to %s...\n", addr_to_str(destination_address));

	alloc_header(&header, destination_address, priority, app, type, payload_length);
	if(payload_length == 0)
		payload = NULL;
	C_route(header, payload);
	free_header(&header);
}

/**
 * @brief Send a packet with application parameters and two payloads.
 *
 * The packet is routed to the destination node.
 *
 * @param destination_address Address of destination node.
 * @param priority Priority of the packet.
 * @param app Application identifier.
 * @param type Packet type identifier.
 * @param payload1 Pointer to first payload.
 * @param payload_length1 Length of first payload in bytes.
 * @param payload2 Pointer to second payload.
 * @param payload_length2 Length of second payload in bytes.
 */
void C_sendto2(C_addr_t destination_address, uint8_t priority, uint8_t app, uint8_t type, const void* payload1, size_t payload_length1, const void* payload2, size_t payload_length2) {
	void* payload = NULL;

	C_ABORT_ON(((payload1 == NULL) && (payload_length1 != 0)) || ((payload1 != NULL) && (payload_length1 == 0)) || ((payload2 == NULL) && (payload_length2 != 0)) || ((payload2 != NULL) && (payload_length2 == 0)), "Trying to assemble packet for sending to %s, but payload parameters are inconsistent.\n", addr_to_str(destination_address));

	MALLOC_JOIN(payload, char, payload1, payload_length1, payload2, payload_length2);
	C_sendto(destination_address, priority, app, type, payload, (payload_length1 + payload_length2));
	FREE(payload);
}

/**
 * @brief Maximal randomized delay before forwarding a packet in seconds.
 *
 * Uncommenting this define causes packet to be forwarded immediately.
 */
//#define FORWARD_DELAY	C_EXPECTED_MULTI_HOP_RTT

#ifdef FORWARD_DELAY
/**
 * @brief Packet forwarding callback to support delayed forwarding.
 *
 * @param parameter Pointer to packet to be forwarded.
 */
static void forward_callback(callback_parameter_t parameter) {
	C_packet_t* packet = (C_packet_t*) parameter;
	C_route(packet->header, packet->payload);
	if(packet->header->payload_length > 0)
		FREE(packet->payload);
	free_header(packet->header);
	FREE(packet);
}
#endif	// FORWARD_DELAY

/**
 * @brief Forward a broadcast packet.
 *
 * May be configured to add a delay for broadcast packets to avoid collisions.
 *
 * @param orig_header Pointer to original header of packet to be forwarded.
 * @param payload Pointer to payload of packet to be forwarded.
 */
void forward(const net_header_t* orig_header, const void* payload) {
	net_header_t* new_header = NULL;
#ifdef FORWARD_DELAY
	C_packet_t* packet = NULL;
#endif	// FORWARD_DELAY

	C_ABORT_ON(!is_valid_address(orig_header->source_address), "Trying to forward packet with invalid source address (%s).\n", addr_to_str(orig_header->source_address));
	C_ABORT_ON(!is_valid_address(orig_header->destination_address), "Trying to forward packet with invalid destination address (%s).\n", addr_to_str(orig_header->destination_address));
	C_ABORT_ON(!is_valid_address(orig_header->last_hop_address), "Trying to forward packet with invalid last hop address (%s).\n", addr_to_str(orig_header->last_hop_address));
	C_ABORT_ON(!is_valid_address(orig_header->next_hop_address), "Trying to forward packet with invalid next hop address (%s).\n", addr_to_str(orig_header->next_hop_address));
	C_ABORT_ON(orig_header->destination_address != C_BROADCAST_ADDRESS, "Trying to forward a unicast packet.\n");
	C_ABORT_ON(orig_header->ttl == 0, "Trying to forward packet with expired TTL.\n");

#ifndef FORWARD_DELAY
	C_DEBUG(NET_DEBUG_LEVEL, C_MSG_VERBOSE, "Forwarding packet...\n");
#else
	C_DEBUG(NET_DEBUG_LEVEL, C_MSG_VERBOSE, "Forwarding packet with delay...\n");
#endif	// FORWARD_DELAY
	alloc_header_from_header(&new_header, orig_header);
	new_header->last_hop_address = C_LOCAL_ADDRESS;

#ifndef FORWARD_DELAY
	C_route(new_header, payload);
	free_header(&new_header);
#else
	MALLOC(packet, C_packet_t);
	packet->header = new_header;
	if(packet->header->payload_length > 0) {
		void* p = NULL;
		MALLOC_ARRAY(p, char, packet->header->payload_length);
		memcpy(p, payload, packet->header->payload_length);
		packet->payload = p;
	} else {
		packet->payload = NULL;
	}
	callback_add(forward_callback, packet, C_random(FORWARD_DELAY));
#endif	// FORWARD_DELAY
}



#ifdef C_ENABLE_BEACONING
#ifdef C_USE_ETX_LINK_METRIC
/**
 * @brief A node address with the corresponding beacon count.
 */
typedef struct {
	C_addr_t	address;		///< Address of the node for which the beacon count is reported.
	uint8_t		received_beacons;	///< Beacon count of this node.
} address_and_beacon_counter_t;

/**
 * @brief Purge beacon counters.
 *
 * @param node Pointer to node whose beacon counters are to be purged.
 */
static void purge_beacon_counters(node_t* node) {
	C_time_t current_time = C_time();

	if((current_time - node->last_beacon_purge_time) >= (2 * C_ETX_WINDOW_SIZE)) {
		node->received_beacons = 0;
		node->received_beacons_counter = 0;
		node->last_beacon_purge_time = current_time;
	} else if((current_time - node->last_beacon_purge_time) >= C_ETX_WINDOW_SIZE) {
		node->received_beacons = node->received_beacons_counter;
		node->received_beacons_counter = 0;
		node->last_beacon_purge_time = current_time;
	} else if((current_time - node->creation_time) < C_ETX_WINDOW_SIZE) {
		node->received_beacons = node->received_beacons_counter;
	}
}

/**
 * @brief Calculate the ETX for a node.
 *
 * This function implements the calcuation of the ETX link metric as described
 * in: Douglas S. J. De Couto, Daniel Aguayo, John Bicket, and Robert Morris.
 * "A High-Throughput Path Metric for Multi-Hop Wireless Routing." Proc. ACM
 * MobiCom, September 2003.
 *
 * @param address Address of the node for which the ETX is to be calculated.
 */
float calculate_ETX(C_addr_t address) {
	node_t* node = NULL;
	C_time_t current_time = C_time();
	float max_beacons = 0.0;
	float df = 0.0;		// Forward delivery ratio.
	float dr = 0.0;		// Reverse delivery ratio.
	float etx = 0.0;

	C_ABORT_ON(!exists_known_node(address), "Trying to calculate ETX for %s, but node is unknown.\n", addr_to_str(address));
	node = get_known_node(address);

	purge_beacon_counters(node);
	if((node->received_beacons == 0) || (node->reported_beacons == 0)) {
		C_DEBUG(NET_DEBUG_LEVEL, C_MSG_VERBOSE, "ETX for the link between this node and %s is infinite, with %d beacon(s) received and %d beacon(s) reported.\n", addr_to_str(address), node->received_beacons, node->reported_beacons);
		return FLT_MAX;
	}

	// Calculate maximal beacon count depending on node creation time.
	if((current_time - node->creation_time) >= C_ETX_WINDOW_SIZE) {
		max_beacons = ((float) C_ETX_WINDOW_SIZE) / ((float) C_BEACON_INTERVAL);
	} else {
		max_beacons = ((float) (current_time - node->creation_time)) / ((float) C_BEACON_INTERVAL);
		max_beacons += 1.0;	// Add 1.0 to be slightly more pessimistic than during normal operation.
	}

	df = (node->reported_beacons / max_beacons);
	dr = (node->received_beacons / max_beacons);
	etx = (1.0 / (df * dr));

	C_DEBUG(NET_DEBUG_LEVEL, C_MSG_VERBOSE, "ETX for the link between this node and %s is (1.0 / (%.4f * %.4f) = %.4f, with %.1f expected beacons.\n", addr_to_str(address), df, dr, etx, max_beacons);
	return etx;
}
#endif	// C_USE_ETX_LINK_METRIC

/**
 * @brief Send a beacon packet.
 *
 * @param destination_address Address of the destination node.
 */
static void send_BEACON(C_addr_t destination_address) {
	net_header_t* header = NULL;
#ifdef C_USE_ETX_LINK_METRIC
	address_and_beacon_counter_t* address_and_beacon_counters = NULL;
	size_t payload_length = sizeof(address_and_beacon_counter_t) * list_size(C_known_nodes);
	unsigned int i = 0;

	if(payload_length > 0) {
		MALLOC_ARRAY(address_and_beacon_counters, address_and_beacon_counter_t, list_size(C_known_nodes));
		node_t* node = NULL;
		list_iterator_t C_known_nodes_iterator;
		list_iterator_init(C_known_nodes, &C_known_nodes_iterator);
		while((node = (node_t*) list_iterator_next(&C_known_nodes_iterator)) != NULL) {
			if(node->minimal_hop_count == 1) {
				purge_beacon_counters(node);
				address_and_beacon_counters[i].address = node->address;
				address_and_beacon_counters[i].received_beacons = node->received_beacons;
				i++;
			}
		}
		payload_length = sizeof(address_and_beacon_counter_t) * i;
	}

	C_DEBUG(C_MAIN_DEBUG_LEVEL, C_MSG_VERBOSE, "Sending BEACON with information about %d link(s)...\n", i);
#endif	// C_USE_ETX_LINK_METRIC

	alloc_header(&header, destination_address, PACKET_PRIORITY_SIGNALLING_LINK, APP_TYPE_INTERNAL, PACKET_TYPE_BEACON, 0);
	header->next_hop_address = destination_address;
	header->ttl = 1;
#ifdef C_USE_ETX_LINK_METRIC
	header->payload_length = payload_length;
	C_send(header, address_and_beacon_counters);
#else
	C_send(header, NULL);
#endif	// C_USE_ETX_LINK_METRIC
	free_header(&header);
#ifdef C_USE_ETX_LINK_METRIC
	FREE(address_and_beacon_counters);
#endif	// C_USE_ETX_LINK_METRIC
}

/**
 * @brief Beaconing callback.
 *
 * @param parameter Not used.
 */
void beacon_callback(callback_parameter_t parameter) {
	float delay = C_BEACON_INTERVAL;
	float jitter = (0.1 * C_BEACON_INTERVAL);

	delay += (C_random(2.0 * jitter) - jitter);
	C_beacon_callback_id = callback_add(beacon_callback, NULL, delay);
	send_BEACON(C_BROADCAST_ADDRESS);
}

/**
 * @brief Handle an incoming BEACON packet.
 *
 * @param header Pointer to header of the incoming packet.
 * @param payload Pointer to the payload.
 */
void handle_BEACON(const net_header_t* header, const void* payload) {
#ifdef C_USE_ETX_LINK_METRIC
	node_t* node = register_known_node(header->source_address);
	purge_beacon_counters(node);
	node->received_beacons_counter++;
	purge_beacon_counters(node);	// We have to purge a second time to ensure that the counter is handled correctly if the node has only been created recently.

	if(header->payload_length > 0) {
		address_and_beacon_counter_t* address_and_beacon_counters = (address_and_beacon_counter_t*) payload;
		unsigned int i = 0;

		C_ABORT_ON(payload == NULL, "Received BEACON from %s with %d byte(s) payload, but the payload is NULL.\n", addr_to_str(header->source_address), header->payload_length);

		for(i = 0; i < (header->payload_length / sizeof(address_and_beacon_counter_t)); i++) {
			if(address_and_beacon_counters[i].address == C_LOCAL_ADDRESS) {
				node->reported_beacons = address_and_beacon_counters[i].received_beacons;
				C_DEBUG(NET_DEBUG_LEVEL, C_MSG_VERBOSE, "Received beacon from %s reporting %d received beacons from this node (own counters: %d, %d). ETX is %.4f.\n", addr_to_str(header->source_address), node->reported_beacons, node->received_beacons, node->received_beacons_counter, calculate_ETX(header->source_address));
				break;
			}
		}
	}

	C_DEBUG(C_MAIN_DEBUG_LEVEL, C_MSG_VERBOSE, "Received BEACON from %s with information about %d link(s). ETX to %s is %.4f.\n", addr_to_str(header->source_address), (header->payload_length / sizeof(address_and_beacon_counter_t)), addr_to_str(header->source_address), calculate_ETX(header->source_address));
#endif	// C_USE_ETX_LINK_METRIC
}
#endif	// C_ENABLE_BEACONING



/**
 * @brief Check for reception of duplicate packets.
 *
 * Check whether a packet has already been received recently and add it to the
 * corresponding data structure for future checks.
 *
 * @param cache Pointer to the packet cache array.
 * @param cache_size Size of the packet cache array.
 * @param header Pointer to header of a recently received packet.
 * @return <code>true</code> if the packet is a duplicate based on its
 * fingerprint, <code>false</code> otherwise.
 */
bool check_for_duplicate(packet_fingerprint_t* cache, uint16_t cache_size, const net_header_t* header) {
	int i, LRU_index = -1;
	C_time_t current_time = C_time();
	C_time_t time_of_oldest_item = current_time;

	C_ABORT_ON(cache == NULL, "Pointer to packet cache is NULL.\n");
	C_ABORT_ON(cache_size == 0, "Packet cache size is 0.\n");
	C_ABORT_ON(header == NULL, "Trying to add packet header at NULL to packet cache.\n");

#ifdef C_ENABLE_BEACONING
	// Ignore beacon packets.
	if((header->app == APP_TYPE_INTERNAL) && (header->type == PACKET_TYPE_BEACON))
		return false;
#endif	// C_ENABLE_BEACONING

	// Do a reverse loop to use the beginning of the cache first and ease debugging.
	for(i = cache_size - 1; i >= 0; i--) {
		// Check whether we have recently seen a similar packet.
		if(
			(cache[i].time != 0)
			&& (cache[i].from == header->source_address)
			&& (cache[i].seqno == header->seqno)
		) {
			C_DEBUG(NET_DEBUG_LEVEL, C_MSG_VERBOSE, "Duplicate packet fingerprint (time=%.2f, from=%s, seqno=%d) found at index %d.\n", current_time, addr_to_str(header->source_address), header->seqno, i);
			cache[i].time = current_time;
			return true;
		}
		// Find the least recently used (LRU) entry in the data
		// structure for storing information about this packet.
		if(
			(cache[i].time == 0)
			|| (cache[i].time <= time_of_oldest_item)
		) {
			time_of_oldest_item = cache[i].time;
			LRU_index = i;
		}
	}
	C_ABORT_ON(LRU_index < 0, "Failed to find slot for packet fingerprint (time=%f, from=%s, seqno=%d).\n", current_time, addr_to_str(header->source_address), header->seqno);
	cache[LRU_index].time = current_time;
	cache[LRU_index].from = header->source_address;
	cache[LRU_index].seqno = header->seqno;
	C_DEBUG(NET_DEBUG_LEVEL, C_MSG_DEBUG, "Added packet fingerprint (time=%.2f, from=%s, seqno=%d) at index %d.\n", cache[LRU_index].time, addr_to_str(cache[LRU_index].from), cache[LRU_index].seqno, LRU_index);
	return false;
}



/*
 * Handling of known nodes.
 */

/**
 * @brief Allocate and initialize a node structure.
 *
 * @param node Pointer to pointer to node structure to allocate.
 * @param address Address of the node.
 */
void alloc_node(node_t** node, C_addr_t address) {
	MALLOC(*node, node_t);
	init_node(*node, address);
}

/**
 * @brief Initialize a node structure.
 *
 * @param node Pointer to node structure to allocate.
 * @param address Address of the node.
 */
void init_node(node_t* node, C_addr_t address) {
	C_ABORT_ON(!is_valid_address(address), "Trying to initialize node with invalid address (%s).\n", addr_to_str(address));

	C_DEBUG(NET_DEBUG_LEVEL, C_MSG_VERBOSE, "Initializing data structure for %s...\n", addr_to_str(address));
	node->address = address;
	node->creation_time = C_time();
	node->minimal_hop_count = (C_MAXIMAL_NETWORK_DIAMETER + 1);
#ifdef C_USE_INTERNAL_LINK_MONITORING
	node->unanswered_packets = 0;
	node->expected_packet_callback_id = 0;
	node->delayed_send_callback_id = 0;
#endif	// C_USE_INTERNAL_LINK_MONITORING
#ifdef C_USE_ETX_LINK_METRIC
	node->received_beacons = 0;
	node->received_beacons_counter = 0;
	node->reported_beacons = 0;
	node->last_beacon_purge_time = C_time();
#endif	// C_USE_ETX_LINK_METRIC
}

/**
 * @brief Free a node structure.
 *
 * @param node Pointer to pointer to node structure to free.
 */
void free_node(node_t** node) {
	free_node_list(*node);
	*node = NULL;
}

/**
 * @brief Free a node structure (list interface).
 *
 * @param p Pointer to node structure to free.
 */
void free_node_list(void* p) {
	node_t* node = (node_t*) p;
	FREE(node);
}

/**
 * @brief Check whether two nodes have matching addresses.
 *
 * @param item Node to check.
 * @param cmp_item Reference node.
 * @return <code>true</code> if the two nodes have the same address,
 * <code>false</code> otherwise.
 */
bool matches_node_address(const void* item, const void* cmp_item) {
	node_t* node = (node_t*) item;
	node_t* cmp_node = (node_t*) cmp_item;
	return (node->address == cmp_node->address);
}

/**
 * @brief Compare the addresses of two nodes.
 *
 * @param item Node to check.
 * @param cmp_item Reference node.
 * @return <code>true</code> if the node has a smaller address than the
 * reference node, <code>false</code> otherwise.
 */
bool node_address_less_than(const void* item, const void* cmp_item) {
	node_t* node = (node_t*) item;
	node_t* cmp_node = (node_t*) cmp_item;
	return (node->address < cmp_node->address);
}

/**
 * @brief Print the address of a node.
 *
 * @param debug_level Numeric debug level of the message.
 * @param msg_type Numeric type of the message.
 * @param item Pointer to node whose address is to be printed.
 */
void print_node_address(uint8_t debug_level, uint8_t msg_type, const void* item) {
	C_DEBUG(debug_level, msg_type, "%s", addr_to_str(((node_t*) item)->address));
}



/*
 * Basic mapping and active link monitoring functions.
 */

/**
 * @brief Register a known node.
 *
 * @param address Address of the node to register.
 * @return Pointer to the node that was registered.
 */
node_t* register_known_node(C_addr_t address) {
	node_t* node = NULL;
	node_t cmp_node;

	C_ABORT_ON(address == C_BROADCAST_ADDRESS, "Trying to register the broadcast address as known node.\n");

	cmp_node.address = address;
	if((node = (node_t*) list_find(C_known_nodes, &cmp_node, matches_node_address)) == NULL) {
		C_ABORT_ON(list_size(C_known_nodes) >= C_MAXIMAL_NUMBER_OF_NODES, "Trying to register %s, but this would exceed the maximal number of supported nodes (%d).\n", addr_to_str(address), C_MAXIMAL_NUMBER_OF_NODES);
		C_DEBUG(NET_DEBUG_LEVEL, C_MSG_INFO, "Adding %s to list of known nodes...\n", addr_to_str(address));
		alloc_node(&node, address);
		list_add(C_known_nodes, node);
	}
	return node;
}

/**
 * @brief Register a known link.
 *
 * @param src_address Address of the source node.
 * @param dst_address Address of the destination node.
 */
void register_known_link(C_addr_t src_address, C_addr_t dst_address) {
	C_ABORT_ON(src_address == C_BROADCAST_ADDRESS, "Trying to register known link from broadcast address...\n");
	C_ABORT_ON(dst_address == C_BROADCAST_ADDRESS, "Trying to register known link to broadcast address...\n");

	C_DEBUG(NET_DEBUG_LEVEL, C_MSG_VERBOSE, "Registering known link between %s and %s...\n", addr_to_str(src_address), addr_to_str(dst_address));
	register_known_node(src_address);
	register_known_node(dst_address);

#ifdef C_USE_NETWORK_MAPPING
	C_network_map[node_address_to_id(src_address) * C_MAXIMAL_NUMBER_OF_NODES + node_address_to_id(dst_address)] = C_time();
#ifdef C_ASSUME_BIDIRECTIONAL_LINKS
	C_network_map[node_address_to_id(dst_address) * C_MAXIMAL_NUMBER_OF_NODES + node_address_to_id(src_address)] = C_time();
#endif	// C_ASSUME_BIDIRECTIONAL_LINKS

	// Invalidate previous edge cost and distance calculations.
	if(C_edge_cost_matrix != NULL)
		FREE(C_edge_cost_matrix);
	if(C_distance_matrix != NULL)
		FREE(C_distance_matrix);
#endif	// C_USE_NETWORK_MAPPING
}

/**
 * @brief Register a reportedly known link.
 *
 * This is to be used for links that have been reported as known by other
 * nodes, but are not to be redistributed.
 *
 * @param src_address Address of the source node.
 * @param dst_address Address of the destination node.
 */
void register_reportedly_known_link(C_addr_t src_address, C_addr_t dst_address) {
	C_ABORT_ON(src_address == C_BROADCAST_ADDRESS, "Trying to register reportedly known link from broadcast address...\n");
	C_ABORT_ON(dst_address == C_BROADCAST_ADDRESS, "Trying to register reportedly known link to broadcast address...\n");

	C_DEBUG(NET_DEBUG_LEVEL, C_MSG_VERBOSE, "Registering reportedly known link between %s and %s...\n", addr_to_str(src_address), addr_to_str(dst_address));
	register_known_node(src_address);
	register_known_node(dst_address);

#ifdef C_USE_NETWORK_MAPPING
	C_network_map[node_address_to_id(src_address) * C_MAXIMAL_NUMBER_OF_NODES + node_address_to_id(dst_address)] = -1.0 * C_time();
#ifdef C_ASSUME_BIDIRECTIONAL_LINKS
	C_network_map[node_address_to_id(dst_address) * C_MAXIMAL_NUMBER_OF_NODES + node_address_to_id(src_address)] = -1.0 * C_time();
#endif	// C_ASSUME_BIDIRECTIONAL_LINKS

	// Invalidate previous edge cost and distance calculations.
	if(C_edge_cost_matrix != NULL)
		FREE(C_edge_cost_matrix);
	if(C_distance_matrix != NULL)
		FREE(C_distance_matrix);
#endif	// C_USE_NETWORK_MAPPING
}

/**
 * @brief Register a broken link.
 *
 * @param src_address Address of the source node.
 * @param dst_address Address of the destination node.
 */
void register_broken_link(C_addr_t src_address, C_addr_t dst_address) {
	C_ABORT_ON(src_address == C_BROADCAST_ADDRESS, "Trying to register broken link from broadcast address...\n");
	C_ABORT_ON(dst_address == C_BROADCAST_ADDRESS, "Trying to register broken link to broadcast address...\n");
	C_ABORT_ON(dst_address == C_LOCAL_ADDRESS, "Trying to register broken link to local address...\n");

	register_known_node(src_address);
	register_known_node(dst_address);

	C_DEBUG(NET_DEBUG_LEVEL, C_MSG_INFO, "Registering broken link between %s and %s...\n", addr_to_str(src_address), addr_to_str(dst_address));

#ifdef C_USE_NETWORK_MAPPING
	if(
		(C_network_map[node_address_to_id(src_address) * C_MAXIMAL_NUMBER_OF_NODES + node_address_to_id(dst_address)] == 0)
#ifdef C_ASSUME_BIDIRECTIONAL_LINKS
		&& (C_network_map[node_address_to_id(dst_address) * C_MAXIMAL_NUMBER_OF_NODES + node_address_to_id(src_address)] == 0)
#endif	// C_ASSUME_BIDIRECTIONAL_LINKS
	) {
		C_DEBUG(NET_DEBUG_LEVEL, C_MSG_DEBUG, "No link between %s and %s registered in network map, ignoring broken link report...\n", addr_to_str(src_address), addr_to_str(dst_address));
		return;
	}

	// Remove link from network map.
	C_network_map[node_address_to_id(src_address) * C_MAXIMAL_NUMBER_OF_NODES + node_address_to_id(dst_address)] = 0;
#ifdef C_ASSUME_BIDIRECTIONAL_LINKS
	C_network_map[node_address_to_id(dst_address) * C_MAXIMAL_NUMBER_OF_NODES + node_address_to_id(src_address)] = 0;
#endif	// C_ASSUME_BIDIRECTIONAL_LINKS

	// Invalidate previous edge cost and distance calculations.
	if(C_edge_cost_matrix != NULL)
		FREE(C_edge_cost_matrix);
	if(C_distance_matrix != NULL)
		FREE(C_distance_matrix);
#endif	// C_USE_NETWORK_MAPPING
}

#ifdef C_USE_INTERNAL_LINK_MONITORING
/**
 * @brief Timeout at which a packet should have been received from a monitored
 * node.
 *
 * If this timeout expires, the link to the monitored node is broken and the
 * route procotol needs remove the affected routes.
 *
 * @param parameter Pointer to monitored node.
 */
static void expected_packet_timout(callback_parameter_t parameter) {
	node_t* node = (node_t*) parameter;
	node->expected_packet_callback_id = 0;
	node->unanswered_packets++;
	if(node->unanswered_packets > INTERNAL_LINK_MONITORING_MAX_UNANSWERED_PACKETS) {
		node->unanswered_packets = 0;
		C_report_broken_link(node->address);
	}
}

/**
 * @brief Timeout at which a packet should have been send to a monitored node.
 *
 * Send an ICMP NOP packet (not an ICMP REPLY as suggested in the draft) in
 * case no other packet has been sent.
 *
 * @param parameter Pointer to monitored node.
 */
static void delayed_send_timeout(callback_parameter_t parameter) {
	node_t* node = (node_t*) parameter;

	C_DEBUG(NET_DEBUG_LEVEL, C_MSG_NOTE, "No packet sent to from %s in the last %.2f seconds, sending empty packet...\n", addr_to_str(node->address), (((float) C_UNICAST_MESSAGE_SENT_TIMEOUT) / 2.0));
	node->delayed_send_callback_id = 0;
	send_BEACON(node->address);
}
#endif	// C_USE_INTERNAL_LINK_MONITORING

/**
 * @brief Register a sent packet.
 *
 * Register a sent packet for active link monitoring. Set timer to wait for
 * packet from next hop node as part of active link monitoring.
 *
 * @param header Pointer to header of packet that has been sent.
 */
void register_sent_packet(const net_header_t* header) {
#ifdef C_USE_INTERNAL_LINK_MONITORING
	node_t* node = NULL;

	if(header->next_hop_address == C_BROADCAST_ADDRESS) {
		C_DEBUG(NET_DEBUG_LEVEL, C_MSG_VERBOSE, "Trying to register sent broadcast packet for active link monitoring, ignoring...\n");
		return;
	}
#endif	// C_USE_INTERNAL_LINK_MONITORING

	C_DEBUG(NET_DEBUG_LEVEL, C_MSG_DEBUG, "Registering sent packet #%d from %s to %s via %s...\n", header->seqno, addr_to_str(header->source_address), addr_to_str(header->destination_address), addr_to_str(header->next_hop_address));

	if(header->next_hop_address != C_BROADCAST_ADDRESS)
		register_known_node(header->next_hop_address);

#ifdef C_USE_INTERNAL_LINK_MONITORING
	node = register_known_node(header->next_hop_address);

	// Start timeout for expected packets.
	if(
		(node->expected_packet_callback_id == 0)	// Don't overwrite any existing timeout.
		&& (header->app != APP_TYPE_NONE)		// Don't start timer because APP_TYPE_NONE is used for signaling.
	) {
		node->expected_packet_callback_id = callback_add(expected_packet_timout, node, C_UNICAST_MESSAGE_SENT_TIMEOUT);
	}

	// Cancel timeout for delayed send.
	if(node->delayed_send_callback_id != 0) {
		callback_cancel(node->delayed_send_callback_id);
		node->delayed_send_callback_id = 0;
	}
#endif	// C_USE_INTERNAL_LINK_MONITORING
}

/**
 * @brief Register a received packet.
 *
 * Register a received packet for active link monitoring, mapping and network
 * statistics. Also cancel timer to wait for packet from last hop node as part
 * of active link monitoring.
 *
 * @param header Pointer to header of packet that has been received.
 */
void register_received_packet(const net_header_t* header) {
#ifdef C_USE_NETWORK_MAPPING
	unsigned int i = 0;
#endif	// C_USE_NETWORK_MAPPING
	node_t* node = NULL;

	C_ABORT_ON(header->last_hop_address == C_BROADCAST_ADDRESS, "Trying to register received packet from broadcast address.\n");

	C_DEBUG(NET_DEBUG_LEVEL, C_MSG_DEBUG, "Registering received packet #%d from %s to %s via last hop %s after %d hop(s)...\n", header->seqno, addr_to_str(header->source_address), addr_to_str(header->destination_address), addr_to_str(header->last_hop_address), header->hop_count);

	/*
	 * Mapping and statistics.
	 */

	node = register_known_node(header->source_address);
	if(node->minimal_hop_count > header->hop_count)
		node->minimal_hop_count = header->hop_count;
	node = register_known_node(header->last_hop_address);
	if(node->minimal_hop_count > 1)
		node->minimal_hop_count = 1;
#ifdef C_USE_ETX_LINK_METRIC
	if(calculate_ETX(header->last_hop_address) < C_ETX_GOOD_LINK_THRESHOLD)
		register_known_link(header->last_hop_address, C_LOCAL_ADDRESS);
#else
	register_known_link(header->last_hop_address, C_LOCAL_ADDRESS);
#endif	// C_USE_ETX_LINK_METRIC

	// Handle packet from direct neighbor.
	if(header->hop_count == 1) {
		C_ABORT_ON(header->source_address != header->last_hop_address, "Hop count to %s is 1, but source address and last hop address are inconsistent (%s != %s).\n", addr_to_str(header->source_address), addr_to_str(header->source_address), addr_to_str(header->last_hop_address));
		C_DEBUG(NET_DEBUG_LEVEL, C_MSG_DEBUG, "Hop count to %s is 1.\n", addr_to_str(header->source_address));
	}

#ifdef C_USE_NETWORK_MAPPING
	// Extract route to each of the nodes on the route.
	if(header->hop_count > 1) {
		C_ABORT_ON(header->hops == NULL, "Hop count is %d, but hop list is NULL.\n", header->hop_count);
		for(i = 0; i < header->hop_count; i++) {
			C_ABORT_ON(!is_valid_address(header->hops[i]), "Trying to register distance to invalid address %s as %d (hop #%d on route from %s).\n", addr_to_str(header->hops[i]), (header->hop_count - 1 - i), (i + 1), addr_to_str(header->source_address));
			register_known_node(header->hops[i]);
			C_DEBUG(NET_DEBUG_LEVEL, C_MSG_DEBUG, "Distance to %s is %d (hop #%d on route from %s).\n", addr_to_str(header->hops[i]), (header->hop_count - 1 - i), (i + 1), addr_to_str(header->source_address));
			register_known_node(header->hops[i]);
			if(i == 0)
				register_known_link(header->source_address, header->hops[0]);
			else
				register_known_link(header->hops[i - 1], header->hops[i]);
		}
	}

	// Extract 1-hop neighborhood of source node.
	if(header->neighbor_count > 1) {
		C_ABORT_ON(header->neighbors == NULL, "Neighbor count is %d, but neighbor list is NULL.\n", header->neighbor_count);
		for(i = 0; i < header->neighbor_count; i++) {
			C_ABORT_ON(!is_valid_address(header->neighbors[i]), "Trying to register distance to invalid address %s as %d (1-hop neighbor of %s).\n", addr_to_str(header->neighbors[i]), (header->hop_count + 1), addr_to_str(header->source_address));
			C_DEBUG(NET_DEBUG_LEVEL, C_MSG_DEBUG, "Distance to %s is %d (1-hop neighbor of %s).\n", addr_to_str(header->neighbors[i]), (header->hop_count + 1), addr_to_str(header->source_address));
			register_known_node(header->neighbors[i]);
			register_known_link(header->source_address, header->neighbors[i]);
		}
	}
#endif	// C_USE_NETWORK_MAPPING

#ifdef C_USE_INTERNAL_LINK_MONITORING
	/*
	 * Active link monitoring.
	 */
	node = register_known_node(header->last_hop_address);

	// Reset unanswered packet count.
	node->unanswered_packets = 0;

	// Cancel timeout for exptected packets.
	if(node->expected_packet_callback_id != 0) {
		callback_cancel(node->expected_packet_callback_id);
		node->expected_packet_callback_id = 0;
	}

	// Start timeout for delayed send.
	if(
		(node->delayed_send_callback_id == 0)	// Don't overwrite any existing timeout.
		&& (header->app != APP_TYPE_NONE)	// Don't start timer because APP_TYPE_NONE is used for signaling.
	) {
		node->delayed_send_callback_id = callback_add(delayed_send_timeout, node, (((float) C_UNICAST_MESSAGE_SENT_TIMEOUT) / 2.0));
	}
#endif	// C_USE_INTERNAL_LINK_MONITORING
}

/**
 * @brief Returns whether a node is known or not.
 *
 * @param address Address of the node.
 * @return <code>true</code> if the node is known, <code>false</code>
 * otherwise.
 */
bool exists_known_node(C_addr_t address) {
	node_t cmp_node;

	memset(&cmp_node, 0, sizeof(node_t));
	cmp_node.address = address;
	return (list_find(C_known_nodes, &cmp_node, matches_node_address) != NULL);
}

/**
 * @brief Returns a known node.
 *
 * @param address Address of the node.
 * @return Pointer to the known node.
 */
node_t* get_known_node(C_addr_t address) {
	node_t* node = NULL;
	node_t cmp_node;

	memset(&cmp_node, 0, sizeof(node_t));
	cmp_node.address = address;
	node = (node_t*) list_find(C_known_nodes, &cmp_node, matches_node_address);
	C_ABORT_ON(node == NULL, "No %s is known.\n", addr_to_str(address));
	return node;
}

/**
 * @brief Convert a node address into a unique ID.
 *
 * These IDs may be used for array indexing.
 *
 * The variable type of this function should correspond to the value of
 * <code>C_UNKNOWN_NODE_ID</code>.
 *
 * NOTE: A node ID is used for internal node identification and does NOT
 * correspond to the node's address.
 *
 * @param address Node address.
 * @return Unique ID (16 bit, i.e. we support 2^16 nodes).
 */
uint16_t node_address_to_id(C_addr_t address) {
	node_t* node = NULL;
	list_iterator_t C_known_nodes_iterator;
	uint16_t node_id = 0;
	bool found = false;

	C_ABORT_ON(!is_valid_address(address), "Trying to convert invalid address to ID (%s).\n", addr_to_str(address));

	list_iterator_init(C_known_nodes, &C_known_nodes_iterator);
	while((node = (node_t*) list_iterator_next(&C_known_nodes_iterator)) != NULL) {
		if(node->address == address) {
			found = true;
			break;
		}
		node_id++;
	}

	C_ABORT_ON(!found, "Trying to convert address %s to a node ID, but address is not known.\n", addr_to_str(address));
	C_ABORT_ON(node_id == C_UNKNOWN_NODE_ID, "Trying to convert address %s to a node ID, but node ID exceeds allocated range.\n", addr_to_str(address));

	return node_id;
}

/**
 * @brief Convert a unique ID into a node address.
 *
 * These IDs may be used for array indexing.
 *
 * @param node_id Unique ID as returned by <code>node_address_to_id()</code>.
 * @return Node address, or <code>C_BROADCAST_ADDRESS</code> if the node ID is
 * <code>C_UNKNOWN_NODE_ID</code>.
 */
C_addr_t node_id_to_address(uint16_t node_id) {
	node_t* node = NULL;

	if(node_id == C_UNKNOWN_NODE_ID)
		return C_BROADCAST_ADDRESS;

	C_ABORT_ON(list_size(C_known_nodes) < node_id, "Trying to convert node ID %d to an address, but only %d nodes are known.\n", node_id, list_size(C_known_nodes));

	node = (node_t*) list_get(C_known_nodes, node_id);

	C_ABORT_ON(!is_valid_address(node->address), "Trying to convert ID to invalid address (%s).\n", addr_to_str(node->address));
	return node->address;
}



/*
 * Calculation of data rate.
 */

/**
 * @brief Number of buckets to use in data rate calculation.
 *
 * We measure the data rate by discretizing time into one second slots. Each
 * time slot has a corresponding bucket and whenever a number of bytes is added
 * to the calculation we increment the count of the current bucket accordingly.
 * The data rate is calculated by averaging the byte count in all buckets.
 * Buckets are purged from old data whenever data is added or the calculation
 * is performed.
 *
 * The minimal number of buckets is 2. Increasing this number will result in a
 * smoother data rate value, but changes take more time to manifest themselves.
 *
 * This value should be set with the frequency of service requests in mind,
 * because too low values will result in the datarate being 0 most of the time.
 */
#define DATARATE_STATS_BUCKETS	60

/**
 * @brief Maximal age of added data rates.
 *
 * A data rate value may be added to calculation of a struct. This value
 * controls when they expire.
 *
 * The duration which is monitored must be larger than the interval in which
 * data is reported (e.g., GCMI_HOST_TRIGGER_MAX_DURATION).
 */
#define DATARATE_STATS_ADDED_DATARATE_MAX_AGE	300

/**
 * @brief Allocate and initialize data rate struct.
 *
 * @param data Pointer to pointer to data rate struct.
 */
void alloc_datarate_stat(datarate_stat_t** data) {
	MALLOC(*data, datarate_stat_t);
	MALLOC_ARRAY((*data)->bytes_per_second_buckets, uint32_t, DATARATE_STATS_BUCKETS);
	list_alloc(&(*data)->added_datarate_stats);
	(*data)->may_aggregate_other_stats = false;
	init_datarate_stat(*data);
}

/**
 * @brief Initialize data rate struct.
 *
 * @param data Pointer to data rate struct.
 */
void init_datarate_stat(datarate_stat_t* data) {
	unsigned int i;

	for(i = 0; i < DATARATE_STATS_BUCKETS; i++)
		data->bytes_per_second_buckets[i] = 0;
	data->valid_buckets = 1;
	data->time_of_last_purge = C_time();
	data->added_datarate = 0;
	data->added_datarate_time = 0;
	data->cached_datarate = 0;
	data->cached_datarate_is_dirty = true;
	list_clear(data->added_datarate_stats);
}

/**
 * @brief Allocate and initialize data rate struct.
 *
 * @param data Pointer to pointer to data rate struct.
 */
void free_datarate_stat(datarate_stat_t** data) {
	list_free(&(*data)->added_datarate_stats);
	FREE((*data)->bytes_per_second_buckets);
	FREE(*data);
}

/**
 * @brief Purge old data from data rate struct.
 *
 * @param data Pointer to data rate struct.
 */
static void purge_datarate_stat(datarate_stat_t* data) {
	C_time_t current_time = C_time();
	unsigned int i;

	C_DEBUG(NET_DEBUG_LEVEL, C_MSG_VERBOSE, "Purging data rate stat at %p...\n", data);
	while((current_time - data->time_of_last_purge) > 1) {
		// More buckets become valid as data is moved to the "older"
		// buckets.
		if(data->valid_buckets < DATARATE_STATS_BUCKETS)
			data->valid_buckets++;

		for(i = (data->valid_buckets - 1); i > 0; i--)
			data->bytes_per_second_buckets[i] = data->bytes_per_second_buckets[i - 1];
		data->bytes_per_second_buckets[0] = 0;
		data->time_of_last_purge++;
		data->cached_datarate_is_dirty = true;
	}
	if(
		(data->added_datarate > 0)
		&& (data->added_datarate_time < (current_time - DATARATE_STATS_ADDED_DATARATE_MAX_AGE))
	) {
		data->added_datarate = 0;
		data->cached_datarate_is_dirty = true;
	}
}

/**
 * @brief Calculate the data rate without recursing into added data rate stats.
 *
 * @param data Pointer to data rate struct.
 * @return Data rate in bytes per second (B/s) given the aggregated data points
 * in the data rate struct.
 */
static uint32_t _calculate_datarate(datarate_stat_t* data) {
	uint32_t total_bytes_per_second = 0;
	unsigned int i;

	if(data->cached_datarate_is_dirty) {
		for(i = 0; i < data->valid_buckets; i++)
			total_bytes_per_second += data->bytes_per_second_buckets[i];
		data->cached_datarate = (total_bytes_per_second / data->valid_buckets);
		data->cached_datarate += data->added_datarate;
		data->cached_datarate_is_dirty = false;
	}
	return data->cached_datarate;
}

/**
 * @brief Add byte count to data rate struct.
 *
 * @param data Pointer to data rate struct.
 * @param bytes Number of bytes to be considered for data rate calculation.
 */
void add_bytes_to_datarate_stat(datarate_stat_t* data, uint32_t bytes) {
#ifdef C_ENABLE_CONSISTENCY_CHECKS
	uint32_t datarate_before_adding_bytes = 0;
#endif	// C_ENABLE_CONSISTENCY_CHECKS

	C_DEBUG(NET_DEBUG_LEVEL, C_MSG_INFO, "Adding %d bytes to data rate stat at %p...\n", bytes, data);
	purge_datarate_stat(data);
#ifdef C_ENABLE_CONSISTENCY_CHECKS
	datarate_before_adding_bytes = _calculate_datarate(data);
#endif	// C_ENABLE_CONSISTENCY_CHECKS

	data->bytes_per_second_buckets[0] += bytes;
	data->cached_datarate_is_dirty = true;

#ifdef C_ENABLE_CONSISTENCY_CHECKS
	C_ABORT_ON(_calculate_datarate(data) < datarate_before_adding_bytes, "Data rate has decreased while adding %d bytes (%d < %d)...\n", bytes, _calculate_datarate(data), datarate_before_adding_bytes);
#endif	// C_ENABLE_CONSISTENCY_CHECKS
}

/**
 * @brief Add data rate to data rate struct.
 *
 * This overwrites any previously added data rate.
 *
 * @param data Pointer to data rate struct.
 * @param datarate Average data rate (B/s) to be added to the data rate
 * calculation.
 */
void set_additional_datarate_for_datarate_stats(datarate_stat_t* data, uint32_t datarate) {
	C_DEBUG(NET_DEBUG_LEVEL, C_MSG_INFO, "Setting additional data rate %d (B/s) for data rate stat at %p...\n", datarate, data);
	data->added_datarate = datarate;
	data->added_datarate_time = C_time();
	data->cached_datarate_is_dirty = true;
}

/**
 * @brief Add two data rate structs.
 *
 * @param dst_data Pointer to first data rate struct.
 * @param src_data Pointer to second data rate struct.
 */
void add_datarate_stats(datarate_stat_t* dst_data, datarate_stat_t* src_data) {
	C_ABORT_ON(!dst_data->may_aggregate_other_stats, "Trying to add data rate stats at %p to %p, but destination is not set to support aggregation...\n", src_data, dst_data);	// Ensures that there is no recursion.

	if(list_contains(dst_data->added_datarate_stats, src_data)) {
		C_DEBUG(NET_DEBUG_LEVEL, C_MSG_DEBUG, "Trying to add data rate stats at %p to %p, but destination already contains this data rate stat. Skipping...\n", src_data, dst_data);
	} else {
		C_ABORT_ON(src_data->may_aggregate_other_stats, "Trying to add data rate stats at %p to %p, but source is set to support aggregation...\n", src_data, dst_data);
		C_DEBUG(NET_DEBUG_LEVEL, C_MSG_INFO, "Adding data rate stats at %p to %p...\n", src_data, dst_data);
		list_add(dst_data->added_datarate_stats, src_data);
	}
}

/**
 * @brief Subtract two data rate structs.
 *
 * @param dst_data Pointer to first data rate struct.
 * @param src_data Pointer to second data rate struct.
 */
void subtract_datarate_stats(datarate_stat_t* dst_data, datarate_stat_t* src_data) {
	if(!list_contains(dst_data->added_datarate_stats, src_data)) {
		C_DEBUG(NET_DEBUG_LEVEL, C_MSG_ERROR, "Trying to subtract data rate stats at %p from %p, but destination does not contain this data rate stat...\n", src_data, dst_data);
	} else {
		C_ABORT_ON(src_data->may_aggregate_other_stats, "Trying to subtract data rate stats at %p from %p, but source wasn't set to support aggregation...\n", src_data, dst_data);
		C_DEBUG(NET_DEBUG_LEVEL, C_MSG_INFO, "Subtracting data rate stats at %p from %p...\n", src_data, dst_data);
		list_remove(dst_data->added_datarate_stats, src_data);
	}
}

/**
 * @brief Calculate the data rate.
 *
 * @param data Pointer to data rate struct.
 * @return Data rate in bytes per second (B/s) given the aggregated data points
 * in the data rate struct.
 */
uint32_t calculate_datarate(datarate_stat_t* data) {
	uint32_t current_datarate = 0;
	datarate_stat_t* datarate_stat = NULL;
	list_iterator_t datarate_stats_iterator;

	C_DEBUG(NET_DEBUG_LEVEL, C_MSG_VERBOSE, "Calculating data rate of data rate stat at %p...\n", data);
	purge_datarate_stat(data);
	current_datarate = _calculate_datarate(data);

	// The added_to_other_datarate_stat ensures that there is no recursion.
	list_iterator_init(data->added_datarate_stats, &datarate_stats_iterator);
	while((datarate_stat = (datarate_stat_t*) list_iterator_next(&datarate_stats_iterator)) != NULL)
		current_datarate += calculate_datarate(datarate_stat);

	C_DEBUG(NET_DEBUG_LEVEL, C_MSG_INFO, "Data rate of data rate stat at %p is %d B/s.\n", data, current_datarate);
	return current_datarate;
}
