/**
 * @file trigger.c
 *
 * @brief Trigger functions.
 *
 * Copyright (c) 2007, 2008, 2009, 2010 Georg Wittenburg, Freie Universitaet
 * Berlin (wittenbu@inf.fu-berlin.de). All rights reserved.
 * Please refer to the COPYING file for detailed licensing information.
 *
 * @author Georg Wittenburg
 *
 * $Id: trigger.c 3177 2010-01-05 16:43:45Z wittenbu $
 */

#include "../config.h"
#include "../macros.h"
#include "../types.h"
#include "../vars.h"
#include "../main.h"
#include "../util/debug.h"



/**
 * @brief Fraction of the trigger duration to consider a trigger event to be at
 * "the end" of a trigger duration.
 */
#define TRIGGER_DURATION_FRACTION_AT_END	0.1



/**
 * @brief Allocate and initialize trigger structure.
 *
 * This implements a time-dependant hysteresis whose trigger values are set
 * closer and closer to a start value as time progresses.
 *
 * TODO: The trigger concept is similar to event-generating ADCs; use correct
 * terminology, e.g. "funnel".
 *
 * @param trigger Pointer to pointer to trigger structure to initialize.
 * @param max_range Maximum range (in both diretions) of the hysteresis at
 * start time.
 * @param max_duration Maximum duration of the trigger; range reaches 0 if
 * exceeded.
 */
void trigger_alloc(trigger_t** trigger, float max_range, C_time_t max_duration) {
	C_ABORT_ON(max_range < 0, "Trying to initialize trigger with negative range (%f)...\n", max_range);
	C_ABORT_ON(max_duration <= 0, "Trying to initialize trigger with non-positive duration (%f)...\n", max_duration);

	MALLOC(*trigger, trigger_t);
	(*trigger)->max_range = max_range;
	(*trigger)->max_duration = max_duration;
	(*trigger)->start_value = 0;
	(*trigger)->start_time = 0;
	(*trigger)->triggered_at_end_count = 0;
}

/**
 * @brief Free trigger structure.
 *
 * @param trigger Pointer to pointer to trigger structure to free.
 */
void trigger_free(trigger_t** trigger) {
	C_ABORT_ON(*trigger == NULL, "Trying to free trigger at NULL...\n");

	FREE(*trigger);
}

/**
 * @brief Reset trigger.
 *
 * @param trigger Pointer to trigger to reset.
 * @param start_value Start value for the time-dependant hysteresis.
 * @param max_range Maximum range (in both diretions) of the hysteresis at
 * start time.
 * @param max_duration Maximum duration of the trigger; range reaches 0 if
 * exceeded.
 */
void trigger_reset(trigger_t* trigger, float start_value, float max_range, C_time_t max_duration) {
	C_ABORT_ON(trigger == NULL, "Trying to re-initialize trigger at NULL...\n");
	C_ABORT_ON(max_range < 0, "Trying to re-initialize trigger with negative range (%f)...\n", max_range);
	C_ABORT_ON(max_duration <= 0, "Trying to re-initialize trigger with non-positive duration (%f)...\n", max_duration);

	trigger->max_range = max_range;
	trigger->max_duration = max_duration;
	trigger->start_value = start_value;
	trigger->start_time = C_time();
}

/**
 * @brief Check trigger.
 *
 * @param trigger Pointer to trigger to check.
 * @param current_value Current value to check against.
 * @return <code>true</code> if the trigger should fire, <code>false</code>
 * otherwise.
 */
bool trigger_check(trigger_t* trigger, float current_value) {
	C_time_t current_time = C_time();
	float remaining_time_fraction = 0.0;
	float current_range_limit = 0;

	C_ABORT_ON(trigger == NULL, "Trying to check trigger at NULL...\n");

	remaining_time_fraction = 1.0 - ((current_time - trigger->start_time) / trigger->max_duration);
	current_range_limit = trigger->max_range * remaining_time_fraction;
	if(
		(trigger->start_time == 0)
		|| (current_time > (trigger->start_time + trigger->max_duration))
		|| (current_value > (trigger->start_value + current_range_limit))
		|| (current_value < (trigger->start_value - current_range_limit))
	) {
		if(remaining_time_fraction < TRIGGER_DURATION_FRACTION_AT_END)
			trigger->triggered_at_end_count++;
		else
			trigger->triggered_at_end_count = 0;
		return true;
	}

	return false;
}

/**
 * @brief Get the number of times a trigger has fired at the end of its duration.
 *
 * @param trigger Pointer to trigger to get info from.
 * @return Number of times a trigger has fired at the end of its duration.
 */
uint16_t trigger_times_fired_at_end(trigger_t* trigger) {
	return trigger->triggered_at_end_count;
}
