#!/usr/bin/perl

##
#
# Generate plots using gnuplot.
#
# Copyright (c) 2010 Georg Wittenburg, Freie Universitaet Berlin
# (wittenbu@inf.fu-berlin.de). All rights reserved.
# Please refer to the COPYING file for detailed licensing information.
#
# $Id: GnuplotWrapper.pm 3664 2010-07-26 10:03:17Z gwitten $
#
##

package GnuplotWrapper;

use strict;

use Exporter;
use base 'Exporter';

our @EXPORT = qw( %VARIABLES %VALUES $PLOT_TYPE_POINTS_PER_PLACEMENT
	$PLOT_TYPE_DATA_SERIES $PLOT_TYPE_DATA_SERIES_PER_PLACEMENT
	$PLOT_TYPE_TIME_SERIES $PLOT_TYPE_HOP_COUNT_HISTOGRAM
	generate_summary_plots generate_time_plots );



my $LEGEND_FONT = 'Helvetica';
my $LEGEND_FONT_SIZE = 16;

my $PLOT_FIVE_NUMBER_SUMMARY = 1;	# See http://en.wikipedia.org/wiki/Five-number_summary
my $PLOT_AVERAGE_AND_STDDEV = 0;

my $HIGHLIGHT_MEAN = 0;
my $HIGHLIGHT_MEDIAN = 1;

my $BOX_WIDTH_FACTOR = (1/40);

my $PLOT_TYPE_NONE				= 0;
our $PLOT_TYPE_POINTS_PER_PLACEMENT		= 1;
our $PLOT_TYPE_DATA_SERIES			= 2;
our $PLOT_TYPE_DATA_SERIES_PER_PLACEMENT	= 3;
our $PLOT_TYPE_TIME_SERIES			= 4;
our $PLOT_TYPE_HOP_COUNT_HISTOGRAM		= 5;




#
# Run 'show palette colornames' for color names.
#
# Line types:
#	0 -> --- (thin)
#	1 -> (continous)
#	2 -> ___
#	3 -> ---
#	4 -> ...
#	5 -> ._._.
#	6 -> .-.-.
#	7 -> -- --
#	8 -> .._..
#	9 -> ---- ----
#
# Point types:
#	0 -> .
#	1 -> +
#	2 -> x
#	3 -> *
#	4 -> (box)
#	5 -> (box, filled)
#	6 -> o
#	7 -> o (filled)
#	8 -> (triangle)
#	9 -> (triangle, filled)
#	10 -> (inverse triangle)
#	11 -> (inverse triangle, filled)
#	12 -> (diamond)
#	13 -> (diamond, filled)
#	14 -> (pentagon)
#	15 -> (pentagon, filled)
#	16 -> o (top center dot)
#	17 -> o (top right dot)
#	18 -> o (top left dot)
#	19 -> o (top filled)
#	20 -> o (bottom left dot)
#	21 -> o (top right, bottom left dots)
#	22 -> o (left filled)
#	23 -> o (filled, bottom right empty)
#	24 -> o (bottom right dot)
#	25 -> o (right filled)
#	26 -> o (top left, bottom right dots)
#	27 -> o (filled, bottom left empty)
#	28 -> o (bottom filled)
#	29 -> o (filled, top left empty)
#	30 -> o (filled, top right empty)
#	31 -> o (filled)
#	32 -> (box, top center dot)
#	33 -> (box, top right dot)
#	[...]
#
my %STYLES = (
	default => {
		offset		=> 10,
		color		=> "blue",
		line_type	=> 0,
		point_type	=> 1,
	},
	value_over_time => {
		offset		=> 20,
		color		=> "blue",
		line_type	=> 0,
		point_type	=> 2,
	},
	optimal_value_over_time => {
		offset		=> 30,
		color		=> "#00CD00",	# "medium-green"
		line_type	=> 0,
		point_type	=> 6,
	},
	cost_over_time => {
		offset		=> 40,
		color		=> "red",
		line_type	=> 0,
		point_type	=> 6,
	},
	contribution_1 => {
		offset		=> 50,
		color		=> "blue",
		line_type	=> 2,
		point_type	=> 3,
	},
	contribution_2 => {
		offset		=> 60,
		color		=> "royalblue",
		line_type	=> 3,
		point_type	=> 2,
	},
	contribution_3 => {
		offset		=> 70,
		color		=> "medium-blue",
		line_type	=> 4,
		point_type	=> 1,
	},
	major_alternative_1 => {
		offset		=> 80,
		color		=> "dark-magenta",
		line_type	=> 5,
		point_type	=> 7,
	},
	major_alternative_2 => {
		offset		=> 90,
		color		=> "light-magenta",
		line_type	=> 6,
		point_type	=> 9,
	},
	major_alternative_3 => {
		offset		=> 100,
		color		=> "purple",
		line_type	=> 8,
		point_type	=> 11,
	},
	minor_alternative_1 => {
		offset		=> 110,
		color		=> "forest-green",
		line_type	=> 7,
		point_type	=> 6,
	},
	minor_alternative_2 => {
		offset		=> 120,
		color		=> "#00CD00",	# "medium-green"
		line_type	=> 9,
		point_type	=> 8,
	},
	minor_alternative_3 => {
		offset		=> 130,
		color		=> "dark-green",
		line_type	=> 4,
		point_type	=> 10,
	},
	base_case_1 => {
		offset		=> 140,
		color		=> "gray40",
		line_type	=> 3,
		point_type	=> 12,
	},
	base_case_2 => {
		offset		=> 150,
		color		=> "gray30",
		line_type	=> 3,
		point_type	=> 13,
	},
	base_case_3 => {
		offset		=> 160,
		color		=> "gray60",
		line_type	=> 4,
		point_type	=> 14,
	},
	base_case_4 => {
		offset		=> 160,
		color		=> "gray50",
		line_type	=> 4,
		point_type	=> 15,
	},
);



# This constant encodes the file format used in all evaluations except for the
# evaluation of routing protocols (see below).
#
# TRACE: current_time(1),nodes(2),area_x(3),area_y(4)
# AGGREGATED: service_size(1),sync_ratio(2),nodes(3),active_node_ratio(4),request_frequency(5)
# OR
# AGGREGATED: service_size(1),active_regions(2),nodes(3),activity_phases(4),request_frequency(5)
our %VARIABLES = (
	current_time => {
		id			=> "current_time",
		trace_index		=> 1,
		aggregated_index	=> -1,	# n/a
		label			=> "Time [minutes]",
		label_y_offset		=> 0,
		default_range		=> "[0:20]",
		scale			=> 1,
		logscale		=> 0,
		box_width		=> (20 * $BOX_WIDTH_FACTOR),
	},
	sync_ratio => {
		id			=> "sync_ratio",
		trace_index		=> -1,	# n/a
		aggregated_index	=> 2,
		label			=> "Synchronization Traffic Ratio [%]",
		label_y_offset		=> 0,
# 		default_range		=> "[-1:21]",
		default_range		=> "[-1:11]",
		scale			=> 100,
		logscale		=> 0,
# 		box_width		=> (22 * $BOX_WIDTH_FACTOR * 0.66),
		box_width		=> (12 * $BOX_WIDTH_FACTOR * 0.66),
	},
	active_regions => {
		id			=> "active_regions",
		trace_index		=> -1,	# n/a
		aggregated_index	=> 2,
		label			=> "Demand Distribution [regions]",
		label_y_offset		=> 0,
		default_range		=> "[0.7:25]",
		scale			=> 1,
		logscale		=> 1,
		box_width		=> 0.04,
	},
	nodes => {
		id			=> "nodes",
		trace_index		=> -1,	# n/a
		aggregated_index	=> 3,
		label			=> "Network Size [nodes]",
		label_y_offset		=> 0,
		default_range		=> "[10:110]",
		scale			=> 1,
		logscale		=> 0,
		box_width		=> (100 * $BOX_WIDTH_FACTOR),
	},
	active_node_ratio => {
		id			=> "active_node_ratio",
		trace_index		=> -1,	# n/a
		aggregated_index	=> 4,
		label			=> "Active Node Ratio [%]",
		label_y_offset		=> 0,
		default_range		=> "[0:105]",
		scale			=> 100,
		logscale		=> 0,
		box_width		=> (110 * $BOX_WIDTH_FACTOR),
	},
	activity_phases => {
		id			=> "activity_phases",
		trace_index		=> -1,	# n/a
		aggregated_index	=> 4,
		label			=> "Demand Volatility [changes/hour]",
		label_y_offset		=> 0,
		default_range		=> "[0.8:120]",
		scale			=> 1,	# NOTE: This is based on a run duration of 60 minutes.
		logscale		=> 1,
		box_width		=> 0.05,
	},
	request_frequency => {
		id			=> "request_frequency",
		trace_index		=> -1,	# n/a
		aggregated_index	=> 5,
		label			=> "Service Demand [requests/s]",
		label_y_offset		=> 0,
		default_range		=> "[0.05:11]",
		scale			=> 1,
		logscale		=> 1,
		box_width		=> 0.06,
	},
	hop_count => {
		id			=> "hop_count",
		trace_index		=> -1,	# n/a
		aggregated_index	=> 20,	# Misused to indicate max_hop_count (see below).
		label			=> "Distance Travelled [hops]",
		label_y_offset		=> 0,
		default_range		=> "[-1:21]",
		scale			=> 1,
		logscale		=> 0,
		box_width		=> (18 * $BOX_WIDTH_FACTOR),
	},
);

# This constant encodes the file format used in evaluation of routing protocols.
#
# AGGREGATED: nodes(1),packet_frequency(2)
our %VARIABLES_ROUTING = (
	nodes => {
		id			=> "nodes",
		trace_index		=> -1,	# n/a
		aggregated_index	=> 1,
		label			=> "Network Size [nodes]",
		label_y_offset		=> 0,
		default_range		=> "[10:110]",
		scale			=> 1,
		logscale		=> 0,
		box_width		=> (100 * $BOX_WIDTH_FACTOR),
	},
	packet_frequency => {
		id			=> "packet_frequency",
		trace_index		=> -1,	# n/a
		aggregated_index	=> 2,
		label			=> "Offered Load [packets/s]",
		label_y_offset		=> 0,
		default_range		=> "[0.05:11]",
		scale			=> 1,
		logscale		=> 1,
		box_width		=> 0.06,
	},
);



# This constant encodes the file format used in all evaluations except for the
# evaluation of routing protocols (see below).
our %VALUES = (
	# TRACE: unicastPacketsPerHopCount_0(12)
	# AGGREGATED: unicastPacketsPerHopCount_0_MEAN(94),unicastPacketsPerHopCount_0_VAR(95),unicastPacketsPerHopCount_0_STDDEV(96),unicastPacketsPerHopCount_0_MIN(97),unicastPacketsPerHopCount_0_LOWER_QUARTILE(98),unicastPacketsPerHopCount_0_MEDIAN(99),unicastPacketsPerHopCount_0_UPPER_QUARTILE(100),unicastPacketsPerHopCount_0_MAX(101)
	unicastPacketsPerHopCount => {
		trace_index		=> 12,
		aggregated_index	=> 94,
		label			=> "Unicast Packets [thousands]",
		label_y_offset		=> 1,
		default_range		=> "[0:150]",
		scale			=> 0.001,
		logscale		=> 0,
		style			=> "default",
	},
	# TRACE: transmittedUnicastPackets(33)
	# AGGREGATED: transmittedUnicastPackets_MEAN(262),transmittedUnicastPackets_VAR(263),transmittedUnicastPackets_STDDEV(264),transmittedUnicastPackets_MIN(265),transmittedUnicastPackets_LOWER_QUARTILE(266),transmittedUnicastPackets_MEDIAN(267),transmittedUnicastPackets_UPPER_QUARTILE(268),transmittedUnicastPackets_MAX(269)
	transmittedUnicastPackets => {
		trace_index		=> 33,
		aggregated_index	=> 262,
		label			=> "Unicast Packets [thousands]",
		label_y_offset		=> 1,
		default_range		=> "[0:27]",
		scale			=> 0.001,
		logscale		=> 0,
		style			=> "default",
	},
	# TRACE: unicastPDR(36)
	# AGGREGATED: unicastPDR_MEAN(286),unicastPDR_VAR(287),unicastPDR_STDDEV(288),unicastPDR_MIN(289),unicastPDR_LOWER_QUARTILE(290),unicastPDR_MEDIAN(291),unicastPDR_UPPER_QUARTILE(292),unicastPDR_MAX(293)
	unicastPDR => {
		trace_index		=> 36,
		aggregated_index	=> 286,
		label			=> "Unicast PDR [%]",
		label_y_offset		=> 2,
		default_range		=> "[0:105]",
		scale			=> 100,
		logscale		=> 0,
		style			=> "default",
	},
	# TRACE: transmittedBroadcastPackets(37)
	# AGGREGATED: transmittedBroadcastPackets_MEAN(294),transmittedBroadcastPackets_VAR(295),transmittedBroadcastPackets_STDDEV(296),transmittedBroadcastPackets_MIN(297),transmittedBroadcastPackets_LOWER_QUARTILE(298),transmittedBroadcastPackets_MEDIAN(299),transmittedBroadcastPackets_UPPER_QUARTILE(300),transmittedBroadcastPackets_MAX(301)
	transmittedBroadcastPackets => {
		trace_index		=> 37,
		aggregated_index	=> 294,
		label			=> "Broadcast Packets [thousands]",
		label_y_offset		=> 1,
		default_range		=> "[0:27]",
		scale			=> 0.001,
		logscale		=> 0,
		style			=> "default",
	},
	# TRACE: transmittedBytes(40)
	# AGGREGATED: transmittedBytes_MEAN(318),transmittedBytes_VAR(319),transmittedBytes_STDDEV(320),transmittedBytes_MIN(321),transmittedBytes_LOWER_QUARTILE(322),transmittedBytes_MEDIAN(323),transmittedBytes_UPPER_QUARTILE(324),transmittedBytes_MAX(325)
	transmittedBytes => {
		trace_index		=> 40,
		aggregated_index	=> 318,
		label			=> "Traffic [MB]",
		label_y_offset		=> 2,
		default_range		=> "[0:790]",
		default_range		=> "",
		scale			=> 0.000001,
		logscale		=> 0,
		style			=> "default",
	},
	# TRACE: currentTransmittedDatarate(42)
	# AGGREGATED: averageTransmittedDatarate_MEAN(350),averageTransmittedDatarate_VAR(351),averageTransmittedDatarate_STDDEV(352),averageTransmittedDatarate_MIN(353),averageTransmittedDatarate_LOWER_QUARTILE(354),averageTransmittedDatarate_MEDIAN(355),averageTransmittedDatarate_UPPER_QUARTILE(356),averageTransmittedDatarate_MAX(357)
	transmittedDatarate => {
		trace_index		=> 42,
		aggregated_index	=> 350,
		label			=> "Network Load [kB/s]",
		label_y_offset		=> 1,
		default_range		=> "[0:550]",
		scale			=> 0.001,
		logscale		=> 0,
		style			=> "default",
	},
	# TRACE: MAC_transmittedBytes(56)
	# AGGREGATED: MAC_transmittedBytes_MEAN(446),MAC_transmittedBytes_VAR(447),MAC_transmittedBytes_STDDEV(448),MAC_transmittedBytes_MIN(449),MAC_transmittedBytes_LOWER_QUARTILE(450),MAC_transmittedBytes_MEDIAN(451),MAC_transmittedBytes_UPPER_QUARTILE(452),MAC_transmittedBytes_MAX(453)
	MAC_transmittedBytes => {
		trace_index		=> 56,
		aggregated_index	=> 446,
		label			=> "Total MAC Traffic [MB]",
		label_y_offset		=> 2,
		default_range		=> "[0:800]",
		scale			=> 0.000001,
		logscale		=> 0,
		style			=> "default",
	},
	# TRACE: DYMO_transmittedRREQs(62)
	# AGGREGATED: DYMO_transmittedRREQs_MEAN(494),DYMO_transmittedRREQs_VAR(495),DYMO_transmittedRREQs_STDDEV(496),DYMO_transmittedRREQs_MIN(497),DYMO_transmittedRREQs_LOWER_QUARTILE(498),DYMO_transmittedRREQs_MEDIAN(499),DYMO_transmittedRREQs_UPPER_QUARTILE(500),DYMO_transmittedRREQs_MAX(501)
	DYMO_transmittedRREQs => {
		trace_index		=> 62,
		aggregated_index	=> 494,
		label			=> "DYMO RREQs [thousands]",
		label_y_offset		=> 2,
		default_range		=> "[0:10.5]",
		scale			=> 0.001,
		logscale		=> 0,
		style			=> "default",
	},
	# TRACE: DYMO_transmittedRREPs(63)
	# AGGREGATED: DYMO_transmittedRREPs_MEAN(502),DYMO_transmittedRREPs_VAR(503),DYMO_transmittedRREPs_STDDEV(504),DYMO_transmittedRREPs_MIN(505),DYMO_transmittedRREPs_LOWER_QUARTILE(506),DYMO_transmittedRREPs_MEDIAN(507),DYMO_transmittedRREPs_UPPER_QUARTILE(508),DYMO_transmittedRREPs_MAX(509)
	DYMO_transmittedRREPs => {
		trace_index		=> 63,
		aggregated_index	=> 502,
		label			=> "DYMO RREPs [thousands]",
		label_y_offset		=> 1,
		default_range		=> "[0:2.5]",
		scale			=> 0.001,
		logscale		=> 0,
		style			=> "default",
	},
	# TRACE: DYMO_transmittedRERRs(64)
	# AGGREGATED: DYMO_transmittedRERRs_MEAN(510),DYMO_transmittedRERRs_VAR(511),DYMO_transmittedRERRs_STDDEV(512),DYMO_transmittedRERRs_MIN(513),DYMO_transmittedRERRs_LOWER_QUARTILE(514),DYMO_transmittedRERRs_MEDIAN(515),DYMO_transmittedRERRs_UPPER_QUARTILE(516),DYMO_transmittedRERRs_MAX(517)
	DYMO_transmittedRERRs => {
		trace_index		=> 64,
		aggregated_index	=> 510,
		label			=> "DYMO RERRs [thousands]",
		label_y_offset		=> 1,
		default_range		=> "[0:2.5]",
		scale			=> 0.001,
		logscale		=> 0,
		style			=> "default",
	},
	# TRACE: DYMO_unicastPdr(67)
	# AGGREGATED: DYMO_unicastPdr_MEAN(534),DYMO_unicastPdr_VAR(535),DYMO_unicastPdr_STDDEV(536),DYMO_unicastPdr_MIN(537),DYMO_unicastPdr_LOWER_QUARTILE(538),DYMO_unicastPdr_MEDIAN(539),DYMO_unicastPdr_UPPER_QUARTILE(540),DYMO_unicastPdr_MAX(541)
	DYMO_unicastPdr => {
		trace_index		=> 67,
		aggregated_index	=> 534,
		label			=> "DYMO Unicast PDR [%]",
		label_y_offset		=> 2,
		default_range		=> "[0:105]",
		scale			=> 100,
		logscale		=> 0,
		style			=> "default",
	},
	# TRACE: RUDP_totalTransmissions(71)
	# AGGREGATED: RUDP_totalTransmissions_MEAN(566),RUDP_totalTransmissions_VAR(567),RUDP_totalTransmissions_STDDEV(568),RUDP_totalTransmissions_MIN(569),RUDP_totalTransmissions_LOWER_QUARTILE(570),RUDP_totalTransmissions_MEDIAN(571),RUDP_totalTransmissions_UPPER_QUARTILE(572),RUDP_totalTransmissions_MAX(573)
	RUDP_totalTransmissions => {
		trace_index		=> 71,
		aggregated_index	=> 566,
		label			=> "RUDP Transm. [thousands]",
		label_y_offset		=> 2,
		default_range		=> "[0:3.1]",
		scale			=> 0.001,
		logscale		=> 0,
		style			=> "default",
	},
	# TRACE: RUDP_pdr(73)
	# AGGREGATED: RUDP_pdr_MEAN(582),RUDP_pdr_VAR(583),RUDP_pdr_STDDEV(584),RUDP_pdr_MIN(585),RUDP_pdr_LOWER_QUARTILE(586),RUDP_pdr_MEDIAN(587),RUDP_pdr_UPPER_QUARTILE(588),RUDP_pdr_MAX(589)
	RUDP_pdr => {
		trace_index		=> 73,
		aggregated_index	=> 582,
		label			=> "RUDP PDR [%]",
		label_y_offset		=> 2,
		default_range		=> "[0:105]",
		scale			=> 100,
		logscale		=> 0,
		style			=> "default",
	},
	# TRACE: SLP_totalQueries(77)
	# AGGREGATED: SLP_totalQueries_MEAN(614),SLP_totalQueries_VAR(615),SLP_totalQueries_STDDEV(616),SLP_totalQueries_MIN(617),SLP_totalQueries_LOWER_QUARTILE(618),SLP_totalQueries_MEDIAN(619),SLP_totalQueries_UPPER_QUARTILE(620),SLP_totalQueries_MAX(621)
	SLP_totalQueries => {
		trace_index		=> 77,
		aggregated_index	=> 614,
		label			=> "SLP Service Requests",
		label_y_offset		=> 2,
		default_range		=> "[0:155]",
		scale			=> 1,
		logscale		=> 0,
		style			=> "default",
	},
	# TRACE: SLP_recall(79)
	# AGGREGATED: SLP_recall_MEAN(630),SLP_recall_VAR(631),SLP_recall_STDDEV(632),SLP_recall_MIN(633),SLP_recall_LOWER_QUARTILE(634),SLP_recall_MEDIAN(635),SLP_recall_UPPER_QUARTILE(636),SLP_recall_MAX(637)
	SLP_recall => {
		trace_index		=> 79,
		aggregated_index	=> 630,
		label			=> "SLP Recall [%]",
		label_y_offset		=> 2,
		default_range		=> "[0:105]",
		scale			=> 100,
		logscale		=> 0,
		style			=> "default",
	},
	# TRACE: SERVICE_averageSuccessfulDataRequestDelay(83)
	# AGGREGATED: SERVICE_averageSuccessfulDataRequestDelay_MEAN(662),SERVICE_averageSuccessfulDataRequestDelay_VAR(663),SERVICE_averageSuccessfulDataRequestDelay_STDDEV(664),SERVICE_averageSuccessfulDataRequestDelay_MIN(665),SERVICE_averageSuccessfulDataRequestDelay_LOWER_QUARTILE(666),SERVICE_averageSuccessfulDataRequestDelay_MEDIAN(667),SERVICE_averageSuccessfulDataRequestDelay_UPPER_QUARTILE(668),SERVICE_averageSuccessfulDataRequestDelay_MAX(669)
	SERVICE_averageSuccessfulDataRequestDelay => {
		trace_index		=> 83,
		aggregated_index	=> 662,
		label			=> "Round-trip Time [s]",
		label_y_offset		=> 3,
		default_range		=> "[0.001:14]",
		scale			=> 1,
		logscale		=> 1,
		style			=> "default",
	},
	# TRACE: SERVICE_requestRecall(95)
	# AGGREGATED: SERVICE_requestRecall_MEAN(758),SERVICE_requestRecall_VAR(759),SERVICE_requestRecall_STDDEV(760),SERVICE_requestRecall_MIN(761),SERVICE_requestRecall_LOWER_QUARTILE(762),SERVICE_requestRecall_MEDIAN(763),SERVICE_requestRecall_UPPER_QUARTILE(764),SERVICE_requestRecall_MAX(765)
	SERVICE_requestRecall => {
		trace_index		=> 95,
		aggregated_index	=> 758,
		label			=> "Service Recall [%]",
		label_y_offset		=> 2,
		default_range		=> "[0:105]",
		scale			=> 100,
		logscale		=> 0,
		style			=> "default",
	},
	# TRACE: SERVICE_currentTransmittedDatarate(98)
	# AGGREGATED: SERVICE_averageTransmittedDatarate_MEAN(798),SERVICE_averageTransmittedDatarate_VAR(799),SERVICE_averageTransmittedDatarate_STDDEV(800),SERVICE_averageTransmittedDatarate_MIN(801),SERVICE_averageTransmittedDatarate_LOWER_QUARTILE(802),SERVICE_averageTransmittedDatarate_MEDIAN(803),SERVICE_averageTransmittedDatarate_UPPER_QUARTILE(804),SERVICE_averageTransmittedDatarate_MAX(805)
	SERVICE_transmittedDatarate => {
		trace_index		=> 98,
		aggregated_index	=> 798,
		label			=> "Service Goodput [kB/s]",
		label_y_offset		=> 1,
		default_range		=> "[0:290]",
		scale			=> 0.001,
		logscale		=> 0,
		style			=> "default",
	},
	# TRACE: SERVICE_averageRequestHopCount(102)
	# AGGREGATED: SERVICE_averageRequestHopCount_MEAN(814),SERVICE_averageRequestHopCount_VAR(815),SERVICE_averageRequestHopCount_STDDEV(816),SERVICE_averageRequestHopCount_MIN(817),SERVICE_averageRequestHopCount_LOWER_QUARTILE(818),SERVICE_averageRequestHopCount_MEDIAN(819),SERVICE_averageRequestHopCount_UPPER_QUARTILE(820),SERVICE_averageRequestHopCount_MAX(821)
	SERVICE_averageRequestHopCount => {
		trace_index		=> 102,
		aggregated_index	=> 814,
		label			=> "Request Distance [hops]",
		label_y_offset		=> 1,
		default_range		=> "[0:8.5]",
		scale			=> 1,
		logscale		=> 0,
		style			=> "default",
	},
	# TRACE: SERVICE_requestPacketsPerHopCount_0(104)
	# AGGREGATED: SERVICE_requestPacketsPerHopCount_0_MEAN(830),SERVICE_requestPacketsPerHopCount_0_VAR(831),SERVICE_requestPacketsPerHopCount_0_STDDEV(832),SERVICE_requestPacketsPerHopCount_0_MIN(833),SERVICE_requestPacketsPerHopCount_0_LOWER_QUARTILE(834),SERVICE_requestPacketsPerHopCount_0_MEDIAN(835),SERVICE_requestPacketsPerHopCount_0_UPPER_QUARTILE(836),SERVICE_requestPacketsPerHopCount_0_MAX(837)
	SERVICE_requestPacketsPerHopCount => {
		trace_index		=> 104,
		aggregated_index	=> 830,
		label			=> "Requests [thousands]",
		label_y_offset		=> 1,
		default_range		=> "[0:170]",
		scale			=> 0.001,
		logscale		=> 0,
		style			=> "default",
	},
	# TRACE: SERVICE_currentlyHostedServices(125)
	# AGGREGATED: SERVICE_averageHostedServices_MEAN(1006),SERVICE_averageHostedServices_VAR(1007),SERVICE_averageHostedServices_STDDEV(1008),SERVICE_averageHostedServices_MIN(1009),SERVICE_averageHostedServices_LOWER_QUARTILE(1010),SERVICE_averageHostedServices_MEDIAN(1011),SERVICE_averageHostedServices_UPPER_QUARTILE(1012),SERVICE_averageHostedServices_MAX(1013)
	SERVICE_hostedServices => {
		trace_index		=> 125,
		aggregated_index	=> 1006,
		label			=> "# Service Instances",
		label_y_offset		=> 1,
		default_range		=> "[0:27.5]",
		scale			=> 1,
		logscale		=> 0,
		style			=> "default",
	},
	# TRACE: TODO
	# AGGREGATED: SERVICE_syncOverhead_absolute_MEAN(1022),SERVICE_syncOverhead_absolute_VAR(1023),SERVICE_syncOverhead_absolute_STDDEV(1024),SERVICE_syncOverhead_absolute_MIN(1025),SERVICE_syncOverhead_absolute_LOWER_QUARTILE(1026),SERVICE_syncOverhead_absolute_MEDIAN(1027),SERVICE_syncOverhead_absolute_UPPER_QUARTILE(1028),SERVICE_syncOverhead_absolute_MAX(1029)
	SERVICE_syncOverhead_absolute => {
		trace_index		=> 0,
		aggregated_index	=> 1022,
		label			=> "Sync. Overhead [MB]",
		label_y_offset		=> 2,
		default_range		=> "",
		scale			=> 0.000001,
		logscale		=> 0,
		style			=> "default",
	},
	# TRACE: TODO
	# AGGREGATED: SERVICE_syncOverhead_relative_MEAN(1030),SERVICE_syncOverhead_relative_VAR(1031),SERVICE_syncOverhead_relative_STDDEV(1032),SERVICE_syncOverhead_relative_MIN(1033),SERVICE_syncOverhead_relative_LOWER_QUARTILE(1034),SERVICE_syncOverhead_relative_MEDIAN(1035),SERVICE_syncOverhead_relative_UPPER_QUARTILE(1036),SERVICE_syncOverhead_relative_MAX(1037)
	SERVICE_syncOverhead_relative => {
		trace_index		=> 0,
		aggregated_index	=> 1030,
		label			=> "Sync. Overhead [%]",
		label_y_offset		=> 2,
		default_range		=> "",
		scale			=> 100,
		logscale		=> 0,
		style			=> "default",
	},
	# TRACE: SERVICE_totalReplications(130)
	# AGGREGATED: SERVICE_totalReplications_MEAN(1038),SERVICE_totalReplications_VAR(1039),SERVICE_totalReplications_STDDEV(1040),SERVICE_totalReplications_MIN(1041),SERVICE_totalReplications_LOWER_QUARTILE(1042),SERVICE_totalReplications_MEDIAN(1043),SERVICE_totalReplications_UPPER_QUARTILE(1044),SERVICE_totalReplications_MAX(1045)
	SERVICE_totalReplications => {
		trace_index		=> 130,
		aggregated_index	=> 1038,
		label			=> "# Service Replications",
		label_y_offset		=> 1,
		default_range		=> "[0:82]",
		scale			=> 1,
		logscale		=> 0,
		style			=> "default",
	},
	# TRACE: SERVICE_totalMigrations(133)
	# AGGREGATED: SERVICE_totalMigrations_MEAN(1062),SERVICE_totalMigrations_VAR(1063),SERVICE_totalMigrations_STDDEV(1064),SERVICE_totalMigrations_MIN(1065),SERVICE_totalMigrations_LOWER_QUARTILE(1066),SERVICE_totalMigrations_MEDIAN(1067),SERVICE_totalMigrations_UPPER_QUARTILE(1068),SERVICE_totalMigrations_MAX(1069)
	SERVICE_totalMigrations => {
		trace_index		=> 133,
		aggregated_index	=> 1062,
		label			=> "# Service Migrations",
		label_y_offset		=> 1,
		default_range		=> "[0:16]",
		scale			=> 1,
		logscale		=> 0,
		style			=> "default",
	},
	# TRACE: PLACEMENT_currentServiceCost(136)
	# AGGREGATED: PLACEMENT_currentServiceCost_MEAN(1086),PLACEMENT_currentServiceCost_VAR(1087),PLACEMENT_currentServiceCost_STDDEV(1088),PLACEMENT_currentServiceCost_MIN(1089),PLACEMENT_currentServiceCost_LOWER_QUARTILE(1090),PLACEMENT_currentServiceCost_MEDIAN(1091),PLACEMENT_currentServiceCost_UPPER_QUARTILE(1092),PLACEMENT_currentServiceCost_MAX(1093)
	PLACEMENT_currentServiceCost => {
		trace_index		=> 136,
		aggregated_index	=> 1086,
		label			=> "Service Cost [kB/s]",
		label_y_offset		=> 1,
		default_range		=> "[0:350]",
		scale			=> 0.001,
		logscale		=> 0,
		style			=> "value_over_time",
	},
	# TRACE: PLACEMENT_currentlyOptimalServiceCost(137)
	# AGGREGATED: PLACEMENT_currentlyOptimalServiceCost_MEAN(1094),PLACEMENT_currentlyOptimalServiceCost_VAR(1095),PLACEMENT_currentlyOptimalServiceCost_STDDEV(1096),PLACEMENT_currentlyOptimalServiceCost_MIN(1097),PLACEMENT_currentlyOptimalServiceCost_LOWER_QUARTILE(1098),PLACEMENT_currentlyOptimalServiceCost_MEDIAN(1099),PLACEMENT_currentlyOptimalServiceCost_UPPER_QUARTILE(1100),PLACEMENT_currentlyOptimalServiceCost_MAX(1101)
	PLACEMENT_currentlyOptimalServiceCost => {
		trace_index		=> 137,
		aggregated_index	=> 1094,
		label			=> "Optimal Service Cost [kB/s]",
		label_y_offset		=> 1,
		default_range		=> "[0:350]",
		scale			=> 0.001,
		logscale		=> 0,
		style			=> "optimal_value_over_time",
	},
	# TRACE: PLACEMENT_currentlyOptimalNumberOfHosts(138)
	# AGGREGATED: PLACEMENT_currentlyOptimalNumberOfHosts_MEAN(1102),PLACEMENT_currentlyOptimalNumberOfHosts_VAR(1103),PLACEMENT_currentlyOptimalNumberOfHosts_STDDEV(1104),PLACEMENT_currentlyOptimalNumberOfHosts_MIN(1105),PLACEMENT_currentlyOptimalNumberOfHosts_LOWER_QUARTILE(1106),PLACEMENT_currentlyOptimalNumberOfHosts_MEDIAN(1107),PLACEMENT_currentlyOptimalNumberOfHosts_UPPER_QUARTILE(1108),PLACEMENT_currentlyOptimalNumberOfHosts_MAX(1109)
	PLACEMENT_currentlyOptimalNumberOfHosts => {
		trace_index		=> 138,
		aggregated_index	=> 1102,
		label			=> "# Optimal Service Instaces",
		label_y_offset		=> 1,
		default_range		=> "[0:10]",
		scale			=> 1,
		logscale		=> 0,
		style			=> "optimal_value_over_time",
	},
	# TRACE: PLACEMENT_currentlyEstimatedAdaptationCost(139)
	# AGGREGATED: PLACEMENT_currentlyEstimatedAdaptationCost_MEAN(1110),PLACEMENT_currentlyEstimatedAdaptationCost_VAR(1111),PLACEMENT_currentlyEstimatedAdaptationCost_STDDEV(1112),PLACEMENT_currentlyEstimatedAdaptationCost_MIN(1113),PLACEMENT_currentlyEstimatedAdaptationCost_LOWER_QUARTILE(1114),PLACEMENT_currentlyEstimatedAdaptationCost_MEDIAN(1115),PLACEMENT_currentlyEstimatedAdaptationCost_UPPER_QUARTILE(1116),PLACEMENT_currentlyEstimatedAdaptationCost_MAX(1117)
	PLACEMENT_currentlyEstimatedAdaptationCost => {
		trace_index		=> 139,
		aggregated_index	=> 1110,
		label			=> "Est. Adaptation Cost [kB/s]",
		label_y_offset		=> 1,
		default_range		=> "[0:350]",
		scale			=> 0.001,
		logscale		=> 0,
		style			=> "cost_over_time",
	},
	# TRACE: PLACEMENT_totalAdaptations(140)
	# AGGREGATED: PLACEMENT_totalAdaptations_MEAN(1118),PLACEMENT_totalAdaptations_VAR(1119),PLACEMENT_totalAdaptations_STDDEV(1120),PLACEMENT_totalAdaptations_MIN(1121),PLACEMENT_totalAdaptations_LOWER_QUARTILE(1122),PLACEMENT_totalAdaptations_MEDIAN(1123),PLACEMENT_totalAdaptations_UPPER_QUARTILE(1124),PLACEMENT_totalAdaptations_MAX(1125)
	PLACEMENT_totalAdaptations => {
		trace_index		=> 140,
		aggregated_index	=> 1118,
		label			=> "# Service Adaptations",
		label_y_offset		=> 1,
		default_range		=> "[0:16]",
		scale			=> 1,
		logscale		=> 0,
		style			=> "default",
	},
);

# This constant encodes the file format used in evaluation of routing protocols.
our %VALUES_ROUTING = (
	# TRACE: TODO
	# AGGREGATED: MAC_transmittedPackets_MEAN(75),MAC_transmittedPackets_VAR(76),MAC_transmittedPackets_STDDEV(77),MAC_transmittedPackets_MIN(78),MAC_transmittedPackets_LOWER_QUARTILE(79),MAC_transmittedPackets_MEDIAN(80),MAC_transmittedPackets_UPPER_QUARTILE(81),MAC_transmittedPackets_MAX(82)
	MAC_transmittedPackets => {
		trace_index		=> 0,
		aggregated_index	=> 75,
		label			=> "Packets [thousands]",
		label_y_offset		=> 1,
		default_range		=> "[0:2100]",
		scale			=> 0.001,
		logscale		=> 0,
		style			=> "default",
	},
	# TRACE: TODO
	# AGGREGATED: MAC_transmittedBytes_MEAN(131),MAC_transmittedBytes_VAR(132),MAC_transmittedBytes_STDDEV(133),MAC_transmittedBytes_MIN(134),MAC_transmittedBytes_LOWER_QUARTILE(135),MAC_transmittedBytes_MEDIAN(136),MAC_transmittedBytes_UPPER_QUARTILE(137),MAC_transmittedBytes_MAX(138)
	MAC_transmittedBytes => {
		trace_index		=> 0,
		aggregated_index	=> 131,
		label			=> "Transmitted Data [MB]",
		label_y_offset		=> 1,
		default_range		=> "[0:1100]",
		scale			=> 0.000001,
		logscale		=> 0,
		style			=> "default",
	},
	# TRACE: TODO
	# AGGREGATED: ICMP_nopPdr_MEAN(307),ICMP_nopPdr_VAR(308),ICMP_nopPdr_STDDEV(309),ICMP_nopPdr_MIN(310),ICMP_nopPdr_LOWER_QUARTILE(311),ICMP_nopPdr_MEDIAN(312),ICMP_nopPdr_UPPER_QUARTILE(313),ICMP_nopPdr_MAX(314)
	ICMP_nopPdr => {
		trace_index		=> 0,
		aggregated_index	=> 307,
		label			=> "Packet Delivery Ratio [%]",
		label_y_offset		=> 2,
		default_range		=> "[0:105]",
		scale			=> 100,
		logscale		=> 0,
		style			=> "default",
	},
);

sub print_gnuplot_header {
	my($GNUPLOT, $filename, $evaluation_ref, $key_offset) = @_;
	my %evaluation = %{$evaluation_ref};

	my $x_axis_label = $evaluation{variable}{label};
	my $x_axis_range = $evaluation{variable}{default_range};
	if(defined $evaluation{x_axis_range}) {$x_axis_range = $evaluation{x_axis_range};}
	my $x_axis_logscale = $evaluation{variable}{logscale};
	my $y_axis_label = $evaluation{value}{label};
	my $y_axis_label_offset = $evaluation{value}{label_y_offset};
	my $y_axis_range = $evaluation{value}{default_range};
	if(defined $evaluation{y_axis_range}) {$y_axis_range = $evaluation{y_axis_range};}
	my $y_axis_logscale = $evaluation{value}{logscale};
	my $key_position = $evaluation{key_position};
	my $box_width = $evaluation{variable}{box_width};

	print $GNUPLOT <<END_HEADER_1 or die "can't print header to gnuplot: $!";

# Output control
set terminal postscript eps enhanced color 24
set output "$filename.eps"

# Input control
set datafile separator ","

# Scale / aspect ratio
set size ratio 0.5 1.0,1.6180339887	# Golden ratio

# Line and point styles
END_HEADER_1

	for my $style (keys %STYLES) {
		print $GNUPLOT "set style line ".($STYLES{$style}{offset} + 1)." lt $STYLES{$style}{line_type} lc rgbcolor \"$STYLES{$style}{color}\" lw 2 pt $STYLES{$style}{point_type}\n";
		print $GNUPLOT "set style line ".($STYLES{$style}{offset} + 2)." lt 1 lc rgbcolor \"$STYLES{$style}{color}\" lw 2 pt 2\n";
		print $GNUPLOT "set style line ".($STYLES{$style}{offset} + 3)." lt 1 lc rgbcolor \"$STYLES{$style}{color}\" lw 4 pt 2\n";
	}

	my $x_axis = "set xlabel \"$x_axis_label\"\n";
	if($x_axis_range ne "") {$x_axis .= "set xrange $x_axis_range\n";}
	if($x_axis_logscale) {$x_axis .= "set logscale x\n";}

	my $y_axis = "set ylabel \"$y_axis_label\" offset $y_axis_label_offset,0\n";
	if($y_axis_range ne "") {$y_axis .= "set yrange $y_axis_range\n";}
	if($y_axis_logscale) {$y_axis .= "set logscale y\n";}

	my $key = "";
	if($key_position ne "") {$key = "set key $key_position\nset key width $key_offset\n";}

	print $GNUPLOT <<END_HEADER_2 or die "can't print header to gnuplot: $!";

# Axis: labels, ranges, scales
$x_axis
$y_axis

# Key
$key

# Plot
set boxwidth $box_width
plot \\
END_HEADER_2
}

sub print_gnuplot_footer {
	my($GNUPLOT) = @_;
	print $GNUPLOT <<END_FOOTER or die "can't print plot footer to gnuplot: $!";

END_FOOTER
}

sub eps2pdf {
	my($filename) = @_;
	print `ps2pdf $filename.eps`;
	print `/usr/bin/pdfcrop $filename.pdf $filename.pdf > /dev/null`;
	unlink "$filename.eps";
}

sub generate_time_plot {
	my($experiment, $evaluation_ref, $trace_filename) = @_;
	my %evaluation = %{$evaluation_ref};

	$trace_filename =~ /(.*)\/$experiment\_(.*)_out\.txt/;
	my $filename = "$experiment\_-_$evaluation{id}_-_-_$2";
	print "Generating $filename.pdf...\n";

	my $variable = $evaluation{variable}{trace_index};
	my $value = $evaluation{value}{trace_index};
	my $value_scale = $evaluation{value}{scale};
	my $style = $evaluation{value}{style};

	my $style_index_1 = $STYLES{$style}{offset} + 1;
	my $style_index_2 = $STYLES{$style}{offset} + 2;
	my $style_index_3 = $STYLES{$style}{offset} + 3;

	open(my $GNUPLOT, "| gnuplot") or die "can't open gnuplot: $!";
	print_gnuplot_header($GNUPLOT, $filename, \%evaluation, 0);

	print $GNUPLOT <<END_TIME_SERIES or die "can't print time series to gnuplot: $!";
"< egrep '^[0-9\.]+,[0-9\.]+,[0-9\.]+,[0-9\.]+,' $trace_filename" \\
using (\$$variable/60):(\$$value*$value_scale) with points ls $style_index_1 notitle
END_TIME_SERIES

# "< head -n 1 $trace_filename" \\
# using (2.5):(1000):(5) with boxes lt 1 lc rgbcolor "#EEEEFF" fs solid notitle, \\
# '' using (10):(1000):(1) with boxes lt 1 lc rgbcolor "#EEEEFF" fs solid notitle, \\
# '' using (20):(1000):(1) with boxes lt 1 lc rgbcolor "#EEEEFF" fs solid notitle, \\
# '' using (1.57154666667):(1000):(0) with boxes lt 1 lc rgbcolor "blue" fs solid notitle, \\
# '' using (5.10386833333):(1000):(0) with boxes lt 1 lc rgbcolor "blue" fs solid notitle, \\
# '' using (11.5740833333):(1000):(0) with boxes lt 1 lc rgbcolor "blue" fs solid notitle, \\
# "< egrep '^[0-9\.]+,[0-9\.]+,[0-9\.]+,[0-9\.]+,' $trace_filename" \\
# using (\$$variable/60):(\$137*$value_scale) with points ls 31 notitle, \\
# '' using (\$$variable/60):(\$$value*$value_scale) with points ls $style_index_1 notitle, \\
# '' using (-100):(-100) with points ls $style_index_1 title \"{/$LEGEND_FONT=$LEGEND_FONT_SIZE Current Cost}\", \\
# '' using (-100):(-100) with points ls 31 title \"{/$LEGEND_FONT=$LEGEND_FONT_SIZE Optimal Cost}\", \\
# '' using (-100):(-100) with line lt 1 lc rgbcolor "blue" title \"{/$LEGEND_FONT=$LEGEND_FONT_SIZE Adaptation}\", \\
# '' using (-100):(-100) with boxes lt 1 lc rgbcolor "#EEEEFF" fs solid title \"{/$LEGEND_FONT=$LEGEND_FONT_SIZE Demand Change}\"

	print_gnuplot_footer($GNUPLOT);
	while(<$GNUPLOT>) {print $_;}
	close $GNUPLOT;

	eps2pdf($filename);
}

sub print_gnuplot_plot_point {
	my($GNUPLOT, $experiment, $evaluation_ref, $placement, $style, $title, $continue) = @_;
	my %evaluation = %{$evaluation_ref};

	my $variable = $evaluation{variable}{aggregated_index};
	my $variable_scale = $evaluation{variable}{scale};
	my $mean = $evaluation{value}{aggregated_index} + 0;
	my $var = $evaluation{value}{aggregated_index} + 1;
	my $stddev = $evaluation{value}{aggregated_index} + 2;
	my $min = $evaluation{value}{aggregated_index} + 3;
	my $lower_quartile = $evaluation{value}{aggregated_index} + 4;
	my $median = $evaluation{value}{aggregated_index} + 5;
	my $upper_quartile = $evaluation{value}{aggregated_index} + 6;
	my $max = $evaluation{value}{aggregated_index} + 7;
	my $value_scale = $evaluation{value}{scale};

	my $style_index_1 = $STYLES{$style}{offset} + 1;
	my $style_index_2 = $STYLES{$style}{offset} + 2;
	my $style_index_3 = $STYLES{$style}{offset} + 3;

	print $GNUPLOT <<END_PLOT_POINT or die "can't print point to gnuplot: $!";
"$experiment\_$placement/$experiment\_STATS.csv" \\
END_PLOT_POINT
	if($HIGHLIGHT_MEAN) {print $GNUPLOT "using (\$$variable*$variable_scale):(\$$mean*$value_scale) with points ls $style_index_1 \\\n";}
	if($HIGHLIGHT_MEDIAN) {print $GNUPLOT "using (\$$variable*$variable_scale):(\$$median*$value_scale) with points ls $style_index_1 \\\n";}
	if($title eq '') {print $GNUPLOT "notitle \\\n";}
	else {print $GNUPLOT "title \"{/$LEGEND_FONT=$LEGEND_FONT_SIZE $title}\" \\\n";}
	if($continue) {
		print $GNUPLOT ", ";
	}
}

sub generate_placement_points_plot {
	my($experiment, $evaluation_ref, $placements_ref, $service_sizes_ref, $sync_ratios_ref, $nodes_ref, $active_node_ratios_ref, $request_frequencies_ref, $context_ref, $key_offset) = @_;
	my %evaluation = %{$evaluation_ref};
	my @placements = @{$placements_ref};
	my @service_sizes = @{$service_sizes_ref};
	my @sync_ratios = @{$sync_ratios_ref};
	my @nodes = @{$nodes_ref};
	my @active_node_ratios = @{$active_node_ratios_ref};
	my @request_frequencies = @{$request_frequencies_ref};
	my %context = %{$context_ref};

	my $filename = "$experiment\_-_$evaluation{id}";
	print "Generating $filename.pdf...\n";

	open(my $GNUPLOT, "| gnuplot") or die "can't open gnuplot: $!";
	print_gnuplot_header($GNUPLOT, $filename, \%evaluation, $key_offset);
	my $iteration = 0;
	for my $placement_ref (@placements) {
		my %placement = %{$placement_ref};
		$iteration++;
		print_gnuplot_plot_point($GNUPLOT, $experiment, \%evaluation, $placement{id}, $placement{style}, $placement{label}, ($iteration < scalar(@placements)));
	}
	print_gnuplot_footer($GNUPLOT);
	while(<$GNUPLOT>) {print $_;}
	close $GNUPLOT;

	eps2pdf($filename);
}

sub print_gnuplot_plot_line {
	my($GNUPLOT, $experiment, $evaluation_ref, $placement, $service_size, $sync_ratio, $nodes, $active_node_ratio, $request_frequency, $style, $title, $continue) = @_;
	my %evaluation = %{$evaluation_ref};

	my $parameter_pattern = "";
	if(
		($evaluation{variable}{id} eq "service_size")
		|| ($evaluation{variable}{id} eq "nodes")		# FIXME: Ugly hack to support variable set for routing evaluation.
	) {
		$parameter_pattern .= "[0-9\.]+,";
	} else {
		$parameter_pattern .= "$service_size,";
# 		$parameter_pattern .= "[0-9\.]+,";			# FIXME: Ugly hack to support comparison of SIM, POOL, and TESTBED.
	}
	if(
		($evaluation{variable}{id} eq "sync_ratio")
		|| ($evaluation{variable}{id} eq "packet_frequency")	# FIXME: Ugly hack to support variable set for routing evaluation.
		|| ($evaluation{variable}{id} eq "active_regions")	# FIXME: Ugly hack to support variable set for variable demand evaluation.
	) {
		$parameter_pattern .= "[0-9\.]+,";
# 		$parameter_pattern .= "(0.000|0.0025|0.005|0.0075|0.010|0.020|0.030|0.040|0.050|0.075|0.100),";	# Explicitly select which sync ratios to consider.
	} else {
		$parameter_pattern .= "$sync_ratio,";
	}
# FIXME: Ugly hack: Don't add the following items to the parameter patter for the routing evaluation.
	if($evaluation{variable}{id} eq "nodes") {
		$parameter_pattern .= "[0-9\.]+,";
	} else {
		$parameter_pattern .= "$nodes,";
	}
	if(
		($evaluation{variable}{id} eq "active_node_ratio")
		|| ($evaluation{variable}{id} eq "activity_phases")	# FIXME: Ugly hack to support variable set for variable demand evaluation.
	) {
		$parameter_pattern .= "[0-9\.]+,";
	} else {
		$parameter_pattern .= "$active_node_ratio,";
	}
	if($evaluation{variable}{id} eq "request_frequency") {
		$parameter_pattern .= "[0-9\.]+,";
	} else {
		$parameter_pattern .= "$request_frequency,";
	}

	my $variable = $evaluation{variable}{aggregated_index};
	my $variable_scale = $evaluation{variable}{scale};
	my $mean = $evaluation{value}{aggregated_index} + 0;
	my $var = $evaluation{value}{aggregated_index} + 1;
	my $stddev = $evaluation{value}{aggregated_index} + 2;
	my $min = $evaluation{value}{aggregated_index} + 3;
	my $lower_quartile = $evaluation{value}{aggregated_index} + 4;
	my $median = $evaluation{value}{aggregated_index} + 5;
	my $upper_quartile = $evaluation{value}{aggregated_index} + 6;
	my $max = $evaluation{value}{aggregated_index} + 7;
	my $value_scale = $evaluation{value}{scale};

	my $style_index_1 = $STYLES{$style}{offset} + 1;
	my $style_index_2 = $STYLES{$style}{offset} + 2;
	my $style_index_3 = $STYLES{$style}{offset} + 3;

	if($PLOT_FIVE_NUMBER_SUMMARY) {
		print $GNUPLOT <<END_PLOT_LINE or die "can't print plot line to gnuplot: $!";
"< egrep '^$parameter_pattern' $experiment\_$placement/$experiment\_STATS.csv" \\
using (\$$variable*$variable_scale):(\$$mean*$value_scale) with points ls $style_index_1 notitle, \\
'' using (\$$variable*$variable_scale):(\$$lower_quartile*$value_scale):(\$$min*$value_scale):(\$$max*$value_scale):(\$$upper_quartile*$value_scale) with candlesticks whiskerbars 0.5 ls $style_index_2 notitle, \\
'' using (\$$variable*$variable_scale):(\$$median*$value_scale):(\$$median*$value_scale):(\$$median*$value_scale):(\$$median*$value_scale) with candlesticks ls $style_index_3 notitle, \\
END_PLOT_LINE
	}
	if($PLOT_AVERAGE_AND_STDDEV) {
		print $GNUPLOT <<END_PLOT_LINE or die "can't print plot line to gnuplot: $!";
"< egrep '^$parameter_pattern' $experiment\_$placement/$experiment\_STATS.csv" \\
using (\$$variable*$variable_scale):(\$$mean*$value_scale):((\$$mean-\$$stddev)*$value_scale):((\$$mean+\$$stddev)*$value_scale) with yerrorbars ls $style_index_1 notitle, \\
END_PLOT_LINE
	}
	if($HIGHLIGHT_MEAN) {print $GNUPLOT "'' using (\$$variable*$variable_scale):(\$$mean*$value_scale) with lines ls $style_index_1 notitle, \\\n";}
	if($HIGHLIGHT_MEDIAN) {print $GNUPLOT "'' using (\$$variable*$variable_scale):(\$$median*$value_scale) with lines ls $style_index_1 notitle, \\\n";}
	print $GNUPLOT "'' using (-100):(-100) with linespoints ls $style_index_1 ";
	if($title eq '') {print $GNUPLOT "notitle \\\n";}
	else {print $GNUPLOT "title \"{/$LEGEND_FONT=$LEGEND_FONT_SIZE $title}\" \\\n";}
	if($continue) {
		print $GNUPLOT ", ";
	}
}

sub generate_single_series_plot {
	my($experiment, $evaluation_ref, $placements_ref, $service_sizes_ref, $sync_ratios_ref, $nodes_ref, $active_node_ratios_ref, $request_frequencies_ref, $context_ref, $key_offset) = @_;
	my %evaluation = %{$evaluation_ref};
	my @placements = @{$placements_ref};
	my @service_sizes = @{$service_sizes_ref};
	my @sync_ratios = @{$sync_ratios_ref};
	my @nodes = @{$nodes_ref};
	my @active_node_ratios = @{$active_node_ratios_ref};
	my @request_frequencies = @{$request_frequencies_ref};
	my %context = %{$context_ref};

	if($evaluation{variable}{id} eq "placement") {
		@placements = (0);
	} elsif($evaluation{variable}{id} eq "service_size") {
		@service_sizes = (0);
	} elsif($evaluation{variable}{id} eq "sync_ratio") {
		@sync_ratios = (0);
	} elsif($evaluation{variable}{id} eq "nodes") {
		@nodes = (0);
	} elsif($evaluation{variable}{id} eq "active_node_ratio") {
		@active_node_ratios = (0);
	} elsif($evaluation{variable}{id} eq "request_frequency") {
		@request_frequencies = (0);
	}

	for my $placement_ref (@placements) {
		my %placement = %{$placement_ref};
	for my $service_size (@service_sizes) {
	for my $sync_ratio (@sync_ratios) {
	for my $nodes (@nodes) {
	for my $active_node_ratio (@active_node_ratios) {
	for my $request_frequency (@request_frequencies) {

		my $filename = "$experiment\_-_$evaluation{id}_-";
		if((scalar(@placements) > 1) || exists($context{placement})) {
			$filename .= "_-_placement_$placement{id}";
		}
		if((scalar(@service_sizes) > 1) || exists($context{service_size})) {
			$filename .= "_-_service_size_$service_size";
		}
		if((scalar(@sync_ratios) > 1) || exists($context{sync_ratio})) {
			$filename .= "_-_sync_ratio_$sync_ratio";
		}
		if((scalar(@nodes) > 1) || exists($context{nodes})) {
			$filename .= "_-_nodes_$nodes";
		}
		if((scalar(@active_node_ratios) > 1) || exists($context{active_node_ratio})) {
			$filename .= "_-_activity_factor_$active_node_ratio";
		}
		if((scalar(@request_frequencies) > 1) || exists($context{request_frequency})) {
			$filename .= "_-_request_frequency_$request_frequency";
		}
		print "Generating $filename.pdf...\n";

		open(my $GNUPLOT, "| gnuplot") or die "can't open gnuplot: $!";
		print_gnuplot_header($GNUPLOT, $filename, \%evaluation, $key_offset);
		print_gnuplot_plot_line($GNUPLOT, $experiment, \%evaluation, $placement{id}, $service_size, $sync_ratio, $nodes, $active_node_ratio, $request_frequency, $placement{style}, '', 0);
		print_gnuplot_footer($GNUPLOT);
		while(<$GNUPLOT>) {print $_;}
		close $GNUPLOT;

		eps2pdf($filename);

	}
	}
	}
	}
	}
	}
}

sub generate_placement_series_plot {
	my($experiment, $evaluation_ref, $placements_ref, $service_sizes_ref, $sync_ratios_ref, $nodes_ref, $active_node_ratios_ref, $request_frequencies_ref, $context_ref, $key_offset) = @_;
	my %evaluation = %{$evaluation_ref};
	my @placements = @{$placements_ref};
	my @service_sizes = @{$service_sizes_ref};
	my @sync_ratios = @{$sync_ratios_ref};
	my @nodes = @{$nodes_ref};
	my @active_node_ratios = @{$active_node_ratios_ref};
	my @request_frequencies = @{$request_frequencies_ref};
	my %context = %{$context_ref};

	if($evaluation{variable}{id} eq "service_size") {
		@service_sizes = (0);
	} elsif($evaluation{variable}{id} eq "nodes") {			# FIXME: Ugly hack to support variable set for routing evaluation.
		@service_sizes = (0);
	} elsif($evaluation{variable}{id} eq "sync_ratio") {
		@sync_ratios = (0);
	} elsif($evaluation{variable}{id} eq "packet_frequency") {	# FIXME: Ugly hack to support variable set for routing evaluation.
		@sync_ratios = (0);
	} elsif($evaluation{variable}{id} eq "active_regions") {	# FIXME: Ugly hack to support variable set for variable demand evaluation.
		@sync_ratios = (0);
	} elsif($evaluation{variable}{id} eq "nodes") {
		@nodes = (0);
	} elsif($evaluation{variable}{id} eq "active_node_ratio") {
		@active_node_ratios = (0);
	} elsif($evaluation{variable}{id} eq "activity_phases") {	# FIXME: Ugly hack to support variable set for variable demand evaluation.
		@active_node_ratios = (0);
	} elsif($evaluation{variable}{id} eq "request_frequency") {
		@request_frequencies = (0);
	}

	for my $service_size (@service_sizes) {
	for my $sync_ratio (@sync_ratios) {
	for my $nodes (@nodes) {
	for my $active_node_ratio (@active_node_ratios) {
	for my $request_frequency (@request_frequencies) {

		my $filename = "$experiment\_-_$evaluation{id}_-";
		if((scalar(@service_sizes) > 1) || exists($context{service_size})) {
			$filename .= "_-_service_size_$service_size";
# 			$filename .= "_-_nodes_$service_size";		# FIXME: Ugly hack to support variable set for routing evaluation.
		}
		if((scalar(@sync_ratios) > 1) || exists($context{sync_ratio})) {
			$filename .= "_-_sync_ratio_$sync_ratio";
# 			$filename .= "_-_active_regions_$sync_ratio";		# FIXME: Ugly hack to support variable set for variable demand evaluation.
		}
		if((scalar(@nodes) > 1) || exists($context{nodes})) {
			$filename .= "_-_nodes_$nodes";
		}
		if((scalar(@active_node_ratios) > 1) || exists($context{active_node_ratio})) {
			$filename .= "_-_activity_factor_$active_node_ratio";
# 			$filename .= "_-_activity_phases_$active_node_ratio";	# FIXME: Ugly hack to support variable set for variable demand evaluation.
		}
		if((scalar(@request_frequencies) > 1) || exists($context{request_frequency})) {
			$filename .= "_-_request_frequency_$request_frequency";
		}
		print "Generating $filename.pdf...\n";

		open(my $GNUPLOT, "| gnuplot") or die "can't open gnuplot: $!";
		print_gnuplot_header($GNUPLOT, $filename, \%evaluation, $key_offset);
		my $iteration = 0;
		for my $placement_ref (@placements) {
			my %placement = %{$placement_ref};
			$iteration++;
			print_gnuplot_plot_line($GNUPLOT, $experiment, \%evaluation, $placement{id}, $service_size, $sync_ratio, $nodes, $active_node_ratio, $request_frequency, $placement{style}, $placement{label}, ($iteration < scalar(@placements)));
		}
		print_gnuplot_footer($GNUPLOT);
		while(<$GNUPLOT>) {print $_;}
		close $GNUPLOT;

		eps2pdf($filename);

	}
	}
	}
	}
	}
}

sub generate_hop_count_histogram_plot {
	my($experiment, $evaluation_ref, $placements_ref, $service_sizes_ref, $sync_ratios_ref, $nodes_ref, $active_node_ratios_ref, $request_frequencies_ref, $context_ref, $key_offset) = @_;
	my %evaluation = %{$evaluation_ref};
	my @placements = @{$placements_ref};
	my @service_sizes = @{$service_sizes_ref};
	my @sync_ratios = @{$sync_ratios_ref};
	my @nodes = @{$nodes_ref};
	my @active_node_ratios = @{$active_node_ratios_ref};
	my @request_frequencies = @{$request_frequencies_ref};
	my %context = %{$context_ref};

	if($evaluation{variable}{id} eq "placement") {
		@placements = (0);
	} elsif($evaluation{variable}{id} eq "service_size") {
		@service_sizes = (0);
	} elsif($evaluation{variable}{id} eq "sync_ratio") {
		@sync_ratios = (0);
	} elsif($evaluation{variable}{id} eq "nodes") {
		@nodes = (0);
	} elsif($evaluation{variable}{id} eq "active_node_ratio") {
		@active_node_ratios = (0);
	} elsif($evaluation{variable}{id} eq "request_frequency") {
		@request_frequencies = (0);
	}

	for my $placement_ref (@placements) {
		my %placement = %{$placement_ref};
	for my $service_size (@service_sizes) {
	for my $sync_ratio (@sync_ratios) {
	for my $nodes (@nodes) {
	for my $active_node_ratio (@active_node_ratios) {
	for my $request_frequency (@request_frequencies) {

		my $filename = "$experiment\_-_$evaluation{id}_-";
		if((scalar(@placements) > 1) || exists($context{placement})) {
			$filename .= "_-_placement_$placement{id}";
		}
		if((scalar(@service_sizes) > 1) || exists($context{service_size})) {
			$filename .= "_-_service_size_$service_size";
		}
		if((scalar(@sync_ratios) > 1) || exists($context{sync_ratio})) {
			$filename .= "_-_sync_ratio_$sync_ratio";
		}
		if((scalar(@nodes) > 1) || exists($context{nodes})) {
			$filename .= "_-_nodes_$nodes";
		}
		if((scalar(@active_node_ratios) > 1) || exists($context{active_node_ratio})) {
			$filename .= "_-_activity_factor_$active_node_ratio";
		}
		if((scalar(@request_frequencies) > 1) || exists($context{request_frequency})) {
			$filename .= "_-_request_frequency_$request_frequency";
		}
		print "Generating $filename.pdf...\n";
	
		open(my $GNUPLOT, "| gnuplot") or die "can't open gnuplot: $!";
		print_gnuplot_header($GNUPLOT, $filename, \%evaluation, $key_offset);
	
		my $max_hop_count = $evaluation{variable}{aggregated_index};
		my $value_scale = $evaluation{value}{scale};
	
		my $style_index_1 = $STYLES{$placement{style}}{offset} + 1;
		my $style_index_2 = $STYLES{$placement{style}}{offset} + 2;
		my $style_index_3 = $STYLES{$placement{style}}{offset} + 3;
	
		print $GNUPLOT <<END_HISTOGRAM_1 or die "can't print histogram to gnuplot: $!";
"< egrep '^$service_size,$sync_ratio,$nodes,$active_node_ratio,$request_frequency,' $experiment\_$placement{id}/$experiment\_STATS.csv" \\
END_HISTOGRAM_1
		for my $hop_count (0..$max_hop_count) {
			my $mean = $evaluation{value}{aggregated_index} + ($hop_count * 8) + 0;
			my $var = $evaluation{value}{aggregated_index} + ($hop_count * 8) + 1;
			my $stddev = $evaluation{value}{aggregated_index} + ($hop_count * 8) + 2;
			my $min = $evaluation{value}{aggregated_index} + ($hop_count * 8) + 3;
			my $lower_quartile = $evaluation{value}{aggregated_index} + ($hop_count * 8) + 4;
			my $median = $evaluation{value}{aggregated_index} + ($hop_count * 8) + 5;
			my $upper_quartile = $evaluation{value}{aggregated_index} + ($hop_count * 8) + 6;
			my $max = $evaluation{value}{aggregated_index} + ($hop_count * 8) + 7;
	
			if($hop_count > 0) {
				print $GNUPLOT "'' ";
			}
			if($HIGHLIGHT_MEAN) {print $GNUPLOT "using ($hop_count):(\$$mean*$value_scale):(0.8) with boxes fs solid 0.1 ls $style_index_2 notitle, \\\n";}
			if($HIGHLIGHT_MEDIAN) {print $GNUPLOT "using ($hop_count):(\$$median*$value_scale):(0.8) with boxes fs solid 0.1 ls $style_index_2 notitle, \\\n";}
			print $GNUPLOT <<END_HISTOGRAM_2 or die "can't print histogram to gnuplot: $!";
'' using ($hop_count):(\$$mean*$value_scale) with points ls $style_index_1 notitle, \\
'' using ($hop_count):(\$$lower_quartile*$value_scale):(\$$min*$value_scale):(\$$max*$value_scale):(\$$upper_quartile*$value_scale) with candlesticks whiskerbars 0.5 ls $style_index_2 notitle, \\
'' using ($hop_count):(\$$median*$value_scale):(\$$median*$value_scale):(\$$median*$value_scale):(\$$median*$value_scale) with candlesticks ls $style_index_3 notitle \\
END_HISTOGRAM_2
			if($hop_count < $max_hop_count) {
				print $GNUPLOT ", ";
			}
		}
		print_gnuplot_footer($GNUPLOT);
		while(<$GNUPLOT>) {print $_;}
		close $GNUPLOT;
	
		eps2pdf($filename);

	}
	}
	}
	}
	}
	}
}

sub generate_summary_plots {
	my($experiment, $evaluations_ref, $placements_ref, $service_sizes_ref, $sync_ratios_ref, $nodes_ref, $active_node_ratios_ref, $request_frequencies_ref, $context_ref) = @_;
	my @evaluations = @{$evaluations_ref};
	my @placements = @{$placements_ref};
	my @service_sizes = @{$service_sizes_ref};
	my @sync_ratios = @{$sync_ratios_ref};
	my @nodes = @{$nodes_ref};
	my @active_node_ratios = @{$active_node_ratios_ref};
	my @request_frequencies = @{$request_frequencies_ref};
	my %context = %{$context_ref};

	my $key_offset = 0;
	for my $placement_ref (@placements) {
		my %placement = %{$placement_ref};
		if($placement{label_offset} < $key_offset) {
			$key_offset = $placement{label_offset};
		}
	}

	for my $evaluation_ref (@evaluations) {
		my %evaluation = %{$evaluation_ref};

		my %evaluation_context = %context;
		delete $evaluation_context{$evaluation{variable}{id}};

		if($evaluation{plot_type} == $PLOT_TYPE_NONE) {
			print "\nERROR: No plot type specified for evaluation \"$evaluation{id}\".\n\n";
			exit(2);
		} elsif($evaluation{plot_type} == $PLOT_TYPE_POINTS_PER_PLACEMENT) {
			generate_placement_points_plot($experiment, \%evaluation, \@placements, \@service_sizes, \@sync_ratios, \@nodes, \@active_node_ratios, \@request_frequencies, \%evaluation_context, $key_offset);
		} elsif($evaluation{plot_type} == $PLOT_TYPE_TIME_SERIES) {
			print "\nERROR: Time series plot type requested by evaluation \"$evaluation{id}\" is not supported for summary plots.\n\n";
			exit(2);
		} elsif($evaluation{plot_type} == $PLOT_TYPE_DATA_SERIES) {
			generate_single_series_plot($experiment, \%evaluation, \@placements, \@service_sizes, \@sync_ratios, \@nodes, \@active_node_ratios, \@request_frequencies, \%evaluation_context, $key_offset);
		} elsif($evaluation{plot_type} == $PLOT_TYPE_DATA_SERIES_PER_PLACEMENT) {
			generate_placement_series_plot($experiment, \%evaluation, \@placements, \@service_sizes, \@sync_ratios, \@nodes, \@active_node_ratios, \@request_frequencies, \%evaluation_context, $key_offset);
		} elsif($evaluation{plot_type} == $PLOT_TYPE_HOP_COUNT_HISTOGRAM) {
			generate_hop_count_histogram_plot($experiment, \%evaluation, \@placements, \@service_sizes, \@sync_ratios, \@nodes, \@active_node_ratios, \@request_frequencies, \%evaluation_context, $key_offset);
		} else {
			print "\nERROR: Invalid plot type \"$evaluation{plot}\" specified for evaluation \"$evaluation{id}\".\n\n";
			exit(2);
		}

	}
}

sub generate_time_plots {
	my($experiment, $evaluations_ref, $trace_filename) = @_;
	my @evaluations = @{$evaluations_ref};

	if(!-e $trace_filename) {
		print "\nERROR: File \"$trace_filename\" does not exist.\n\n";
		exit(2);
	}
	if(!-r $trace_filename) {
		print "\nERROR: File \"$trace_filename\" is not readable.\n\n";
		exit(2);
	}

	for my $evaluation_ref (@evaluations) {
		my %evaluation = %{$evaluation_ref};

		if($evaluation{plot_type} == $PLOT_TYPE_NONE) {
			print "\nERROR: No plot type specified for evaluation \"$evaluation{id}\".\n\n";
			exit(2);
		} elsif($evaluation{plot_type} == $PLOT_TYPE_POINTS_PER_PLACEMENT) {
			print "\nERROR: Placement points plot type requested by evaluation \"$evaluation{id}\" is not supported for time plots.\n\n";
			exit(2);
		} elsif($evaluation{plot_type} == $PLOT_TYPE_TIME_SERIES) {
			generate_time_plot($experiment, \%evaluation, $trace_filename);
		} elsif($evaluation{plot_type} == $PLOT_TYPE_DATA_SERIES) {
			print "\nERROR: Single series plot type requested by evaluation \"$evaluation{id}\" is not supported for time plots.\n\n";
			exit(2);
		} elsif($evaluation{plot_type} == $PLOT_TYPE_DATA_SERIES_PER_PLACEMENT) {
			print "\nERROR: Placement series plot type requested by evaluation \"$evaluation{id}\" is not supported for time plots.\n\n";
			exit(2);
		} elsif($evaluation{plot_type} == $PLOT_TYPE_HOP_COUNT_HISTOGRAM) {
			print "\nERROR: Hop count histogram plot type requested by evaluation \"$evaluation{id}\" is not supported for time plots.\n\n";
			exit(2);
		} else {
			print "\nERROR: Invalid plot type \"$evaluation{plot}\" specified for evaluation \"$evaluation{id}\".\n\n";
			exit(2);
		}

	}
}
