#!/usr/bin/perl

##
#
# This script takes the collected simulation results from several runs and
# generates statistics based on the raw data.
#
# The Statistics::Descriptive module is used to do the math (see
# http://search.cpan.org/~colink/Statistics-Descriptive-2.6/Descriptive.pm)
#
# Example:	calculate_statistics.pl sim.out nodes,position run
#
# Copyright (c) 2007, 2008, 2009, 2010 Georg Wittenburg, Freie Universitaet
# Berlin (wittenbu@inf.fu-berlin.de). All rights reserved.
# Please refer to the COPYING file for detailed licensing information.
#
# $Id: calculate_statistics.pl 3324 2010-02-19 16:59:08Z gwitten $
#
##

use strict;

use Statistics::Descriptive;

my $OUT_FILE;
my $global_num_values = -1;

sub do_stats {
	my($current_variable_string, $array_of_values_ref) = @_;
	my @array_of_values = @{$array_of_values_ref};
	my $num_runs = scalar(@array_of_values);
	my $num_values = @{$array_of_values[0]};

	if($num_values != $global_num_values) {
		if($global_num_values = -1) {
			$global_num_values = $num_values;
		} else {
			die "The number of values has changed unexpectedly ($num_values != $global_num_values)";
		}
	}

	print "Processing $num_runs runs of $num_values values for variables $current_variable_string...\n";
	print $OUT_FILE $current_variable_string;
	for(my $value_index = 0; $value_index < $num_values; $value_index++) {
		if($array_of_values[0][$value_index] ne 'n/a') {
			my $stat = Statistics::Descriptive::Full->new();
			for(my $run_index = 0; $run_index < $num_runs; $run_index++) {
				$stat->add_data($array_of_values[$run_index][$value_index]);
			}
			print $OUT_FILE ','.$stat->mean();
			print $OUT_FILE ','.$stat->variance();
			print $OUT_FILE ','.$stat->standard_deviation();
			print $OUT_FILE ','.$stat->min();
			print $OUT_FILE ','.$stat->percentile(25);
			print $OUT_FILE ','.$stat->median();
			print $OUT_FILE ','.$stat->percentile(75);
			print $OUT_FILE ','.$stat->max();
		} else {
			print $OUT_FILE ',n/a';
			print $OUT_FILE ',n/a';
			print $OUT_FILE ',n/a';
			print $OUT_FILE ',n/a';
			print $OUT_FILE ',n/a';
			print $OUT_FILE ',n/a';
			print $OUT_FILE ',n/a';
			print $OUT_FILE ',n/a';
		}
	}
	print $OUT_FILE "\n";
}

if(scalar(@ARGV) != 3) {
	print "Usage:\tcalculate_statistics.pl collected_output_file variable_name(s) aggregation_variable_name\n\n";
	exit(1);
}

my $csv_file = $ARGV[0];
my @variable_names = split(/,/, $ARGV[1]);
my $aggregation_variable_name = $ARGV[2];

if($csv_file !~ /(.+)\.csv\Z/o) {
	print "CSV file \"$csv_file\" does not end with \".csv\".\n";
	exit(1);
}
my $base_name = $1;

print "CSV file is $csv_file.\n";
print "Variables names are [".join(',', @variable_names)."], aggregation variable name is \"$aggregation_variable_name\".\n";
print "\n";

my $header = "";
my $current_variable_string = "";
my @array_of_values;
open(my $IN_FILE, "<$csv_file") or die $!;
open($OUT_FILE, ">$base_name".'_STATS.csv') or die $!;
while(<$IN_FILE>) {
	chomp;

	# Process header.
	if($header eq "") {
		$header = $_;
		my $expected_header = join(',', @variable_names).",$aggregation_variable_name,";
		if($header !~ /\A$expected_header(.+)/) {
			print "ERROR: Header in collected output file doesn't match the expected header \"$expected_header\".\n";
			exit(1);
		}
		my @column_titles = split(',', $1);

		my $column = 1;
		for my $variable_name (@variable_names) {
			if($variable_name eq $variable_names[0]) {
				print $OUT_FILE "$variable_name($column)";
			} else {
				print $OUT_FILE ",$variable_name($column)";
			}
			$column++;
		}
		for my $column_title (@column_titles) {
			for my $suffix ("_MEAN", "_VAR", "_STDDEV", "_MIN", "_LOWER_QUARTILE", "_MEDIAN", "_UPPER_QUARTILE", "_MAX") {
				print $OUT_FILE ",$column_title$suffix($column)";
				$column++;
			}
		}
		print $OUT_FILE "\n";
		next;
	}

	# Process data.
	if(($current_variable_string eq "") || ($_ !~ /\A$current_variable_string,(\d+),(.+)/)) {
		if($current_variable_string ne "") {
			do_stats($current_variable_string, \@array_of_values);
		}

		my @current_variable_values = split(',', $_);
		splice(@current_variable_values, scalar(@variable_names));
		$current_variable_string = join(',', @current_variable_values);

		@array_of_values = ();
	}

	$_ !~ /\A$current_variable_string,(\d+),(.+)/;
	#print "Adding run #$1 for variables $current_variable_string...\n";
	my @values = split(',', $2);
	push(@array_of_values, \@values);
}
do_stats($current_variable_string, \@array_of_values);
close($IN_FILE);
close($OUT_FILE);
