#!/usr/bin/perl

##
#
# Generate scenarios for evaluations on a testbed.
#
# Copyright (c) 2010 Georg Wittenburg, Freie Universitaet Berlin
# (wittenbu@inf.fu-berlin.de). All rights reserved.
# Please refer to the COPYING file for detailed licensing information.
#
# $Id: eval_generate_scenario.pl 3484 2010-04-28 08:42:28Z gwitten $
#
##

use strict;

my $SERVICE_ID = "srvA";
# my $REQUEST_SIZE = 1024;	# Used for POOL.
my $REQUEST_SIZE = 256;	# Used for TESTBED.

if(scalar(@ARGV) < 10) {
	print "Usage:\tgenerate_scenario.pl scenario_name runs number_of_nodes\n";
	print "\tsetup_duration run_duration service_size sync_traffic_ratio\n";
	print "\tactive_node_ratio request_frequency placement_algorithm\n";
	print "\t[topology_directory]\n";
	print "\n";
	print "Example:\t./eval_generate_scenario.pl test_eval 10 100 300 900 100000 \\\n";
	print "\t\t0.05 0.5 5.0 graph_cost_multiple_instances\n";
	print "\n";
	exit(1);
}

my $scenario_name = $ARGV[0];
my $runs = $ARGV[1];
my $number_of_nodes = $ARGV[2];
my $setup_duration = $ARGV[3];
my $run_duration = $ARGV[4];
my $service_size = $ARGV[5];
my $sync_traffic_ratio = $ARGV[6];
my $active_node_ratio = $ARGV[7];
my $request_frequency = $ARGV[8];
my $placement_algorithm = $ARGV[9];
my $topology_directory = "";
if(scalar(@ARGV) == 11) {
	$topology_directory = $ARGV[10];
}

# TODO: Clean up and document the distinctions between the phases.
my $start_phase_begin = 10.0;	# Add 10s at the beginning to make sure the program is running on all hosts.
				# This is removed later on when the results are collected in eval_collect_results.pl.
my $start_phase_end = $start_phase_begin + $setup_duration;
my $run_begin = $start_phase_end;
my $run_end = $run_begin + $run_duration;
my $stop_phase_begin = $run_end + 60.0;
my $stop_phase_end = $stop_phase_begin + 60.0;
my $finalize_phase_begin = $stop_phase_end + 60.0;
my $finalize_phase_end = $finalize_phase_begin + 300.0;

if(-e $scenario_name) {
	print "ERROR: A directory for scenario \"$scenario_name\" already exists.\n";
	exit(2);
}

if($topology_directory eq "") {
	print "\nGenerating $runs variation(s) of scenario \"$scenario_name\" for $number_of_nodes nodes...\n\n";
} else {
	print "\nGenerating $runs variation(s) of scenario \"$scenario_name\" for $number_of_nodes nodes with topologies from $topology_directory...\n";
}
mkdir $scenario_name;

for my $run (1..$runs) {
	my $run_str = $run;
	if($run < 10) {
		$run_str = '0'.$run;
	}

	my $initial_service_host = (int(rand($number_of_nodes)) + 1);
	print "Generating run #$run_str with initial service host is node #$initial_service_host...\n";
	mkdir "$scenario_name/run_$run_str";

	for my $node (1..$number_of_nodes) {
		my $node_str = $node;
		if($node < 10) {
			$node_str = '00'.$node;
		} elsif($node < 100) {
			$node_str = '0'.$node;
		}

		my $is_active = 0;
		if(rand(1.0) <= $active_node_ratio) {
			$is_active = 1;
		}
		my $start_time = rand($start_phase_end - $start_phase_begin) + $start_phase_begin;
		# my $stop_time = rand($stop_phase_end - $stop_phase_begin) + $stop_phase_begin;
		my $stop_time = $stop_phase_begin;
		my $command_file = "$scenario_name/run_$run_str/node_$node_str\_cmnds.txt";

		open(COMMANDS, ">$command_file") or die "Can't open file \"$command_file\" for writing";
	
		if($topology_directory ne "") {
			my $topology_file = "$topology_directory/$number_of_nodes\_nodes/topology_$run_str.txt";
			if(!-e $topology_file) {
				print "ERROR: Topology file \"$topology_file\" does not exist.\n";
				exit(3);
			}
			if(!-r $topology_file) {
				print "ERROR: Topology file \"$topology_file\" is not readable.\n";
				exit(3);
			}
			print COMMANDS "0.0 set_topology $topology_file\n";
		}

		if($node == $initial_service_host) {
			if($is_active) {
				print "Generating node #$node_str of variation #$run_str (initial service host)...\n";
			} else {
				print "Generating node #$node_str of variation #$run_str (initial service host, inactive)...\n";
			}
			print COMMANDS "$start_phase_begin SERVICE_create_service $SERVICE_ID $service_size $sync_traffic_ratio\n";
			print COMMANDS "$start_phase_begin SERVICE_set_placement_algorithm $SERVICE_ID $placement_algorithm\n";
		} elsif($is_active) {
			print "Generating node #$node_str of variation #$run_str...\n";
		} else {
			print "Generating node #$node_str of variation #$run_str (inactive)...\n";
		}

		print COMMANDS "$start_time start\n";

		if($is_active) {
			for(my $t = $start_time; $t < ($stop_time - (1.0 / $request_frequency)); $t += (1.0 / $request_frequency)) {
				my $t_rand = $t + rand(1.0 / $request_frequency);
				if(rand(1.0) <=  $sync_traffic_ratio) {
					print COMMANDS "$t_rand SERVICE_update_data srvA 0x0000 $REQUEST_SIZE\n";
				} else {
					print COMMANDS "$t_rand SERVICE_request_data srvA 0x0000 $REQUEST_SIZE\n";
				}
			}
		}

		print COMMANDS "$stop_time stop\n";
		print COMMANDS "$finalize_phase_begin finalize\n";

		close(COMMANDS);
	}

	print "\n";
}

exit();
