#!/usr/bin/perl

##
#
# This script gathers simulation results from all OUT files in a directory. The
# results are commonly the last line of each file as a comma-separated-values
# (CSV) list. Additionally, the script parses the file names according to the
# variable names given as a command line argument. The entire data is written
# to a new CSV file.
#
# Example:	sim_collect_results.pl . my_sim nodes,position,run
#
# Copyright (c) 2007, 2008, 2009, 2010 Georg Wittenburg, Freie Universitaet
# Berlin (wittenbu@inf.fu-berlin.de). All rights reserved.
# Please refer to the COPYING file for detailed licensing information.
#
# $Id: sim_collect_results.pl 3287 2010-02-12 17:27:34Z gwitten $
#
##

use strict;

sub byfields {
	my $string_a = $a;
	my $string_b = $b;
	my $cmp;
	do {
		$string_a =~ /^(.+?)_(.+)$/;
		my $field_a = $1;
		$string_a = $2;
		$string_b =~ /^(.+?)_(.+)$/;
		my $field_b = $1;
		$string_b = $2;
		$cmp = ($field_a <=> $field_b);
	} while($cmp == 0);
	return $cmp;
}

if((scalar(@ARGV) < 3) || (scalar(@ARGV) > 4)) {
	print "Usage:\tsim_collect_results.pl out_directory base_name variable_name(s) [result_section]\n\n";
	exit(1);
}

my $out_directory = $ARGV[0];
my $base_name = $ARGV[1];
my @variable_names = split(/,/, $ARGV[2]);
my $result_section = 1;
if(scalar(@ARGV) == 4) {
	$result_section = $ARGV[3];
}

if($out_directory !~ /\/\Z/o) {
	$out_directory .= '/';
}

print "Output files are in $out_directory.\n";
print "Base name is \"$base_name\", variables names are [".join(',', @variable_names)."].\n";
print "\n";

if(-e "$base_name.csv") {
	die "Output file $base_name.csv already exists";
}

print "Scanning for output files...\n";
my @output_files;
opendir(my $DIR, $out_directory) or die $!;
while(my $file = readdir($DIR)) {
	if($file =~ /$base_name.+_out\.txt\Z/) {
		push(@output_files, $out_directory.'/'.$file);
	}
}
closedir($DIR);
@output_files = sort byfields @output_files;
print "Found ".scalar(@output_files)." output file(s).\n";
print "\n";

my $file_count = 0;
my $print_header = 1;
open(my $CSV_FILE, ">$base_name.csv") or die $!;
for my $output_file (@output_files) {
	my $simulation_complete_flag = 0;
	my $current_result_section = 0;
	$file_count++;

	print "Processing $output_file ($file_count/".scalar(@output_files).")...\n";
	open(my $OUTPUT_FILE, "<$output_file") or die $!;
	OUTPUT_LINE: while(<$OUTPUT_FILE>) {
		if($_ =~ /Simulation complete/) {
			$simulation_complete_flag = 1;
		}
		if($_ =~ /^###$/) {

			# Skip to the correct result section.
			$current_result_section++;
			if($current_result_section != $result_section) {
				next OUTPUT_LINE;
			}

			# Check whether the simulation has been completed.
			if(!$simulation_complete_flag) {
				die "No simulation complete flag before results section #$result_section in $output_file";
			}

			my $header_line = 1;
			# Process result section.
			DATA_LINE: while(<$OUTPUT_FILE>) {

				# Break on empty line.
				if(($_ =~ /^$/) || ($_ =~ /^.$/)) {
					last DATA_LINE;
				}

				# Print header, but only once.
				if($header_line) {
					if($print_header) {
						print $CSV_FILE join(',', @variable_names).",".$_;
						$print_header = 0;
					}
					$header_line = 0;
					next DATA_LINE;
				}

				# Print multi-line output data.
				for my $variable_name (@variable_names) {
					$output_file =~ /_$variable_name\_(\d+\.\d+|\d+)/;
					print $CSV_FILE "$1,";
				}
				print $CSV_FILE $_;

			}
		}
	}
	close($OUTPUT_FILE);

	# Check whether the simulation has been completed.
	if(!$simulation_complete_flag) {
		die "No simulation complete flag in $output_file";
	}
}
close($CSV_FILE);
print "Done.\n";
